/* Automatically @generated by tree-sitter v0.25.6 */

#include "tree_sitter/parser.h"

#if defined(__GNUC__) || defined(__clang__)
#pragma GCC diagnostic ignored "-Wmissing-field-initializers"
#endif

#ifdef _MSC_VER
#pragma optimize("", off)
#elif defined(__clang__)
#pragma clang optimize off
#elif defined(__GNUC__)
#pragma GCC optimize ("O0")
#endif

#define LANGUAGE_VERSION 15
#define STATE_COUNT 311
#define LARGE_STATE_COUNT 5
#define SYMBOL_COUNT 102
#define ALIAS_COUNT 0
#define TOKEN_COUNT 56
#define EXTERNAL_TOKEN_COUNT 6
#define FIELD_COUNT 12
#define MAX_ALIAS_SEQUENCE_LENGTH 6
#define MAX_RESERVED_WORD_SET_SIZE 0
#define PRODUCTION_ID_COUNT 19
#define SUPERTYPE_COUNT 0

enum ts_symbol_identifiers {
  sym_identifier = 1,
  sym_comma = 2,
  sym_semi_colon = 3,
  anon_sym_COLON = 4,
  sym_equals = 5,
  sym_token = 6,
  sym_static_token = 7,
  sym_static_argument = 8,
  anon_sym_ATmacro = 9,
  aux_sym_macro_call_token1 = 10,
  anon_sym_ATpriority = 11,
  anon_sym_ATname = 12,
  anon_sym_ATderive = 13,
  sym_name_selector = 14,
  sym_tag_selector = 15,
  aux_sym_state_selector_token1 = 16,
  sym_pseudo_selector = 17,
  sym_scope_open = 18,
  sym_scope_close = 19,
  sym_tuple_open = 20,
  sym_tuple_close = 21,
  sym_number_ambiguous = 22,
  sym_number_percent = 23,
  sym_number_offset = 24,
  sym_pow_operator = 25,
  sym_div_operator = 26,
  sym_floor_div_operator = 27,
  sym_mod_operator = 28,
  sym_mult_operator = 29,
  sym_add_operator = 30,
  sym_sub_operator = 31,
  sym_single_quote_string = 32,
  sym_double_quote_string = 33,
  sym__color_code_shade = 34,
  aux_sym_color_code_tailwind_token1 = 35,
  aux_sym_color_code_tailwind_token2 = 36,
  aux_sym_color_code_skin_token1 = 37,
  aux_sym_color_code_skin_token2 = 38,
  aux_sym_color_code_brick_token1 = 39,
  aux_sym_color_code_brick_token2 = 40,
  aux_sym_color_code_css_token1 = 41,
  aux_sym_color_code_css_token2 = 42,
  sym_color_hex = 43,
  anon_sym_Enum = 44,
  anon_sym_DOT = 45,
  sym_rbx_asset = 46,
  sym_rbx_content = 47,
  anon_sym_DASH_DASH = 48,
  aux_sym_comment_token1 = 49,
  sym__block_comment_start = 50,
  sym__block_comment_content = 51,
  sym__block_comment_end = 52,
  sym__block_string_start = 53,
  sym__block_string_content = 54,
  sym__block_string_end = 55,
  sym_source_file = 56,
  sym__definition_no_macro = 57,
  sym__definition = 58,
  sym__definition_scope_body = 59,
  sym_colon = 60,
  sym_reference = 61,
  sym_property_assignment = 62,
  sym_token_assignment = 63,
  sym_static_token_assignment = 64,
  sym_macro_declaration = 65,
  sym_macro_args = 66,
  sym_macro_scope = 67,
  sym_macro_body = 68,
  sym_macro_call = 69,
  sym_priority_declaration = 70,
  sym_name_declaration = 71,
  sym_derive_declaration = 72,
  sym_derive_value = 73,
  sym_class_selector = 74,
  sym_state_selector = 75,
  sym_selector = 76,
  sym_rule_scope = 77,
  sym_rule_scope_body = 78,
  sym_datatype = 79,
  sym_tuple = 80,
  sym_number = 81,
  sym_operator = 82,
  sym_operation = 83,
  sym_string = 84,
  sym_block_string = 85,
  sym_color = 86,
  sym_color_code_tailwind = 87,
  sym_color_code_skin = 88,
  sym_color_code_brick = 89,
  sym_color_code_css = 90,
  sym_enum = 91,
  sym_enum_full = 92,
  sym_enum_shorthand = 93,
  sym_comment = 94,
  aux_sym_source_file_repeat1 = 95,
  aux_sym_macro_args_repeat1 = 96,
  aux_sym_macro_scope_repeat1 = 97,
  aux_sym_macro_call_repeat1 = 98,
  aux_sym_derive_value_repeat1 = 99,
  aux_sym_rule_scope_repeat1 = 100,
  aux_sym_rule_scope_repeat2 = 101,
};

static const char * const ts_symbol_names[] = {
  [ts_builtin_sym_end] = "end",
  [sym_identifier] = "identifier",
  [sym_comma] = "comma",
  [sym_semi_colon] = "semi_colon",
  [anon_sym_COLON] = ":",
  [sym_equals] = "equals",
  [sym_token] = "token",
  [sym_static_token] = "static_token",
  [sym_static_argument] = "static_argument",
  [anon_sym_ATmacro] = "@macro",
  [aux_sym_macro_call_token1] = "macro_call_token1",
  [anon_sym_ATpriority] = "@priority",
  [anon_sym_ATname] = "@name",
  [anon_sym_ATderive] = "@derive",
  [sym_name_selector] = "name_selector",
  [sym_tag_selector] = "tag_selector",
  [aux_sym_state_selector_token1] = "state_selector_token1",
  [sym_pseudo_selector] = "pseudo_selector",
  [sym_scope_open] = "scope_open",
  [sym_scope_close] = "scope_close",
  [sym_tuple_open] = "tuple_open",
  [sym_tuple_close] = "tuple_close",
  [sym_number_ambiguous] = "number_ambiguous",
  [sym_number_percent] = "number_percent",
  [sym_number_offset] = "number_offset",
  [sym_pow_operator] = "pow_operator",
  [sym_div_operator] = "div_operator",
  [sym_floor_div_operator] = "floor_div_operator",
  [sym_mod_operator] = "mod_operator",
  [sym_mult_operator] = "mult_operator",
  [sym_add_operator] = "add_operator",
  [sym_sub_operator] = "sub_operator",
  [sym_single_quote_string] = "single_quote_string",
  [sym_double_quote_string] = "double_quote_string",
  [sym__color_code_shade] = "_color_code_shade",
  [aux_sym_color_code_tailwind_token1] = "color_code_tailwind_token1",
  [aux_sym_color_code_tailwind_token2] = "color_code_tailwind_token2",
  [aux_sym_color_code_skin_token1] = "color_code_skin_token1",
  [aux_sym_color_code_skin_token2] = "color_code_skin_token2",
  [aux_sym_color_code_brick_token1] = "color_code_brick_token1",
  [aux_sym_color_code_brick_token2] = "color_code_brick_token2",
  [aux_sym_color_code_css_token1] = "color_code_css_token1",
  [aux_sym_color_code_css_token2] = "color_code_css_token2",
  [sym_color_hex] = "color_hex",
  [anon_sym_Enum] = "Enum",
  [anon_sym_DOT] = ".",
  [sym_rbx_asset] = "rbx_asset",
  [sym_rbx_content] = "rbx_content",
  [anon_sym_DASH_DASH] = "--",
  [aux_sym_comment_token1] = "comment_content",
  [sym__block_comment_start] = "[[",
  [sym__block_comment_content] = "comment_content",
  [sym__block_comment_end] = "]]",
  [sym__block_string_start] = "[[",
  [sym__block_string_content] = "string_content",
  [sym__block_string_end] = "]]",
  [sym_source_file] = "source_file",
  [sym__definition_no_macro] = "_definition_no_macro",
  [sym__definition] = "_definition",
  [sym__definition_scope_body] = "_definition_scope_body",
  [sym_colon] = "colon",
  [sym_reference] = "reference",
  [sym_property_assignment] = "property_assignment",
  [sym_token_assignment] = "token_assignment",
  [sym_static_token_assignment] = "static_token_assignment",
  [sym_macro_declaration] = "macro_declaration",
  [sym_macro_args] = "macro_args",
  [sym_macro_scope] = "macro_scope",
  [sym_macro_body] = "macro_body",
  [sym_macro_call] = "macro_call",
  [sym_priority_declaration] = "priority_declaration",
  [sym_name_declaration] = "name_declaration",
  [sym_derive_declaration] = "derive_declaration",
  [sym_derive_value] = "derive_value",
  [sym_class_selector] = "class_selector",
  [sym_state_selector] = "state_selector",
  [sym_selector] = "selector",
  [sym_rule_scope] = "rule_scope",
  [sym_rule_scope_body] = "rule_scope_body",
  [sym_datatype] = "datatype",
  [sym_tuple] = "tuple",
  [sym_number] = "number",
  [sym_operator] = "operator",
  [sym_operation] = "operation",
  [sym_string] = "string",
  [sym_block_string] = "block_string",
  [sym_color] = "color",
  [sym_color_code_tailwind] = "color_code_tailwind",
  [sym_color_code_skin] = "color_code_skin",
  [sym_color_code_brick] = "color_code_brick",
  [sym_color_code_css] = "color_code_css",
  [sym_enum] = "enum",
  [sym_enum_full] = "enum_full",
  [sym_enum_shorthand] = "enum_shorthand",
  [sym_comment] = "comment",
  [aux_sym_source_file_repeat1] = "source_file_repeat1",
  [aux_sym_macro_args_repeat1] = "macro_args_repeat1",
  [aux_sym_macro_scope_repeat1] = "macro_scope_repeat1",
  [aux_sym_macro_call_repeat1] = "macro_call_repeat1",
  [aux_sym_derive_value_repeat1] = "derive_value_repeat1",
  [aux_sym_rule_scope_repeat1] = "rule_scope_repeat1",
  [aux_sym_rule_scope_repeat2] = "rule_scope_repeat2",
};

static const TSSymbol ts_symbol_map[] = {
  [ts_builtin_sym_end] = ts_builtin_sym_end,
  [sym_identifier] = sym_identifier,
  [sym_comma] = sym_comma,
  [sym_semi_colon] = sym_semi_colon,
  [anon_sym_COLON] = anon_sym_COLON,
  [sym_equals] = sym_equals,
  [sym_token] = sym_token,
  [sym_static_token] = sym_static_token,
  [sym_static_argument] = sym_static_argument,
  [anon_sym_ATmacro] = anon_sym_ATmacro,
  [aux_sym_macro_call_token1] = aux_sym_macro_call_token1,
  [anon_sym_ATpriority] = anon_sym_ATpriority,
  [anon_sym_ATname] = anon_sym_ATname,
  [anon_sym_ATderive] = anon_sym_ATderive,
  [sym_name_selector] = sym_name_selector,
  [sym_tag_selector] = sym_tag_selector,
  [aux_sym_state_selector_token1] = aux_sym_state_selector_token1,
  [sym_pseudo_selector] = sym_pseudo_selector,
  [sym_scope_open] = sym_scope_open,
  [sym_scope_close] = sym_scope_close,
  [sym_tuple_open] = sym_tuple_open,
  [sym_tuple_close] = sym_tuple_close,
  [sym_number_ambiguous] = sym_number_ambiguous,
  [sym_number_percent] = sym_number_percent,
  [sym_number_offset] = sym_number_offset,
  [sym_pow_operator] = sym_pow_operator,
  [sym_div_operator] = sym_div_operator,
  [sym_floor_div_operator] = sym_floor_div_operator,
  [sym_mod_operator] = sym_mod_operator,
  [sym_mult_operator] = sym_mult_operator,
  [sym_add_operator] = sym_add_operator,
  [sym_sub_operator] = sym_sub_operator,
  [sym_single_quote_string] = sym_single_quote_string,
  [sym_double_quote_string] = sym_double_quote_string,
  [sym__color_code_shade] = sym__color_code_shade,
  [aux_sym_color_code_tailwind_token1] = aux_sym_color_code_tailwind_token1,
  [aux_sym_color_code_tailwind_token2] = aux_sym_color_code_tailwind_token2,
  [aux_sym_color_code_skin_token1] = aux_sym_color_code_skin_token1,
  [aux_sym_color_code_skin_token2] = aux_sym_color_code_skin_token2,
  [aux_sym_color_code_brick_token1] = aux_sym_color_code_brick_token1,
  [aux_sym_color_code_brick_token2] = aux_sym_color_code_brick_token2,
  [aux_sym_color_code_css_token1] = aux_sym_color_code_css_token1,
  [aux_sym_color_code_css_token2] = aux_sym_color_code_css_token2,
  [sym_color_hex] = sym_color_hex,
  [anon_sym_Enum] = anon_sym_Enum,
  [anon_sym_DOT] = anon_sym_DOT,
  [sym_rbx_asset] = sym_rbx_asset,
  [sym_rbx_content] = sym_rbx_content,
  [anon_sym_DASH_DASH] = anon_sym_DASH_DASH,
  [aux_sym_comment_token1] = sym__block_comment_content,
  [sym__block_comment_start] = sym__block_comment_start,
  [sym__block_comment_content] = sym__block_comment_content,
  [sym__block_comment_end] = sym__block_comment_end,
  [sym__block_string_start] = sym__block_comment_start,
  [sym__block_string_content] = sym__block_string_content,
  [sym__block_string_end] = sym__block_comment_end,
  [sym_source_file] = sym_source_file,
  [sym__definition_no_macro] = sym__definition_no_macro,
  [sym__definition] = sym__definition,
  [sym__definition_scope_body] = sym__definition_scope_body,
  [sym_colon] = sym_colon,
  [sym_reference] = sym_reference,
  [sym_property_assignment] = sym_property_assignment,
  [sym_token_assignment] = sym_token_assignment,
  [sym_static_token_assignment] = sym_static_token_assignment,
  [sym_macro_declaration] = sym_macro_declaration,
  [sym_macro_args] = sym_macro_args,
  [sym_macro_scope] = sym_macro_scope,
  [sym_macro_body] = sym_macro_body,
  [sym_macro_call] = sym_macro_call,
  [sym_priority_declaration] = sym_priority_declaration,
  [sym_name_declaration] = sym_name_declaration,
  [sym_derive_declaration] = sym_derive_declaration,
  [sym_derive_value] = sym_derive_value,
  [sym_class_selector] = sym_class_selector,
  [sym_state_selector] = sym_state_selector,
  [sym_selector] = sym_selector,
  [sym_rule_scope] = sym_rule_scope,
  [sym_rule_scope_body] = sym_rule_scope_body,
  [sym_datatype] = sym_datatype,
  [sym_tuple] = sym_tuple,
  [sym_number] = sym_number,
  [sym_operator] = sym_operator,
  [sym_operation] = sym_operation,
  [sym_string] = sym_string,
  [sym_block_string] = sym_block_string,
  [sym_color] = sym_color,
  [sym_color_code_tailwind] = sym_color_code_tailwind,
  [sym_color_code_skin] = sym_color_code_skin,
  [sym_color_code_brick] = sym_color_code_brick,
  [sym_color_code_css] = sym_color_code_css,
  [sym_enum] = sym_enum,
  [sym_enum_full] = sym_enum_full,
  [sym_enum_shorthand] = sym_enum_shorthand,
  [sym_comment] = sym_comment,
  [aux_sym_source_file_repeat1] = aux_sym_source_file_repeat1,
  [aux_sym_macro_args_repeat1] = aux_sym_macro_args_repeat1,
  [aux_sym_macro_scope_repeat1] = aux_sym_macro_scope_repeat1,
  [aux_sym_macro_call_repeat1] = aux_sym_macro_call_repeat1,
  [aux_sym_derive_value_repeat1] = aux_sym_derive_value_repeat1,
  [aux_sym_rule_scope_repeat1] = aux_sym_rule_scope_repeat1,
  [aux_sym_rule_scope_repeat2] = aux_sym_rule_scope_repeat2,
};

static const TSSymbolMetadata ts_symbol_metadata[] = {
  [ts_builtin_sym_end] = {
    .visible = false,
    .named = true,
  },
  [sym_identifier] = {
    .visible = true,
    .named = true,
  },
  [sym_comma] = {
    .visible = true,
    .named = true,
  },
  [sym_semi_colon] = {
    .visible = true,
    .named = true,
  },
  [anon_sym_COLON] = {
    .visible = true,
    .named = false,
  },
  [sym_equals] = {
    .visible = true,
    .named = true,
  },
  [sym_token] = {
    .visible = true,
    .named = true,
  },
  [sym_static_token] = {
    .visible = true,
    .named = true,
  },
  [sym_static_argument] = {
    .visible = true,
    .named = true,
  },
  [anon_sym_ATmacro] = {
    .visible = true,
    .named = false,
  },
  [aux_sym_macro_call_token1] = {
    .visible = false,
    .named = false,
  },
  [anon_sym_ATpriority] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_ATname] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_ATderive] = {
    .visible = true,
    .named = false,
  },
  [sym_name_selector] = {
    .visible = true,
    .named = true,
  },
  [sym_tag_selector] = {
    .visible = true,
    .named = true,
  },
  [aux_sym_state_selector_token1] = {
    .visible = false,
    .named = false,
  },
  [sym_pseudo_selector] = {
    .visible = true,
    .named = true,
  },
  [sym_scope_open] = {
    .visible = true,
    .named = true,
  },
  [sym_scope_close] = {
    .visible = true,
    .named = true,
  },
  [sym_tuple_open] = {
    .visible = true,
    .named = true,
  },
  [sym_tuple_close] = {
    .visible = true,
    .named = true,
  },
  [sym_number_ambiguous] = {
    .visible = true,
    .named = true,
  },
  [sym_number_percent] = {
    .visible = true,
    .named = true,
  },
  [sym_number_offset] = {
    .visible = true,
    .named = true,
  },
  [sym_pow_operator] = {
    .visible = true,
    .named = true,
  },
  [sym_div_operator] = {
    .visible = true,
    .named = true,
  },
  [sym_floor_div_operator] = {
    .visible = true,
    .named = true,
  },
  [sym_mod_operator] = {
    .visible = true,
    .named = true,
  },
  [sym_mult_operator] = {
    .visible = true,
    .named = true,
  },
  [sym_add_operator] = {
    .visible = true,
    .named = true,
  },
  [sym_sub_operator] = {
    .visible = true,
    .named = true,
  },
  [sym_single_quote_string] = {
    .visible = true,
    .named = true,
  },
  [sym_double_quote_string] = {
    .visible = true,
    .named = true,
  },
  [sym__color_code_shade] = {
    .visible = false,
    .named = true,
  },
  [aux_sym_color_code_tailwind_token1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_color_code_tailwind_token2] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_color_code_skin_token1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_color_code_skin_token2] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_color_code_brick_token1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_color_code_brick_token2] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_color_code_css_token1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_color_code_css_token2] = {
    .visible = false,
    .named = false,
  },
  [sym_color_hex] = {
    .visible = true,
    .named = true,
  },
  [anon_sym_Enum] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_DOT] = {
    .visible = true,
    .named = false,
  },
  [sym_rbx_asset] = {
    .visible = true,
    .named = true,
  },
  [sym_rbx_content] = {
    .visible = true,
    .named = true,
  },
  [anon_sym_DASH_DASH] = {
    .visible = true,
    .named = false,
  },
  [aux_sym_comment_token1] = {
    .visible = true,
    .named = true,
  },
  [sym__block_comment_start] = {
    .visible = true,
    .named = false,
  },
  [sym__block_comment_content] = {
    .visible = true,
    .named = true,
  },
  [sym__block_comment_end] = {
    .visible = true,
    .named = false,
  },
  [sym__block_string_start] = {
    .visible = true,
    .named = false,
  },
  [sym__block_string_content] = {
    .visible = true,
    .named = true,
  },
  [sym__block_string_end] = {
    .visible = true,
    .named = false,
  },
  [sym_source_file] = {
    .visible = true,
    .named = true,
  },
  [sym__definition_no_macro] = {
    .visible = false,
    .named = true,
  },
  [sym__definition] = {
    .visible = false,
    .named = true,
  },
  [sym__definition_scope_body] = {
    .visible = false,
    .named = true,
  },
  [sym_colon] = {
    .visible = true,
    .named = true,
  },
  [sym_reference] = {
    .visible = true,
    .named = true,
  },
  [sym_property_assignment] = {
    .visible = true,
    .named = true,
  },
  [sym_token_assignment] = {
    .visible = true,
    .named = true,
  },
  [sym_static_token_assignment] = {
    .visible = true,
    .named = true,
  },
  [sym_macro_declaration] = {
    .visible = true,
    .named = true,
  },
  [sym_macro_args] = {
    .visible = true,
    .named = true,
  },
  [sym_macro_scope] = {
    .visible = true,
    .named = true,
  },
  [sym_macro_body] = {
    .visible = true,
    .named = true,
  },
  [sym_macro_call] = {
    .visible = true,
    .named = true,
  },
  [sym_priority_declaration] = {
    .visible = true,
    .named = true,
  },
  [sym_name_declaration] = {
    .visible = true,
    .named = true,
  },
  [sym_derive_declaration] = {
    .visible = true,
    .named = true,
  },
  [sym_derive_value] = {
    .visible = true,
    .named = true,
  },
  [sym_class_selector] = {
    .visible = true,
    .named = true,
  },
  [sym_state_selector] = {
    .visible = true,
    .named = true,
  },
  [sym_selector] = {
    .visible = true,
    .named = true,
  },
  [sym_rule_scope] = {
    .visible = true,
    .named = true,
  },
  [sym_rule_scope_body] = {
    .visible = true,
    .named = true,
  },
  [sym_datatype] = {
    .visible = true,
    .named = true,
  },
  [sym_tuple] = {
    .visible = true,
    .named = true,
  },
  [sym_number] = {
    .visible = true,
    .named = true,
  },
  [sym_operator] = {
    .visible = true,
    .named = true,
  },
  [sym_operation] = {
    .visible = true,
    .named = true,
  },
  [sym_string] = {
    .visible = true,
    .named = true,
  },
  [sym_block_string] = {
    .visible = true,
    .named = true,
  },
  [sym_color] = {
    .visible = true,
    .named = true,
  },
  [sym_color_code_tailwind] = {
    .visible = true,
    .named = true,
  },
  [sym_color_code_skin] = {
    .visible = true,
    .named = true,
  },
  [sym_color_code_brick] = {
    .visible = true,
    .named = true,
  },
  [sym_color_code_css] = {
    .visible = true,
    .named = true,
  },
  [sym_enum] = {
    .visible = true,
    .named = true,
  },
  [sym_enum_full] = {
    .visible = true,
    .named = true,
  },
  [sym_enum_shorthand] = {
    .visible = true,
    .named = true,
  },
  [sym_comment] = {
    .visible = true,
    .named = true,
  },
  [aux_sym_source_file_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_macro_args_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_macro_scope_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_macro_call_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_derive_value_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_rule_scope_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_rule_scope_repeat2] = {
    .visible = false,
    .named = false,
  },
};

enum ts_field_identifiers {
  field_annotation = 1,
  field_body = 2,
  field_color = 3,
  field_content = 4,
  field_end = 5,
  field_name = 6,
  field_property_assignment = 7,
  field_selector = 8,
  field_shade = 9,
  field_start = 10,
  field_static_token_assignment = 11,
  field_token_assignment = 12,
};

static const char * const ts_field_names[] = {
  [0] = NULL,
  [field_annotation] = "annotation",
  [field_body] = "body",
  [field_color] = "color",
  [field_content] = "content",
  [field_end] = "end",
  [field_name] = "name",
  [field_property_assignment] = "property_assignment",
  [field_selector] = "selector",
  [field_shade] = "shade",
  [field_start] = "start",
  [field_static_token_assignment] = "static_token_assignment",
  [field_token_assignment] = "token_assignment",
};

static const TSMapSlice ts_field_map_slices[PRODUCTION_ID_COUNT] = {
  [1] = {.index = 0, .length = 2},
  [2] = {.index = 2, .length = 3},
  [3] = {.index = 5, .length = 1},
  [4] = {.index = 6, .length = 3},
  [5] = {.index = 9, .length = 3},
  [6] = {.index = 12, .length = 1},
  [7] = {.index = 13, .length = 1},
  [8] = {.index = 14, .length = 2},
  [9] = {.index = 16, .length = 2},
  [10] = {.index = 18, .length = 4},
  [11] = {.index = 22, .length = 4},
  [12] = {.index = 26, .length = 3},
  [13] = {.index = 29, .length = 2},
  [14] = {.index = 31, .length = 3},
  [15] = {.index = 34, .length = 3},
  [16] = {.index = 37, .length = 4},
  [17] = {.index = 41, .length = 3},
  [18] = {.index = 44, .length = 4},
};

static const TSFieldMapEntry ts_field_map_entries[] = {
  [0] =
    {field_content, 1},
    {field_start, 0},
  [2] =
    {field_content, 1},
    {field_end, 2},
    {field_start, 0},
  [5] =
    {field_annotation, 0},
  [6] =
    {field_token_assignment, 0},
    {field_token_assignment, 1},
    {field_token_assignment, 2},
  [9] =
    {field_static_token_assignment, 0},
    {field_static_token_assignment, 1},
    {field_static_token_assignment, 2},
  [12] =
    {field_body, 1},
  [13] =
    {field_selector, 0},
  [14] =
    {field_color, 1},
    {field_name, 0},
  [16] =
    {field_selector, 0},
    {field_selector, 1},
  [18] =
    {field_token_assignment, 0},
    {field_token_assignment, 1},
    {field_token_assignment, 2},
    {field_token_assignment, 3},
  [22] =
    {field_static_token_assignment, 0},
    {field_static_token_assignment, 1},
    {field_static_token_assignment, 2},
    {field_static_token_assignment, 3},
  [26] =
    {field_property_assignment, 0},
    {field_property_assignment, 1},
    {field_property_assignment, 2},
  [29] =
    {field_body, 2},
    {field_selector, 0},
  [31] =
    {field_body, 3},
    {field_selector, 0},
    {field_selector, 1},
  [34] =
    {field_selector, 0},
    {field_selector, 1},
    {field_selector, 2},
  [37] =
    {field_property_assignment, 0},
    {field_property_assignment, 1},
    {field_property_assignment, 2},
    {field_property_assignment, 3},
  [41] =
    {field_color, 1},
    {field_name, 0},
    {field_shade, 3},
  [44] =
    {field_body, 4},
    {field_selector, 0},
    {field_selector, 1},
    {field_selector, 2},
};

static const TSSymbol ts_alias_sequences[PRODUCTION_ID_COUNT][MAX_ALIAS_SEQUENCE_LENGTH] = {
  [0] = {0},
};

static const uint16_t ts_non_terminal_alias_map[] = {
  0,
};

static const TSStateId ts_primary_state_ids[STATE_COUNT] = {
  [0] = 0,
  [1] = 1,
  [2] = 2,
  [3] = 3,
  [4] = 4,
  [5] = 5,
  [6] = 6,
  [7] = 7,
  [8] = 8,
  [9] = 9,
  [10] = 10,
  [11] = 11,
  [12] = 12,
  [13] = 13,
  [14] = 14,
  [15] = 15,
  [16] = 16,
  [17] = 17,
  [18] = 18,
  [19] = 19,
  [20] = 20,
  [21] = 21,
  [22] = 22,
  [23] = 23,
  [24] = 24,
  [25] = 25,
  [26] = 26,
  [27] = 27,
  [28] = 28,
  [29] = 29,
  [30] = 30,
  [31] = 31,
  [32] = 32,
  [33] = 33,
  [34] = 34,
  [35] = 35,
  [36] = 36,
  [37] = 37,
  [38] = 38,
  [39] = 39,
  [40] = 40,
  [41] = 41,
  [42] = 42,
  [43] = 43,
  [44] = 39,
  [45] = 45,
  [46] = 46,
  [47] = 47,
  [48] = 48,
  [49] = 49,
  [50] = 50,
  [51] = 51,
  [52] = 52,
  [53] = 53,
  [54] = 54,
  [55] = 55,
  [56] = 56,
  [57] = 57,
  [58] = 58,
  [59] = 59,
  [60] = 60,
  [61] = 61,
  [62] = 62,
  [63] = 63,
  [64] = 64,
  [65] = 65,
  [66] = 66,
  [67] = 67,
  [68] = 68,
  [69] = 69,
  [70] = 70,
  [71] = 71,
  [72] = 72,
  [73] = 39,
  [74] = 74,
  [75] = 22,
  [76] = 19,
  [77] = 77,
  [78] = 78,
  [79] = 79,
  [80] = 80,
  [81] = 81,
  [82] = 82,
  [83] = 83,
  [84] = 78,
  [85] = 85,
  [86] = 86,
  [87] = 79,
  [88] = 88,
  [89] = 80,
  [90] = 90,
  [91] = 85,
  [92] = 92,
  [93] = 93,
  [94] = 86,
  [95] = 79,
  [96] = 80,
  [97] = 81,
  [98] = 82,
  [99] = 88,
  [100] = 100,
  [101] = 90,
  [102] = 93,
  [103] = 86,
  [104] = 100,
  [105] = 105,
  [106] = 105,
  [107] = 107,
  [108] = 108,
  [109] = 109,
  [110] = 110,
  [111] = 111,
  [112] = 112,
  [113] = 113,
  [114] = 114,
  [115] = 108,
  [116] = 110,
  [117] = 117,
  [118] = 118,
  [119] = 113,
  [120] = 112,
  [121] = 114,
  [122] = 111,
  [123] = 123,
  [124] = 109,
  [125] = 125,
  [126] = 47,
  [127] = 8,
  [128] = 64,
  [129] = 58,
  [130] = 59,
  [131] = 49,
  [132] = 50,
  [133] = 133,
  [134] = 27,
  [135] = 34,
  [136] = 35,
  [137] = 48,
  [138] = 54,
  [139] = 40,
  [140] = 140,
  [141] = 141,
  [142] = 142,
  [143] = 38,
  [144] = 144,
  [145] = 25,
  [146] = 146,
  [147] = 147,
  [148] = 37,
  [149] = 149,
  [150] = 150,
  [151] = 151,
  [152] = 52,
  [153] = 53,
  [154] = 154,
  [155] = 36,
  [156] = 156,
  [157] = 7,
  [158] = 158,
  [159] = 159,
  [160] = 160,
  [161] = 161,
  [162] = 162,
  [163] = 42,
  [164] = 43,
  [165] = 165,
  [166] = 165,
  [167] = 161,
  [168] = 162,
  [169] = 169,
  [170] = 156,
  [171] = 171,
  [172] = 158,
  [173] = 159,
  [174] = 160,
  [175] = 165,
  [176] = 161,
  [177] = 156,
  [178] = 169,
  [179] = 171,
  [180] = 30,
  [181] = 41,
  [182] = 182,
  [183] = 183,
  [184] = 60,
  [185] = 77,
  [186] = 45,
  [187] = 70,
  [188] = 188,
  [189] = 71,
  [190] = 28,
  [191] = 29,
  [192] = 192,
  [193] = 182,
  [194] = 72,
  [195] = 31,
  [196] = 61,
  [197] = 33,
  [198] = 51,
  [199] = 62,
  [200] = 63,
  [201] = 201,
  [202] = 65,
  [203] = 66,
  [204] = 67,
  [205] = 68,
  [206] = 206,
  [207] = 26,
  [208] = 55,
  [209] = 69,
  [210] = 56,
  [211] = 201,
  [212] = 32,
  [213] = 213,
  [214] = 214,
  [215] = 214,
  [216] = 216,
  [217] = 217,
  [218] = 217,
  [219] = 219,
  [220] = 220,
  [221] = 221,
  [222] = 222,
  [223] = 24,
  [224] = 224,
  [225] = 225,
  [226] = 226,
  [227] = 220,
  [228] = 219,
  [229] = 229,
  [230] = 230,
  [231] = 231,
  [232] = 232,
  [233] = 232,
  [234] = 234,
  [235] = 235,
  [236] = 235,
  [237] = 237,
  [238] = 237,
  [239] = 239,
  [240] = 240,
  [241] = 241,
  [242] = 242,
  [243] = 243,
  [244] = 244,
  [245] = 245,
  [246] = 246,
  [247] = 247,
  [248] = 248,
  [249] = 249,
  [250] = 250,
  [251] = 251,
  [252] = 252,
  [253] = 251,
  [254] = 243,
  [255] = 255,
  [256] = 256,
  [257] = 257,
  [258] = 258,
  [259] = 259,
  [260] = 260,
  [261] = 261,
  [262] = 262,
  [263] = 263,
  [264] = 264,
  [265] = 260,
  [266] = 261,
  [267] = 267,
  [268] = 268,
  [269] = 262,
  [270] = 263,
  [271] = 271,
  [272] = 272,
  [273] = 273,
  [274] = 274,
  [275] = 275,
  [276] = 276,
  [277] = 277,
  [278] = 278,
  [279] = 279,
  [280] = 280,
  [281] = 281,
  [282] = 282,
  [283] = 46,
  [284] = 274,
  [285] = 285,
  [286] = 282,
  [287] = 280,
  [288] = 288,
  [289] = 289,
  [290] = 290,
  [291] = 291,
  [292] = 292,
  [293] = 290,
  [294] = 294,
  [295] = 281,
  [296] = 273,
  [297] = 274,
  [298] = 298,
  [299] = 299,
  [300] = 294,
  [301] = 301,
  [302] = 292,
  [303] = 303,
  [304] = 279,
  [305] = 71,
  [306] = 72,
  [307] = 307,
  [308] = 70,
  [309] = 69,
  [310] = 310,
};

static bool ts_lex(TSLexer *lexer, TSStateId state) {
  START_LEXER();
  eof = lexer->eof(lexer);
  switch (state) {
    case 0:
      if (eof) ADVANCE(1185);
      ADVANCE_MAP(
        '"', 1000,
        '#', 1174,
        '$', 997,
        '%', 1359,
        '&', 1179,
        '\'', 1005,
        '(', 1348,
        ')', 1349,
        '*', 1360,
        '+', 1361,
        ',', 1325,
        '-', 1362,
        '.', 1397,
        '/', 1357,
        '1', 1015,
        '2', 1015,
        '3', 1015,
        '4', 1015,
        '5', 1013,
        '6', 1015,
        '7', 1015,
        '8', 1015,
        '9', 1016,
        ':', 1328,
        ';', 1326,
        '=', 1329,
        '@', 1024,
        'C', 1309,
        'E', 1214,
        'R', 1258,
        '^', 1356,
        'c', 1218,
        'e', 1289,
        'h', 1230,
        'r', 1199,
        '{', 1346,
        '}', 1347,
        'A', 1288,
        'a', 1288,
        'B', 1282,
        'b', 1282,
        'F', 1317,
        'f', 1317,
        'G', 1299,
        'g', 1299,
        'I', 1293,
        'i', 1293,
        'L', 1279,
        'l', 1279,
        'N', 1262,
        'n', 1262,
        'O', 1285,
        'o', 1285,
        'P', 1259,
        'p', 1259,
        'S', 1238,
        's', 1238,
        'T', 1266,
        't', 1266,
        'V', 1274,
        'v', 1274,
        'Y', 1260,
        'y', 1260,
        'Z', 1275,
        'z', 1275,
      );
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(0);
      if (('D' <= lookahead && lookahead <= 'X') ||
          lookahead == '_' ||
          ('d' <= lookahead && lookahead <= 'x')) ADVANCE(1324);
      END_STATE();
    case 1:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'x') ADVANCE(1355);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 2:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 0x212a) ADVANCE(1368);
      if (lookahead == 'K' ||
          lookahead == 'k') ADVANCE(1369);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 3:
      ADVANCE_MAP(
        '!', 1334,
        0x212a, 1164,
        'K', 973,
        'k', 973,
        'L', 39,
        'l', 39,
        'T', 668,
        't', 668,
      );
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 4:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 0x212a) ADVANCE(1373);
      if (lookahead == 'E' ||
          lookahead == 'e') ADVANCE(174);
      if (lookahead == 'K' ||
          lookahead == 'k') ADVANCE(1381);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 5:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 0x212a) ADVANCE(1373);
      if (lookahead == 'K' ||
          lookahead == 'k') ADVANCE(1381);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 6:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 0x212a) ADVANCE(1060);
      if (lookahead == 'K' ||
          lookahead == 'k') ADVANCE(135);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 7:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 0x212a) ADVANCE(1382);
      if (lookahead == 'K' ||
          lookahead == 'k') ADVANCE(1380);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 8:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 0x212a) ADVANCE(1062);
      if (lookahead == 'K' ||
          lookahead == 'k') ADVANCE(19);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 9:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 0x212a) ADVANCE(1384);
      if (lookahead == 'K' ||
          lookahead == 'k') ADVANCE(1394);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 10:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 0x212a) ADVANCE(1099);
      if (lookahead == 'K' ||
          lookahead == 'k') ADVANCE(413);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 11:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 0x212a) ADVANCE(1166);
      if (lookahead == 'K' ||
          lookahead == 'k') ADVANCE(987);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 12:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 0x212a) ADVANCE(1168);
      if (lookahead == 'K' ||
          lookahead == 'k') ADVANCE(988);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 13:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 0x212a) ADVANCE(1069);
      if (lookahead == 'K' ||
          lookahead == 'k') ADVANCE(224);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 14:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 0x212a) ADVANCE(1101);
      if (lookahead == 'K' ||
          lookahead == 'k') ADVANCE(415);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 15:
      ADVANCE_MAP(
        '!', 1334,
        0x212a, 1171,
        'A', 192,
        'a', 192,
        'E', 113,
        'e', 113,
        'I', 302,
        'i', 302,
        'K', 993,
        'k', 993,
        'L', 105,
        'l', 105,
        'N', 677,
        'n', 677,
        'P', 802,
        'p', 802,
        'T', 332,
        't', 332,
      );
      if (('0' <= lookahead && lookahead <= '9') ||
          ('B' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('b' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 16:
      ADVANCE_MAP(
        '!', 1334,
        0x212a, 1171,
        'A', 570,
        'a', 570,
        'E', 116,
        'e', 116,
        'K', 993,
        'k', 993,
        'L', 110,
        'l', 110,
        'T', 332,
        't', 332,
      );
      if (('0' <= lookahead && lookahead <= '9') ||
          ('B' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('b' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 17:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 0x212a) ADVANCE(1171);
      if (lookahead == 'K' ||
          lookahead == 'k') ADVANCE(993);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 18:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 0x212a) ADVANCE(1072);
      if (lookahead == 'K' ||
          lookahead == 'k') ADVANCE(231);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 19:
      ADVANCE_MAP(
        '!', 1334,
        0x212a, 1096,
        'B', 563,
        'b', 563,
        'C', 986,
        'c', 986,
        'G', 696,
        'g', 696,
        'K', 395,
        'k', 395,
        'M', 107,
        'm', 107,
        'O', 545,
        'o', 545,
        'R', 271,
        'r', 271,
        'S', 84,
        's', 84,
        'T', 941,
        't', 941,
        'V', 479,
        'v', 479,
      );
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 20:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'A' ||
          lookahead == 'a') ADVANCE(1369);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('B' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('b' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 21:
      ADVANCE_MAP(
        '!', 1334,
        'A', 191,
        'a', 191,
        'G', 43,
        'g', 43,
        'L', 671,
        'l', 671,
        'O', 158,
        'o', 158,
        'R', 421,
        'r', 421,
        'U', 769,
        'u', 769,
        'Y', 65,
        'y', 65,
      );
      if (('0' <= lookahead && lookahead <= '9') ||
          ('B' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('b' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 22:
      ADVANCE_MAP(
        '!', 1334,
        'A', 424,
        'a', 424,
        'E', 244,
        'e', 244,
        'O', 946,
        'o', 946,
        'U', 824,
        'u', 824,
      );
      if (('0' <= lookahead && lookahead <= '9') ||
          ('B' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('b' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 23:
      ADVANCE_MAP(
        '!', 1334,
        'A', 193,
        'a', 193,
        'I', 777,
        'i', 777,
        'L', 96,
        'l', 96,
        'O', 342,
        'o', 342,
      );
      if (('0' <= lookahead && lookahead <= '9') ||
          ('B' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('b' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 24:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'A' ||
          lookahead == 'a') ADVANCE(728);
      if (lookahead == 'E' ||
          lookahead == 'e') ADVANCE(595);
      if (lookahead == 'I' ||
          lookahead == 'i') ADVANCE(343);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('B' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('b' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 25:
      ADVANCE_MAP(
        '!', 1334,
        'A', 351,
        'a', 351,
        'E', 180,
        'e', 180,
        'I', 187,
        'i', 187,
        'O', 825,
        'o', 825,
        'U', 541,
        'u', 541,
      );
      if (('0' <= lookahead && lookahead <= '9') ||
          ('B' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('b' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 26:
      ADVANCE_MAP(
        '!', 1334,
        'A', 765,
        'a', 765,
        'E', 69,
        'e', 69,
        'H', 678,
        'h', 678,
        'I', 608,
        'i', 608,
        'L', 912,
        'l', 912,
      );
      if (('0' <= lookahead && lookahead <= '9') ||
          ('B' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('b' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 27:
      ADVANCE_MAP(
        '!', 1334,
        'A', 350,
        'a', 350,
        'E', 29,
        'e', 29,
        'H', 52,
        'h', 52,
        'I', 558,
        'i', 558,
        'L', 469,
        'l', 469,
        'M', 666,
        'm', 666,
        'T', 282,
        't', 282,
        'U', 638,
        'u', 638,
      );
      if (('0' <= lookahead && lookahead <= '9') ||
          ('B' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('b' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 28:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'A' ||
          lookahead == 'a') ADVANCE(1381);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('B' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('b' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 29:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'A' ||
          lookahead == 'a') ADVANCE(365);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('B' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('b' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 30:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'A' ||
          lookahead == 'a') ADVANCE(952);
      if (lookahead == 'E' ||
          lookahead == 'e') ADVANCE(599);
      if (lookahead == 'I' ||
          lookahead == 'i') ADVANCE(368);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('B' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('b' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 31:
      ADVANCE_MAP(
        '!', 1334,
        'A', 354,
        'a', 354,
        'E', 201,
        'e', 201,
        'I', 198,
        'i', 198,
        'O', 159,
        'o', 159,
      );
      if (('0' <= lookahead && lookahead <= '9') ||
          ('B' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('b' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 32:
      ADVANCE_MAP(
        '!', 1334,
        'A', 537,
        'a', 537,
        'E', 100,
        'e', 100,
        'I', 624,
        'i', 624,
        'L', 919,
        'l', 919,
        'O', 970,
        'o', 970,
        'U', 812,
        'u', 812,
      );
      if (('0' <= lookahead && lookahead <= '9') ||
          ('B' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('b' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 33:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'A' ||
          lookahead == 'a') ADVANCE(156);
      if (lookahead == 'U' ||
          lookahead == 'u') ADVANCE(232);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('B' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('b' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 34:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'A' ||
          lookahead == 'a') ADVANCE(1394);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('B' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('b' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 35:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'A' ||
          lookahead == 'a') ADVANCE(1389);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('B' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('b' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 36:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'A' ||
          lookahead == 'a') ADVANCE(483);
      if (lookahead == 'Y' ||
          lookahead == 'y') ADVANCE(1394);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('B' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('b' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 37:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'A' ||
          lookahead == 'a') ADVANCE(973);
      if (lookahead == 'E' ||
          lookahead == 'e') ADVANCE(248);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('B' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('b' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 38:
      ADVANCE_MAP(
        '!', 1334,
        'A', 120,
        'a', 120,
        'E', 456,
        'e', 456,
        'L', 44,
        'l', 44,
        'R', 412,
        'r', 412,
        'U', 758,
        'u', 758,
      );
      if (('0' <= lookahead && lookahead <= '9') ||
          ('B' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('b' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 39:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'A' ||
          lookahead == 'a') ADVANCE(867);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('B' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('b' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 40:
      ADVANCE_MAP(
        '!', 1334,
        'A', 962,
        'a', 962,
        'E', 67,
        'e', 67,
        'O', 676,
        'o', 676,
        'R', 72,
        'r', 72,
        'U', 753,
        'u', 753,
      );
      if (('0' <= lookahead && lookahead <= '9') ||
          ('B' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('b' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 41:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'A' ||
          lookahead == 'a') ADVANCE(943);
      if (lookahead == 'E' ||
          lookahead == 'e') ADVANCE(684);
      if (lookahead == 'O' ||
          lookahead == 'o') ADVANCE(913);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('B' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('b' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 42:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'A' ||
          lookahead == 'a') ADVANCE(484);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('B' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('b' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 43:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'A' ||
          lookahead == 'a') ADVANCE(126);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('B' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('b' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 44:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'A' ||
          lookahead == 'a') ADVANCE(152);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('B' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('b' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 45:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'A' ||
          lookahead == 'a') ADVANCE(974);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('B' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('b' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 46:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'A' ||
          lookahead == 'a') ADVANCE(606);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('B' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('b' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 47:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'A' ||
          lookahead == 'a') ADVANCE(10);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('B' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('b' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 48:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'A' ||
          lookahead == 'a') ADVANCE(615);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('B' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('b' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 49:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'A' ||
          lookahead == 'a') ADVANCE(148);
      if (lookahead == 'Y' ||
          lookahead == 'y') ADVANCE(960);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('B' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('b' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 50:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'A' ||
          lookahead == 'a') ADVANCE(148);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('B' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('b' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 51:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'A' ||
          lookahead == 'a') ADVANCE(942);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('B' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('b' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 52:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'A' ||
          lookahead == 'a') ADVANCE(585);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('B' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('b' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 53:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'A' ||
          lookahead == 'a') ADVANCE(492);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('B' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('b' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 54:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'A' ||
          lookahead == 'a') ADVANCE(119);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('B' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('b' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 55:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'A' ||
          lookahead == 'a') ADVANCE(724);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('B' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('b' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 56:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'A' ||
          lookahead == 'a') ADVANCE(764);
      if (lookahead == 'G' ||
          lookahead == 'g') ADVANCE(359);
      if (lookahead == 'L' ||
          lookahead == 'l') ADVANCE(250);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('B' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('b' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 57:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'A' ||
          lookahead == 'a') ADVANCE(162);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('B' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('b' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 58:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'A' ||
          lookahead == 'a') ADVANCE(858);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('B' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('b' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 59:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'A' ||
          lookahead == 'a') ADVANCE(150);
      if (lookahead == 'U' ||
          lookahead == 'u') ADVANCE(224);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('B' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('b' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 60:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'A' ||
          lookahead == 'a') ADVANCE(976);
      if (lookahead == 'E' ||
          lookahead == 'e') ADVANCE(288);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('B' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('b' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 61:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'A' ||
          lookahead == 'a') ADVANCE(976);
      if (lookahead == 'E' ||
          lookahead == 'e') ADVANCE(976);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('B' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('b' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 62:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'A' ||
          lookahead == 'a') ADVANCE(976);
      if (lookahead == 'E' ||
          lookahead == 'e') ADVANCE(284);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('B' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('b' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 63:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'A' ||
          lookahead == 'a') ADVANCE(14);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('B' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('b' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 64:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'A' ||
          lookahead == 'a') ADVANCE(763);
      if (lookahead == 'H' ||
          lookahead == 'h') ADVANCE(298);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('B' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('b' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 65:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'A' ||
          lookahead == 'a') ADVANCE(607);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('B' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('b' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 66:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'A' ||
          lookahead == 'a') ADVANCE(578);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('B' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('b' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 67:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'A' ||
          lookahead == 'a') ADVANCE(485);
      if (lookahead == 'R' ||
          lookahead == 'r') ADVANCE(774);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('B' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('b' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 68:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'A' ||
          lookahead == 'a') ADVANCE(932);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('B' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('b' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 69:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'A' ||
          lookahead == 'a') ADVANCE(770);
      if (lookahead == 'R' ||
          lookahead == 'r') ADVANCE(848);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('B' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('b' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 70:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'A' ||
          lookahead == 'a') ADVANCE(621);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('B' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('b' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 71:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'A' ||
          lookahead == 'a') ADVANCE(589);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('B' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('b' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 72:
      ADVANCE_MAP(
        '!', 1334,
        'A', 625,
        'a', 625,
        'B', 542,
        'b', 542,
        'F', 507,
        'f', 507,
        'G', 785,
        'g', 785,
        'L', 345,
        'l', 345,
        'M', 296,
        'm', 296,
        'R', 254,
        'r', 254,
        'Y', 308,
        'y', 308,
      );
      if (('0' <= lookahead && lookahead <= '9') ||
          ('C' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('c' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 73:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'A' ||
          lookahead == 'a') ADVANCE(486);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('B' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('b' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 74:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'A' ||
          lookahead == 'a') ADVANCE(497);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('B' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('b' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 75:
      ADVANCE_MAP(
        '!', 1334,
        'A', 766,
        'a', 766,
        'E', 259,
        'e', 259,
        'I', 583,
        'i', 583,
        'O', 211,
        'o', 211,
      );
      if (('0' <= lookahead && lookahead <= '9') ||
          ('B' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('b' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 76:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'A' ||
          lookahead == 'a') ADVANCE(496);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('B' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('b' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 77:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'A' ||
          lookahead == 'a') ADVANCE(800);
      if (lookahead == 'O' ||
          lookahead == 'o') ADVANCE(171);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('B' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('b' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 78:
      ADVANCE_MAP(
        '!', 1334,
        'A', 611,
        'a', 611,
        'E', 83,
        'e', 83,
        'H', 470,
        'h', 470,
        'O', 598,
        'o', 598,
        'U', 816,
        'u', 816,
      );
      if (('0' <= lookahead && lookahead <= '9') ||
          ('B' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('b' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 79:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'A' ||
          lookahead == 'a') ADVANCE(611);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('B' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('b' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 80:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'A' ||
          lookahead == 'a') ADVANCE(885);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('B' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('b' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 81:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'A' ||
          lookahead == 'a') ADVANCE(862);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('B' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('b' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 82:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'A' ||
          lookahead == 'a') ADVANCE(490);
      if (lookahead == 'N' ||
          lookahead == 'n') ADVANCE(341);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('B' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('b' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 83:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'A' ||
          lookahead == 'a') ADVANCE(490);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('B' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('b' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 84:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'A' ||
          lookahead == 'a') ADVANCE(570);
      if (lookahead == 'E' ||
          lookahead == 'e') ADVANCE(116);
      if (lookahead == 'L' ||
          lookahead == 'l') ADVANCE(105);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('B' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('b' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 85:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'A' ||
          lookahead == 'a') ADVANCE(548);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('B' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('b' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 86:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'A' ||
          lookahead == 'a') ADVANCE(651);
      if (lookahead == 'G' ||
          lookahead == 'g') ADVANCE(661);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('B' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('b' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 87:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'A' ||
          lookahead == 'a') ADVANCE(149);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('B' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('b' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 88:
      ADVANCE_MAP(
        '!', 1334,
        'A', 215,
        'a', 215,
        'H', 77,
        'h', 77,
        'O', 755,
        'o', 755,
        'R', 481,
        'r', 481,
        'Y', 79,
        'y', 79,
      );
      if (('0' <= lookahead && lookahead <= '9') ||
          ('B' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('b' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 89:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'A' ||
          lookahead == 'a') ADVANCE(636);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('B' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('b' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 90:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'A' ||
          lookahead == 'a') ADVANCE(743);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('B' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('b' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 91:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'A' ||
          lookahead == 'a') ADVANCE(166);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('B' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('b' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 92:
      ADVANCE_MAP(
        '!', 1334,
        'A', 449,
        'a', 449,
        'H', 711,
        'h', 711,
        'O', 513,
        'o', 513,
        'R', 60,
        'r', 60,
      );
      if (('0' <= lookahead && lookahead <= '9') ||
          ('B' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('b' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 93:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'A' ||
          lookahead == 'a') ADVANCE(531);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('B' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('b' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 94:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'A' ||
          lookahead == 'a') ADVANCE(982);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('B' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('b' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 95:
      ADVANCE_MAP(
        '!', 1334,
        'A', 745,
        'a', 745,
        'B', 563,
        'b', 563,
        'O', 806,
        'o', 806,
        'P', 940,
        'p', 940,
        'S', 292,
        's', 292,
        'T', 941,
        't', 941,
        'V', 480,
        'v', 480,
      );
      if (('0' <= lookahead && lookahead <= '9') ||
          ('C' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('c' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 96:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'A' ||
          lookahead == 'a') ADVANCE(588);
      if (lookahead == 'I' ||
          lookahead == 'i') ADVANCE(621);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('B' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('b' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 97:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'A' ||
          lookahead == 'a') ADVANCE(880);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('B' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('b' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 98:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'A' ||
          lookahead == 'a') ADVANCE(642);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('B' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('b' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 99:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'A' ||
          lookahead == 'a') ADVANCE(653);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('B' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('b' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 100:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'A' ||
          lookahead == 'a') ADVANCE(160);
      if (lookahead == 'R' ||
          lookahead == 'r') ADVANCE(909);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('B' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('b' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 101:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'A' ||
          lookahead == 'a') ADVANCE(518);
      if (lookahead == 'D' ||
          lookahead == 'd') ADVANCE(194);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('B' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('b' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 102:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'A' ||
          lookahead == 'a') ADVANCE(853);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('B' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('b' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 103:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'A' ||
          lookahead == 'a') ADVANCE(649);
      if (lookahead == 'C' ||
          lookahead == 'c') ADVANCE(398);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('B' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('b' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 104:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'A' ||
          lookahead == 'a') ADVANCE(643);
      if (lookahead == 'C' ||
          lookahead == 'c') ADVANCE(398);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('B' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('b' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 105:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'A' ||
          lookahead == 'a') ADVANCE(888);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('B' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('b' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 106:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'A' ||
          lookahead == 'a') ADVANCE(791);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('B' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('b' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 107:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'A' ||
          lookahead == 'a') ADVANCE(355);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('B' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('b' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 108:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'A' ||
          lookahead == 'a') ADVANCE(887);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('B' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('b' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 109:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'A' ||
          lookahead == 'a') ADVANCE(792);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('B' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('b' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 110:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'A' ||
          lookahead == 'a') ADVANCE(890);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('B' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('b' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 111:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'A' ||
          lookahead == 'a') ADVANCE(903);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('B' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('b' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 112:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'A' ||
          lookahead == 'a') ADVANCE(550);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('B' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('b' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 113:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'A' ||
          lookahead == 'a') ADVANCE(371);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('B' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('b' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 114:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'A' ||
          lookahead == 'a') ADVANCE(968);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('B' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('b' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 115:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'A' ||
          lookahead == 'a') ADVANCE(852);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('B' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('b' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 116:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'A' ||
          lookahead == 'a') ADVANCE(372);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('B' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('b' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 117:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'A' ||
          lookahead == 'a') ADVANCE(571);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('B' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('b' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 118:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'A' ||
          lookahead == 'a') ADVANCE(574);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('B' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('b' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 119:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'B' ||
          lookahead == 'b') ADVANCE(1394);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 120:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'B' ||
          lookahead == 'b') ADVANCE(977);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 121:
      ADVANCE_MAP(
        '!', 1334,
        'B', 516,
        'b', 516,
        'G', 787,
        'g', 787,
        'L', 445,
        'l', 445,
        'O', 805,
        'o', 805,
        'P', 441,
        'p', 441,
        'R', 267,
        'r', 267,
        'S', 900,
        's', 900,
        'Y', 308,
        'y', 308,
      );
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 122:
      ADVANCE_MAP(
        '!', 1334,
        'B', 549,
        'b', 549,
        'G', 785,
        'g', 785,
        'L', 445,
        'l', 445,
        'O', 803,
        'o', 803,
        'R', 253,
        'r', 253,
        'S', 900,
        's', 900,
      );
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 123:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'B' ||
          lookahead == 'b') ADVANCE(251);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 124:
      ADVANCE_MAP(
        '!', 1334,
        'B', 559,
        'b', 559,
        'G', 785,
        'g', 785,
        'L', 476,
        'l', 476,
        'O', 803,
        'o', 803,
        'V', 478,
        'v', 478,
        'Y', 308,
        'y', 308,
      );
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 125:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'B' ||
          lookahead == 'b') ADVANCE(775);
      if (lookahead == 'L' ||
          lookahead == 'l') ADVANCE(445);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 126:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'B' ||
          lookahead == 'b') ADVANCE(775);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 127:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'B' ||
          lookahead == 'b') ADVANCE(488);
      if (lookahead == 'R' ||
          lookahead == 'r') ADVANCE(254);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 128:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'B' ||
          lookahead == 'b') ADVANCE(255);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 129:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'B' ||
          lookahead == 'b') ADVANCE(707);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 130:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'B' ||
          lookahead == 'b') ADVANCE(321);
      if (lookahead == 'D' ||
          lookahead == 'd') ADVANCE(1394);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 131:
      ADVANCE_MAP(
        '!', 1334,
        'B', 544,
        'b', 544,
        'I', 916,
        'i', 916,
        'R', 320,
        'r', 320,
        'Y', 305,
        'y', 305,
      );
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 132:
      ADVANCE_MAP(
        '!', 1334,
        'B', 517,
        'b', 517,
        'G', 785,
        'g', 785,
        'O', 803,
        'o', 803,
        'P', 937,
        'p', 937,
        'R', 303,
        'r', 303,
        'V', 478,
        'v', 478,
        'Y', 308,
        'y', 308,
      );
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 133:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'B' ||
          lookahead == 'b') ADVANCE(554);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 134:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'B' ||
          lookahead == 'b') ADVANCE(811);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 135:
      ADVANCE_MAP(
        '!', 1334,
        'B', 543,
        'b', 543,
        'C', 915,
        'c', 915,
        'G', 787,
        'g', 787,
        'I', 652,
        'i', 652,
        'N', 664,
        'n', 664,
        'O', 803,
        'o', 803,
        'R', 253,
        'r', 253,
        'S', 900,
        's', 900,
        'T', 68,
        't', 68,
      );
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 136:
      ADVANCE_MAP(
        '!', 1334,
        'B', 543,
        'b', 543,
        'G', 785,
        'g', 785,
        'O', 803,
        'o', 803,
        'Y', 308,
        'y', 308,
      );
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 137:
      ADVANCE_MAP(
        '!', 1334,
        'B', 543,
        'b', 543,
        'G', 785,
        'g', 785,
        'R', 301,
        'r', 301,
        'Y', 308,
        'y', 308,
      );
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 138:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'B' ||
          lookahead == 'b') ADVANCE(543);
      if (lookahead == 'O' ||
          lookahead == 'o') ADVANCE(803);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 139:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'B' ||
          lookahead == 'b') ADVANCE(543);
      if (lookahead == 'P' ||
          lookahead == 'p') ADVANCE(937);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 140:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'B' ||
          lookahead == 'b') ADVANCE(543);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 141:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'B' ||
          lookahead == 'b') ADVANCE(810);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 142:
      ADVANCE_MAP(
        '!', 1334,
        'B', 553,
        'b', 553,
        'G', 785,
        'g', 785,
        'R', 254,
        'r', 254,
        'V', 475,
        'v', 475,
        'Y', 311,
        'y', 311,
      );
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 143:
      ADVANCE_MAP(
        '!', 1334,
        'B', 563,
        'b', 563,
        'C', 721,
        'c', 721,
        'G', 723,
        'g', 723,
        'P', 447,
        'p', 447,
        'S', 16,
        's', 16,
        'Y', 316,
        'y', 316,
      );
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 144:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'B' ||
          lookahead == 'b') ADVANCE(563);
      if (lookahead == 'G' ||
          lookahead == 'g') ADVANCE(760);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 145:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'B' ||
          lookahead == 'b') ADVANCE(563);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 146:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'C' ||
          lookahead == 'c') ADVANCE(382);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 147:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'C' ||
          lookahead == 'c') ADVANCE(1369);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 148:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'C' ||
          lookahead == 'c') ADVANCE(1381);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 149:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'C' ||
          lookahead == 'c') ADVANCE(375);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 150:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'C' ||
          lookahead == 'c') ADVANCE(5);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 151:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'C' ||
          lookahead == 'c') ADVANCE(140);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 152:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'C' ||
          lookahead == 'c') ADVANCE(7);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 153:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'C' ||
          lookahead == 'c') ADVANCE(12);
      if (lookahead == 'G' ||
          lookahead == 'g') ADVANCE(392);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 154:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'C' ||
          lookahead == 'c') ADVANCE(12);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 155:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'C' ||
          lookahead == 'c') ADVANCE(394);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 156:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'C' ||
          lookahead == 'c') ADVANCE(9);
      if (lookahead == 'N' ||
          lookahead == 'n') ADVANCE(161);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 157:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'C' ||
          lookahead == 'c') ADVANCE(9);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 158:
      ADVANCE_MAP(
        '!', 1334,
        'C', 682,
        'c', 682,
        'O', 572,
        'o', 572,
        'P', 730,
        'p', 730,
        'R', 5,
        'r', 5,
      );
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 159:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'C' ||
          lookahead == 'c') ADVANCE(163);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 160:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'C' ||
          lookahead == 'c') ADVANCE(389);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 161:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'C' ||
          lookahead == 'c') ADVANCE(401);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 162:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'C' ||
          lookahead == 'c') ADVANCE(688);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 163:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'C' ||
          lookahead == 'c') ADVANCE(102);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 164:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'C' ||
          lookahead == 'c') ADVANCE(801);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 165:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'C' ||
          lookahead == 'c') ADVANCE(90);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 166:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'C' ||
          lookahead == 'c') ADVANCE(231);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 167:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'C' ||
          lookahead == 'c') ADVANCE(894);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 168:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'C' ||
          lookahead == 'c') ADVANCE(393);
      if (lookahead == 'S' ||
          lookahead == 's') ADVANCE(387);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 169:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'C' ||
          lookahead == 'c') ADVANCE(398);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 170:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'C' ||
          lookahead == 'c') ADVANCE(165);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 171:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'C' ||
          lookahead == 'c') ADVANCE(693);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 172:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'C' ||
          lookahead == 'c') ADVANCE(399);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 173:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'C' ||
          lookahead == 'c') ADVANCE(99);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 174:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'C' ||
          lookahead == 'c') ADVANCE(701);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 175:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'C' ||
          lookahead == 'c') ADVANCE(404);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 176:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'C' ||
          lookahead == 'c') ADVANCE(328);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 177:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'D' ||
          lookahead == 'd') ADVANCE(1369);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 178:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'D' ||
          lookahead == 'd') ADVANCE(1371);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 179:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'D' ||
          lookahead == 'd') ADVANCE(1381);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 180:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'D' ||
          lookahead == 'd') ADVANCE(131);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 181:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'D' ||
          lookahead == 'd') ADVANCE(142);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 182:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'D' ||
          lookahead == 'd') ADVANCE(1376);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 183:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'D' ||
          lookahead == 'd') ADVANCE(1394);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 184:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'D' ||
          lookahead == 'd') ADVANCE(1387);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 185:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'D' ||
          lookahead == 'd') ADVANCE(974);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 186:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'D' ||
          lookahead == 'd') ADVANCE(410);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 187:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'D' ||
          lookahead == 'd') ADVANCE(362);
      if (lookahead == 'N' ||
          lookahead == 'n') ADVANCE(858);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 188:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'D' ||
          lookahead == 'd') ADVANCE(367);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 189:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'D' ||
          lookahead == 'd') ADVANCE(987);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 190:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'D' ||
          lookahead == 'd') ADVANCE(269);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 191:
      ADVANCE_MAP(
        '!', 1334,
        'D', 263,
        'd', 263,
        'M', 659,
        'm', 659,
        'R', 641,
        'r', 641,
        'S', 403,
        's', 403,
      );
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 192:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'D' ||
          lookahead == 'd') ADVANCE(214);
      if (lookahead == 'L' ||
          lookahead == 'l') ADVANCE(600);
      if (lookahead == 'N' ||
          lookahead == 'n') ADVANCE(205);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 193:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'D' ||
          lookahead == 'd') ADVANCE(262);
      if (lookahead == 'W' ||
          lookahead == 'w') ADVANCE(616);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 194:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'D' ||
          lookahead == 'd') ADVANCE(434);
      if (lookahead == 'F' ||
          lookahead == 'f') ADVANCE(519);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 195:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'D' ||
          lookahead == 'd') ADVANCE(472);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 196:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'D' ||
          lookahead == 'd') ADVANCE(435);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 197:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'D' ||
          lookahead == 'd') ADVANCE(527);
      if (lookahead == 'I' ||
          lookahead == 'i') ADVANCE(948);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 198:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'D' ||
          lookahead == 'd') ADVANCE(656);
      if (lookahead == 'N' ||
          lookahead == 'n') ADVANCE(877);
      if (lookahead == 'S' ||
          lookahead == 's') ADVANCE(905);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 199:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'D' ||
          lookahead == 'd') ADVANCE(437);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 200:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'D' ||
          lookahead == 'd') ADVANCE(416);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 201:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'D' ||
          lookahead == 'd') ADVANCE(444);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 202:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'D' ||
          lookahead == 'd') ADVANCE(261);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 203:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'D' ||
          lookahead == 'd') ADVANCE(300);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 204:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'D' ||
          lookahead == 'd') ADVANCE(1375);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 205:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'D' ||
          lookahead == 'd') ADVANCE(992);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 206:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'D' ||
          lookahead == 'd') ADVANCE(454);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 207:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'D' ||
          lookahead == 'd') ADVANCE(196);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 208:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'D' ||
          lookahead == 'd') ADVANCE(1377);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 209:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'D' ||
          lookahead == 'd') ADVANCE(199);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 210:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'D' ||
          lookahead == 'd') ADVANCE(287);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 211:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'D' ||
          lookahead == 'd') ADVANCE(369);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 212:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'D' ||
          lookahead == 'd') ADVANCE(290);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 213:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'D' ||
          lookahead == 'd') ADVANCE(85);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 214:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'D' ||
          lookahead == 'd') ADVANCE(552);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 215:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'D' ||
          lookahead == 'd') ADVANCE(330);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 216:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'D' ||
          lookahead == 'd') ADVANCE(324);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 217:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'D' ||
          lookahead == 'd') ADVANCE(991);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 218:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'E' ||
          lookahead == 'e') ADVANCE(1369);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 219:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'E' ||
          lookahead == 'e') ADVANCE(1371);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 220:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'E' ||
          lookahead == 'e') ADVANCE(101);
      if (lookahead == 'O' ||
          lookahead == 'o') ADVANCE(983);
      if (lookahead == 'U' ||
          lookahead == 'u') ADVANCE(826);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 221:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'E' ||
          lookahead == 'e') ADVANCE(1381);
      if (lookahead == 'I' ||
          lookahead == 'i') ADVANCE(832);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 222:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'E' ||
          lookahead == 'e') ADVANCE(1381);
      if (lookahead == 'I' ||
          lookahead == 'i') ADVANCE(834);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 223:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'E' ||
          lookahead == 'e') ADVANCE(1381);
      if (lookahead == 'I' ||
          lookahead == 'i') ADVANCE(631);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 224:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'E' ||
          lookahead == 'e') ADVANCE(1381);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 225:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'E' ||
          lookahead == 'e') ADVANCE(1380);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 226:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'E' ||
          lookahead == 'e') ADVANCE(772);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 227:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'E' ||
          lookahead == 'e') ADVANCE(265);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 228:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'E' ||
          lookahead == 'e') ADVANCE(1379);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 229:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'E' ||
          lookahead == 'e') ADVANCE(1374);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 230:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'E' ||
          lookahead == 'e') ADVANCE(130);
      if (lookahead == 'O' ||
          lookahead == 'o') ADVANCE(843);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 231:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'E' ||
          lookahead == 'e') ADVANCE(1394);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 232:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'E' ||
          lookahead == 'e') ADVANCE(1392);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 233:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'E' ||
          lookahead == 'e') ADVANCE(1388);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 234:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'E' ||
          lookahead == 'e') ADVANCE(344);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 235:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'E' ||
          lookahead == 'e') ADVANCE(1386);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 236:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'E' ||
          lookahead == 'e') ADVANCE(144);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 237:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'E' ||
          lookahead == 'e') ADVANCE(1391);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 238:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'E' ||
          lookahead == 'e') ADVANCE(1390);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 239:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'E' ||
          lookahead == 'e') ADVANCE(499);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 240:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'E' ||
          lookahead == 'e') ADVANCE(177);
      if (lookahead == 'O' ||
          lookahead == 'o') ADVANCE(823);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 241:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'E' ||
          lookahead == 'e') ADVANCE(907);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 242:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'E' ||
          lookahead == 'e') ADVANCE(42);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 243:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'E' ||
          lookahead == 'e') ADVANCE(857);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 244:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'E' ||
          lookahead == 'e') ADVANCE(725);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 245:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'E' ||
          lookahead == 'e') ADVANCE(974);
      if (lookahead == 'I' ||
          lookahead == 'i') ADVANCE(587);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 246:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'E' ||
          lookahead == 'e') ADVANCE(974);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 247:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'E' ||
          lookahead == 'e') ADVANCE(762);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 248:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'E' ||
          lookahead == 'e') ADVANCE(606);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 249:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'E' ||
          lookahead == 'e') ADVANCE(140);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 250:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'E' ||
          lookahead == 'e') ADVANCE(167);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 251:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'E' ||
          lookahead == 'e') ADVANCE(751);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 252:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'E' ||
          lookahead == 'e') ADVANCE(87);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 253:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'E' ||
          lookahead == 'e') ADVANCE(179);
      if (lookahead == 'O' ||
          lookahead == 'o') ADVANCE(984);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 254:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'E' ||
          lookahead == 'e') ADVANCE(179);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 255:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'E' ||
          lookahead == 'e') ADVANCE(769);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 256:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'E' ||
          lookahead == 'e') ADVANCE(367);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 257:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'E' ||
          lookahead == 'e') ADVANCE(207);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 258:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'E' ||
          lookahead == 'e') ADVANCE(858);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 259:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'E' ||
          lookahead == 'e') ADVANCE(727);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 260:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'E' ||
          lookahead == 'e') ADVANCE(358);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 261:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'E' ||
          lookahead == 'e') ADVANCE(958);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 262:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'E' ||
          lookahead == 'e') ADVANCE(188);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 263:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'E' ||
          lookahead == 'e') ADVANCE(864);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 264:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'E' ||
          lookahead == 'e') ADVANCE(985);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 265:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'E' ||
          lookahead == 'e') ADVANCE(607);
      if (lookahead == 'Y' ||
          lookahead == 'y') ADVANCE(1380);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 266:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'E' ||
          lookahead == 'e') ADVANCE(607);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 267:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'E' ||
          lookahead == 'e') ADVANCE(182);
      if (lookahead == 'O' ||
          lookahead == 'o') ADVANCE(984);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 268:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'E' ||
          lookahead == 'e') ADVANCE(621);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 269:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'E' ||
          lookahead == 'e') ADVANCE(752);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 270:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'E' ||
          lookahead == 'e') ADVANCE(878);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 271:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'E' ||
          lookahead == 'e') ADVANCE(183);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 272:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'E' ||
          lookahead == 'e') ADVANCE(628);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 273:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'E' ||
          lookahead == 'e') ADVANCE(213);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 274:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'E' ||
          lookahead == 'e') ADVANCE(767);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 275:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'E' ||
          lookahead == 'e') ADVANCE(895);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 276:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'E' ||
          lookahead == 'e') ADVANCE(497);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 277:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'E' ||
          lookahead == 'e') ADVANCE(630);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 278:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'E' ||
          lookahead == 'e') ADVANCE(487);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 279:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'E' ||
          lookahead == 'e') ADVANCE(754);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 280:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'E' ||
          lookahead == 'e') ADVANCE(266);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 281:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'E' ||
          lookahead == 'e') ADVANCE(859);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 282:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'E' ||
          lookahead == 'e') ADVANCE(276);
      if (lookahead == 'O' ||
          lookahead == 'o') ADVANCE(797);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 283:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'E' ||
          lookahead == 'e') ADVANCE(935);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 284:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'E' ||
          lookahead == 'e') ADVANCE(611);
      if (lookahead == 'Y' ||
          lookahead == 'y') ADVANCE(1394);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 285:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'E' ||
          lookahead == 'e') ADVANCE(611);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 286:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'E' ||
          lookahead == 'e') ADVANCE(862);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 287:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'E' ||
          lookahead == 'e') ADVANCE(814);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 288:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'E' ||
          lookahead == 'e') ADVANCE(612);
      if (lookahead == 'Y' ||
          lookahead == 'y') ADVANCE(1394);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 289:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'E' ||
          lookahead == 'e') ADVANCE(756);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 290:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'E' ||
          lookahead == 'e') ADVANCE(757);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 291:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'E' ||
          lookahead == 'e') ADVANCE(639);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 292:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'E' ||
          lookahead == 'e') ADVANCE(116);
      if (lookahead == 'L' ||
          lookahead == 'l') ADVANCE(111);
      if (lookahead == 'P' ||
          lookahead == 'p') ADVANCE(802);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 293:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'E' ||
          lookahead == 'e') ADVANCE(66);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 294:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'E' ||
          lookahead == 'e') ADVANCE(285);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 295:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'E' ||
          lookahead == 'e') ADVANCE(618);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 296:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'E' ||
          lookahead == 'e') ADVANCE(195);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 297:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'E' ||
          lookahead == 'e') ADVANCE(512);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 298:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'E' ||
          lookahead == 'e') ADVANCE(58);
      if (lookahead == 'I' ||
          lookahead == 'i') ADVANCE(881);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 299:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'E' ||
          lookahead == 'e') ADVANCE(134);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 300:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'E' ||
          lookahead == 'e') ADVANCE(640);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 301:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'E' ||
          lookahead == 'e') ADVANCE(204);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 302:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'E' ||
          lookahead == 'e') ADVANCE(633);
      if (lookahead == 'L' ||
          lookahead == 'l') ADVANCE(950);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 303:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'E' ||
          lookahead == 'e') ADVANCE(208);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 304:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'E' ||
          lookahead == 'e') ADVANCE(897);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 305:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'E' ||
          lookahead == 'e') ADVANCE(547);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 306:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'E' ||
          lookahead == 'e') ADVANCE(81);
      if (lookahead == 'I' ||
          lookahead == 'i') ADVANCE(889);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 307:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'E' ||
          lookahead == 'e') ADVANCE(781);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 308:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'E' ||
          lookahead == 'e') ADVANCE(555);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 309:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'E' ||
          lookahead == 'e') ADVANCE(562);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 310:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'E' ||
          lookahead == 'e') ADVANCE(560);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 311:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'E' ||
          lookahead == 'e') ADVANCE(564);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 312:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'E' ||
          lookahead == 'e') ADVANCE(511);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 313:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'E' ||
          lookahead == 'e') ADVANCE(565);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 314:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'E' ||
          lookahead == 'e') ADVANCE(566);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 315:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'E' ||
          lookahead == 'e') ADVANCE(528);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 316:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'E' ||
          lookahead == 'e') ADVANCE(567);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 317:
      ADVANCE_MAP(
        '!', 1334,
        'E', 465,
        'e', 465,
        'I', 842,
        'i', 842,
        'L', 33,
        'l', 33,
        'R', 704,
        'r', 704,
        'U', 782,
        'u', 782,
      );
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 318:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'E' ||
          lookahead == 'e') ADVANCE(370);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 319:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'E' ||
          lookahead == 'e') ADVANCE(978);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 320:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'E' ||
          lookahead == 'e') ADVANCE(209);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 321:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'E' ||
          lookahead == 'e') ADVANCE(170);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 322:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'E' ||
          lookahead == 'e') ADVANCE(503);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 323:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'E' ||
          lookahead == 'e') ADVANCE(654);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 324:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'E' ||
          lookahead == 'e') ADVANCE(644);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 325:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'E' ||
          lookahead == 'e') ADVANCE(855);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 326:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'E' ||
          lookahead == 'e') ADVANCE(372);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 327:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'E' ||
          lookahead == 'e') ADVANCE(141);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 328:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'E' ||
          lookahead == 'e') ADVANCE(145);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 329:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'E' ||
          lookahead == 'e') ADVANCE(988);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 330:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'E' ||
          lookahead == 'e') ADVANCE(902);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 331:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'E' ||
          lookahead == 'e') ADVANCE(571);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 332:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'E' ||
          lookahead == 'e') ADVANCE(331);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 333:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'E' ||
          lookahead == 'e') ADVANCE(856);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 334:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'E' ||
          lookahead == 'e') ADVANCE(971);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 335:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'F' ||
          lookahead == 'f') ADVANCE(1394);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 336:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'F' ||
          lookahead == 'f') ADVANCE(335);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 337:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'F' ||
          lookahead == 'f') ADVANCE(339);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 338:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'F' ||
          lookahead == 'f') ADVANCE(519);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 339:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'F' ||
          lookahead == 'f') ADVANCE(691);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 340:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'F' ||
          lookahead == 'f') ADVANCE(525);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 341:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'F' ||
          lookahead == 'f') ADVANCE(534);
      if (lookahead == 'S' ||
          lookahead == 's') ADVANCE(451);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 342:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'G' ||
          lookahead == 'g') ADVANCE(1381);
      if (lookahead == 'R' ||
          lookahead == 'r') ADVANCE(325);
      if (lookahead == 'S' ||
          lookahead == 's') ADVANCE(828);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 343:
      ADVANCE_MAP(
        '!', 1334,
        'G', 391,
        'g', 391,
        'L', 49,
        'l', 49,
        'M', 256,
        'm', 256,
        'N', 266,
        'n', 266,
      );
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 344:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'G' ||
          lookahead == 'g') ADVANCE(695);
      if (lookahead == 'T' ||
          lookahead == 't') ADVANCE(941);
      if (lookahead == 'V' ||
          lookahead == 'v') ADVANCE(480);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 345:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'G' ||
          lookahead == 'g') ADVANCE(140);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 346:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'G' ||
          lookahead == 'g') ADVANCE(657);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 347:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'G' ||
          lookahead == 'g') ADVANCE(218);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 348:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'G' ||
          lookahead == 'g') ADVANCE(934);
      if (lookahead == 'L' ||
          lookahead == 'l') ADVANCE(55);
      if (lookahead == 'N' ||
          lookahead == 'n') ADVANCE(893);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 349:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'G' ||
          lookahead == 'g') ADVANCE(659);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 350:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'G' ||
          lookahead == 'g') ADVANCE(256);
      if (lookahead == 'L' ||
          lookahead == 'l') ADVANCE(596);
      if (lookahead == 'N' ||
          lookahead == 'n') ADVANCE(181);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 351:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'G' ||
          lookahead == 'g') ADVANCE(272);
      if (lookahead == 'R' ||
          lookahead == 'r') ADVANCE(703);
      if (lookahead == 'U' ||
          lookahead == 'u') ADVANCE(947);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 352:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'G' ||
          lookahead == 'g') ADVANCE(224);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 353:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'G' ||
          lookahead == 'g') ADVANCE(229);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 354:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'G' ||
          lookahead == 'g') ADVANCE(291);
      if (lookahead == 'R' ||
          lookahead == 'r') ADVANCE(714);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 355:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'G' ||
          lookahead == 'g') ADVANCE(291);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 356:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'G' ||
          lookahead == 'g') ADVANCE(231);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 357:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'G' ||
          lookahead == 'g') ADVANCE(238);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 358:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'G' ||
          lookahead == 'g') ADVANCE(760);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 359:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'G' ||
          lookahead == 'g') ADVANCE(732);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 360:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'G' ||
          lookahead == 'g') ADVANCE(58);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 361:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'G' ||
          lookahead == 'g') ADVANCE(402);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 362:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'G' ||
          lookahead == 'g') ADVANCE(776);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 363:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'G' ||
          lookahead == 'g') ADVANCE(390);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 364:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'G' ||
          lookahead == 'g') ADVANCE(785);
      if (lookahead == 'O' ||
          lookahead == 'o') ADVANCE(803);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 365:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'G' ||
          lookahead == 'g') ADVANCE(785);
      if (lookahead == 'S' ||
          lookahead == 's') ADVANCE(407);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 366:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'G' ||
          lookahead == 'g') ADVANCE(785);
      if (lookahead == 'V' ||
          lookahead == 'v') ADVANCE(478);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 367:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'G' ||
          lookahead == 'g') ADVANCE(785);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 368:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'G' ||
          lookahead == 'g') ADVANCE(396);
      if (lookahead == 'M' ||
          lookahead == 'm') ADVANCE(233);
      if (lookahead == 'N' ||
          lookahead == 'n') ADVANCE(285);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 369:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'G' ||
          lookahead == 'g') ADVANCE(287);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 370:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'G' ||
          lookahead == 'g') ADVANCE(789);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 371:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'G' ||
          lookahead == 'g') ADVANCE(796);
      if (lookahead == 'S' ||
          lookahead == 's') ADVANCE(408);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 372:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'G' ||
          lookahead == 'g') ADVANCE(796);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 373:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'G' ||
          lookahead == 'g') ADVANCE(372);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 374:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'G' ||
          lookahead == 'g') ADVANCE(409);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 375:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'H' ||
          lookahead == 'h') ADVANCE(1371);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 376:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'H' ||
          lookahead == 'h') ADVANCE(364);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 377:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'H' ||
          lookahead == 'h') ADVANCE(136);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 378:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'H' ||
          lookahead == 'h') ADVANCE(125);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 379:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'H' ||
          lookahead == 'h') ADVANCE(366);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 380:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'H' ||
          lookahead == 'h') ADVANCE(546);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 381:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'H' ||
          lookahead == 'h') ADVANCE(1394);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 382:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'H' ||
          lookahead == 'h') ADVANCE(819);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 383:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'H' ||
          lookahead == 'h') ADVANCE(960);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 384:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'H' ||
          lookahead == 'h') ADVANCE(953);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 385:
      ADVANCE_MAP(
        '!', 1334,
        'H', 710,
        'h', 710,
        'O', 502,
        'o', 502,
        'R', 245,
        'r', 245,
        'U', 646,
        'u', 646,
      );
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 386:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'H' ||
          lookahead == 'h') ADVANCE(733);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 387:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'H' ||
          lookahead == 'h') ADVANCE(367);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 388:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'H' ||
          lookahead == 'h') ADVANCE(47);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 389:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'H' ||
          lookahead == 'h') ADVANCE(735);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 390:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'H' ||
          lookahead == 'h') ADVANCE(864);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 391:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'H' ||
          lookahead == 'h') ADVANCE(860);
      if (lookahead == 'Y' ||
          lookahead == 'y') ADVANCE(309);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 392:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'H' ||
          lookahead == 'h') ADVANCE(861);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 393:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'H' ||
          lookahead == 'h') ADVANCE(717);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 394:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'H' ||
          lookahead == 'h') ADVANCE(674);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 395:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'H' ||
          lookahead == 'h') ADVANCE(63);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 396:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'H' ||
          lookahead == 'h') ADVANCE(863);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 397:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'H' ||
          lookahead == 'h') ADVANCE(468);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 398:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'H' ||
          lookahead == 'h') ADVANCE(438);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 399:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'H' ||
          lookahead == 'h') ADVANCE(457);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 400:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'H' ||
          lookahead == 'h') ADVANCE(431);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 401:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'H' ||
          lookahead == 'h') ADVANCE(273);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 402:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'H' ||
          lookahead == 'h') ADVANCE(891);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 403:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'H' ||
          lookahead == 'h') ADVANCE(591);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 404:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'H' ||
          lookahead == 'h') ADVANCE(845);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 405:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'H' ||
          lookahead == 'h') ADVANCE(306);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 406:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'H' ||
          lookahead == 'h') ADVANCE(473);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 407:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'H' ||
          lookahead == 'h') ADVANCE(312);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 408:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'H' ||
          lookahead == 'h') ADVANCE(315);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 409:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'H' ||
          lookahead == 'h') ADVANCE(902);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 410:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'I' ||
          lookahead == 'i') ADVANCE(346);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 411:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'I' ||
          lookahead == 'i') ADVANCE(945);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 412:
      ADVANCE_MAP(
        '!', 1334,
        'I', 153,
        'i', 153,
        'O', 609,
        'o', 609,
        'R', 257,
        'r', 257,
        'Y', 305,
        'y', 305,
      );
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 413:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'I' ||
          lookahead == 'i') ADVANCE(1381);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 414:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'I' ||
          lookahead == 'i') ADVANCE(168);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 415:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'I' ||
          lookahead == 'i') ADVANCE(1394);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 416:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'I' ||
          lookahead == 'i') ADVANCE(86);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 417:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'I' ||
          lookahead == 'i') ADVANCE(605);
      if (lookahead == 'U' ||
          lookahead == 'u') ADVANCE(750);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 418:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'I' ||
          lookahead == 'i') ADVANCE(662);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 419:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'I' ||
          lookahead == 'i') ADVANCE(613);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 420:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'I' ||
          lookahead == 'i') ADVANCE(944);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 421:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'I' ||
          lookahead == 'i') ADVANCE(584);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 422:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'I' ||
          lookahead == 'i') ADVANCE(20);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 423:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'I' ||
          lookahead == 'i') ADVANCE(148);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 424:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'I' ||
          lookahead == 'i') ADVANCE(822);
      if (lookahead == 'R' ||
          lookahead == 'r') ADVANCE(6);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 425:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'I' ||
          lookahead == 'i') ADVANCE(173);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 426:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'I' ||
          lookahead == 'i') ADVANCE(818);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 427:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'I' ||
          lookahead == 'i') ADVANCE(737);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 428:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'I' ||
          lookahead == 'i') ADVANCE(361);
      if (lookahead == 'O' ||
          lookahead == 'o') ADVANCE(961);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 429:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'I' ||
          lookahead == 'i') ADVANCE(601);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 430:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'I' ||
          lookahead == 'i') ADVANCE(151);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 431:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'I' ||
          lookahead == 'i') ADVANCE(726);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 432:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'I' ||
          lookahead == 'i') ADVANCE(176);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 433:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'I' ||
          lookahead == 'i') ADVANCE(485);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 434:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'I' ||
          lookahead == 'i') ADVANCE(829);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 435:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'I' ||
          lookahead == 'i') ADVANCE(847);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 436:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'I' ||
          lookahead == 'i') ADVANCE(831);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 437:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'I' ||
          lookahead == 'i') ADVANCE(832);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 438:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'I' ||
          lookahead == 'i') ADVANCE(183);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 439:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'I' ||
          lookahead == 'i') ADVANCE(833);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 440:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'I' ||
          lookahead == 'i') ADVANCE(875);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 441:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'I' ||
          lookahead == 'i') ADVANCE(614);
      if (lookahead == 'U' ||
          lookahead == 'u') ADVANCE(807);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 442:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'I' ||
          lookahead == 'i') ADVANCE(614);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 443:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'I' ||
          lookahead == 'i') ADVANCE(835);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 444:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'I' ||
          lookahead == 'i') ADVANCE(920);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 445:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'I' ||
          lookahead == 'i') ADVANCE(520);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 446:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'I' ||
          lookahead == 'i') ADVANCE(493);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 447:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'I' ||
          lookahead == 'i') ADVANCE(624);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 448:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'I' ||
          lookahead == 'i') ADVANCE(611);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 449:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'I' ||
          lookahead == 'i') ADVANCE(629);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 450:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'I' ||
          lookahead == 'i') ADVANCE(34);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 451:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'I' ||
          lookahead == 'i') ADVANCE(509);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 452:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'I' ||
          lookahead == 'i') ADVANCE(622);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 453:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'I' ||
          lookahead == 'i') ADVANCE(277);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 454:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'I' ||
          lookahead == 'i') ADVANCE(349);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 455:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'I' ||
          lookahead == 'i') ADVANCE(582);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 456:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'I' ||
          lookahead == 'i') ADVANCE(352);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 457:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'I' ||
          lookahead == 'i') ADVANCE(337);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 458:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'I' ||
          lookahead == 'i') ADVANCE(154);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 459:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'I' ||
          lookahead == 'i') ADVANCE(155);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 460:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'I' ||
          lookahead == 'i') ADVANCE(840);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 461:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'I' ||
          lookahead == 'i') ADVANCE(515);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 462:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'I' ||
          lookahead == 'i') ADVANCE(157);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 463:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'I' ||
          lookahead == 'i') ADVANCE(841);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 464:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'I' ||
          lookahead == 'i') ADVANCE(689);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 465:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'I' ||
          lookahead == 'i') ADVANCE(356);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 466:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'I' ||
          lookahead == 'i') ADVANCE(635);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 467:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'I' ||
          lookahead == 'i') ADVANCE(690);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 468:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'I' ||
          lookahead == 'i') ADVANCE(881);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 469:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'I' ||
          lookahead == 'i') ADVANCE(586);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 470:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'I' ||
          lookahead == 'i') ADVANCE(839);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 471:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'I' ||
          lookahead == 'i') ADVANCE(951);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 472:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'I' ||
          lookahead == 'i') ADVANCE(809);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 473:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'I' ||
          lookahead == 'i') ADVANCE(887);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 474:
      ADVANCE_MAP(
        '!', 1334,
        'I', 793,
        'i', 793,
        'L', 718,
        'l', 718,
        'O', 815,
        'o', 815,
        'U', 175,
        'u', 175,
      );
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 475:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'I' ||
          lookahead == 'i') ADVANCE(712);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 476:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'I' ||
          lookahead == 'i') ADVANCE(363);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 477:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'I' ||
          lookahead == 'i') ADVANCE(747);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 478:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'I' ||
          lookahead == 'i') ADVANCE(713);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 479:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'I' ||
          lookahead == 'i') ADVANCE(715);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 480:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'I' ||
          lookahead == 'i') ADVANCE(716);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 481:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'I' ||
          lookahead == 'i') ADVANCE(602);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 482:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'I' ||
          lookahead == 'i') ADVANCE(374);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 483:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'J' ||
          lookahead == 'j') ADVANCE(722);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 484:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'L' ||
          lookahead == 'l') ADVANCE(1369);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 485:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'L' ||
          lookahead == 'l') ADVANCE(1381);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 486:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'L' ||
          lookahead == 'l') ADVANCE(139);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 487:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'L' ||
          lookahead == 'l') ADVANCE(124);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 488:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'L' ||
          lookahead == 'l') ADVANCE(59);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 489:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'L' ||
          lookahead == 'l') ADVANCE(197);
      if (lookahead == 'R' ||
          lookahead == 'r') ADVANCE(103);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 490:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'L' ||
          lookahead == 'l') ADVANCE(1394);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 491:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'L' ||
          lookahead == 'l') ADVANCE(914);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 492:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'L' ||
          lookahead == 'l') ADVANCE(177);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 493:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'L' ||
          lookahead == 'l') ADVANCE(5);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 494:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'L' ||
          lookahead == 'l') ADVANCE(178);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 495:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'L' ||
          lookahead == 'l') ADVANCE(658);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 496:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'L' ||
          lookahead == 'l') ADVANCE(960);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 497:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'L' ||
          lookahead == 'l') ADVANCE(140);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 498:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'L' ||
          lookahead == 'l') ADVANCE(190);
      if (lookahead == 'R' ||
          lookahead == 'r') ADVANCE(872);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 499:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'L' ||
          lookahead == 'l') ADVANCE(495);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 500:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'L' ||
          lookahead == 'l') ADVANCE(411);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 501:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'L' ||
          lookahead == 'l') ADVANCE(218);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 502:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'L' ||
          lookahead == 'l') ADVANCE(179);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 503:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'L' ||
          lookahead == 'l') ADVANCE(367);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 504:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'L' ||
          lookahead == 'l') ADVANCE(420);
      if (lookahead == 'Y' ||
          lookahead == 'y') ADVANCE(850);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 505:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'L' ||
          lookahead == 'l') ADVANCE(243);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 506:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'L' ||
          lookahead == 'l') ADVANCE(975);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 507:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'L' ||
          lookahead == 'l') ADVANCE(908);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 508:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'L' ||
          lookahead == 'l') ADVANCE(979);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 509:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'L' ||
          lookahead == 'l') ADVANCE(9);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 510:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'L' ||
          lookahead == 'l') ADVANCE(269);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 511:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'L' ||
          lookahead == 'l') ADVANCE(485);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 512:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'L' ||
          lookahead == 'l') ADVANCE(514);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 513:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'L' ||
          lookahead == 'l') ADVANCE(184);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 514:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'L' ||
          lookahead == 'l') ADVANCE(663);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 515:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'L' ||
          lookahead == 'l') ADVANCE(568);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 516:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'L' ||
          lookahead == 'l') ADVANCE(911);
      if (lookahead == 'R' ||
          lookahead == 'r') ADVANCE(458);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 517:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'L' ||
          lookahead == 'l') ADVANCE(911);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 518:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'L' ||
          lookahead == 'l') ADVANCE(506);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 519:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'L' ||
          lookahead == 'l') ADVANCE(427);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 520:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'L' ||
          lookahead == 'l') ADVANCE(50);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 521:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'L' ||
          lookahead == 'l') ADVANCE(667);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 522:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'L' ||
          lookahead == 'l') ADVANCE(670);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 523:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'L' ||
          lookahead == 'l') ADVANCE(672);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 524:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'L' ||
          lookahead == 'l') ADVANCE(922);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 525:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'L' ||
          lookahead == 'l') ADVANCE(669);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 526:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'L' ||
          lookahead == 'l') ADVANCE(224);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 527:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'L' ||
          lookahead == 'l') ADVANCE(91);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 528:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'L' ||
          lookahead == 'l') ADVANCE(490);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 529:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'L' ||
          lookahead == 'l') ADVANCE(677);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 530:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'L' ||
          lookahead == 'l') ADVANCE(679);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 531:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'L' ||
          lookahead == 'l') ADVANCE(423);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 532:
      ADVANCE_MAP(
        '!', 1334,
        'L', 432,
        'l', 432,
        'N', 886,
        'n', 886,
        'Q', 923,
        'q', 923,
        'Z', 936,
        'z', 936,
      );
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Y') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'y')) ADVANCE(996);
      END_STATE();
    case 533:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'L' ||
          lookahead == 'l') ADVANCE(319);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 534:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'L' ||
          lookahead == 'l') ADVANCE(681);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 535:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'L' ||
          lookahead == 'l') ADVANCE(281);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 536:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'L' ||
          lookahead == 'l') ADVANCE(258);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 537:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'L' ||
          lookahead == 'l') ADVANCE(234);
      if (lookahead == 'P' ||
          lookahead == 'p') ADVANCE(94);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 538:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'L' ||
          lookahead == 'l') ADVANCE(231);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 539:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'L' ||
          lookahead == 'l') ADVANCE(286);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 540:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'L' ||
          lookahead == 'l') ADVANCE(304);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 541:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'L' ||
          lookahead == 'l') ADVANCE(128);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 542:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'L' ||
          lookahead == 'l') ADVANCE(925);
      if (lookahead == 'R' ||
          lookahead == 'r') ADVANCE(428);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 543:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'L' ||
          lookahead == 'l') ADVANCE(925);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 544:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'L' ||
          lookahead == 'l') ADVANCE(924);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 545:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'L' ||
          lookahead == 'l') ADVANCE(471);
      if (lookahead == 'R' ||
          lookahead == 'r') ADVANCE(104);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 546:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'L' ||
          lookahead == 'l') ADVANCE(445);
      if (lookahead == 'V' ||
          lookahead == 'v') ADVANCE(478);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 547:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'L' ||
          lookahead == 'l') ADVANCE(521);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 548:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'L' ||
          lookahead == 'l') ADVANCE(594);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 549:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'L' ||
          lookahead == 'l') ADVANCE(933);
      if (lookahead == 'R' ||
          lookahead == 'r') ADVANCE(699);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 550:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'L' ||
          lookahead == 'l') ADVANCE(531);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 551:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'L' ||
          lookahead == 'l') ADVANCE(700);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 552:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'L' ||
          lookahead == 'l') ADVANCE(327);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 553:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'L' ||
          lookahead == 'l') ADVANCE(928);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 554:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'L' ||
          lookahead == 'l') ADVANCE(926);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 555:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'L' ||
          lookahead == 'l') ADVANCE(522);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 556:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'L' ||
          lookahead == 'l') ADVANCE(705);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 557:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'L' ||
          lookahead == 'l') ADVANCE(70);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 558:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'L' ||
          lookahead == 'l') ADVANCE(949);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 559:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'L' ||
          lookahead == 'l') ADVANCE(929);
      if (lookahead == 'R' ||
          lookahead == 'r') ADVANCE(699);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 560:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'L' ||
          lookahead == 'l') ADVANCE(510);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 561:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'L' ||
          lookahead == 'l') ADVANCE(203);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 562:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'L' ||
          lookahead == 'l') ADVANCE(551);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 563:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'L' ||
          lookahead == 'l') ADVANCE(930);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 564:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'L' ||
          lookahead == 'l') ADVANCE(523);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 565:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'L' ||
          lookahead == 'l') ADVANCE(556);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 566:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'L' ||
          lookahead == 'l') ADVANCE(530);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 567:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'L' ||
          lookahead == 'l') ADVANCE(529);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 568:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'L' ||
          lookahead == 'l') ADVANCE(370);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 569:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'L' ||
          lookahead == 'l') ADVANCE(108);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 570:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'L' ||
          lookahead == 'l') ADVANCE(600);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 571:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'L' ||
          lookahead == 'l') ADVANCE(145);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 572:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'L' ||
          lookahead == 'l') ADVANCE(988);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 573:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'L' ||
          lookahead == 'l') ADVANCE(216);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 574:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'L' ||
          lookahead == 'l') ADVANCE(971);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 575:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'M' ||
          lookahead == 'm') ADVANCE(123);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 576:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'M' ||
          lookahead == 'm') ADVANCE(1381);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 577:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'M' ||
          lookahead == 'm') ADVANCE(122);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 578:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'M' ||
          lookahead == 'm') ADVANCE(1394);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 579:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'M' ||
          lookahead == 'm') ADVANCE(95);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 580:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'M' ||
          lookahead == 'm') ADVANCE(247);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 581:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'M' ||
          lookahead == 'm') ADVANCE(140);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 582:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'M' ||
          lookahead == 'm') ADVANCE(218);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 583:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'M' ||
          lookahead == 'm') ADVANCE(358);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 584:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'M' ||
          lookahead == 'm') ADVANCE(849);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 585:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'M' ||
          lookahead == 'm') ADVANCE(779);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 586:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'M' ||
          lookahead == 'm') ADVANCE(256);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 587:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'M' ||
          lookahead == 'm') ADVANCE(224);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 588:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'M' ||
          lookahead == 'm') ADVANCE(226);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 589:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'M' ||
          lookahead == 'm') ADVANCE(106);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 590:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'M' ||
          lookahead == 'm') ADVANCE(270);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 591:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'M' ||
          lookahead == 'm') ADVANCE(307);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 592:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'M' ||
          lookahead == 'm') ADVANCE(275);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 593:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'M' ||
          lookahead == 'm') ADVANCE(683);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 594:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'M' ||
          lookahead == 'm') ADVANCE(708);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 595:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'M' ||
          lookahead == 'm') ADVANCE(719);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 596:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'M' ||
          lookahead == 'm') ADVANCE(680);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 597:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'M' ||
          lookahead == 'm') ADVANCE(446);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 598:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'M' ||
          lookahead == 'm') ADVANCE(80);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 599:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'M' ||
          lookahead == 'm') ADVANCE(694);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 600:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'M' ||
          lookahead == 'm') ADVANCE(691);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 601:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'M' ||
          lookahead == 'm') ADVANCE(596);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 602:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'M' ||
          lookahead == 'm') ADVANCE(854);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 603:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'M' ||
          lookahead == 'm') ADVANCE(990);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 604:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'N' ||
          lookahead == 'n') ADVANCE(186);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 605:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'N' ||
          lookahead == 'n') ADVANCE(2);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 606:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'N' ||
          lookahead == 'n') ADVANCE(1369);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 607:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'N' ||
          lookahead == 'n') ADVANCE(1381);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 608:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'N' ||
          lookahead == 'n') ADVANCE(4);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 609:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'N' ||
          lookahead == 'n') ADVANCE(995);
      if (lookahead == 'W' ||
          lookahead == 'w') ADVANCE(607);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 610:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'N' ||
          lookahead == 'n') ADVANCE(364);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 611:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'N' ||
          lookahead == 'n') ADVANCE(1394);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 612:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'N' ||
          lookahead == 'n') ADVANCE(1393);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 613:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'N' ||
          lookahead == 'n') ADVANCE(147);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 614:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'N' ||
          lookahead == 'n') ADVANCE(5);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 615:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'N' ||
          lookahead == 'n') ADVANCE(347);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 616:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'N' ||
          lookahead == 'n') ADVANCE(126);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 617:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'N' ||
          lookahead == 'n') ADVANCE(974);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 618:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'N' ||
          lookahead == 'n') ADVANCE(190);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 619:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'N' ||
          lookahead == 'n') ADVANCE(731);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 620:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'N' ||
          lookahead == 'n') ADVANCE(218);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 621:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'N' ||
          lookahead == 'n') ADVANCE(858);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 622:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'N' ||
          lookahead == 'n') ADVANCE(373);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 623:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'N' ||
          lookahead == 'n') ADVANCE(185);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 624:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'N' ||
          lookahead == 'n') ADVANCE(9);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 625:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'N' ||
          lookahead == 'n') ADVANCE(846);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 626:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'N' ||
          lookahead == 'n') ADVANCE(28);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 627:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'N' ||
          lookahead == 'n') ADVANCE(183);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 628:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'N' ||
          lookahead == 'n') ADVANCE(874);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 629:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'N' ||
          lookahead == 'n') ADVANCE(820);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 630:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'N' ||
          lookahead == 'n') ADVANCE(626);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 631:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'N' ||
          lookahead == 'n') ADVANCE(224);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 632:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'N' ||
          lookahead == 'n') ADVANCE(34);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 633:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'N' ||
          lookahead == 'n') ADVANCE(632);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 634:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'N' ||
          lookahead == 'n') ADVANCE(264);
      if (lookahead == 'T' ||
          lookahead == 't') ADVANCE(740);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 635:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'N' ||
          lookahead == 'n') ADVANCE(231);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 636:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'N' ||
          lookahead == 'n') ADVANCE(352);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 637:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'N' ||
          lookahead == 'n') ADVANCE(844);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 638:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'N' ||
          lookahead == 'n') ADVANCE(778);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 639:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'N' ||
          lookahead == 'n') ADVANCE(883);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 640:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'N' ||
          lookahead == 'n') ADVANCE(788);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 641:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'N' ||
          lookahead == 'n') ADVANCE(97);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 642:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'N' ||
          lookahead == 'n') ADVANCE(353);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 643:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'N' ||
          lookahead == 'n') ADVANCE(356);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 644:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'N' ||
          lookahead == 'n') ADVANCE(790);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 645:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'N' ||
          lookahead == 'n') ADVANCE(200);
      if (lookahead == 'V' ||
          lookahead == 'v') ADVANCE(692);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 646:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'N' ||
          lookahead == 'n') ADVANCE(590);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 647:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'N' ||
          lookahead == 'n') ADVANCE(76);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 648:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'N' ||
          lookahead == 'n') ADVANCE(172);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 649:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'N' ||
          lookahead == 'n') ADVANCE(357);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 650:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'N' ||
          lookahead == 'n') ADVANCE(592);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 651:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'N' ||
          lookahead == 'n') ADVANCE(795);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 652:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'N' ||
          lookahead == 'n') ADVANCE(206);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 653:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'N' ||
          lookahead == 'n') ADVANCE(318);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 654:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'N' ||
          lookahead == 'n') ADVANCE(212);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 655:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'N' ||
          lookahead == 'n') ADVANCE(372);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 656:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'N' ||
          lookahead == 'n') ADVANCE(482);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 657:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'O' ||
          lookahead == 'o') ADVANCE(1369);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 658:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'O' ||
          lookahead == 'o') ADVANCE(954);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 659:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'O' ||
          lookahead == 'o') ADVANCE(1381);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 660:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'O' ||
          lookahead == 'o') ADVANCE(634);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 661:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'O' ||
          lookahead == 'o') ADVANCE(1394);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 662:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'O' ||
          lookahead == 'o') ADVANCE(505);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 663:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'O' ||
          lookahead == 'o') ADVANCE(957);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 664:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'O' ||
          lookahead == 'o') ADVANCE(913);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 665:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'O' ||
          lookahead == 'o') ADVANCE(865);
      if (lookahead == 'U' ||
          lookahead == 'u') ADVANCE(771);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 666:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'O' ||
          lookahead == 'o') ADVANCE(11);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 667:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'O' ||
          lookahead == 'o') ADVANCE(965);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 668:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'O' ||
          lookahead == 'o') ADVANCE(620);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 669:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'O' ||
          lookahead == 'o') ADVANCE(724);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 670:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'O' ||
          lookahead == 'o') ADVANCE(955);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 671:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'O' ||
          lookahead == 'o') ADVANCE(918);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 672:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'O' ||
          lookahead == 'o') ADVANCE(956);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 673:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'O' ||
          lookahead == 'o') ADVANCE(494);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 674:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'O' ||
          lookahead == 'o') ADVANCE(13);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 675:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'O' ||
          lookahead == 'o') ADVANCE(150);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 676:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'O' ||
          lookahead == 'o') ADVANCE(871);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 677:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'O' ||
          lookahead == 'o') ADVANCE(958);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 678:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'O' ||
          lookahead == 'o') ADVANCE(821);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 679:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'O' ||
          lookahead == 'o') ADVANCE(959);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 680:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'O' ||
          lookahead == 'o') ADVANCE(607);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 681:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'O' ||
          lookahead == 'o') ADVANCE(969);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 682:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'O' ||
          lookahead == 'o') ADVANCE(28);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 683:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'O' ||
          lookahead == 'o') ADVANCE(18);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 684:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'O' ||
          lookahead == 'o') ADVANCE(610);
      if (lookahead == 'W' ||
          lookahead == 'w') ADVANCE(989);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 685:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'O' ||
          lookahead == 'o') ADVANCE(183);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 686:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'O' ||
          lookahead == 'o') ADVANCE(460);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 687:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'O' ||
          lookahead == 'o') ADVANCE(217);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 688:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'O' ||
          lookahead == 'o') ADVANCE(879);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 689:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'O' ||
          lookahead == 'o') ADVANCE(619);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 690:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'O' ||
          lookahead == 'o') ADVANCE(647);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 691:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'O' ||
          lookahead == 'o') ADVANCE(611);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 692:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'O' ||
          lookahead == 'o') ADVANCE(768);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 693:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'O' ||
          lookahead == 'o') ADVANCE(569);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 694:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'O' ||
          lookahead == 'o') ADVANCE(648);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 695:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'O' ||
          lookahead == 'o') ADVANCE(561);
      if (lookahead == 'R' ||
          lookahead == 'r') ADVANCE(294);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 696:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'O' ||
          lookahead == 'o') ADVANCE(561);
      if (lookahead == 'R' ||
          lookahead == 'r') ADVANCE(62);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 697:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'O' ||
          lookahead == 'o') ADVANCE(685);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 698:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'O' ||
          lookahead == 'o') ADVANCE(827);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 699:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'O' ||
          lookahead == 'o') ADVANCE(961);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 700:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'O' ||
          lookahead == 'o') ADVANCE(966);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 701:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'O' ||
          lookahead == 'o') ADVANCE(631);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 702:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'O' ||
          lookahead == 'o') ADVANCE(840);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 703:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'O' ||
          lookahead == 'o') ADVANCE(680);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 704:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'O' ||
          lookahead == 'o') ADVANCE(964);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 705:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'O' ||
          lookahead == 'o') ADVANCE(963);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 706:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'O' ||
          lookahead == 'o') ADVANCE(653);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 707:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'O' ||
          lookahead == 'o') ADVANCE(784);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 708:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'O' ||
          lookahead == 'o') ADVANCE(627);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 709:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'O' ||
          lookahead == 'o') ADVANCE(841);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 710:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'O' ||
          lookahead == 'o') ADVANCE(830);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 711:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'O' ||
          lookahead == 'o') ADVANCE(838);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 712:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'O' ||
          lookahead == 'o') ADVANCE(535);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 713:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'O' ||
          lookahead == 'o') ADVANCE(536);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 714:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'O' ||
          lookahead == 'o') ADVANCE(691);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 715:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'O' ||
          lookahead == 'o') ADVANCE(539);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 716:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'O' ||
          lookahead == 'o') ADVANCE(540);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 717:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'O' ||
          lookahead == 'o') ADVANCE(803);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 718:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'O' ||
          lookahead == 'o') ADVANCE(817);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 719:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'O' ||
          lookahead == 'o') ADVANCE(650);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 720:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'O' ||
          lookahead == 'o') ADVANCE(463);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 721:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'O' ||
          lookahead == 'o') ADVANCE(804);
      if (lookahead == 'Y' ||
          lookahead == 'y') ADVANCE(79);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 722:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'O' ||
          lookahead == 'o') ADVANCE(971);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 723:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'O' ||
          lookahead == 'o') ADVANCE(573);
      if (lookahead == 'R' ||
          lookahead == 'r') ADVANCE(62);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 724:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'P' ||
          lookahead == 'p') ADVANCE(1381);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 725:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'P' ||
          lookahead == 'p') ADVANCE(138);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 726:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'P' ||
          lookahead == 'p') ADVANCE(1394);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 727:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'P' ||
          lookahead == 'p') ADVANCE(739);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 728:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'P' ||
          lookahead == 'p') ADVANCE(426);
      if (lookahead == 'U' ||
          lookahead == 'u') ADVANCE(783);
      if (lookahead == 'V' ||
          lookahead == 'v') ADVANCE(295);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 729:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'P' ||
          lookahead == 'p') ADVANCE(383);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 730:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'P' ||
          lookahead == 'p') ADVANCE(269);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 731:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'P' ||
          lookahead == 'p') ADVANCE(442);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 732:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'P' ||
          lookahead == 'p') ADVANCE(557);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 733:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'P' ||
          lookahead == 'p') ADVANCE(115);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 734:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'P' ||
          lookahead == 'p') ADVANCE(109);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 735:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'P' ||
          lookahead == 'p') ADVANCE(910);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 736:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'P' ||
          lookahead == 'p') ADVANCE(224);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 737:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'P' ||
          lookahead == 'p') ADVANCE(340);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 738:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'P' ||
          lookahead == 'p') ADVANCE(501);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 739:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'P' ||
          lookahead == 'p') ADVANCE(447);
      if (lookahead == 'S' ||
          lookahead == 's') ADVANCE(17);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 740:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'P' ||
          lookahead == 'p') ADVANCE(447);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 741:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'P' ||
          lookahead == 'p') ADVANCE(526);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 742:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'P' ||
          lookahead == 'p') ADVANCE(538);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 743:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'P' ||
          lookahead == 'p') ADVANCE(940);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 744:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'Q' ||
          lookahead == 'q') ADVANCE(921);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 745:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'Q' ||
          lookahead == 'q') ADVANCE(927);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 746:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'Q' ||
          lookahead == 'q') ADVANCE(930);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 747:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'Q' ||
          lookahead == 'q') ADVANCE(931);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 748:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'Q' ||
          lookahead == 'q') ADVANCE(938);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 749:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'R' ||
          lookahead == 'r') ADVANCE(37);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 750:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'R' ||
          lookahead == 'r') ADVANCE(738);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 751:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'R' ||
          lookahead == 'r') ADVANCE(1369);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 752:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'R' ||
          lookahead == 'r') ADVANCE(1381);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 753:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'R' ||
          lookahead == 'r') ADVANCE(744);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 754:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'R' ||
          lookahead == 'r') ADVANCE(1378);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 755:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'R' ||
          lookahead == 'r') ADVANCE(82);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 756:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'R' ||
          lookahead == 'r') ADVANCE(1394);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 757:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'R' ||
          lookahead == 'r') ADVANCE(1385);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 758:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'R' ||
          lookahead == 'r') ADVANCE(348);
      if (lookahead == 'T' ||
          lookahead == 't') ADVANCE(892);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 759:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'R' ||
          lookahead == 'r') ADVANCE(42);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 760:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'R' ||
          lookahead == 'r') ADVANCE(61);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 761:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'R' ||
          lookahead == 'r') ADVANCE(974);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 762:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'R' ||
          lookahead == 'r') ADVANCE(53);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 763:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'R' ||
          lookahead == 'r') ADVANCE(603);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 764:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'R' ||
          lookahead == 'r') ADVANCE(868);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 765:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'R' ||
          lookahead == 'r') ADVANCE(851);
      if (lookahead == 'S' ||
          lookahead == 's') ADVANCE(884);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 766:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'R' ||
          lookahead == 'r') ADVANCE(8);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 767:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'R' ||
          lookahead == 'r') ADVANCE(597);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 768:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'R' ||
          lookahead == 'r') ADVANCE(976);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 769:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'R' ||
          lookahead == 'r') ADVANCE(761);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 770:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'R' ||
          lookahead == 'r') ADVANCE(485);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 771:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'R' ||
          lookahead == 'r') ADVANCE(773);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 772:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'R' ||
          lookahead == 'r') ADVANCE(257);
      if (lookahead == 'Y' ||
          lookahead == 'y') ADVANCE(313);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 773:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'R' ||
          lookahead == 'r') ADVANCE(425);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 774:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'R' ||
          lookahead == 'r') ADVANCE(57);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 775:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'R' ||
          lookahead == 'r') ADVANCE(699);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 776:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'R' ||
          lookahead == 'r') ADVANCE(45);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 777:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'R' ||
          lookahead == 'r') ADVANCE(329);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 778:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'R' ||
          lookahead == 'r') ADVANCE(460);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 779:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'R' ||
          lookahead == 'r') ADVANCE(675);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 780:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'R' ||
          lookahead == 'r') ADVANCE(430);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 781:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'R' ||
          lookahead == 'r') ADVANCE(224);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 782:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'R' ||
          lookahead == 'r') ADVANCE(508);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 783:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'R' ||
          lookahead == 'r') ADVANCE(322);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 784:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'R' ||
          lookahead == 'r') ADVANCE(661);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 785:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'R' ||
          lookahead == 'r') ADVANCE(280);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 786:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'R' ||
          lookahead == 'r') ADVANCE(54);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 787:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'R' ||
          lookahead == 'r') ADVANCE(227);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 788:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'R' ||
          lookahead == 'r') ADVANCE(685);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 789:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'R' ||
          lookahead == 'r') ADVANCE(246);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 790:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'R' ||
          lookahead == 'r') ADVANCE(687);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 791:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'R' ||
          lookahead == 'r') ADVANCE(466);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 792:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'R' ||
          lookahead == 'r') ADVANCE(268);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 793:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'R' ||
          lookahead == 'r') ADVANCE(299);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 794:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'R' ||
          lookahead == 'r') ADVANCE(231);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 795:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'R' ||
          lookahead == 'r') ADVANCE(271);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 796:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'R' ||
          lookahead == 'r') ADVANCE(294);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 797:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'R' ||
          lookahead == 'r') ADVANCE(581);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 798:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'R' ||
          lookahead == 'r') ADVANCE(283);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 799:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'R' ||
          lookahead == 'r') ADVANCE(48);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 800:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'R' ||
          lookahead == 'r') ADVANCE(882);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 801:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'R' ||
          lookahead == 'r') ADVANCE(293);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 802:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'R' ||
          lookahead == 'r') ADVANCE(452);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 803:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'R' ||
          lookahead == 'r') ADVANCE(89);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 804:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'R' ||
          lookahead == 'r') ADVANCE(83);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 805:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'R' ||
          lookahead == 'r') ADVANCE(98);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 806:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'R' ||
          lookahead == 'r') ADVANCE(169);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 807:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'R' ||
          lookahead == 'r') ADVANCE(741);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 808:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'R' ||
          lookahead == 'r') ADVANCE(702);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 809:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'R' ||
          lookahead == 'r') ADVANCE(320);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 810:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'R' ||
          lookahead == 'r') ADVANCE(704);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 811:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'R' ||
          lookahead == 'r') ADVANCE(462);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 812:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'R' ||
          lookahead == 'r') ADVANCE(742);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 813:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'R' ||
          lookahead == 'r') ADVANCE(709);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 814:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'R' ||
          lookahead == 'r') ADVANCE(145);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 815:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'R' ||
          lookahead == 'r') ADVANCE(333);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 816:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'R' ||
          lookahead == 'r') ADVANCE(748);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 817:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'R' ||
          lookahead == 'r') ADVANCE(118);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 818:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'S' ||
          lookahead == 's') ADVANCE(1381);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 819:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'S' ||
          lookahead == 's') ADVANCE(422);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 820:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'S' ||
          lookahead == 's') ADVANCE(129);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 821:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'S' ||
          lookahead == 's') ADVANCE(729);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 822:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'S' ||
          lookahead == 's') ADVANCE(980);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 823:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'S' ||
          lookahead == 's') ADVANCE(218);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 824:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'S' ||
          lookahead == 's') ADVANCE(869);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 825:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'S' ||
          lookahead == 's') ADVANCE(818);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 826:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'S' ||
          lookahead == 's') ADVANCE(858);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 827:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'S' ||
          lookahead == 's') ADVANCE(219);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 828:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'S' ||
          lookahead == 's') ADVANCE(433);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 829:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'S' ||
          lookahead == 's') ADVANCE(378);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 830:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'S' ||
          lookahead == 's') ADVANCE(898);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 831:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'S' ||
          lookahead == 's') ADVANCE(387);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 832:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'S' ||
          lookahead == 's') ADVANCE(384);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 833:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'S' ||
          lookahead == 's') ADVANCE(376);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 834:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'S' ||
          lookahead == 's') ADVANCE(379);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 835:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'S' ||
          lookahead == 's') ADVANCE(380);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 836:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'S' ||
          lookahead == 's') ADVANCE(381);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 837:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'S' ||
          lookahead == 's') ADVANCE(453);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 838:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'S' ||
          lookahead == 's') ADVANCE(904);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 839:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'S' ||
          lookahead == 's') ADVANCE(896);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 840:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'S' ||
          lookahead == 's') ADVANCE(224);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 841:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'S' ||
          lookahead == 's') ADVANCE(231);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 842:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'S' ||
          lookahead == 's') ADVANCE(746);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 843:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'S' ||
          lookahead == 's') ADVANCE(992);
      if (lookahead == 'Y' ||
          lookahead == 'y') ADVANCE(117);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 844:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'S' ||
          lookahead == 's') ADVANCE(876);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 845:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'S' ||
          lookahead == 's') ADVANCE(450);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 846:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'S' ||
          lookahead == 's') ADVANCE(734);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 847:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'S' ||
          lookahead == 's') ADVANCE(393);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 848:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'S' ||
          lookahead == 's') ADVANCE(429);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 849:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'S' ||
          lookahead == 's') ADVANCE(680);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 850:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'S' ||
          lookahead == 's') ADVANCE(873);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 851:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'S' ||
          lookahead == 's') ADVANCE(533);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 852:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'S' ||
          lookahead == 's') ADVANCE(881);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 853:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'S' ||
          lookahead == 's') ADVANCE(448);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 854:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'S' ||
          lookahead == 's') ADVANCE(691);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 855:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'S' ||
          lookahead == 's') ADVANCE(870);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 856:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'S' ||
          lookahead == 's') ADVANCE(901);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 857:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'T' ||
          lookahead == 't') ADVANCE(1369);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 858:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'T' ||
          lookahead == 't') ADVANCE(1381);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 859:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'T' ||
          lookahead == 't') ADVANCE(1380);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 860:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'T' ||
          lookahead == 't') ADVANCE(121);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 861:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'T' ||
          lookahead == 't') ADVANCE(132);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 862:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'T' ||
          lookahead == 't') ADVANCE(1394);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 863:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'T' ||
          lookahead == 't') ADVANCE(143);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 864:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'T' ||
          lookahead == 't') ADVANCE(140);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 865:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'T' ||
          lookahead == 't') ADVANCE(731);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 866:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'T' ||
          lookahead == 't') ADVANCE(759);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 867:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'T' ||
          lookahead == 't') ADVANCE(218);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 868:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'T' ||
          lookahead == 't') ADVANCE(377);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 869:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'T' ||
          lookahead == 't') ADVANCE(981);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 870:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'T' ||
          lookahead == 't') ADVANCE(367);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 871:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'T' ||
          lookahead == 't') ADVANCE(386);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 872:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'T' ||
          lookahead == 't') ADVANCE(459);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 873:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'T' ||
          lookahead == 't') ADVANCE(269);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 874:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'T' ||
          lookahead == 't') ADVANCE(28);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 875:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'T' ||
          lookahead == 't') ADVANCE(939);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 876:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'T' ||
          lookahead == 't') ADVANCE(440);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 877:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'T' ||
          lookahead == 't') ADVANCE(164);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 878:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'T' ||
          lookahead == 't') ADVANCE(112);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 879:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'T' ||
          lookahead == 't') ADVANCE(874);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 880:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'T' ||
          lookahead == 't') ADVANCE(464);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 881:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'T' ||
          lookahead == 't') ADVANCE(224);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 882:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'T' ||
          lookahead == 't') ADVANCE(798);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 883:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'T' ||
          lookahead == 't') ADVANCE(34);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 884:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'T' ||
          lookahead == 't') ADVANCE(278);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 885:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'T' ||
          lookahead == 't') ADVANCE(661);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 886:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'T' ||
          lookahead == 't') ADVANCE(477);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 887:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'T' ||
          lookahead == 't') ADVANCE(231);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 888:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'T' ||
          lookahead == 't') ADVANCE(236);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 889:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'T' ||
          lookahead == 't') ADVANCE(237);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 890:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'T' ||
          lookahead == 't') ADVANCE(260);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 891:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'T' ||
          lookahead == 't') ADVANCE(133);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 892:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'T' ||
          lookahead == 't') ADVANCE(274);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 893:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'T' ||
          lookahead == 't') ADVANCE(837);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 894:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'T' ||
          lookahead == 't') ADVANCE(780);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 895:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'T' ||
          lookahead == 't') ADVANCE(93);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 896:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'T' ||
          lookahead == 't') ADVANCE(538);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 897:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'T' ||
          lookahead == 't') ADVANCE(795);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 898:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'T' ||
          lookahead == 't') ADVANCE(370);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 899:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'T' ||
          lookahead == 't') ADVANCE(467);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 900:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'T' ||
          lookahead == 't') ADVANCE(706);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 901:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'T' ||
          lookahead == 't') ADVANCE(372);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 902:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'T' ||
          lookahead == 't') ADVANCE(145);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 903:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'T' ||
          lookahead == 't') ADVANCE(328);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 904:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'T' ||
          lookahead == 't') ADVANCE(971);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 905:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'T' ||
          lookahead == 't') ADVANCE(994);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 906:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'U' ||
          lookahead == 'u') ADVANCE(146);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 907:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'U' ||
          lookahead == 'u') ADVANCE(866);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 908:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'U' ||
          lookahead == 'u') ADVANCE(137);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 909:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'U' ||
          lookahead == 'u') ADVANCE(1394);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 910:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'U' ||
          lookahead == 'u') ADVANCE(336);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 911:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'U' ||
          lookahead == 'u') ADVANCE(222);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 912:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'U' ||
          lookahead == 'u') ADVANCE(576);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 913:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'U' ||
          lookahead == 'u') ADVANCE(360);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 914:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'U' ||
          lookahead == 'u') ADVANCE(218);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 915:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'U' ||
          lookahead == 'u') ADVANCE(769);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 916:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'U' ||
          lookahead == 'u') ADVANCE(577);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 917:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'U' ||
          lookahead == 'u') ADVANCE(461);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 918:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'U' ||
          lookahead == 'u') ADVANCE(189);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 919:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'U' ||
          lookahead == 'u') ADVANCE(578);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 920:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'U' ||
          lookahead == 'u') ADVANCE(579);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 921:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'U' ||
          lookahead == 'u') ADVANCE(686);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 922:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'U' ||
          lookahead == 'u') ADVANCE(836);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 923:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'U' ||
          lookahead == 'u') ADVANCE(35);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 924:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'U' ||
          lookahead == 'u') ADVANCE(436);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 925:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'U' ||
          lookahead == 'u') ADVANCE(224);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 926:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'U' ||
          lookahead == 'u') ADVANCE(437);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 927:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'U' ||
          lookahead == 'u') ADVANCE(71);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 928:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'U' ||
          lookahead == 'u') ADVANCE(225);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 929:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'U' ||
          lookahead == 'u') ADVANCE(228);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 930:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'U' ||
          lookahead == 'u') ADVANCE(231);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 931:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'U' ||
          lookahead == 'u') ADVANCE(334);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 932:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'U' ||
          lookahead == 'u') ADVANCE(736);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 933:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'U' ||
          lookahead == 'u') ADVANCE(221);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 934:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'U' ||
          lookahead == 'u') ADVANCE(623);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 935:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'U' ||
          lookahead == 'u') ADVANCE(841);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 936:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'U' ||
          lookahead == 'u') ADVANCE(794);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 937:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'U' ||
          lookahead == 'u') ADVANCE(807);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 938:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'U' ||
          lookahead == 'u') ADVANCE(720);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 939:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'U' ||
          lookahead == 'u') ADVANCE(899);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 940:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'U' ||
          lookahead == 'u') ADVANCE(812);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 941:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'U' ||
          lookahead == 'u') ADVANCE(816);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 942:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'V' ||
          lookahead == 'v') ADVANCE(36);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 943:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'V' ||
          lookahead == 'v') ADVANCE(977);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 944:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'V' ||
          lookahead == 'v') ADVANCE(223);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 945:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'V' ||
          lookahead == 'v') ADVANCE(219);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 946:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'V' ||
          lookahead == 'v') ADVANCE(249);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 947:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'V' ||
          lookahead == 'v') ADVANCE(224);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 948:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'V' ||
          lookahead == 'v') ADVANCE(235);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 949:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'V' ||
          lookahead == 'v') ADVANCE(279);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 950:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'V' ||
          lookahead == 'v') ADVANCE(289);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 951:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'V' ||
          lookahead == 'v') ADVANCE(326);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 952:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'V' ||
          lookahead == 'v') ADVANCE(323);
      if (lookahead == 'W' ||
          lookahead == 'w') ADVANCE(655);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 953:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'V' ||
          lookahead == 'v') ADVANCE(478);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 954:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'W' ||
          lookahead == 'w') ADVANCE(1369);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 955:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'W' ||
          lookahead == 'w') ADVANCE(1381);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 956:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'W' ||
          lookahead == 'w') ADVANCE(1380);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 957:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'W' ||
          lookahead == 'w') ADVANCE(338);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 958:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'W' ||
          lookahead == 'w') ADVANCE(1394);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 959:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'W' ||
          lookahead == 'w') ADVANCE(1388);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 960:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'W' ||
          lookahead == 'w') ADVANCE(397);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 961:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'W' ||
          lookahead == 'w') ADVANCE(607);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 962:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'W' ||
          lookahead == 'w') ADVANCE(617);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 963:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'W' ||
          lookahead == 'w') ADVANCE(435);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 964:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'W' ||
          lookahead == 'w') ADVANCE(611);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 965:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'W' ||
          lookahead == 'w') ADVANCE(439);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 966:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'W' ||
          lookahead == 'w') ADVANCE(414);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 967:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'W' ||
          lookahead == 'w') ADVANCE(697);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 968:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'W' ||
          lookahead == 'w') ADVANCE(400);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 969:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'W' ||
          lookahead == 'w') ADVANCE(287);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 970:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'W' ||
          lookahead == 'w') ADVANCE(210);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 971:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'W' ||
          lookahead == 'w') ADVANCE(406);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 972:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'Y' ||
          lookahead == 'y') ADVANCE(46);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 973:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'Y' ||
          lookahead == 'y') ADVANCE(1369);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 974:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'Y' ||
          lookahead == 'y') ADVANCE(1381);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 975:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'Y' ||
          lookahead == 'y') ADVANCE(127);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 976:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'Y' ||
          lookahead == 'y') ADVANCE(1394);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 977:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'Y' ||
          lookahead == 'y') ADVANCE(140);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 978:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'Y' ||
          lookahead == 'y') ADVANCE(367);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 979:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'Y' ||
          lookahead == 'y') ADVANCE(967);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 980:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'Y' ||
          lookahead == 'y') ADVANCE(717);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 981:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'Y' ||
          lookahead == 'y') ADVANCE(808);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 982:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'Y' ||
          lookahead == 'y') ADVANCE(114);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 983:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'Y' ||
          lookahead == 'y') ADVANCE(73);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 984:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'Y' ||
          lookahead == 'y') ADVANCE(74);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 985:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'Y' ||
          lookahead == 'y') ADVANCE(202);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 986:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'Y' ||
          lookahead == 'y') ADVANCE(79);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 987:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'Y' ||
          lookahead == 'y') ADVANCE(370);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 988:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'Y' ||
          lookahead == 'y') ADVANCE(308);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 989:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'Y' ||
          lookahead == 'y') ADVANCE(310);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 990:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'Y' ||
          lookahead == 'y') ADVANCE(313);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 991:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'Y' ||
          lookahead == 'y') ADVANCE(316);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 992:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'Y' ||
          lookahead == 'y') ADVANCE(141);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 993:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'Y' ||
          lookahead == 'y') ADVANCE(145);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 994:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'Y' ||
          lookahead == 'y') ADVANCE(813);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 995:
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'Z' ||
          lookahead == 'z') ADVANCE(224);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Y') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'y')) ADVANCE(996);
      END_STATE();
    case 996:
      if (lookahead == '!') ADVANCE(1334);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 997:
      if (lookahead == '!') ADVANCE(1181);
      if (('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(1330);
      END_STATE();
    case 998:
      ADVANCE_MAP(
        '"', 1000,
        '#', 1174,
        '$', 997,
        '%', 1359,
        '&', 1179,
        '\'', 1005,
        '(', 1348,
        '*', 1360,
        '+', 1361,
        ',', 1325,
        '-', 1362,
        '.', 1020,
        '/', 1357,
        ':', 1328,
        ';', 1326,
        '=', 1329,
        '@', 1025,
        'C', 1310,
        'E', 1215,
        '^', 1356,
        '_', 1186,
        'c', 1219,
        'h', 1230,
        'r', 1200,
        '{', 1346,
        '}', 1347,
        'B', 1306,
        'b', 1306,
        'S', 1240,
        's', 1240,
        'T', 1322,
        't', 1322,
      );
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(998);
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(1351);
      if (('A' <= lookahead && lookahead <= 'Z') ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(1324);
      END_STATE();
    case 999:
      ADVANCE_MAP(
        '"', 1000,
        '#', 1177,
        '$', 997,
        '&', 1179,
        '\'', 1005,
        '(', 1348,
        ')', 1349,
        ',', 1325,
        '-', 1006,
        '.', 1176,
        ':', 1180,
        ';', 1326,
        'C', 1310,
        'E', 1215,
        '_', 1186,
        'c', 1219,
        'h', 1230,
        'r', 1200,
        'B', 1306,
        'b', 1306,
        'S', 1240,
        's', 1240,
        'T', 1322,
        't', 1322,
      );
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(999);
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(1351);
      if (('A' <= lookahead && lookahead <= 'Z') ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(1324);
      END_STATE();
    case 1000:
      if (lookahead == '"') ADVANCE(1364);
      if (lookahead != 0 &&
          lookahead != '\t' &&
          lookahead != '\n') ADVANCE(1000);
      END_STATE();
    case 1001:
      ADVANCE_MAP(
        '%', 1359,
        '&', 1179,
        ')', 1349,
        '*', 1360,
        '+', 1361,
        ',', 1325,
        '-', 1362,
        '.', 1176,
        '/', 1357,
        ':', 1327,
        ';', 1326,
        '^', 1356,
        '_', 1350,
        'A', 575,
        'a', 575,
        'B', 491,
        'b', 491,
        'C', 972,
        'c', 972,
        'E', 580,
        'e', 580,
        'F', 906,
        'f', 906,
        'G', 749,
        'g', 749,
        'I', 604,
        'i', 604,
        'L', 455,
        'l', 455,
        'N', 241,
        'n', 241,
        'O', 799,
        'o', 799,
        'P', 417,
        'p', 417,
        'R', 240,
        'r', 240,
        'S', 3,
        's', 3,
        'T', 242,
        't', 242,
        'V', 418,
        'v', 418,
        'Y', 239,
        'y', 239,
        'Z', 419,
        'z', 419,
      );
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(1001);
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(1351);
      if (('D' <= lookahead && lookahead <= 'X') ||
          ('d' <= lookahead && lookahead <= 'x')) ADVANCE(996);
      END_STATE();
    case 1002:
      ADVANCE_MAP(
        '&', 1179,
        '-', 1006,
        '.', 1397,
        ':', 1327,
        'G', 673,
        'g', 673,
        'O', 500,
        'o', 500,
        'P', 252,
        'p', 252,
        'R', 698,
        'r', 698,
      );
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(1002);
      if (('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 1003:
      ADVANCE_MAP(
        '&', 1179,
        '-', 1006,
        0x212a, 1094,
        'A', 498,
        'a', 498,
        'B', 38,
        'b', 38,
        'C', 21,
        'c', 21,
        'D', 22,
        'd', 22,
        'E', 56,
        'e', 56,
        'F', 23,
        'f', 23,
        'G', 385,
        'g', 385,
        'H', 665,
        'h', 665,
        'I', 637,
        'i', 637,
        'K', 388,
        'k', 388,
        'L', 24,
        'l', 24,
        'M', 25,
        'm', 25,
        'N', 41,
        'n', 41,
        'O', 504,
        'o', 504,
        'P', 26,
        'p', 26,
        'Q', 917,
        'q', 917,
        'R', 220,
        'r', 220,
        'S', 27,
        's', 27,
        'T', 40,
        't', 40,
        'W', 64,
        'w', 64,
        'Y', 297,
        'y', 297,
      );
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(1003);
      if (('J' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('j' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 1004:
      ADVANCE_MAP(
        '&', 1179,
        '-', 1006,
        0x212a, 1096,
        'A', 532,
        'a', 532,
        'B', 317,
        'b', 317,
        'C', 88,
        'c', 88,
        'D', 75,
        'd', 75,
        'F', 474,
        'f', 474,
        'G', 92,
        'g', 92,
        'H', 660,
        'h', 660,
        'I', 645,
        'i', 645,
        'K', 395,
        'k', 395,
        'L', 30,
        'l', 30,
        'M', 31,
        'm', 31,
        'N', 51,
        'n', 51,
        'O', 489,
        'o', 489,
        'P', 32,
        'p', 32,
        'R', 230,
        'r', 230,
        'S', 15,
        's', 15,
        'T', 78,
        't', 78,
        'V', 479,
        'v', 479,
        'W', 405,
        'w', 405,
        'Y', 314,
        'y', 314,
      );
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(1004);
      if (('E' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('e' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 1005:
      if (lookahead == '\'') ADVANCE(1363);
      if (lookahead != 0 &&
          lookahead != '\t' &&
          lookahead != '\n') ADVANCE(1005);
      END_STATE();
    case 1006:
      if (lookahead == '-') ADVANCE(1401);
      END_STATE();
    case 1007:
      if (lookahead == '/') ADVANCE(1399);
      END_STATE();
    case 1008:
      if (lookahead == '/') ADVANCE(1175);
      END_STATE();
    case 1009:
      if (lookahead == '/') ADVANCE(1398);
      END_STATE();
    case 1010:
      if (lookahead == '/') ADVANCE(1007);
      END_STATE();
    case 1011:
      if (lookahead == '/') ADVANCE(1008);
      END_STATE();
    case 1012:
      if (lookahead == '/') ADVANCE(1009);
      END_STATE();
    case 1013:
      if (lookahead == '0') ADVANCE(1366);
      END_STATE();
    case 1014:
      if (lookahead == '0') ADVANCE(1365);
      END_STATE();
    case 1015:
      if (lookahead == '0') ADVANCE(1014);
      END_STATE();
    case 1016:
      if (lookahead == '0') ADVANCE(1014);
      if (lookahead == '5') ADVANCE(1014);
      END_STATE();
    case 1017:
      if (lookahead == ':') ADVANCE(1370);
      END_STATE();
    case 1018:
      if (lookahead == ':') ADVANCE(1372);
      END_STATE();
    case 1019:
      if (lookahead == ':') ADVANCE(1182);
      if (('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(1344);
      END_STATE();
    case 1020:
      if (lookahead == '_') ADVANCE(1341);
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(1352);
      if (('A' <= lookahead && lookahead <= 'Z') ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(1343);
      END_STATE();
    case 1021:
      if (lookahead == 'a') ADVANCE(1023);
      END_STATE();
    case 1022:
      if (lookahead == 'a') ADVANCE(1032);
      END_STATE();
    case 1023:
      if (lookahead == 'c') ADVANCE(1037);
      END_STATE();
    case 1024:
      if (lookahead == 'd') ADVANCE(1028);
      if (lookahead == 'm') ADVANCE(1021);
      if (lookahead == 'n') ADVANCE(1022);
      if (lookahead == 'p') ADVANCE(1035);
      END_STATE();
    case 1025:
      if (lookahead == 'd') ADVANCE(1028);
      if (lookahead == 'n') ADVANCE(1022);
      if (lookahead == 'p') ADVANCE(1035);
      END_STATE();
    case 1026:
      if (lookahead == 'e') ADVANCE(1337);
      END_STATE();
    case 1027:
      if (lookahead == 'e') ADVANCE(1338);
      END_STATE();
    case 1028:
      if (lookahead == 'e') ADVANCE(1036);
      END_STATE();
    case 1029:
      if (lookahead == 'i') ADVANCE(1034);
      END_STATE();
    case 1030:
      if (lookahead == 'i') ADVANCE(1040);
      END_STATE();
    case 1031:
      if (lookahead == 'i') ADVANCE(1039);
      END_STATE();
    case 1032:
      if (lookahead == 'm') ADVANCE(1026);
      END_STATE();
    case 1033:
      if (lookahead == 'o') ADVANCE(1333);
      END_STATE();
    case 1034:
      if (lookahead == 'o') ADVANCE(1038);
      END_STATE();
    case 1035:
      if (lookahead == 'r') ADVANCE(1029);
      END_STATE();
    case 1036:
      if (lookahead == 'r') ADVANCE(1030);
      END_STATE();
    case 1037:
      if (lookahead == 'r') ADVANCE(1033);
      END_STATE();
    case 1038:
      if (lookahead == 'r') ADVANCE(1031);
      END_STATE();
    case 1039:
      if (lookahead == 't') ADVANCE(1042);
      END_STATE();
    case 1040:
      if (lookahead == 'v') ADVANCE(1027);
      END_STATE();
    case 1041:
      if (lookahead == 'x') ADVANCE(1354);
      END_STATE();
    case 1042:
      if (lookahead == 'y') ADVANCE(1336);
      END_STATE();
    case 1043:
      if (lookahead == 'A' ||
          lookahead == 'a') ADVANCE(1172);
      END_STATE();
    case 1044:
      if (lookahead == 'A' ||
          lookahead == 'a') ADVANCE(1384);
      END_STATE();
    case 1045:
      if (lookahead == 'A' ||
          lookahead == 'a') ADVANCE(1123);
      END_STATE();
    case 1046:
      if (lookahead == 'A' ||
          lookahead == 'a') ADVANCE(1149);
      END_STATE();
    case 1047:
      if (lookahead == 'A' ||
          lookahead == 'a') ADVANCE(1107);
      END_STATE();
    case 1048:
      if (lookahead == 'A' ||
          lookahead == 'a') ADVANCE(1155);
      END_STATE();
    case 1049:
      if (lookahead == 'A' ||
          lookahead == 'a') ADVANCE(1154);
      END_STATE();
    case 1050:
      if (lookahead == 'A' ||
          lookahead == 'a') ADVANCE(1122);
      END_STATE();
    case 1051:
      if (lookahead == 'A' ||
          lookahead == 'a') ADVANCE(1167);
      if (lookahead == 'E' ||
          lookahead == 'e') ADVANCE(1082);
      END_STATE();
    case 1052:
      if (lookahead == 'A' ||
          lookahead == 'a') ADVANCE(1167);
      if (lookahead == 'E' ||
          lookahead == 'e') ADVANCE(1167);
      END_STATE();
    case 1053:
      if (lookahead == 'A' ||
          lookahead == 'a') ADVANCE(1108);
      if (lookahead == 'E' ||
          lookahead == 'e') ADVANCE(1057);
      if (lookahead == 'L' ||
          lookahead == 'l') ADVANCE(1049);
      END_STATE();
    case 1054:
      if (lookahead == 'A' ||
          lookahead == 'a') ADVANCE(1173);
      END_STATE();
    case 1055:
      if (lookahead == 'A' ||
          lookahead == 'a') ADVANCE(1114);
      END_STATE();
    case 1056:
      if (lookahead == 'A' ||
          lookahead == 'a') ADVANCE(1091);
      END_STATE();
    case 1057:
      if (lookahead == 'A' ||
          lookahead == 'a') ADVANCE(1093);
      END_STATE();
    case 1058:
      if (lookahead == 'A' ||
          lookahead == 'a') ADVANCE(1127);
      if (lookahead == 'C' ||
          lookahead == 'c') ADVANCE(1095);
      END_STATE();
    case 1059:
      if (lookahead == 'B' ||
          lookahead == 'b') ADVANCE(1109);
      END_STATE();
    case 1060:
      ADVANCE_MAP(
        'B', 1109,
        'b', 1109,
        'C', 1157,
        'c', 1157,
        'G', 1142,
        'g', 1142,
        'I', 1121,
        'i', 1121,
        'N', 1130,
        'n', 1130,
        'O', 1141,
        'o', 1141,
        'R', 1076,
        'r', 1076,
        'S', 1151,
        's', 1151,
        'T', 1048,
        't', 1048,
      );
      END_STATE();
    case 1061:
      if (lookahead == 'B' ||
          lookahead == 'b') ADVANCE(1117);
      END_STATE();
    case 1062:
      ADVANCE_MAP(
        'B', 1117,
        'b', 1117,
        'C', 1170,
        'c', 1170,
        'G', 1134,
        'g', 1134,
        'M', 1056,
        'm', 1056,
        'O', 1115,
        'o', 1115,
        'R', 1077,
        'r', 1077,
        'S', 1053,
        's', 1053,
        'T', 1159,
        't', 1159,
        'V', 1106,
        'v', 1106,
        'K', 1096,
        'k', 1096,
        0x212a, 1096,
      );
      END_STATE();
    case 1063:
      if (lookahead == 'B' ||
          lookahead == 'b') ADVANCE(1117);
      if (lookahead == 'G' ||
          lookahead == 'g') ADVANCE(1140);
      END_STATE();
    case 1064:
      if (lookahead == 'C' ||
          lookahead == 'c') ADVANCE(1373);
      END_STATE();
    case 1065:
      if (lookahead == 'D' ||
          lookahead == 'd') ADVANCE(1373);
      END_STATE();
    case 1066:
      if (lookahead == 'D' ||
          lookahead == 'd') ADVANCE(1384);
      END_STATE();
    case 1067:
      if (lookahead == 'D' ||
          lookahead == 'd') ADVANCE(1104);
      END_STATE();
    case 1068:
      if (lookahead == 'D' ||
          lookahead == 'd') ADVANCE(1084);
      END_STATE();
    case 1069:
      if (lookahead == 'E' ||
          lookahead == 'e') ADVANCE(1373);
      END_STATE();
    case 1070:
      if (lookahead == 'E' ||
          lookahead == 'e') ADVANCE(1092);
      END_STATE();
    case 1071:
      if (lookahead == 'E' ||
          lookahead == 'e') ADVANCE(1078);
      END_STATE();
    case 1072:
      if (lookahead == 'E' ||
          lookahead == 'e') ADVANCE(1384);
      END_STATE();
    case 1073:
      if (lookahead == 'E' ||
          lookahead == 'e') ADVANCE(1063);
      END_STATE();
    case 1074:
      if (lookahead == 'E' ||
          lookahead == 'e') ADVANCE(1110);
      END_STATE();
    case 1075:
      if (lookahead == 'E' ||
          lookahead == 'e') ADVANCE(1152);
      END_STATE();
    case 1076:
      if (lookahead == 'E' ||
          lookahead == 'e') ADVANCE(1065);
      if (lookahead == 'O' ||
          lookahead == 'o') ADVANCE(1169);
      END_STATE();
    case 1077:
      if (lookahead == 'E' ||
          lookahead == 'e') ADVANCE(1066);
      END_STATE();
    case 1078:
      if (lookahead == 'E' ||
          lookahead == 'e') ADVANCE(1120);
      if (lookahead == 'Y' ||
          lookahead == 'y') ADVANCE(1382);
      END_STATE();
    case 1079:
      if (lookahead == 'E' ||
          lookahead == 'e') ADVANCE(1165);
      END_STATE();
    case 1080:
      if (lookahead == 'E' ||
          lookahead == 'e') ADVANCE(1150);
      END_STATE();
    case 1081:
      if (lookahead == 'E' ||
          lookahead == 'e') ADVANCE(1122);
      END_STATE();
    case 1082:
      if (lookahead == 'E' ||
          lookahead == 'e') ADVANCE(1122);
      if (lookahead == 'Y' ||
          lookahead == 'y') ADVANCE(1384);
      END_STATE();
    case 1083:
      if (lookahead == 'E' ||
          lookahead == 'e') ADVANCE(1126);
      END_STATE();
    case 1084:
      if (lookahead == 'E' ||
          lookahead == 'e') ADVANCE(1124);
      END_STATE();
    case 1085:
      if (lookahead == 'E' ||
          lookahead == 'e') ADVANCE(1081);
      END_STATE();
    case 1086:
      if (lookahead == 'E' ||
          lookahead == 'e') ADVANCE(1093);
      END_STATE();
    case 1087:
      if (lookahead == 'G' ||
          lookahead == 'g') ADVANCE(1128);
      END_STATE();
    case 1088:
      if (lookahead == 'G' ||
          lookahead == 'g') ADVANCE(1069);
      END_STATE();
    case 1089:
      if (lookahead == 'G' ||
          lookahead == 'g') ADVANCE(1046);
      END_STATE();
    case 1090:
      if (lookahead == 'G' ||
          lookahead == 'g') ADVANCE(1072);
      END_STATE();
    case 1091:
      if (lookahead == 'G' ||
          lookahead == 'g') ADVANCE(1083);
      END_STATE();
    case 1092:
      if (lookahead == 'G' ||
          lookahead == 'g') ADVANCE(1145);
      END_STATE();
    case 1093:
      if (lookahead == 'G' ||
          lookahead == 'g') ADVANCE(1147);
      END_STATE();
    case 1094:
      if (lookahead == 'H' ||
          lookahead == 'h') ADVANCE(1043);
      END_STATE();
    case 1095:
      if (lookahead == 'H' ||
          lookahead == 'h') ADVANCE(1105);
      END_STATE();
    case 1096:
      if (lookahead == 'H' ||
          lookahead == 'h') ADVANCE(1054);
      END_STATE();
    case 1097:
      if (lookahead == 'I' ||
          lookahead == 'i') ADVANCE(1119);
      END_STATE();
    case 1098:
      if (lookahead == 'I' ||
          lookahead == 'i') ADVANCE(1119);
      if (lookahead == 'Y' ||
          lookahead == 'y') ADVANCE(1368);
      END_STATE();
    case 1099:
      if (lookahead == 'I' ||
          lookahead == 'i') ADVANCE(1373);
      END_STATE();
    case 1100:
      if (lookahead == 'I' ||
          lookahead == 'i') ADVANCE(1064);
      END_STATE();
    case 1101:
      if (lookahead == 'I' ||
          lookahead == 'i') ADVANCE(1384);
      END_STATE();
    case 1102:
      if (lookahead == 'I' ||
          lookahead == 'i') ADVANCE(1162);
      END_STATE();
    case 1103:
      if (lookahead == 'I' ||
          lookahead == 'i') ADVANCE(1148);
      END_STATE();
    case 1104:
      if (lookahead == 'I' ||
          lookahead == 'i') ADVANCE(1087);
      END_STATE();
    case 1105:
      if (lookahead == 'I' ||
          lookahead == 'i') ADVANCE(1066);
      END_STATE();
    case 1106:
      if (lookahead == 'I' ||
          lookahead == 'i') ADVANCE(1135);
      END_STATE();
    case 1107:
      if (lookahead == 'L' ||
          lookahead == 'l') ADVANCE(1059);
      END_STATE();
    case 1108:
      if (lookahead == 'L' ||
          lookahead == 'l') ADVANCE(1118);
      END_STATE();
    case 1109:
      if (lookahead == 'L' ||
          lookahead == 'l') ADVANCE(1158);
      END_STATE();
    case 1110:
      if (lookahead == 'L' ||
          lookahead == 'l') ADVANCE(1111);
      END_STATE();
    case 1111:
      if (lookahead == 'L' ||
          lookahead == 'l') ADVANCE(1129);
      END_STATE();
    case 1112:
      if (lookahead == 'L' ||
          lookahead == 'l') ADVANCE(1100);
      END_STATE();
    case 1113:
      if (lookahead == 'L' ||
          lookahead == 'l') ADVANCE(1068);
      END_STATE();
    case 1114:
      if (lookahead == 'L' ||
          lookahead == 'l') ADVANCE(1112);
      END_STATE();
    case 1115:
      if (lookahead == 'L' ||
          lookahead == 'l') ADVANCE(1102);
      if (lookahead == 'R' ||
          lookahead == 'r') ADVANCE(1058);
      END_STATE();
    case 1116:
      if (lookahead == 'L' ||
          lookahead == 'l') ADVANCE(1080);
      END_STATE();
    case 1117:
      if (lookahead == 'L' ||
          lookahead == 'l') ADVANCE(1161);
      END_STATE();
    case 1118:
      if (lookahead == 'M' ||
          lookahead == 'm') ADVANCE(1133);
      END_STATE();
    case 1119:
      if (lookahead == 'N' ||
          lookahead == 'n') ADVANCE(1017);
      END_STATE();
    case 1120:
      if (lookahead == 'N' ||
          lookahead == 'n') ADVANCE(1373);
      END_STATE();
    case 1121:
      if (lookahead == 'N' ||
          lookahead == 'n') ADVANCE(1067);
      END_STATE();
    case 1122:
      if (lookahead == 'N' ||
          lookahead == 'n') ADVANCE(1384);
      END_STATE();
    case 1123:
      if (lookahead == 'N' ||
          lookahead == 'n') ADVANCE(1088);
      END_STATE();
    case 1124:
      if (lookahead == 'N' ||
          lookahead == 'n') ADVANCE(1146);
      END_STATE();
    case 1125:
      if (lookahead == 'N' ||
          lookahead == 'n') ADVANCE(1070);
      END_STATE();
    case 1126:
      if (lookahead == 'N' ||
          lookahead == 'n') ADVANCE(1153);
      END_STATE();
    case 1127:
      if (lookahead == 'N' ||
          lookahead == 'n') ADVANCE(1090);
      END_STATE();
    case 1128:
      if (lookahead == 'O' ||
          lookahead == 'o') ADVANCE(1373);
      END_STATE();
    case 1129:
      if (lookahead == 'O' ||
          lookahead == 'o') ADVANCE(1163);
      END_STATE();
    case 1130:
      if (lookahead == 'O' ||
          lookahead == 'o') ADVANCE(1156);
      END_STATE();
    case 1131:
      if (lookahead == 'O' ||
          lookahead == 'o') ADVANCE(1066);
      END_STATE();
    case 1132:
      if (lookahead == 'O' ||
          lookahead == 'o') ADVANCE(1125);
      END_STATE();
    case 1133:
      if (lookahead == 'O' ||
          lookahead == 'o') ADVANCE(1122);
      END_STATE();
    case 1134:
      if (lookahead == 'O' ||
          lookahead == 'o') ADVANCE(1113);
      if (lookahead == 'R' ||
          lookahead == 'r') ADVANCE(1051);
      END_STATE();
    case 1135:
      if (lookahead == 'O' ||
          lookahead == 'o') ADVANCE(1116);
      END_STATE();
    case 1136:
      if (lookahead == 'O' ||
          lookahead == 'o') ADVANCE(1103);
      END_STATE();
    case 1137:
      if (lookahead == 'P' ||
          lookahead == 'p') ADVANCE(1069);
      END_STATE();
    case 1138:
      if (lookahead == 'Q' ||
          lookahead == 'q') ADVANCE(1160);
      END_STATE();
    case 1139:
      if (lookahead == 'R' ||
          lookahead == 'r') ADVANCE(1138);
      END_STATE();
    case 1140:
      if (lookahead == 'R' ||
          lookahead == 'r') ADVANCE(1052);
      END_STATE();
    case 1141:
      if (lookahead == 'R' ||
          lookahead == 'r') ADVANCE(1045);
      END_STATE();
    case 1142:
      if (lookahead == 'R' ||
          lookahead == 'r') ADVANCE(1071);
      END_STATE();
    case 1143:
      if (lookahead == 'R' ||
          lookahead == 'r') ADVANCE(1144);
      END_STATE();
    case 1144:
      if (lookahead == 'R' ||
          lookahead == 'r') ADVANCE(1165);
      END_STATE();
    case 1145:
      if (lookahead == 'R' ||
          lookahead == 'r') ADVANCE(1079);
      END_STATE();
    case 1146:
      if (lookahead == 'R' ||
          lookahead == 'r') ADVANCE(1131);
      END_STATE();
    case 1147:
      if (lookahead == 'R' ||
          lookahead == 'r') ADVANCE(1085);
      END_STATE();
    case 1148:
      if (lookahead == 'S' ||
          lookahead == 's') ADVANCE(1072);
      END_STATE();
    case 1149:
      if (lookahead == 'T' ||
          lookahead == 't') ADVANCE(1373);
      END_STATE();
    case 1150:
      if (lookahead == 'T' ||
          lookahead == 't') ADVANCE(1384);
      END_STATE();
    case 1151:
      if (lookahead == 'T' ||
          lookahead == 't') ADVANCE(1132);
      END_STATE();
    case 1152:
      if (lookahead == 'T' ||
          lookahead == 't') ADVANCE(1055);
      END_STATE();
    case 1153:
      if (lookahead == 'T' ||
          lookahead == 't') ADVANCE(1044);
      END_STATE();
    case 1154:
      if (lookahead == 'T' ||
          lookahead == 't') ADVANCE(1073);
      END_STATE();
    case 1155:
      if (lookahead == 'U' ||
          lookahead == 'u') ADVANCE(1137);
      END_STATE();
    case 1156:
      if (lookahead == 'U' ||
          lookahead == 'u') ADVANCE(1089);
      END_STATE();
    case 1157:
      if (lookahead == 'U' ||
          lookahead == 'u') ADVANCE(1143);
      END_STATE();
    case 1158:
      if (lookahead == 'U' ||
          lookahead == 'u') ADVANCE(1069);
      END_STATE();
    case 1159:
      if (lookahead == 'U' ||
          lookahead == 'u') ADVANCE(1139);
      END_STATE();
    case 1160:
      if (lookahead == 'U' ||
          lookahead == 'u') ADVANCE(1136);
      END_STATE();
    case 1161:
      if (lookahead == 'U' ||
          lookahead == 'u') ADVANCE(1072);
      END_STATE();
    case 1162:
      if (lookahead == 'V' ||
          lookahead == 'v') ADVANCE(1086);
      END_STATE();
    case 1163:
      if (lookahead == 'W' ||
          lookahead == 'w') ADVANCE(1373);
      END_STATE();
    case 1164:
      if (lookahead == 'Y' ||
          lookahead == 'y') ADVANCE(1368);
      END_STATE();
    case 1165:
      if (lookahead == 'Y' ||
          lookahead == 'y') ADVANCE(1373);
      END_STATE();
    case 1166:
      if (lookahead == 'Y' ||
          lookahead == 'y') ADVANCE(1092);
      END_STATE();
    case 1167:
      if (lookahead == 'Y' ||
          lookahead == 'y') ADVANCE(1384);
      END_STATE();
    case 1168:
      if (lookahead == 'Y' ||
          lookahead == 'y') ADVANCE(1074);
      END_STATE();
    case 1169:
      if (lookahead == 'Y' ||
          lookahead == 'y') ADVANCE(1047);
      END_STATE();
    case 1170:
      if (lookahead == 'Y' ||
          lookahead == 'y') ADVANCE(1050);
      END_STATE();
    case 1171:
      if (lookahead == 'Y' ||
          lookahead == 'y') ADVANCE(1061);
      END_STATE();
    case 1172:
      if (lookahead == 'K' ||
          lookahead == 'k' ||
          lookahead == 0x212a) ADVANCE(1099);
      END_STATE();
    case 1173:
      if (lookahead == 'K' ||
          lookahead == 'k' ||
          lookahead == 0x212a) ADVANCE(1101);
      END_STATE();
    case 1174:
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(1395);
      if (('A' <= lookahead && lookahead <= 'F') ||
          ('a' <= lookahead && lookahead <= 'f')) ADVANCE(1339);
      if (('G' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('g' <= lookahead && lookahead <= 'z')) ADVANCE(1340);
      END_STATE();
    case 1175:
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(1400);
      END_STATE();
    case 1176:
      if (('0' <= lookahead && lookahead <= '9') ||
          lookahead == '_') ADVANCE(1352);
      END_STATE();
    case 1177:
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'F') ||
          ('a' <= lookahead && lookahead <= 'f')) ADVANCE(1395);
      END_STATE();
    case 1178:
      if (('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(1340);
      END_STATE();
    case 1179:
      if (('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(1332);
      END_STATE();
    case 1180:
      if (('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(1344);
      END_STATE();
    case 1181:
      if (('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(1331);
      END_STATE();
    case 1182:
      if (('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(1345);
      END_STATE();
    case 1183:
      if (('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(1343);
      END_STATE();
    case 1184:
      if (eof) ADVANCE(1185);
      ADVANCE_MAP(
        '#', 1178,
        '$', 997,
        '&', 1179,
        ',', 1325,
        '-', 1006,
        '.', 1183,
        ':', 1019,
        ';', 1326,
        '=', 1329,
        '@', 1024,
        '{', 1346,
        '}', 1347,
      );
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(1184);
      if (('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(1324);
      END_STATE();
    case 1185:
      ACCEPT_TOKEN(ts_builtin_sym_end);
      END_STATE();
    case 1186:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == '%') ADVANCE(1353);
      if (lookahead == '.') ADVANCE(1176);
      if (lookahead == 'p') ADVANCE(1236);
      if (('0' <= lookahead && lookahead <= '9') ||
          lookahead == '_') ADVANCE(1186);
      if (('A' <= lookahead && lookahead <= 'Z') ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(1324);
      END_STATE();
    case 1187:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == ':') ADVANCE(1367);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(1324);
      END_STATE();
    case 1188:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == ':') ADVANCE(1383);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(1324);
      END_STATE();
    case 1189:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == ':') ADVANCE(1010);
      if (lookahead == 'i') ADVANCE(1203);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(1324);
      END_STATE();
    case 1190:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == ':') ADVANCE(1010);
      if (lookahead == 's') ADVANCE(1191);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(1324);
      END_STATE();
    case 1191:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == ':') ADVANCE(1010);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(1324);
      END_STATE();
    case 1192:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == ':') ADVANCE(1370);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(1324);
      END_STATE();
    case 1193:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == ':') ADVANCE(1372);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(1324);
      END_STATE();
    case 1194:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == ':') ADVANCE(1011);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(1324);
      END_STATE();
    case 1195:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == ':') ADVANCE(1012);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(1324);
      END_STATE();
    case 1196:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'a') ADVANCE(1223);
      if (lookahead == 'g') ADVANCE(1197);
      if (lookahead == 'h') ADVANCE(1232);
      if (lookahead == 't') ADVANCE(1205);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('b' <= lookahead && lookahead <= 'z')) ADVANCE(1324);
      END_STATE();
    case 1197:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'a') ADVANCE(1213);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('b' <= lookahead && lookahead <= 'z')) ADVANCE(1324);
      END_STATE();
    case 1198:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'a') ADVANCE(1225);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('b' <= lookahead && lookahead <= 'z')) ADVANCE(1324);
      END_STATE();
    case 1199:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'b') ADVANCE(1237);
      if (lookahead == 'E' ||
          lookahead == 'e') ADVANCE(1255);
      if (lookahead == 'O' ||
          lookahead == 'o') ADVANCE(1312);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(1324);
      END_STATE();
    case 1200:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'b') ADVANCE(1237);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(1324);
      END_STATE();
    case 1201:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'b') ADVANCE(1191);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(1324);
      END_STATE();
    case 1202:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'd') ADVANCE(1194);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(1324);
      END_STATE();
    case 1203:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'd') ADVANCE(1195);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(1324);
      END_STATE();
    case 1204:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'e') ADVANCE(1198);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(1324);
      END_STATE();
    case 1205:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'e') ADVANCE(1212);
      if (lookahead == 'h') ADVANCE(1235);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(1324);
      END_STATE();
    case 1206:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'e') ADVANCE(1229);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(1324);
      END_STATE();
    case 1207:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'e') ADVANCE(1233);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(1324);
      END_STATE();
    case 1208:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'e') ADVANCE(1217);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(1324);
      END_STATE();
    case 1209:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'i') ADVANCE(1202);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(1324);
      END_STATE();
    case 1210:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'm') ADVANCE(1396);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(1324);
      END_STATE();
    case 1211:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'm') ADVANCE(1201);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(1324);
      END_STATE();
    case 1212:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'm') ADVANCE(1221);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(1324);
      END_STATE();
    case 1213:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'm') ADVANCE(1204);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(1324);
      END_STATE();
    case 1214:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'n') ADVANCE(1234);
      if (lookahead == 'M' ||
          lookahead == 'm') ADVANCE(1261);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(1324);
      END_STATE();
    case 1215:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'n') ADVANCE(1234);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(1324);
      END_STATE();
    case 1216:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'n') ADVANCE(1227);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(1324);
      END_STATE();
    case 1217:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'n') ADVANCE(1228);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(1324);
      END_STATE();
    case 1218:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'o') ADVANCE(1216);
      if (lookahead == 'S' ||
          lookahead == 's') ADVANCE(1311);
      if (lookahead == 'Y' ||
          lookahead == 'y') ADVANCE(1247);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(1324);
      END_STATE();
    case 1219:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'o') ADVANCE(1216);
      if (lookahead == 'S' ||
          lookahead == 's') ADVANCE(1311);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(1324);
      END_STATE();
    case 1220:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'p') ADVANCE(1190);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(1324);
      END_STATE();
    case 1221:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'p') ADVANCE(1191);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(1324);
      END_STATE();
    case 1222:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 's') ADVANCE(1206);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(1324);
      END_STATE();
    case 1223:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 's') ADVANCE(1222);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(1324);
      END_STATE();
    case 1224:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 's') ADVANCE(1207);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(1324);
      END_STATE();
    case 1225:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 's') ADVANCE(1224);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(1324);
      END_STATE();
    case 1226:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 't') ADVANCE(1220);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(1324);
      END_STATE();
    case 1227:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 't') ADVANCE(1208);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(1324);
      END_STATE();
    case 1228:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 't') ADVANCE(1209);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(1324);
      END_STATE();
    case 1229:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 't') ADVANCE(1189);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(1324);
      END_STATE();
    case 1230:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 't') ADVANCE(1226);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(1324);
      END_STATE();
    case 1231:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 't') ADVANCE(1221);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(1324);
      END_STATE();
    case 1232:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 't') ADVANCE(1231);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(1324);
      END_STATE();
    case 1233:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 't') ADVANCE(1191);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(1324);
      END_STATE();
    case 1234:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'u') ADVANCE(1210);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(1324);
      END_STATE();
    case 1235:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'u') ADVANCE(1211);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(1324);
      END_STATE();
    case 1236:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'x') ADVANCE(1324);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(1324);
      END_STATE();
    case 1237:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'x') ADVANCE(1196);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(1324);
      END_STATE();
    case 1238:
      ACCEPT_TOKEN(sym_identifier);
      ADVANCE_MAP(
        '!', 1334,
        0x212a, 1098,
        'K', 1276,
        'k', 1276,
        'L', 1245,
        'l', 1245,
        'T', 1302,
        't', 1302,
      );
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(1324);
      END_STATE();
    case 1239:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 0x212a) ADVANCE(1368);
      if (lookahead == 'K' ||
          lookahead == 'k') ADVANCE(1324);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(1324);
      END_STATE();
    case 1240:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 0x212a) ADVANCE(1097);
      if (lookahead == 'K' ||
          lookahead == 'k') ADVANCE(1277);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(1324);
      END_STATE();
    case 1241:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 0x212a) ADVANCE(1018);
      if (lookahead == 'K' ||
          lookahead == 'k') ADVANCE(1193);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(1324);
      END_STATE();
    case 1242:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'A' ||
          lookahead == 'a') ADVANCE(1324);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('B' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('b' <= lookahead && lookahead <= 'z')) ADVANCE(1324);
      END_STATE();
    case 1243:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'A' ||
          lookahead == 'a') ADVANCE(1281);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('B' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('b' <= lookahead && lookahead <= 'z')) ADVANCE(1324);
      END_STATE();
    case 1244:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'A' ||
          lookahead == 'a') ADVANCE(1323);
      if (lookahead == 'E' ||
          lookahead == 'e') ADVANCE(1265);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('B' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('b' <= lookahead && lookahead <= 'z')) ADVANCE(1324);
      END_STATE();
    case 1245:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'A' ||
          lookahead == 'a') ADVANCE(1315);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('B' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('b' <= lookahead && lookahead <= 'z')) ADVANCE(1324);
      END_STATE();
    case 1246:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'A' ||
          lookahead == 'a') ADVANCE(1280);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('B' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('b' <= lookahead && lookahead <= 'z')) ADVANCE(1324);
      END_STATE();
    case 1247:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'A' ||
          lookahead == 'a') ADVANCE(1291);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('B' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('b' <= lookahead && lookahead <= 'z')) ADVANCE(1324);
      END_STATE();
    case 1248:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'A' ||
          lookahead == 'a') ADVANCE(1294);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('B' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('b' <= lookahead && lookahead <= 'z')) ADVANCE(1324);
      END_STATE();
    case 1249:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'A' ||
          lookahead == 'a') ADVANCE(1253);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('B' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('b' <= lookahead && lookahead <= 'z')) ADVANCE(1324);
      END_STATE();
    case 1250:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'B' ||
          lookahead == 'b') ADVANCE(1263);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(1324);
      END_STATE();
    case 1251:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'C' ||
          lookahead == 'c') ADVANCE(1324);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(1324);
      END_STATE();
    case 1252:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'C' ||
          lookahead == 'c') ADVANCE(1270);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(1324);
      END_STATE();
    case 1253:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'C' ||
          lookahead == 'c') ADVANCE(1269);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(1324);
      END_STATE();
    case 1254:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'C' ||
          lookahead == 'c') ADVANCE(1241);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(1324);
      END_STATE();
    case 1255:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'D' ||
          lookahead == 'd') ADVANCE(1324);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(1324);
      END_STATE();
    case 1256:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'D' ||
          lookahead == 'd') ADVANCE(1272);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(1324);
      END_STATE();
    case 1257:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'E' ||
          lookahead == 'e') ADVANCE(1324);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(1324);
      END_STATE();
    case 1258:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'E' ||
          lookahead == 'e') ADVANCE(1255);
      if (lookahead == 'O' ||
          lookahead == 'o') ADVANCE(1312);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(1324);
      END_STATE();
    case 1259:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'E' ||
          lookahead == 'e') ADVANCE(1249);
      if (lookahead == 'I' ||
          lookahead == 'i') ADVANCE(1292);
      if (lookahead == 'U' ||
          lookahead == 'u') ADVANCE(1305);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(1324);
      END_STATE();
    case 1260:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'E' ||
          lookahead == 'e') ADVANCE(1286);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(1324);
      END_STATE();
    case 1261:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'E' ||
          lookahead == 'e') ADVANCE(1308);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(1324);
      END_STATE();
    case 1262:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'E' ||
          lookahead == 'e') ADVANCE(1318);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(1324);
      END_STATE();
    case 1263:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'E' ||
          lookahead == 'e') ADVANCE(1304);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(1324);
      END_STATE();
    case 1264:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'E' ||
          lookahead == 'e') ADVANCE(1314);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(1324);
      END_STATE();
    case 1265:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'E' ||
          lookahead == 'e') ADVANCE(1291);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(1324);
      END_STATE();
    case 1266:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'E' ||
          lookahead == 'e') ADVANCE(1246);
      if (lookahead == 'W' ||
          lookahead == 'w') ADVANCE(1187);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(1324);
      END_STATE();
    case 1267:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'G' ||
          lookahead == 'g') ADVANCE(1257);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(1324);
      END_STATE();
    case 1268:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'G' ||
          lookahead == 'g') ADVANCE(1298);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(1324);
      END_STATE();
    case 1269:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'H' ||
          lookahead == 'h') ADVANCE(1324);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(1324);
      END_STATE();
    case 1270:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'H' ||
          lookahead == 'h') ADVANCE(1313);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(1324);
      END_STATE();
    case 1271:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'I' ||
          lookahead == 'i') ADVANCE(1320);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(1324);
      END_STATE();
    case 1272:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'I' ||
          lookahead == 'i') ADVANCE(1268);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(1324);
      END_STATE();
    case 1273:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'I' ||
          lookahead == 'i') ADVANCE(1254);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(1324);
      END_STATE();
    case 1274:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'I' ||
          lookahead == 'i') ADVANCE(1301);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(1324);
      END_STATE();
    case 1275:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'I' ||
          lookahead == 'i') ADVANCE(1295);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(1324);
      END_STATE();
    case 1276:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'I' ||
          lookahead == 'i') ADVANCE(1296);
      if (lookahead == 'Y' ||
          lookahead == 'y') ADVANCE(1324);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(1324);
      END_STATE();
    case 1277:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'I' ||
          lookahead == 'i') ADVANCE(1296);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(1324);
      END_STATE();
    case 1278:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'I' ||
          lookahead == 'i') ADVANCE(1242);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(1324);
      END_STATE();
    case 1279:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'I' ||
          lookahead == 'i') ADVANCE(1290);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(1324);
      END_STATE();
    case 1280:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'L' ||
          lookahead == 'l') ADVANCE(1324);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(1324);
      END_STATE();
    case 1281:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'L' ||
          lookahead == 'l') ADVANCE(1255);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(1324);
      END_STATE();
    case 1282:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'L' ||
          lookahead == 'l') ADVANCE(1319);
      if (lookahead == 'R' ||
          lookahead == 'r') ADVANCE(1273);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(1324);
      END_STATE();
    case 1283:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'L' ||
          lookahead == 'l') ADVANCE(1300);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(1324);
      END_STATE();
    case 1284:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'L' ||
          lookahead == 'l') ADVANCE(1257);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(1324);
      END_STATE();
    case 1285:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'L' ||
          lookahead == 'l') ADVANCE(1271);
      if (lookahead == 'R' ||
          lookahead == 'r') ADVANCE(1248);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(1324);
      END_STATE();
    case 1286:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'L' ||
          lookahead == 'l') ADVANCE(1283);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(1324);
      END_STATE();
    case 1287:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'L' ||
          lookahead == 'l') ADVANCE(1264);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(1324);
      END_STATE();
    case 1288:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'M' ||
          lookahead == 'm') ADVANCE(1250);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(1324);
      END_STATE();
    case 1289:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'M' ||
          lookahead == 'm') ADVANCE(1261);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(1324);
      END_STATE();
    case 1290:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'M' ||
          lookahead == 'm') ADVANCE(1257);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(1324);
      END_STATE();
    case 1291:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'N' ||
          lookahead == 'n') ADVANCE(1324);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(1324);
      END_STATE();
    case 1292:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'N' ||
          lookahead == 'n') ADVANCE(1239);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(1324);
      END_STATE();
    case 1293:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'N' ||
          lookahead == 'n') ADVANCE(1256);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(1324);
      END_STATE();
    case 1294:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'N' ||
          lookahead == 'n') ADVANCE(1267);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(1324);
      END_STATE();
    case 1295:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'N' ||
          lookahead == 'n') ADVANCE(1251);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(1324);
      END_STATE();
    case 1296:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'N' ||
          lookahead == 'n') ADVANCE(1192);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(1324);
      END_STATE();
    case 1297:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'N' ||
          lookahead == 'n') ADVANCE(1257);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(1324);
      END_STATE();
    case 1298:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'O' ||
          lookahead == 'o') ADVANCE(1324);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(1324);
      END_STATE();
    case 1299:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'O' ||
          lookahead == 'o') ADVANCE(1281);
      if (lookahead == 'R' ||
          lookahead == 'r') ADVANCE(1244);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(1324);
      END_STATE();
    case 1300:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'O' ||
          lookahead == 'o') ADVANCE(1321);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(1324);
      END_STATE();
    case 1301:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'O' ||
          lookahead == 'o') ADVANCE(1287);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(1324);
      END_STATE();
    case 1302:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'O' ||
          lookahead == 'o') ADVANCE(1297);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(1324);
      END_STATE();
    case 1303:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'P' ||
          lookahead == 'p') ADVANCE(1284);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(1324);
      END_STATE();
    case 1304:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'R' ||
          lookahead == 'r') ADVANCE(1324);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(1324);
      END_STATE();
    case 1305:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'R' ||
          lookahead == 'r') ADVANCE(1303);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(1324);
      END_STATE();
    case 1306:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'R' ||
          lookahead == 'r') ADVANCE(1273);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(1324);
      END_STATE();
    case 1307:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'R' ||
          lookahead == 'r') ADVANCE(1246);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(1324);
      END_STATE();
    case 1308:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'R' ||
          lookahead == 'r') ADVANCE(1243);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(1324);
      END_STATE();
    case 1309:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'S' ||
          lookahead == 's') ADVANCE(1311);
      if (lookahead == 'Y' ||
          lookahead == 'y') ADVANCE(1247);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(1324);
      END_STATE();
    case 1310:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'S' ||
          lookahead == 's') ADVANCE(1311);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(1324);
      END_STATE();
    case 1311:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'S' ||
          lookahead == 's') ADVANCE(1188);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(1324);
      END_STATE();
    case 1312:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'S' ||
          lookahead == 's') ADVANCE(1257);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(1324);
      END_STATE();
    case 1313:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'S' ||
          lookahead == 's') ADVANCE(1278);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(1324);
      END_STATE();
    case 1314:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'T' ||
          lookahead == 't') ADVANCE(1324);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(1324);
      END_STATE();
    case 1315:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'T' ||
          lookahead == 't') ADVANCE(1257);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(1324);
      END_STATE();
    case 1316:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'T' ||
          lookahead == 't') ADVANCE(1307);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(1324);
      END_STATE();
    case 1317:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'U' ||
          lookahead == 'u') ADVANCE(1252);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(1324);
      END_STATE();
    case 1318:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'U' ||
          lookahead == 'u') ADVANCE(1316);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(1324);
      END_STATE();
    case 1319:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'U' ||
          lookahead == 'u') ADVANCE(1257);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(1324);
      END_STATE();
    case 1320:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'V' ||
          lookahead == 'v') ADVANCE(1257);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(1324);
      END_STATE();
    case 1321:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'W' ||
          lookahead == 'w') ADVANCE(1324);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(1324);
      END_STATE();
    case 1322:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'W' ||
          lookahead == 'w') ADVANCE(1187);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(1324);
      END_STATE();
    case 1323:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'Y' ||
          lookahead == 'y') ADVANCE(1324);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(1324);
      END_STATE();
    case 1324:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == '!') ADVANCE(1334);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(1324);
      END_STATE();
    case 1325:
      ACCEPT_TOKEN(sym_comma);
      END_STATE();
    case 1326:
      ACCEPT_TOKEN(sym_semi_colon);
      END_STATE();
    case 1327:
      ACCEPT_TOKEN(anon_sym_COLON);
      END_STATE();
    case 1328:
      ACCEPT_TOKEN(anon_sym_COLON);
      if (lookahead == ':') ADVANCE(1182);
      if (('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(1344);
      END_STATE();
    case 1329:
      ACCEPT_TOKEN(sym_equals);
      END_STATE();
    case 1330:
      ACCEPT_TOKEN(sym_token);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(1330);
      END_STATE();
    case 1331:
      ACCEPT_TOKEN(sym_static_token);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(1331);
      END_STATE();
    case 1332:
      ACCEPT_TOKEN(sym_static_argument);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(1332);
      END_STATE();
    case 1333:
      ACCEPT_TOKEN(anon_sym_ATmacro);
      END_STATE();
    case 1334:
      ACCEPT_TOKEN(aux_sym_macro_call_token1);
      END_STATE();
    case 1335:
      ACCEPT_TOKEN(aux_sym_macro_call_token1);
      if (lookahead != 0 &&
          lookahead != '\n' &&
          lookahead != '\r') ADVANCE(1406);
      END_STATE();
    case 1336:
      ACCEPT_TOKEN(anon_sym_ATpriority);
      END_STATE();
    case 1337:
      ACCEPT_TOKEN(anon_sym_ATname);
      END_STATE();
    case 1338:
      ACCEPT_TOKEN(anon_sym_ATderive);
      END_STATE();
    case 1339:
      ACCEPT_TOKEN(sym_name_selector);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'F') ||
          ('a' <= lookahead && lookahead <= 'f')) ADVANCE(1339);
      if (('G' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('g' <= lookahead && lookahead <= 'z')) ADVANCE(1340);
      END_STATE();
    case 1340:
      ACCEPT_TOKEN(sym_name_selector);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(1340);
      END_STATE();
    case 1341:
      ACCEPT_TOKEN(sym_tag_selector);
      if (lookahead == '%') ADVANCE(1353);
      if (lookahead == 'p') ADVANCE(1342);
      if (('0' <= lookahead && lookahead <= '9') ||
          lookahead == '_') ADVANCE(1341);
      if (('A' <= lookahead && lookahead <= 'Z') ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(1343);
      END_STATE();
    case 1342:
      ACCEPT_TOKEN(sym_tag_selector);
      if (lookahead == 'x') ADVANCE(1343);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(1343);
      END_STATE();
    case 1343:
      ACCEPT_TOKEN(sym_tag_selector);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(1343);
      END_STATE();
    case 1344:
      ACCEPT_TOKEN(aux_sym_state_selector_token1);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(1344);
      END_STATE();
    case 1345:
      ACCEPT_TOKEN(sym_pseudo_selector);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(1345);
      END_STATE();
    case 1346:
      ACCEPT_TOKEN(sym_scope_open);
      END_STATE();
    case 1347:
      ACCEPT_TOKEN(sym_scope_close);
      END_STATE();
    case 1348:
      ACCEPT_TOKEN(sym_tuple_open);
      END_STATE();
    case 1349:
      ACCEPT_TOKEN(sym_tuple_close);
      END_STATE();
    case 1350:
      ACCEPT_TOKEN(sym_number_ambiguous);
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == '%') ADVANCE(1353);
      if (lookahead == '.') ADVANCE(1176);
      if (lookahead == 'p') ADVANCE(1);
      if (('0' <= lookahead && lookahead <= '9') ||
          lookahead == '_') ADVANCE(1350);
      if (('A' <= lookahead && lookahead <= 'Z') ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 1351:
      ACCEPT_TOKEN(sym_number_ambiguous);
      if (lookahead == '%') ADVANCE(1353);
      if (lookahead == '.') ADVANCE(1176);
      if (lookahead == 'p') ADVANCE(1041);
      if (('0' <= lookahead && lookahead <= '9') ||
          lookahead == '_') ADVANCE(1351);
      END_STATE();
    case 1352:
      ACCEPT_TOKEN(sym_number_ambiguous);
      if (lookahead == '%') ADVANCE(1353);
      if (lookahead == 'p') ADVANCE(1041);
      if (('0' <= lookahead && lookahead <= '9') ||
          lookahead == '_') ADVANCE(1352);
      END_STATE();
    case 1353:
      ACCEPT_TOKEN(sym_number_percent);
      END_STATE();
    case 1354:
      ACCEPT_TOKEN(sym_number_offset);
      END_STATE();
    case 1355:
      ACCEPT_TOKEN(sym_number_offset);
      if (lookahead == '!') ADVANCE(1334);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 1356:
      ACCEPT_TOKEN(sym_pow_operator);
      END_STATE();
    case 1357:
      ACCEPT_TOKEN(sym_div_operator);
      if (lookahead == '/') ADVANCE(1358);
      END_STATE();
    case 1358:
      ACCEPT_TOKEN(sym_floor_div_operator);
      END_STATE();
    case 1359:
      ACCEPT_TOKEN(sym_mod_operator);
      END_STATE();
    case 1360:
      ACCEPT_TOKEN(sym_mult_operator);
      END_STATE();
    case 1361:
      ACCEPT_TOKEN(sym_add_operator);
      END_STATE();
    case 1362:
      ACCEPT_TOKEN(sym_sub_operator);
      if (lookahead == '-') ADVANCE(1401);
      END_STATE();
    case 1363:
      ACCEPT_TOKEN(sym_single_quote_string);
      END_STATE();
    case 1364:
      ACCEPT_TOKEN(sym_double_quote_string);
      END_STATE();
    case 1365:
      ACCEPT_TOKEN(sym__color_code_shade);
      END_STATE();
    case 1366:
      ACCEPT_TOKEN(sym__color_code_shade);
      if (lookahead == '0') ADVANCE(1365);
      END_STATE();
    case 1367:
      ACCEPT_TOKEN(aux_sym_color_code_tailwind_token1);
      END_STATE();
    case 1368:
      ACCEPT_TOKEN(aux_sym_color_code_tailwind_token2);
      END_STATE();
    case 1369:
      ACCEPT_TOKEN(aux_sym_color_code_tailwind_token2);
      if (lookahead == '!') ADVANCE(1334);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 1370:
      ACCEPT_TOKEN(aux_sym_color_code_skin_token1);
      END_STATE();
    case 1371:
      ACCEPT_TOKEN(aux_sym_color_code_skin_token2);
      if (lookahead == '!') ADVANCE(1334);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 1372:
      ACCEPT_TOKEN(aux_sym_color_code_brick_token1);
      END_STATE();
    case 1373:
      ACCEPT_TOKEN(aux_sym_color_code_brick_token2);
      END_STATE();
    case 1374:
      ACCEPT_TOKEN(aux_sym_color_code_brick_token2);
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'B' ||
          lookahead == 'b') ADVANCE(775);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 1375:
      ACCEPT_TOKEN(aux_sym_color_code_brick_token2);
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'D' ||
          lookahead == 'd') ADVANCE(435);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 1376:
      ACCEPT_TOKEN(aux_sym_color_code_brick_token2);
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'D' ||
          lookahead == 'd') ADVANCE(437);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 1377:
      ACCEPT_TOKEN(aux_sym_color_code_brick_token2);
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'D' ||
          lookahead == 'd') ADVANCE(443);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 1378:
      ACCEPT_TOKEN(aux_sym_color_code_brick_token2);
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'F' ||
          lookahead == 'f') ADVANCE(519);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 1379:
      ACCEPT_TOKEN(aux_sym_color_code_brick_token2);
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'G' ||
          lookahead == 'g') ADVANCE(785);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 1380:
      ACCEPT_TOKEN(aux_sym_color_code_brick_token2);
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'M' ||
          lookahead == 'm') ADVANCE(270);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 1381:
      ACCEPT_TOKEN(aux_sym_color_code_brick_token2);
      if (lookahead == '!') ADVANCE(1334);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 1382:
      ACCEPT_TOKEN(aux_sym_color_code_brick_token2);
      if (lookahead == 'M' ||
          lookahead == 'm') ADVANCE(1075);
      END_STATE();
    case 1383:
      ACCEPT_TOKEN(aux_sym_color_code_css_token1);
      END_STATE();
    case 1384:
      ACCEPT_TOKEN(aux_sym_color_code_css_token2);
      END_STATE();
    case 1385:
      ACCEPT_TOKEN(aux_sym_color_code_css_token2);
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'B' ||
          lookahead == 'b') ADVANCE(524);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 1386:
      ACCEPT_TOKEN(aux_sym_color_code_css_token2);
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'D' ||
          lookahead == 'd') ADVANCE(786);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 1387:
      ACCEPT_TOKEN(aux_sym_color_code_css_token2);
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'E' ||
          lookahead == 'e') ADVANCE(640);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 1388:
      ACCEPT_TOKEN(aux_sym_color_code_css_token2);
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'G' ||
          lookahead == 'g') ADVANCE(796);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 1389:
      ACCEPT_TOKEN(aux_sym_color_code_css_token2);
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'M' ||
          lookahead == 'm') ADVANCE(106);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 1390:
      ACCEPT_TOKEN(aux_sym_color_code_css_token2);
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'R' ||
          lookahead == 'r') ADVANCE(271);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 1391:
      ACCEPT_TOKEN(aux_sym_color_code_css_token2);
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'S' ||
          lookahead == 's') ADVANCE(593);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 1392:
      ACCEPT_TOKEN(aux_sym_color_code_css_token2);
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'V' ||
          lookahead == 'v') ADVANCE(479);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 1393:
      ACCEPT_TOKEN(aux_sym_color_code_css_token2);
      if (lookahead == '!') ADVANCE(1334);
      if (lookahead == 'Y' ||
          lookahead == 'y') ADVANCE(316);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 1394:
      ACCEPT_TOKEN(aux_sym_color_code_css_token2);
      if (lookahead == '!') ADVANCE(1334);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(996);
      END_STATE();
    case 1395:
      ACCEPT_TOKEN(sym_color_hex);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'F') ||
          ('a' <= lookahead && lookahead <= 'f')) ADVANCE(1395);
      END_STATE();
    case 1396:
      ACCEPT_TOKEN(anon_sym_Enum);
      if (lookahead == '!') ADVANCE(1334);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(1324);
      END_STATE();
    case 1397:
      ACCEPT_TOKEN(anon_sym_DOT);
      END_STATE();
    case 1398:
      ACCEPT_TOKEN(sym_rbx_asset);
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(1398);
      END_STATE();
    case 1399:
      ACCEPT_TOKEN(sym_rbx_asset);
      if (lookahead != 0 &&
          lookahead != ' ' &&
          lookahead != ')') ADVANCE(1399);
      END_STATE();
    case 1400:
      ACCEPT_TOKEN(sym_rbx_content);
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(1400);
      END_STATE();
    case 1401:
      ACCEPT_TOKEN(anon_sym_DASH_DASH);
      END_STATE();
    case 1402:
      ACCEPT_TOKEN(anon_sym_DASH_DASH);
      if (lookahead != 0 &&
          lookahead != '\n' &&
          lookahead != '\r') ADVANCE(1406);
      END_STATE();
    case 1403:
      ACCEPT_TOKEN(aux_sym_comment_token1);
      if (lookahead == '!') ADVANCE(1335);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(1403);
      if (lookahead != 0 &&
          lookahead != '\n' &&
          lookahead != '\r') ADVANCE(1406);
      END_STATE();
    case 1404:
      ACCEPT_TOKEN(aux_sym_comment_token1);
      if (lookahead == '-') ADVANCE(1405);
      if (lookahead == '\t' ||
          lookahead == 0x0b ||
          lookahead == '\f' ||
          lookahead == ' ') ADVANCE(1404);
      if (('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(1403);
      if (lookahead != 0 &&
          (lookahead < '\t' || '\r' < lookahead)) ADVANCE(1406);
      END_STATE();
    case 1405:
      ACCEPT_TOKEN(aux_sym_comment_token1);
      if (lookahead == '-') ADVANCE(1402);
      if (lookahead != 0 &&
          lookahead != '\n' &&
          lookahead != '\r') ADVANCE(1406);
      END_STATE();
    case 1406:
      ACCEPT_TOKEN(aux_sym_comment_token1);
      if (lookahead != 0 &&
          lookahead != '\n' &&
          lookahead != '\r') ADVANCE(1406);
      END_STATE();
    default:
      return false;
  }
}

static const TSLexerMode ts_lex_modes[STATE_COUNT] = {
  [0] = {.lex_state = 0, .external_lex_state = 1},
  [1] = {.lex_state = 1184, .external_lex_state = 2},
  [2] = {.lex_state = 998, .external_lex_state = 3},
  [3] = {.lex_state = 998, .external_lex_state = 3},
  [4] = {.lex_state = 998, .external_lex_state = 3},
  [5] = {.lex_state = 998, .external_lex_state = 3},
  [6] = {.lex_state = 998, .external_lex_state = 3},
  [7] = {.lex_state = 998, .external_lex_state = 3},
  [8] = {.lex_state = 998, .external_lex_state = 3},
  [9] = {.lex_state = 998, .external_lex_state = 3},
  [10] = {.lex_state = 998, .external_lex_state = 3},
  [11] = {.lex_state = 998, .external_lex_state = 3},
  [12] = {.lex_state = 998, .external_lex_state = 3},
  [13] = {.lex_state = 998, .external_lex_state = 3},
  [14] = {.lex_state = 998, .external_lex_state = 3},
  [15] = {.lex_state = 998, .external_lex_state = 3},
  [16] = {.lex_state = 998, .external_lex_state = 3},
  [17] = {.lex_state = 998, .external_lex_state = 3},
  [18] = {.lex_state = 998, .external_lex_state = 3},
  [19] = {.lex_state = 999, .external_lex_state = 3},
  [20] = {.lex_state = 998, .external_lex_state = 3},
  [21] = {.lex_state = 998, .external_lex_state = 3},
  [22] = {.lex_state = 999, .external_lex_state = 3},
  [23] = {.lex_state = 998, .external_lex_state = 3},
  [24] = {.lex_state = 998, .external_lex_state = 3},
  [25] = {.lex_state = 998, .external_lex_state = 3},
  [26] = {.lex_state = 998, .external_lex_state = 3},
  [27] = {.lex_state = 998, .external_lex_state = 3},
  [28] = {.lex_state = 998, .external_lex_state = 3},
  [29] = {.lex_state = 998, .external_lex_state = 3},
  [30] = {.lex_state = 998, .external_lex_state = 3},
  [31] = {.lex_state = 998, .external_lex_state = 3},
  [32] = {.lex_state = 998, .external_lex_state = 3},
  [33] = {.lex_state = 998, .external_lex_state = 3},
  [34] = {.lex_state = 998, .external_lex_state = 3},
  [35] = {.lex_state = 998, .external_lex_state = 3},
  [36] = {.lex_state = 998, .external_lex_state = 3},
  [37] = {.lex_state = 998, .external_lex_state = 3},
  [38] = {.lex_state = 998, .external_lex_state = 3},
  [39] = {.lex_state = 999, .external_lex_state = 3},
  [40] = {.lex_state = 998, .external_lex_state = 3},
  [41] = {.lex_state = 998, .external_lex_state = 3},
  [42] = {.lex_state = 998, .external_lex_state = 3},
  [43] = {.lex_state = 998, .external_lex_state = 3},
  [44] = {.lex_state = 999, .external_lex_state = 3},
  [45] = {.lex_state = 998, .external_lex_state = 3},
  [46] = {.lex_state = 998, .external_lex_state = 3},
  [47] = {.lex_state = 998, .external_lex_state = 3},
  [48] = {.lex_state = 998, .external_lex_state = 3},
  [49] = {.lex_state = 998, .external_lex_state = 3},
  [50] = {.lex_state = 998, .external_lex_state = 3},
  [51] = {.lex_state = 998, .external_lex_state = 3},
  [52] = {.lex_state = 998, .external_lex_state = 3},
  [53] = {.lex_state = 998, .external_lex_state = 3},
  [54] = {.lex_state = 998, .external_lex_state = 3},
  [55] = {.lex_state = 998, .external_lex_state = 3},
  [56] = {.lex_state = 998, .external_lex_state = 3},
  [57] = {.lex_state = 998, .external_lex_state = 3},
  [58] = {.lex_state = 998, .external_lex_state = 3},
  [59] = {.lex_state = 998, .external_lex_state = 3},
  [60] = {.lex_state = 998, .external_lex_state = 3},
  [61] = {.lex_state = 998, .external_lex_state = 3},
  [62] = {.lex_state = 998, .external_lex_state = 3},
  [63] = {.lex_state = 998, .external_lex_state = 3},
  [64] = {.lex_state = 998, .external_lex_state = 3},
  [65] = {.lex_state = 998, .external_lex_state = 3},
  [66] = {.lex_state = 998, .external_lex_state = 3},
  [67] = {.lex_state = 998, .external_lex_state = 3},
  [68] = {.lex_state = 998, .external_lex_state = 3},
  [69] = {.lex_state = 998, .external_lex_state = 3},
  [70] = {.lex_state = 998, .external_lex_state = 3},
  [71] = {.lex_state = 998, .external_lex_state = 3},
  [72] = {.lex_state = 998, .external_lex_state = 3},
  [73] = {.lex_state = 999, .external_lex_state = 3},
  [74] = {.lex_state = 998, .external_lex_state = 3},
  [75] = {.lex_state = 999, .external_lex_state = 3},
  [76] = {.lex_state = 999, .external_lex_state = 3},
  [77] = {.lex_state = 998, .external_lex_state = 3},
  [78] = {.lex_state = 999, .external_lex_state = 3},
  [79] = {.lex_state = 999, .external_lex_state = 3},
  [80] = {.lex_state = 999, .external_lex_state = 3},
  [81] = {.lex_state = 999, .external_lex_state = 3},
  [82] = {.lex_state = 999, .external_lex_state = 3},
  [83] = {.lex_state = 999, .external_lex_state = 3},
  [84] = {.lex_state = 999, .external_lex_state = 3},
  [85] = {.lex_state = 999, .external_lex_state = 3},
  [86] = {.lex_state = 999, .external_lex_state = 3},
  [87] = {.lex_state = 999, .external_lex_state = 3},
  [88] = {.lex_state = 999, .external_lex_state = 3},
  [89] = {.lex_state = 999, .external_lex_state = 3},
  [90] = {.lex_state = 999, .external_lex_state = 3},
  [91] = {.lex_state = 999, .external_lex_state = 3},
  [92] = {.lex_state = 999, .external_lex_state = 3},
  [93] = {.lex_state = 999, .external_lex_state = 3},
  [94] = {.lex_state = 999, .external_lex_state = 3},
  [95] = {.lex_state = 999, .external_lex_state = 3},
  [96] = {.lex_state = 999, .external_lex_state = 3},
  [97] = {.lex_state = 999, .external_lex_state = 3},
  [98] = {.lex_state = 999, .external_lex_state = 3},
  [99] = {.lex_state = 999, .external_lex_state = 3},
  [100] = {.lex_state = 999, .external_lex_state = 3},
  [101] = {.lex_state = 999, .external_lex_state = 3},
  [102] = {.lex_state = 999, .external_lex_state = 3},
  [103] = {.lex_state = 999, .external_lex_state = 3},
  [104] = {.lex_state = 999, .external_lex_state = 3},
  [105] = {.lex_state = 999, .external_lex_state = 3},
  [106] = {.lex_state = 999, .external_lex_state = 3},
  [107] = {.lex_state = 1184, .external_lex_state = 2},
  [108] = {.lex_state = 1184, .external_lex_state = 2},
  [109] = {.lex_state = 1184, .external_lex_state = 2},
  [110] = {.lex_state = 1184, .external_lex_state = 2},
  [111] = {.lex_state = 1184, .external_lex_state = 2},
  [112] = {.lex_state = 1184, .external_lex_state = 2},
  [113] = {.lex_state = 1184, .external_lex_state = 2},
  [114] = {.lex_state = 1184, .external_lex_state = 2},
  [115] = {.lex_state = 1184, .external_lex_state = 2},
  [116] = {.lex_state = 1184, .external_lex_state = 2},
  [117] = {.lex_state = 1184, .external_lex_state = 2},
  [118] = {.lex_state = 1184, .external_lex_state = 2},
  [119] = {.lex_state = 1184, .external_lex_state = 2},
  [120] = {.lex_state = 1184, .external_lex_state = 2},
  [121] = {.lex_state = 1184, .external_lex_state = 2},
  [122] = {.lex_state = 1184, .external_lex_state = 2},
  [123] = {.lex_state = 1184, .external_lex_state = 2},
  [124] = {.lex_state = 1184, .external_lex_state = 2},
  [125] = {.lex_state = 999, .external_lex_state = 3},
  [126] = {.lex_state = 999, .external_lex_state = 3},
  [127] = {.lex_state = 1184, .external_lex_state = 2},
  [128] = {.lex_state = 1184, .external_lex_state = 2},
  [129] = {.lex_state = 1184, .external_lex_state = 2},
  [130] = {.lex_state = 1184, .external_lex_state = 2},
  [131] = {.lex_state = 1184, .external_lex_state = 2},
  [132] = {.lex_state = 1184, .external_lex_state = 2},
  [133] = {.lex_state = 1184, .external_lex_state = 2},
  [134] = {.lex_state = 1184, .external_lex_state = 2},
  [135] = {.lex_state = 1184, .external_lex_state = 2},
  [136] = {.lex_state = 1184, .external_lex_state = 2},
  [137] = {.lex_state = 1184, .external_lex_state = 2},
  [138] = {.lex_state = 1184, .external_lex_state = 2},
  [139] = {.lex_state = 1184, .external_lex_state = 2},
  [140] = {.lex_state = 1184, .external_lex_state = 2},
  [141] = {.lex_state = 1184, .external_lex_state = 2},
  [142] = {.lex_state = 1184, .external_lex_state = 2},
  [143] = {.lex_state = 1184, .external_lex_state = 2},
  [144] = {.lex_state = 1184, .external_lex_state = 2},
  [145] = {.lex_state = 1184, .external_lex_state = 2},
  [146] = {.lex_state = 1184, .external_lex_state = 2},
  [147] = {.lex_state = 1184, .external_lex_state = 2},
  [148] = {.lex_state = 1184, .external_lex_state = 2},
  [149] = {.lex_state = 1184, .external_lex_state = 2},
  [150] = {.lex_state = 1184, .external_lex_state = 2},
  [151] = {.lex_state = 1184, .external_lex_state = 2},
  [152] = {.lex_state = 1184, .external_lex_state = 2},
  [153] = {.lex_state = 1184, .external_lex_state = 2},
  [154] = {.lex_state = 1184, .external_lex_state = 2},
  [155] = {.lex_state = 1184, .external_lex_state = 2},
  [156] = {.lex_state = 0, .external_lex_state = 2},
  [157] = {.lex_state = 0, .external_lex_state = 2},
  [158] = {.lex_state = 0, .external_lex_state = 2},
  [159] = {.lex_state = 0, .external_lex_state = 2},
  [160] = {.lex_state = 0, .external_lex_state = 2},
  [161] = {.lex_state = 0, .external_lex_state = 2},
  [162] = {.lex_state = 0, .external_lex_state = 2},
  [163] = {.lex_state = 1001, .external_lex_state = 2},
  [164] = {.lex_state = 1001, .external_lex_state = 2},
  [165] = {.lex_state = 0, .external_lex_state = 2},
  [166] = {.lex_state = 0, .external_lex_state = 2},
  [167] = {.lex_state = 0, .external_lex_state = 2},
  [168] = {.lex_state = 0, .external_lex_state = 2},
  [169] = {.lex_state = 0, .external_lex_state = 2},
  [170] = {.lex_state = 0, .external_lex_state = 2},
  [171] = {.lex_state = 0, .external_lex_state = 2},
  [172] = {.lex_state = 0, .external_lex_state = 2},
  [173] = {.lex_state = 0, .external_lex_state = 2},
  [174] = {.lex_state = 0, .external_lex_state = 2},
  [175] = {.lex_state = 0, .external_lex_state = 2},
  [176] = {.lex_state = 0, .external_lex_state = 2},
  [177] = {.lex_state = 0, .external_lex_state = 2},
  [178] = {.lex_state = 0, .external_lex_state = 2},
  [179] = {.lex_state = 0, .external_lex_state = 2},
  [180] = {.lex_state = 0, .external_lex_state = 2},
  [181] = {.lex_state = 0, .external_lex_state = 2},
  [182] = {.lex_state = 1184, .external_lex_state = 2},
  [183] = {.lex_state = 0, .external_lex_state = 2},
  [184] = {.lex_state = 0, .external_lex_state = 2},
  [185] = {.lex_state = 0, .external_lex_state = 2},
  [186] = {.lex_state = 0, .external_lex_state = 2},
  [187] = {.lex_state = 0, .external_lex_state = 2},
  [188] = {.lex_state = 1184, .external_lex_state = 2},
  [189] = {.lex_state = 0, .external_lex_state = 2},
  [190] = {.lex_state = 0, .external_lex_state = 2},
  [191] = {.lex_state = 0, .external_lex_state = 2},
  [192] = {.lex_state = 1184, .external_lex_state = 2},
  [193] = {.lex_state = 1184, .external_lex_state = 2},
  [194] = {.lex_state = 0, .external_lex_state = 2},
  [195] = {.lex_state = 0, .external_lex_state = 2},
  [196] = {.lex_state = 0, .external_lex_state = 2},
  [197] = {.lex_state = 0, .external_lex_state = 2},
  [198] = {.lex_state = 0, .external_lex_state = 2},
  [199] = {.lex_state = 0, .external_lex_state = 2},
  [200] = {.lex_state = 0, .external_lex_state = 2},
  [201] = {.lex_state = 1184, .external_lex_state = 2},
  [202] = {.lex_state = 0, .external_lex_state = 2},
  [203] = {.lex_state = 0, .external_lex_state = 2},
  [204] = {.lex_state = 0, .external_lex_state = 2},
  [205] = {.lex_state = 0, .external_lex_state = 2},
  [206] = {.lex_state = 0, .external_lex_state = 3},
  [207] = {.lex_state = 0, .external_lex_state = 2},
  [208] = {.lex_state = 0, .external_lex_state = 2},
  [209] = {.lex_state = 0, .external_lex_state = 2},
  [210] = {.lex_state = 0, .external_lex_state = 2},
  [211] = {.lex_state = 1184, .external_lex_state = 2},
  [212] = {.lex_state = 0, .external_lex_state = 2},
  [213] = {.lex_state = 0, .external_lex_state = 2},
  [214] = {.lex_state = 0, .external_lex_state = 2},
  [215] = {.lex_state = 0, .external_lex_state = 2},
  [216] = {.lex_state = 1184, .external_lex_state = 2},
  [217] = {.lex_state = 0, .external_lex_state = 2},
  [218] = {.lex_state = 0, .external_lex_state = 2},
  [219] = {.lex_state = 1184, .external_lex_state = 2},
  [220] = {.lex_state = 1184, .external_lex_state = 2},
  [221] = {.lex_state = 1184, .external_lex_state = 2},
  [222] = {.lex_state = 0, .external_lex_state = 3},
  [223] = {.lex_state = 1184, .external_lex_state = 2},
  [224] = {.lex_state = 0, .external_lex_state = 3},
  [225] = {.lex_state = 0, .external_lex_state = 3},
  [226] = {.lex_state = 1184, .external_lex_state = 2},
  [227] = {.lex_state = 1184, .external_lex_state = 2},
  [228] = {.lex_state = 1184, .external_lex_state = 2},
  [229] = {.lex_state = 1184, .external_lex_state = 2},
  [230] = {.lex_state = 0, .external_lex_state = 3},
  [231] = {.lex_state = 1184, .external_lex_state = 2},
  [232] = {.lex_state = 0, .external_lex_state = 3},
  [233] = {.lex_state = 0, .external_lex_state = 3},
  [234] = {.lex_state = 1184, .external_lex_state = 2},
  [235] = {.lex_state = 0, .external_lex_state = 3},
  [236] = {.lex_state = 0, .external_lex_state = 3},
  [237] = {.lex_state = 1001, .external_lex_state = 2},
  [238] = {.lex_state = 1001, .external_lex_state = 2},
  [239] = {.lex_state = 0, .external_lex_state = 2},
  [240] = {.lex_state = 0, .external_lex_state = 2},
  [241] = {.lex_state = 0, .external_lex_state = 3},
  [242] = {.lex_state = 0, .external_lex_state = 2},
  [243] = {.lex_state = 1002, .external_lex_state = 2},
  [244] = {.lex_state = 0, .external_lex_state = 2},
  [245] = {.lex_state = 0, .external_lex_state = 2},
  [246] = {.lex_state = 0, .external_lex_state = 2},
  [247] = {.lex_state = 0, .external_lex_state = 2},
  [248] = {.lex_state = 0, .external_lex_state = 2},
  [249] = {.lex_state = 0, .external_lex_state = 2},
  [250] = {.lex_state = 0, .external_lex_state = 2},
  [251] = {.lex_state = 1002, .external_lex_state = 2},
  [252] = {.lex_state = 0, .external_lex_state = 2},
  [253] = {.lex_state = 1002, .external_lex_state = 2},
  [254] = {.lex_state = 1002, .external_lex_state = 2},
  [255] = {.lex_state = 0, .external_lex_state = 2},
  [256] = {.lex_state = 0, .external_lex_state = 2},
  [257] = {.lex_state = 0, .external_lex_state = 2},
  [258] = {.lex_state = 0, .external_lex_state = 2},
  [259] = {.lex_state = 0, .external_lex_state = 2},
  [260] = {.lex_state = 1001, .external_lex_state = 2},
  [261] = {.lex_state = 1002, .external_lex_state = 2},
  [262] = {.lex_state = 1003, .external_lex_state = 2},
  [263] = {.lex_state = 1004, .external_lex_state = 2},
  [264] = {.lex_state = 0, .external_lex_state = 2},
  [265] = {.lex_state = 1001, .external_lex_state = 2},
  [266] = {.lex_state = 1002, .external_lex_state = 2},
  [267] = {.lex_state = 0, .external_lex_state = 2},
  [268] = {.lex_state = 0, .external_lex_state = 2},
  [269] = {.lex_state = 1003, .external_lex_state = 2},
  [270] = {.lex_state = 1004, .external_lex_state = 2},
  [271] = {.lex_state = 0, .external_lex_state = 2},
  [272] = {.lex_state = 0, .external_lex_state = 2},
  [273] = {.lex_state = 0, .external_lex_state = 2},
  [274] = {.lex_state = 0, .external_lex_state = 2},
  [275] = {.lex_state = 0, .external_lex_state = 2},
  [276] = {.lex_state = 0, .external_lex_state = 2},
  [277] = {.lex_state = 1404, .external_lex_state = 2},
  [278] = {.lex_state = 0, .external_lex_state = 4},
  [279] = {.lex_state = 1184, .external_lex_state = 2},
  [280] = {.lex_state = 0, .external_lex_state = 2},
  [281] = {.lex_state = 0, .external_lex_state = 2},
  [282] = {.lex_state = 0, .external_lex_state = 5},
  [283] = {.lex_state = 1184, .external_lex_state = 2},
  [284] = {.lex_state = 0, .external_lex_state = 2},
  [285] = {.lex_state = 0, .external_lex_state = 2},
  [286] = {.lex_state = 0, .external_lex_state = 5},
  [287] = {.lex_state = 0, .external_lex_state = 2},
  [288] = {.lex_state = 0, .external_lex_state = 6},
  [289] = {.lex_state = 0, .external_lex_state = 2},
  [290] = {.lex_state = 0, .external_lex_state = 2},
  [291] = {.lex_state = 0, .external_lex_state = 2},
  [292] = {.lex_state = 1184, .external_lex_state = 2},
  [293] = {.lex_state = 0, .external_lex_state = 2},
  [294] = {.lex_state = 0, .external_lex_state = 7},
  [295] = {.lex_state = 0, .external_lex_state = 2},
  [296] = {.lex_state = 0, .external_lex_state = 2},
  [297] = {.lex_state = 0, .external_lex_state = 2},
  [298] = {.lex_state = 0, .external_lex_state = 2},
  [299] = {.lex_state = 1184, .external_lex_state = 2},
  [300] = {.lex_state = 0, .external_lex_state = 7},
  [301] = {.lex_state = 0, .external_lex_state = 2},
  [302] = {.lex_state = 1184, .external_lex_state = 2},
  [303] = {.lex_state = 0, .external_lex_state = 2},
  [304] = {.lex_state = 1184, .external_lex_state = 2},
  [305] = {(TSStateId)(-1),},
  [306] = {(TSStateId)(-1),},
  [307] = {(TSStateId)(-1),},
  [308] = {(TSStateId)(-1),},
  [309] = {(TSStateId)(-1),},
  [310] = {(TSStateId)(-1),},
};

static const uint16_t ts_parse_table[LARGE_STATE_COUNT][SYMBOL_COUNT] = {
  [STATE(0)] = {
    [sym_macro_call] = STATE(0),
    [sym_comment] = STATE(0),
    [ts_builtin_sym_end] = ACTIONS(1),
    [sym_identifier] = ACTIONS(1),
    [sym_comma] = ACTIONS(1),
    [sym_semi_colon] = ACTIONS(1),
    [anon_sym_COLON] = ACTIONS(1),
    [sym_equals] = ACTIONS(1),
    [sym_token] = ACTIONS(1),
    [sym_static_token] = ACTIONS(1),
    [sym_static_argument] = ACTIONS(1),
    [anon_sym_ATmacro] = ACTIONS(1),
    [aux_sym_macro_call_token1] = ACTIONS(3),
    [anon_sym_ATpriority] = ACTIONS(1),
    [anon_sym_ATname] = ACTIONS(1),
    [anon_sym_ATderive] = ACTIONS(1),
    [sym_name_selector] = ACTIONS(1),
    [aux_sym_state_selector_token1] = ACTIONS(1),
    [sym_pseudo_selector] = ACTIONS(1),
    [sym_scope_open] = ACTIONS(1),
    [sym_scope_close] = ACTIONS(1),
    [sym_tuple_open] = ACTIONS(1),
    [sym_tuple_close] = ACTIONS(1),
    [sym_pow_operator] = ACTIONS(1),
    [sym_div_operator] = ACTIONS(1),
    [sym_floor_div_operator] = ACTIONS(1),
    [sym_mod_operator] = ACTIONS(1),
    [sym_mult_operator] = ACTIONS(1),
    [sym_add_operator] = ACTIONS(1),
    [sym_sub_operator] = ACTIONS(1),
    [sym_single_quote_string] = ACTIONS(1),
    [sym_double_quote_string] = ACTIONS(1),
    [sym__color_code_shade] = ACTIONS(1),
    [aux_sym_color_code_tailwind_token1] = ACTIONS(1),
    [aux_sym_color_code_tailwind_token2] = ACTIONS(1),
    [aux_sym_color_code_skin_token1] = ACTIONS(1),
    [aux_sym_color_code_skin_token2] = ACTIONS(1),
    [aux_sym_color_code_brick_token1] = ACTIONS(1),
    [aux_sym_color_code_css_token1] = ACTIONS(1),
    [sym_color_hex] = ACTIONS(1),
    [anon_sym_Enum] = ACTIONS(1),
    [anon_sym_DOT] = ACTIONS(1),
    [sym_rbx_asset] = ACTIONS(1),
    [sym_rbx_content] = ACTIONS(1),
    [anon_sym_DASH_DASH] = ACTIONS(5),
    [sym__block_comment_start] = ACTIONS(7),
    [sym__block_comment_content] = ACTIONS(1),
    [sym__block_comment_end] = ACTIONS(1),
    [sym__block_string_start] = ACTIONS(1),
    [sym__block_string_content] = ACTIONS(1),
    [sym__block_string_end] = ACTIONS(1),
  },
  [STATE(1)] = {
    [sym_source_file] = STATE(276),
    [sym__definition_no_macro] = STATE(147),
    [sym__definition] = STATE(151),
    [sym__definition_scope_body] = STATE(145),
    [sym_token_assignment] = STATE(145),
    [sym_static_token_assignment] = STATE(145),
    [sym_macro_declaration] = STATE(147),
    [sym_macro_call] = STATE(1),
    [sym_derive_declaration] = STATE(145),
    [sym_class_selector] = STATE(223),
    [sym_state_selector] = STATE(223),
    [sym_selector] = STATE(221),
    [sym_rule_scope] = STATE(145),
    [sym_comment] = STATE(1),
    [aux_sym_source_file_repeat1] = STATE(123),
    [aux_sym_rule_scope_repeat1] = STATE(201),
    [ts_builtin_sym_end] = ACTIONS(9),
    [sym_identifier] = ACTIONS(11),
    [sym_comma] = ACTIONS(13),
    [sym_semi_colon] = ACTIONS(15),
    [sym_token] = ACTIONS(17),
    [sym_static_token] = ACTIONS(19),
    [sym_static_argument] = ACTIONS(21),
    [anon_sym_ATmacro] = ACTIONS(23),
    [aux_sym_macro_call_token1] = ACTIONS(3),
    [anon_sym_ATderive] = ACTIONS(25),
    [sym_name_selector] = ACTIONS(21),
    [sym_tag_selector] = ACTIONS(21),
    [aux_sym_state_selector_token1] = ACTIONS(27),
    [sym_pseudo_selector] = ACTIONS(21),
    [sym_scope_open] = ACTIONS(29),
    [anon_sym_DASH_DASH] = ACTIONS(5),
    [sym__block_comment_start] = ACTIONS(7),
  },
  [STATE(2)] = {
    [sym__definition_no_macro] = STATE(12),
    [sym__definition_scope_body] = STATE(13),
    [sym_colon] = STATE(14),
    [sym_reference] = STATE(32),
    [sym_token_assignment] = STATE(25),
    [sym_static_token_assignment] = STATE(25),
    [sym_macro_body] = STATE(15),
    [sym_macro_call] = STATE(2),
    [sym_priority_declaration] = STATE(16),
    [sym_name_declaration] = STATE(17),
    [sym_derive_declaration] = STATE(25),
    [sym_class_selector] = STATE(24),
    [sym_state_selector] = STATE(24),
    [sym_selector] = STATE(18),
    [sym_rule_scope] = STATE(25),
    [sym_datatype] = STATE(6),
    [sym_tuple] = STATE(32),
    [sym_number] = STATE(32),
    [sym_operator] = STATE(20),
    [sym_operation] = STATE(32),
    [sym_string] = STATE(32),
    [sym_block_string] = STATE(26),
    [sym_color] = STATE(32),
    [sym_color_code_tailwind] = STATE(31),
    [sym_color_code_skin] = STATE(31),
    [sym_color_code_brick] = STATE(31),
    [sym_color_code_css] = STATE(31),
    [sym_enum] = STATE(32),
    [sym_enum_full] = STATE(33),
    [sym_enum_shorthand] = STATE(33),
    [sym_comment] = STATE(2),
    [aux_sym_macro_scope_repeat1] = STATE(3),
    [aux_sym_rule_scope_repeat1] = STATE(211),
    [sym_identifier] = ACTIONS(31),
    [sym_comma] = ACTIONS(33),
    [sym_semi_colon] = ACTIONS(35),
    [anon_sym_COLON] = ACTIONS(37),
    [sym_equals] = ACTIONS(39),
    [sym_token] = ACTIONS(41),
    [sym_static_token] = ACTIONS(43),
    [sym_static_argument] = ACTIONS(45),
    [aux_sym_macro_call_token1] = ACTIONS(47),
    [anon_sym_ATpriority] = ACTIONS(49),
    [anon_sym_ATname] = ACTIONS(51),
    [anon_sym_ATderive] = ACTIONS(53),
    [sym_name_selector] = ACTIONS(55),
    [sym_tag_selector] = ACTIONS(57),
    [aux_sym_state_selector_token1] = ACTIONS(59),
    [sym_pseudo_selector] = ACTIONS(55),
    [sym_scope_open] = ACTIONS(61),
    [sym_scope_close] = ACTIONS(63),
    [sym_tuple_open] = ACTIONS(65),
    [sym_number_ambiguous] = ACTIONS(67),
    [sym_number_percent] = ACTIONS(69),
    [sym_number_offset] = ACTIONS(67),
    [sym_pow_operator] = ACTIONS(71),
    [sym_div_operator] = ACTIONS(73),
    [sym_floor_div_operator] = ACTIONS(71),
    [sym_mod_operator] = ACTIONS(71),
    [sym_mult_operator] = ACTIONS(71),
    [sym_add_operator] = ACTIONS(71),
    [sym_sub_operator] = ACTIONS(73),
    [sym_single_quote_string] = ACTIONS(75),
    [sym_double_quote_string] = ACTIONS(75),
    [aux_sym_color_code_tailwind_token1] = ACTIONS(77),
    [aux_sym_color_code_skin_token1] = ACTIONS(79),
    [aux_sym_color_code_brick_token1] = ACTIONS(81),
    [aux_sym_color_code_css_token1] = ACTIONS(83),
    [sym_color_hex] = ACTIONS(85),
    [anon_sym_Enum] = ACTIONS(87),
    [sym_rbx_asset] = ACTIONS(89),
    [sym_rbx_content] = ACTIONS(89),
    [anon_sym_DASH_DASH] = ACTIONS(5),
    [sym__block_comment_start] = ACTIONS(7),
    [sym__block_string_start] = ACTIONS(91),
  },
  [STATE(3)] = {
    [sym__definition_no_macro] = STATE(12),
    [sym__definition_scope_body] = STATE(13),
    [sym_colon] = STATE(14),
    [sym_reference] = STATE(32),
    [sym_token_assignment] = STATE(25),
    [sym_static_token_assignment] = STATE(25),
    [sym_macro_body] = STATE(15),
    [sym_macro_call] = STATE(3),
    [sym_priority_declaration] = STATE(16),
    [sym_name_declaration] = STATE(17),
    [sym_derive_declaration] = STATE(25),
    [sym_class_selector] = STATE(24),
    [sym_state_selector] = STATE(24),
    [sym_selector] = STATE(18),
    [sym_rule_scope] = STATE(25),
    [sym_datatype] = STATE(6),
    [sym_tuple] = STATE(32),
    [sym_number] = STATE(32),
    [sym_operator] = STATE(20),
    [sym_operation] = STATE(32),
    [sym_string] = STATE(32),
    [sym_block_string] = STATE(26),
    [sym_color] = STATE(32),
    [sym_color_code_tailwind] = STATE(31),
    [sym_color_code_skin] = STATE(31),
    [sym_color_code_brick] = STATE(31),
    [sym_color_code_css] = STATE(31),
    [sym_enum] = STATE(32),
    [sym_enum_full] = STATE(33),
    [sym_enum_shorthand] = STATE(33),
    [sym_comment] = STATE(3),
    [aux_sym_macro_scope_repeat1] = STATE(4),
    [aux_sym_rule_scope_repeat1] = STATE(211),
    [sym_identifier] = ACTIONS(31),
    [sym_comma] = ACTIONS(33),
    [sym_semi_colon] = ACTIONS(35),
    [anon_sym_COLON] = ACTIONS(37),
    [sym_equals] = ACTIONS(39),
    [sym_token] = ACTIONS(41),
    [sym_static_token] = ACTIONS(43),
    [sym_static_argument] = ACTIONS(45),
    [aux_sym_macro_call_token1] = ACTIONS(47),
    [anon_sym_ATpriority] = ACTIONS(49),
    [anon_sym_ATname] = ACTIONS(51),
    [anon_sym_ATderive] = ACTIONS(53),
    [sym_name_selector] = ACTIONS(55),
    [sym_tag_selector] = ACTIONS(57),
    [aux_sym_state_selector_token1] = ACTIONS(59),
    [sym_pseudo_selector] = ACTIONS(55),
    [sym_scope_open] = ACTIONS(61),
    [sym_scope_close] = ACTIONS(93),
    [sym_tuple_open] = ACTIONS(65),
    [sym_number_ambiguous] = ACTIONS(67),
    [sym_number_percent] = ACTIONS(69),
    [sym_number_offset] = ACTIONS(67),
    [sym_pow_operator] = ACTIONS(71),
    [sym_div_operator] = ACTIONS(73),
    [sym_floor_div_operator] = ACTIONS(71),
    [sym_mod_operator] = ACTIONS(71),
    [sym_mult_operator] = ACTIONS(71),
    [sym_add_operator] = ACTIONS(71),
    [sym_sub_operator] = ACTIONS(73),
    [sym_single_quote_string] = ACTIONS(75),
    [sym_double_quote_string] = ACTIONS(75),
    [aux_sym_color_code_tailwind_token1] = ACTIONS(77),
    [aux_sym_color_code_skin_token1] = ACTIONS(79),
    [aux_sym_color_code_brick_token1] = ACTIONS(81),
    [aux_sym_color_code_css_token1] = ACTIONS(83),
    [sym_color_hex] = ACTIONS(85),
    [anon_sym_Enum] = ACTIONS(87),
    [sym_rbx_asset] = ACTIONS(89),
    [sym_rbx_content] = ACTIONS(89),
    [anon_sym_DASH_DASH] = ACTIONS(5),
    [sym__block_comment_start] = ACTIONS(7),
    [sym__block_string_start] = ACTIONS(91),
  },
  [STATE(4)] = {
    [sym__definition_no_macro] = STATE(12),
    [sym__definition_scope_body] = STATE(13),
    [sym_colon] = STATE(14),
    [sym_reference] = STATE(32),
    [sym_token_assignment] = STATE(25),
    [sym_static_token_assignment] = STATE(25),
    [sym_macro_body] = STATE(15),
    [sym_macro_call] = STATE(4),
    [sym_priority_declaration] = STATE(16),
    [sym_name_declaration] = STATE(17),
    [sym_derive_declaration] = STATE(25),
    [sym_class_selector] = STATE(24),
    [sym_state_selector] = STATE(24),
    [sym_selector] = STATE(18),
    [sym_rule_scope] = STATE(25),
    [sym_datatype] = STATE(6),
    [sym_tuple] = STATE(32),
    [sym_number] = STATE(32),
    [sym_operator] = STATE(20),
    [sym_operation] = STATE(32),
    [sym_string] = STATE(32),
    [sym_block_string] = STATE(26),
    [sym_color] = STATE(32),
    [sym_color_code_tailwind] = STATE(31),
    [sym_color_code_skin] = STATE(31),
    [sym_color_code_brick] = STATE(31),
    [sym_color_code_css] = STATE(31),
    [sym_enum] = STATE(32),
    [sym_enum_full] = STATE(33),
    [sym_enum_shorthand] = STATE(33),
    [sym_comment] = STATE(4),
    [aux_sym_macro_scope_repeat1] = STATE(4),
    [aux_sym_rule_scope_repeat1] = STATE(211),
    [sym_identifier] = ACTIONS(95),
    [sym_comma] = ACTIONS(98),
    [sym_semi_colon] = ACTIONS(101),
    [anon_sym_COLON] = ACTIONS(104),
    [sym_equals] = ACTIONS(107),
    [sym_token] = ACTIONS(110),
    [sym_static_token] = ACTIONS(113),
    [sym_static_argument] = ACTIONS(116),
    [aux_sym_macro_call_token1] = ACTIONS(119),
    [anon_sym_ATpriority] = ACTIONS(122),
    [anon_sym_ATname] = ACTIONS(125),
    [anon_sym_ATderive] = ACTIONS(128),
    [sym_name_selector] = ACTIONS(131),
    [sym_tag_selector] = ACTIONS(134),
    [aux_sym_state_selector_token1] = ACTIONS(137),
    [sym_pseudo_selector] = ACTIONS(131),
    [sym_scope_open] = ACTIONS(140),
    [sym_scope_close] = ACTIONS(143),
    [sym_tuple_open] = ACTIONS(145),
    [sym_number_ambiguous] = ACTIONS(148),
    [sym_number_percent] = ACTIONS(151),
    [sym_number_offset] = ACTIONS(148),
    [sym_pow_operator] = ACTIONS(154),
    [sym_div_operator] = ACTIONS(157),
    [sym_floor_div_operator] = ACTIONS(154),
    [sym_mod_operator] = ACTIONS(154),
    [sym_mult_operator] = ACTIONS(154),
    [sym_add_operator] = ACTIONS(154),
    [sym_sub_operator] = ACTIONS(157),
    [sym_single_quote_string] = ACTIONS(160),
    [sym_double_quote_string] = ACTIONS(160),
    [aux_sym_color_code_tailwind_token1] = ACTIONS(163),
    [aux_sym_color_code_skin_token1] = ACTIONS(166),
    [aux_sym_color_code_brick_token1] = ACTIONS(169),
    [aux_sym_color_code_css_token1] = ACTIONS(172),
    [sym_color_hex] = ACTIONS(175),
    [anon_sym_Enum] = ACTIONS(178),
    [sym_rbx_asset] = ACTIONS(181),
    [sym_rbx_content] = ACTIONS(181),
    [anon_sym_DASH_DASH] = ACTIONS(5),
    [sym__block_comment_start] = ACTIONS(7),
    [sym__block_string_start] = ACTIONS(184),
  },
};

static const uint16_t ts_small_parse_table[] = {
  [0] = 8,
    ACTIONS(5), 1,
      anon_sym_DASH_DASH,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(193), 1,
      aux_sym_rule_scope_repeat1,
    STATE(228), 1,
      sym_selector,
    STATE(5), 2,
      sym_macro_call,
      sym_comment,
    STATE(223), 2,
      sym_class_selector,
      sym_state_selector,
    ACTIONS(187), 9,
      sym_identifier,
      anon_sym_COLON,
      sym_tag_selector,
      sym_number_ambiguous,
      sym_number_offset,
      sym_div_operator,
      sym_sub_operator,
      sym_color_hex,
      anon_sym_Enum,
    ACTIONS(189), 31,
      sym__block_string_start,
      sym_comma,
      sym_semi_colon,
      sym_equals,
      sym_token,
      sym_static_token,
      sym_static_argument,
      aux_sym_macro_call_token1,
      anon_sym_ATpriority,
      anon_sym_ATname,
      anon_sym_ATderive,
      sym_name_selector,
      aux_sym_state_selector_token1,
      sym_pseudo_selector,
      sym_scope_open,
      sym_scope_close,
      sym_tuple_open,
      sym_number_percent,
      sym_pow_operator,
      sym_floor_div_operator,
      sym_mod_operator,
      sym_mult_operator,
      sym_add_operator,
      sym_single_quote_string,
      sym_double_quote_string,
      aux_sym_color_code_tailwind_token1,
      aux_sym_color_code_skin_token1,
      aux_sym_color_code_brick_token1,
      aux_sym_color_code_css_token1,
      sym_rbx_asset,
      sym_rbx_content,
  [65] = 6,
    ACTIONS(5), 1,
      anon_sym_DASH_DASH,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(105), 1,
      sym_operator,
    STATE(6), 2,
      sym_macro_call,
      sym_comment,
    ACTIONS(187), 9,
      sym_identifier,
      anon_sym_COLON,
      sym_tag_selector,
      sym_number_ambiguous,
      sym_number_offset,
      sym_div_operator,
      sym_sub_operator,
      sym_color_hex,
      anon_sym_Enum,
    ACTIONS(189), 31,
      sym__block_string_start,
      sym_comma,
      sym_semi_colon,
      sym_equals,
      sym_token,
      sym_static_token,
      sym_static_argument,
      aux_sym_macro_call_token1,
      anon_sym_ATpriority,
      anon_sym_ATname,
      anon_sym_ATderive,
      sym_name_selector,
      aux_sym_state_selector_token1,
      sym_pseudo_selector,
      sym_scope_open,
      sym_scope_close,
      sym_tuple_open,
      sym_number_percent,
      sym_pow_operator,
      sym_floor_div_operator,
      sym_mod_operator,
      sym_mult_operator,
      sym_add_operator,
      sym_single_quote_string,
      sym_double_quote_string,
      aux_sym_color_code_tailwind_token1,
      aux_sym_color_code_skin_token1,
      aux_sym_color_code_brick_token1,
      aux_sym_color_code_css_token1,
      sym_rbx_asset,
      sym_rbx_content,
  [123] = 6,
    ACTIONS(5), 1,
      anon_sym_DASH_DASH,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(105), 1,
      sym_operator,
    STATE(7), 2,
      sym_macro_call,
      sym_comment,
    ACTIONS(191), 9,
      sym_identifier,
      anon_sym_COLON,
      sym_tag_selector,
      sym_number_ambiguous,
      sym_number_offset,
      sym_div_operator,
      sym_sub_operator,
      sym_color_hex,
      anon_sym_Enum,
    ACTIONS(193), 31,
      sym__block_string_start,
      sym_comma,
      sym_semi_colon,
      sym_equals,
      sym_token,
      sym_static_token,
      sym_static_argument,
      aux_sym_macro_call_token1,
      anon_sym_ATpriority,
      anon_sym_ATname,
      anon_sym_ATderive,
      sym_name_selector,
      aux_sym_state_selector_token1,
      sym_pseudo_selector,
      sym_scope_open,
      sym_scope_close,
      sym_tuple_open,
      sym_number_percent,
      sym_pow_operator,
      sym_floor_div_operator,
      sym_mod_operator,
      sym_mult_operator,
      sym_add_operator,
      sym_single_quote_string,
      sym_double_quote_string,
      aux_sym_color_code_tailwind_token1,
      aux_sym_color_code_skin_token1,
      aux_sym_color_code_brick_token1,
      aux_sym_color_code_css_token1,
      sym_rbx_asset,
      sym_rbx_content,
  [181] = 5,
    ACTIONS(5), 1,
      anon_sym_DASH_DASH,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(8), 2,
      sym_macro_call,
      sym_comment,
    ACTIONS(195), 9,
      sym_identifier,
      anon_sym_COLON,
      sym_tag_selector,
      sym_number_ambiguous,
      sym_number_offset,
      sym_div_operator,
      sym_sub_operator,
      sym_color_hex,
      anon_sym_Enum,
    ACTIONS(197), 31,
      sym__block_string_start,
      sym_comma,
      sym_semi_colon,
      sym_equals,
      sym_token,
      sym_static_token,
      sym_static_argument,
      aux_sym_macro_call_token1,
      anon_sym_ATpriority,
      anon_sym_ATname,
      anon_sym_ATderive,
      sym_name_selector,
      aux_sym_state_selector_token1,
      sym_pseudo_selector,
      sym_scope_open,
      sym_scope_close,
      sym_tuple_open,
      sym_number_percent,
      sym_pow_operator,
      sym_floor_div_operator,
      sym_mod_operator,
      sym_mult_operator,
      sym_add_operator,
      sym_single_quote_string,
      sym_double_quote_string,
      aux_sym_color_code_tailwind_token1,
      aux_sym_color_code_skin_token1,
      aux_sym_color_code_brick_token1,
      aux_sym_color_code_css_token1,
      sym_rbx_asset,
      sym_rbx_content,
  [236] = 6,
    ACTIONS(5), 1,
      anon_sym_DASH_DASH,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(203), 1,
      sym_equals,
    STATE(9), 2,
      sym_macro_call,
      sym_comment,
    ACTIONS(199), 9,
      sym_identifier,
      anon_sym_COLON,
      sym_tag_selector,
      sym_number_ambiguous,
      sym_number_offset,
      sym_div_operator,
      sym_sub_operator,
      sym_color_hex,
      anon_sym_Enum,
    ACTIONS(201), 30,
      sym__block_string_start,
      sym_comma,
      sym_semi_colon,
      sym_token,
      sym_static_token,
      sym_static_argument,
      aux_sym_macro_call_token1,
      anon_sym_ATpriority,
      anon_sym_ATname,
      anon_sym_ATderive,
      sym_name_selector,
      aux_sym_state_selector_token1,
      sym_pseudo_selector,
      sym_scope_open,
      sym_scope_close,
      sym_tuple_open,
      sym_number_percent,
      sym_pow_operator,
      sym_floor_div_operator,
      sym_mod_operator,
      sym_mult_operator,
      sym_add_operator,
      sym_single_quote_string,
      sym_double_quote_string,
      aux_sym_color_code_tailwind_token1,
      aux_sym_color_code_skin_token1,
      aux_sym_color_code_brick_token1,
      aux_sym_color_code_css_token1,
      sym_rbx_asset,
      sym_rbx_content,
  [293] = 5,
    ACTIONS(5), 1,
      anon_sym_DASH_DASH,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(10), 2,
      sym_macro_call,
      sym_comment,
    ACTIONS(187), 9,
      sym_identifier,
      anon_sym_COLON,
      sym_tag_selector,
      sym_number_ambiguous,
      sym_number_offset,
      sym_div_operator,
      sym_sub_operator,
      sym_color_hex,
      anon_sym_Enum,
    ACTIONS(189), 31,
      sym__block_string_start,
      sym_comma,
      sym_semi_colon,
      sym_equals,
      sym_token,
      sym_static_token,
      sym_static_argument,
      aux_sym_macro_call_token1,
      anon_sym_ATpriority,
      anon_sym_ATname,
      anon_sym_ATderive,
      sym_name_selector,
      aux_sym_state_selector_token1,
      sym_pseudo_selector,
      sym_scope_open,
      sym_scope_close,
      sym_tuple_open,
      sym_number_percent,
      sym_pow_operator,
      sym_floor_div_operator,
      sym_mod_operator,
      sym_mult_operator,
      sym_add_operator,
      sym_single_quote_string,
      sym_double_quote_string,
      aux_sym_color_code_tailwind_token1,
      aux_sym_color_code_skin_token1,
      aux_sym_color_code_brick_token1,
      aux_sym_color_code_css_token1,
      sym_rbx_asset,
      sym_rbx_content,
  [348] = 5,
    ACTIONS(5), 1,
      anon_sym_DASH_DASH,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(11), 2,
      sym_macro_call,
      sym_comment,
    ACTIONS(205), 9,
      sym_identifier,
      anon_sym_COLON,
      sym_tag_selector,
      sym_number_ambiguous,
      sym_number_offset,
      sym_div_operator,
      sym_sub_operator,
      sym_color_hex,
      anon_sym_Enum,
    ACTIONS(207), 31,
      sym__block_string_start,
      sym_comma,
      sym_semi_colon,
      sym_equals,
      sym_token,
      sym_static_token,
      sym_static_argument,
      aux_sym_macro_call_token1,
      anon_sym_ATpriority,
      anon_sym_ATname,
      anon_sym_ATderive,
      sym_name_selector,
      aux_sym_state_selector_token1,
      sym_pseudo_selector,
      sym_scope_open,
      sym_scope_close,
      sym_tuple_open,
      sym_number_percent,
      sym_pow_operator,
      sym_floor_div_operator,
      sym_mod_operator,
      sym_mult_operator,
      sym_add_operator,
      sym_single_quote_string,
      sym_double_quote_string,
      aux_sym_color_code_tailwind_token1,
      aux_sym_color_code_skin_token1,
      aux_sym_color_code_brick_token1,
      aux_sym_color_code_css_token1,
      sym_rbx_asset,
      sym_rbx_content,
  [403] = 5,
    ACTIONS(5), 1,
      anon_sym_DASH_DASH,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(12), 2,
      sym_macro_call,
      sym_comment,
    ACTIONS(187), 9,
      sym_identifier,
      anon_sym_COLON,
      sym_tag_selector,
      sym_number_ambiguous,
      sym_number_offset,
      sym_div_operator,
      sym_sub_operator,
      sym_color_hex,
      anon_sym_Enum,
    ACTIONS(189), 31,
      sym__block_string_start,
      sym_comma,
      sym_semi_colon,
      sym_equals,
      sym_token,
      sym_static_token,
      sym_static_argument,
      aux_sym_macro_call_token1,
      anon_sym_ATpriority,
      anon_sym_ATname,
      anon_sym_ATderive,
      sym_name_selector,
      aux_sym_state_selector_token1,
      sym_pseudo_selector,
      sym_scope_open,
      sym_scope_close,
      sym_tuple_open,
      sym_number_percent,
      sym_pow_operator,
      sym_floor_div_operator,
      sym_mod_operator,
      sym_mult_operator,
      sym_add_operator,
      sym_single_quote_string,
      sym_double_quote_string,
      aux_sym_color_code_tailwind_token1,
      aux_sym_color_code_skin_token1,
      aux_sym_color_code_brick_token1,
      aux_sym_color_code_css_token1,
      sym_rbx_asset,
      sym_rbx_content,
  [458] = 5,
    ACTIONS(5), 1,
      anon_sym_DASH_DASH,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(13), 2,
      sym_macro_call,
      sym_comment,
    ACTIONS(187), 9,
      sym_identifier,
      anon_sym_COLON,
      sym_tag_selector,
      sym_number_ambiguous,
      sym_number_offset,
      sym_div_operator,
      sym_sub_operator,
      sym_color_hex,
      anon_sym_Enum,
    ACTIONS(189), 31,
      sym__block_string_start,
      sym_comma,
      sym_semi_colon,
      sym_equals,
      sym_token,
      sym_static_token,
      sym_static_argument,
      aux_sym_macro_call_token1,
      anon_sym_ATpriority,
      anon_sym_ATname,
      anon_sym_ATderive,
      sym_name_selector,
      aux_sym_state_selector_token1,
      sym_pseudo_selector,
      sym_scope_open,
      sym_scope_close,
      sym_tuple_open,
      sym_number_percent,
      sym_pow_operator,
      sym_floor_div_operator,
      sym_mod_operator,
      sym_mult_operator,
      sym_add_operator,
      sym_single_quote_string,
      sym_double_quote_string,
      aux_sym_color_code_tailwind_token1,
      aux_sym_color_code_skin_token1,
      aux_sym_color_code_brick_token1,
      aux_sym_color_code_css_token1,
      sym_rbx_asset,
      sym_rbx_content,
  [513] = 5,
    ACTIONS(5), 1,
      anon_sym_DASH_DASH,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(14), 2,
      sym_macro_call,
      sym_comment,
    ACTIONS(187), 9,
      sym_identifier,
      anon_sym_COLON,
      sym_tag_selector,
      sym_number_ambiguous,
      sym_number_offset,
      sym_div_operator,
      sym_sub_operator,
      sym_color_hex,
      anon_sym_Enum,
    ACTIONS(189), 31,
      sym__block_string_start,
      sym_comma,
      sym_semi_colon,
      sym_equals,
      sym_token,
      sym_static_token,
      sym_static_argument,
      aux_sym_macro_call_token1,
      anon_sym_ATpriority,
      anon_sym_ATname,
      anon_sym_ATderive,
      sym_name_selector,
      aux_sym_state_selector_token1,
      sym_pseudo_selector,
      sym_scope_open,
      sym_scope_close,
      sym_tuple_open,
      sym_number_percent,
      sym_pow_operator,
      sym_floor_div_operator,
      sym_mod_operator,
      sym_mult_operator,
      sym_add_operator,
      sym_single_quote_string,
      sym_double_quote_string,
      aux_sym_color_code_tailwind_token1,
      aux_sym_color_code_skin_token1,
      aux_sym_color_code_brick_token1,
      aux_sym_color_code_css_token1,
      sym_rbx_asset,
      sym_rbx_content,
  [568] = 5,
    ACTIONS(5), 1,
      anon_sym_DASH_DASH,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(15), 2,
      sym_macro_call,
      sym_comment,
    ACTIONS(209), 9,
      sym_identifier,
      anon_sym_COLON,
      sym_tag_selector,
      sym_number_ambiguous,
      sym_number_offset,
      sym_div_operator,
      sym_sub_operator,
      sym_color_hex,
      anon_sym_Enum,
    ACTIONS(211), 31,
      sym__block_string_start,
      sym_comma,
      sym_semi_colon,
      sym_equals,
      sym_token,
      sym_static_token,
      sym_static_argument,
      aux_sym_macro_call_token1,
      anon_sym_ATpriority,
      anon_sym_ATname,
      anon_sym_ATderive,
      sym_name_selector,
      aux_sym_state_selector_token1,
      sym_pseudo_selector,
      sym_scope_open,
      sym_scope_close,
      sym_tuple_open,
      sym_number_percent,
      sym_pow_operator,
      sym_floor_div_operator,
      sym_mod_operator,
      sym_mult_operator,
      sym_add_operator,
      sym_single_quote_string,
      sym_double_quote_string,
      aux_sym_color_code_tailwind_token1,
      aux_sym_color_code_skin_token1,
      aux_sym_color_code_brick_token1,
      aux_sym_color_code_css_token1,
      sym_rbx_asset,
      sym_rbx_content,
  [623] = 5,
    ACTIONS(5), 1,
      anon_sym_DASH_DASH,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(16), 2,
      sym_macro_call,
      sym_comment,
    ACTIONS(187), 9,
      sym_identifier,
      anon_sym_COLON,
      sym_tag_selector,
      sym_number_ambiguous,
      sym_number_offset,
      sym_div_operator,
      sym_sub_operator,
      sym_color_hex,
      anon_sym_Enum,
    ACTIONS(189), 31,
      sym__block_string_start,
      sym_comma,
      sym_semi_colon,
      sym_equals,
      sym_token,
      sym_static_token,
      sym_static_argument,
      aux_sym_macro_call_token1,
      anon_sym_ATpriority,
      anon_sym_ATname,
      anon_sym_ATderive,
      sym_name_selector,
      aux_sym_state_selector_token1,
      sym_pseudo_selector,
      sym_scope_open,
      sym_scope_close,
      sym_tuple_open,
      sym_number_percent,
      sym_pow_operator,
      sym_floor_div_operator,
      sym_mod_operator,
      sym_mult_operator,
      sym_add_operator,
      sym_single_quote_string,
      sym_double_quote_string,
      aux_sym_color_code_tailwind_token1,
      aux_sym_color_code_skin_token1,
      aux_sym_color_code_brick_token1,
      aux_sym_color_code_css_token1,
      sym_rbx_asset,
      sym_rbx_content,
  [678] = 5,
    ACTIONS(5), 1,
      anon_sym_DASH_DASH,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(17), 2,
      sym_macro_call,
      sym_comment,
    ACTIONS(187), 9,
      sym_identifier,
      anon_sym_COLON,
      sym_tag_selector,
      sym_number_ambiguous,
      sym_number_offset,
      sym_div_operator,
      sym_sub_operator,
      sym_color_hex,
      anon_sym_Enum,
    ACTIONS(189), 31,
      sym__block_string_start,
      sym_comma,
      sym_semi_colon,
      sym_equals,
      sym_token,
      sym_static_token,
      sym_static_argument,
      aux_sym_macro_call_token1,
      anon_sym_ATpriority,
      anon_sym_ATname,
      anon_sym_ATderive,
      sym_name_selector,
      aux_sym_state_selector_token1,
      sym_pseudo_selector,
      sym_scope_open,
      sym_scope_close,
      sym_tuple_open,
      sym_number_percent,
      sym_pow_operator,
      sym_floor_div_operator,
      sym_mod_operator,
      sym_mult_operator,
      sym_add_operator,
      sym_single_quote_string,
      sym_double_quote_string,
      aux_sym_color_code_tailwind_token1,
      aux_sym_color_code_skin_token1,
      aux_sym_color_code_brick_token1,
      aux_sym_color_code_css_token1,
      sym_rbx_asset,
      sym_rbx_content,
  [733] = 5,
    ACTIONS(5), 1,
      anon_sym_DASH_DASH,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(18), 2,
      sym_macro_call,
      sym_comment,
    ACTIONS(187), 9,
      sym_identifier,
      anon_sym_COLON,
      sym_tag_selector,
      sym_number_ambiguous,
      sym_number_offset,
      sym_div_operator,
      sym_sub_operator,
      sym_color_hex,
      anon_sym_Enum,
    ACTIONS(189), 31,
      sym__block_string_start,
      sym_comma,
      sym_semi_colon,
      sym_equals,
      sym_token,
      sym_static_token,
      sym_static_argument,
      aux_sym_macro_call_token1,
      anon_sym_ATpriority,
      anon_sym_ATname,
      anon_sym_ATderive,
      sym_name_selector,
      aux_sym_state_selector_token1,
      sym_pseudo_selector,
      sym_scope_open,
      sym_scope_close,
      sym_tuple_open,
      sym_number_percent,
      sym_pow_operator,
      sym_floor_div_operator,
      sym_mod_operator,
      sym_mult_operator,
      sym_add_operator,
      sym_single_quote_string,
      sym_double_quote_string,
      aux_sym_color_code_tailwind_token1,
      aux_sym_color_code_skin_token1,
      aux_sym_color_code_brick_token1,
      aux_sym_color_code_css_token1,
      sym_rbx_asset,
      sym_rbx_content,
  [788] = 27,
    ACTIONS(5), 1,
      anon_sym_DASH_DASH,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(213), 1,
      sym_identifier,
    ACTIONS(219), 1,
      aux_sym_macro_call_token1,
    ACTIONS(221), 1,
      aux_sym_state_selector_token1,
    ACTIONS(223), 1,
      sym_tuple_open,
    ACTIONS(225), 1,
      sym_tuple_close,
    ACTIONS(229), 1,
      sym_number_percent,
    ACTIONS(233), 1,
      aux_sym_color_code_tailwind_token1,
    ACTIONS(235), 1,
      aux_sym_color_code_skin_token1,
    ACTIONS(237), 1,
      aux_sym_color_code_brick_token1,
    ACTIONS(239), 1,
      aux_sym_color_code_css_token1,
    ACTIONS(241), 1,
      sym_color_hex,
    ACTIONS(243), 1,
      anon_sym_Enum,
    ACTIONS(247), 1,
      sym__block_string_start,
    STATE(104), 1,
      aux_sym_macro_call_repeat1,
    STATE(178), 1,
      sym_datatype,
    STATE(207), 1,
      sym_block_string,
    ACTIONS(215), 2,
      sym_comma,
      sym_semi_colon,
    ACTIONS(227), 2,
      sym_number_ambiguous,
      sym_number_offset,
    ACTIONS(231), 2,
      sym_single_quote_string,
      sym_double_quote_string,
    ACTIONS(245), 2,
      sym_rbx_asset,
      sym_rbx_content,
    STATE(19), 2,
      sym_macro_call,
      sym_comment,
    STATE(197), 2,
      sym_enum_full,
      sym_enum_shorthand,
    ACTIONS(217), 3,
      sym_token,
      sym_static_token,
      sym_static_argument,
    STATE(195), 4,
      sym_color_code_tailwind,
      sym_color_code_skin,
      sym_color_code_brick,
      sym_color_code_css,
    STATE(212), 7,
      sym_reference,
      sym_tuple,
      sym_number,
      sym_operation,
      sym_string,
      sym_color,
      sym_enum,
  [887] = 5,
    ACTIONS(5), 1,
      anon_sym_DASH_DASH,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(20), 2,
      sym_macro_call,
      sym_comment,
    ACTIONS(187), 9,
      sym_identifier,
      anon_sym_COLON,
      sym_tag_selector,
      sym_number_ambiguous,
      sym_number_offset,
      sym_div_operator,
      sym_sub_operator,
      sym_color_hex,
      anon_sym_Enum,
    ACTIONS(189), 31,
      sym__block_string_start,
      sym_comma,
      sym_semi_colon,
      sym_equals,
      sym_token,
      sym_static_token,
      sym_static_argument,
      aux_sym_macro_call_token1,
      anon_sym_ATpriority,
      anon_sym_ATname,
      anon_sym_ATderive,
      sym_name_selector,
      aux_sym_state_selector_token1,
      sym_pseudo_selector,
      sym_scope_open,
      sym_scope_close,
      sym_tuple_open,
      sym_number_percent,
      sym_pow_operator,
      sym_floor_div_operator,
      sym_mod_operator,
      sym_mult_operator,
      sym_add_operator,
      sym_single_quote_string,
      sym_double_quote_string,
      aux_sym_color_code_tailwind_token1,
      aux_sym_color_code_skin_token1,
      aux_sym_color_code_brick_token1,
      aux_sym_color_code_css_token1,
      sym_rbx_asset,
      sym_rbx_content,
  [942] = 5,
    ACTIONS(5), 1,
      anon_sym_DASH_DASH,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(21), 2,
      sym_macro_call,
      sym_comment,
    ACTIONS(187), 9,
      sym_identifier,
      anon_sym_COLON,
      sym_tag_selector,
      sym_number_ambiguous,
      sym_number_offset,
      sym_div_operator,
      sym_sub_operator,
      sym_color_hex,
      anon_sym_Enum,
    ACTIONS(189), 31,
      sym__block_string_start,
      sym_comma,
      sym_semi_colon,
      sym_equals,
      sym_token,
      sym_static_token,
      sym_static_argument,
      aux_sym_macro_call_token1,
      anon_sym_ATpriority,
      anon_sym_ATname,
      anon_sym_ATderive,
      sym_name_selector,
      aux_sym_state_selector_token1,
      sym_pseudo_selector,
      sym_scope_open,
      sym_scope_close,
      sym_tuple_open,
      sym_number_percent,
      sym_pow_operator,
      sym_floor_div_operator,
      sym_mod_operator,
      sym_mult_operator,
      sym_add_operator,
      sym_single_quote_string,
      sym_double_quote_string,
      aux_sym_color_code_tailwind_token1,
      aux_sym_color_code_skin_token1,
      aux_sym_color_code_brick_token1,
      aux_sym_color_code_css_token1,
      sym_rbx_asset,
      sym_rbx_content,
  [997] = 27,
    ACTIONS(5), 1,
      anon_sym_DASH_DASH,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(213), 1,
      sym_identifier,
    ACTIONS(219), 1,
      aux_sym_macro_call_token1,
    ACTIONS(221), 1,
      aux_sym_state_selector_token1,
    ACTIONS(223), 1,
      sym_tuple_open,
    ACTIONS(229), 1,
      sym_number_percent,
    ACTIONS(233), 1,
      aux_sym_color_code_tailwind_token1,
    ACTIONS(235), 1,
      aux_sym_color_code_skin_token1,
    ACTIONS(237), 1,
      aux_sym_color_code_brick_token1,
    ACTIONS(239), 1,
      aux_sym_color_code_css_token1,
    ACTIONS(241), 1,
      sym_color_hex,
    ACTIONS(243), 1,
      anon_sym_Enum,
    ACTIONS(247), 1,
      sym__block_string_start,
    ACTIONS(251), 1,
      sym_tuple_close,
    STATE(82), 1,
      aux_sym_macro_call_repeat1,
    STATE(162), 1,
      sym_datatype,
    STATE(207), 1,
      sym_block_string,
    ACTIONS(227), 2,
      sym_number_ambiguous,
      sym_number_offset,
    ACTIONS(231), 2,
      sym_single_quote_string,
      sym_double_quote_string,
    ACTIONS(245), 2,
      sym_rbx_asset,
      sym_rbx_content,
    ACTIONS(249), 2,
      sym_comma,
      sym_semi_colon,
    STATE(22), 2,
      sym_macro_call,
      sym_comment,
    STATE(197), 2,
      sym_enum_full,
      sym_enum_shorthand,
    ACTIONS(217), 3,
      sym_token,
      sym_static_token,
      sym_static_argument,
    STATE(195), 4,
      sym_color_code_tailwind,
      sym_color_code_skin,
      sym_color_code_brick,
      sym_color_code_css,
    STATE(212), 7,
      sym_reference,
      sym_tuple,
      sym_number,
      sym_operation,
      sym_string,
      sym_color,
      sym_enum,
  [1096] = 5,
    ACTIONS(5), 1,
      anon_sym_DASH_DASH,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(23), 2,
      sym_macro_call,
      sym_comment,
    ACTIONS(187), 9,
      sym_identifier,
      anon_sym_COLON,
      sym_tag_selector,
      sym_number_ambiguous,
      sym_number_offset,
      sym_div_operator,
      sym_sub_operator,
      sym_color_hex,
      anon_sym_Enum,
    ACTIONS(189), 31,
      sym__block_string_start,
      sym_comma,
      sym_semi_colon,
      sym_equals,
      sym_token,
      sym_static_token,
      sym_static_argument,
      aux_sym_macro_call_token1,
      anon_sym_ATpriority,
      anon_sym_ATname,
      anon_sym_ATderive,
      sym_name_selector,
      aux_sym_state_selector_token1,
      sym_pseudo_selector,
      sym_scope_open,
      sym_scope_close,
      sym_tuple_open,
      sym_number_percent,
      sym_pow_operator,
      sym_floor_div_operator,
      sym_mod_operator,
      sym_mult_operator,
      sym_add_operator,
      sym_single_quote_string,
      sym_double_quote_string,
      aux_sym_color_code_tailwind_token1,
      aux_sym_color_code_skin_token1,
      aux_sym_color_code_brick_token1,
      aux_sym_color_code_css_token1,
      sym_rbx_asset,
      sym_rbx_content,
  [1151] = 5,
    ACTIONS(5), 1,
      anon_sym_DASH_DASH,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(24), 2,
      sym_macro_call,
      sym_comment,
    ACTIONS(253), 9,
      sym_identifier,
      anon_sym_COLON,
      sym_tag_selector,
      sym_number_ambiguous,
      sym_number_offset,
      sym_div_operator,
      sym_sub_operator,
      sym_color_hex,
      anon_sym_Enum,
    ACTIONS(255), 31,
      sym__block_string_start,
      sym_comma,
      sym_semi_colon,
      sym_equals,
      sym_token,
      sym_static_token,
      sym_static_argument,
      aux_sym_macro_call_token1,
      anon_sym_ATpriority,
      anon_sym_ATname,
      anon_sym_ATderive,
      sym_name_selector,
      aux_sym_state_selector_token1,
      sym_pseudo_selector,
      sym_scope_open,
      sym_scope_close,
      sym_tuple_open,
      sym_number_percent,
      sym_pow_operator,
      sym_floor_div_operator,
      sym_mod_operator,
      sym_mult_operator,
      sym_add_operator,
      sym_single_quote_string,
      sym_double_quote_string,
      aux_sym_color_code_tailwind_token1,
      aux_sym_color_code_skin_token1,
      aux_sym_color_code_brick_token1,
      aux_sym_color_code_css_token1,
      sym_rbx_asset,
      sym_rbx_content,
  [1206] = 5,
    ACTIONS(5), 1,
      anon_sym_DASH_DASH,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(25), 2,
      sym_macro_call,
      sym_comment,
    ACTIONS(257), 9,
      sym_identifier,
      anon_sym_COLON,
      sym_tag_selector,
      sym_number_ambiguous,
      sym_number_offset,
      sym_div_operator,
      sym_sub_operator,
      sym_color_hex,
      anon_sym_Enum,
    ACTIONS(259), 31,
      sym__block_string_start,
      sym_comma,
      sym_semi_colon,
      sym_equals,
      sym_token,
      sym_static_token,
      sym_static_argument,
      aux_sym_macro_call_token1,
      anon_sym_ATpriority,
      anon_sym_ATname,
      anon_sym_ATderive,
      sym_name_selector,
      aux_sym_state_selector_token1,
      sym_pseudo_selector,
      sym_scope_open,
      sym_scope_close,
      sym_tuple_open,
      sym_number_percent,
      sym_pow_operator,
      sym_floor_div_operator,
      sym_mod_operator,
      sym_mult_operator,
      sym_add_operator,
      sym_single_quote_string,
      sym_double_quote_string,
      aux_sym_color_code_tailwind_token1,
      aux_sym_color_code_skin_token1,
      aux_sym_color_code_brick_token1,
      aux_sym_color_code_css_token1,
      sym_rbx_asset,
      sym_rbx_content,
  [1261] = 5,
    ACTIONS(5), 1,
      anon_sym_DASH_DASH,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(26), 2,
      sym_macro_call,
      sym_comment,
    ACTIONS(261), 9,
      sym_identifier,
      anon_sym_COLON,
      sym_tag_selector,
      sym_number_ambiguous,
      sym_number_offset,
      sym_div_operator,
      sym_sub_operator,
      sym_color_hex,
      anon_sym_Enum,
    ACTIONS(263), 31,
      sym__block_string_start,
      sym_comma,
      sym_semi_colon,
      sym_equals,
      sym_token,
      sym_static_token,
      sym_static_argument,
      aux_sym_macro_call_token1,
      anon_sym_ATpriority,
      anon_sym_ATname,
      anon_sym_ATderive,
      sym_name_selector,
      aux_sym_state_selector_token1,
      sym_pseudo_selector,
      sym_scope_open,
      sym_scope_close,
      sym_tuple_open,
      sym_number_percent,
      sym_pow_operator,
      sym_floor_div_operator,
      sym_mod_operator,
      sym_mult_operator,
      sym_add_operator,
      sym_single_quote_string,
      sym_double_quote_string,
      aux_sym_color_code_tailwind_token1,
      aux_sym_color_code_skin_token1,
      aux_sym_color_code_brick_token1,
      aux_sym_color_code_css_token1,
      sym_rbx_asset,
      sym_rbx_content,
  [1316] = 5,
    ACTIONS(5), 1,
      anon_sym_DASH_DASH,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(27), 2,
      sym_macro_call,
      sym_comment,
    ACTIONS(265), 9,
      sym_identifier,
      anon_sym_COLON,
      sym_tag_selector,
      sym_number_ambiguous,
      sym_number_offset,
      sym_div_operator,
      sym_sub_operator,
      sym_color_hex,
      anon_sym_Enum,
    ACTIONS(267), 31,
      sym__block_string_start,
      sym_comma,
      sym_semi_colon,
      sym_equals,
      sym_token,
      sym_static_token,
      sym_static_argument,
      aux_sym_macro_call_token1,
      anon_sym_ATpriority,
      anon_sym_ATname,
      anon_sym_ATderive,
      sym_name_selector,
      aux_sym_state_selector_token1,
      sym_pseudo_selector,
      sym_scope_open,
      sym_scope_close,
      sym_tuple_open,
      sym_number_percent,
      sym_pow_operator,
      sym_floor_div_operator,
      sym_mod_operator,
      sym_mult_operator,
      sym_add_operator,
      sym_single_quote_string,
      sym_double_quote_string,
      aux_sym_color_code_tailwind_token1,
      aux_sym_color_code_skin_token1,
      aux_sym_color_code_brick_token1,
      aux_sym_color_code_css_token1,
      sym_rbx_asset,
      sym_rbx_content,
  [1371] = 5,
    ACTIONS(5), 1,
      anon_sym_DASH_DASH,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(28), 2,
      sym_macro_call,
      sym_comment,
    ACTIONS(199), 9,
      sym_identifier,
      anon_sym_COLON,
      sym_tag_selector,
      sym_number_ambiguous,
      sym_number_offset,
      sym_div_operator,
      sym_sub_operator,
      sym_color_hex,
      anon_sym_Enum,
    ACTIONS(201), 31,
      sym__block_string_start,
      sym_comma,
      sym_semi_colon,
      sym_equals,
      sym_token,
      sym_static_token,
      sym_static_argument,
      aux_sym_macro_call_token1,
      anon_sym_ATpriority,
      anon_sym_ATname,
      anon_sym_ATderive,
      sym_name_selector,
      aux_sym_state_selector_token1,
      sym_pseudo_selector,
      sym_scope_open,
      sym_scope_close,
      sym_tuple_open,
      sym_number_percent,
      sym_pow_operator,
      sym_floor_div_operator,
      sym_mod_operator,
      sym_mult_operator,
      sym_add_operator,
      sym_single_quote_string,
      sym_double_quote_string,
      aux_sym_color_code_tailwind_token1,
      aux_sym_color_code_skin_token1,
      aux_sym_color_code_brick_token1,
      aux_sym_color_code_css_token1,
      sym_rbx_asset,
      sym_rbx_content,
  [1426] = 5,
    ACTIONS(5), 1,
      anon_sym_DASH_DASH,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(29), 2,
      sym_macro_call,
      sym_comment,
    ACTIONS(205), 9,
      sym_identifier,
      anon_sym_COLON,
      sym_tag_selector,
      sym_number_ambiguous,
      sym_number_offset,
      sym_div_operator,
      sym_sub_operator,
      sym_color_hex,
      anon_sym_Enum,
    ACTIONS(207), 31,
      sym__block_string_start,
      sym_comma,
      sym_semi_colon,
      sym_equals,
      sym_token,
      sym_static_token,
      sym_static_argument,
      aux_sym_macro_call_token1,
      anon_sym_ATpriority,
      anon_sym_ATname,
      anon_sym_ATderive,
      sym_name_selector,
      aux_sym_state_selector_token1,
      sym_pseudo_selector,
      sym_scope_open,
      sym_scope_close,
      sym_tuple_open,
      sym_number_percent,
      sym_pow_operator,
      sym_floor_div_operator,
      sym_mod_operator,
      sym_mult_operator,
      sym_add_operator,
      sym_single_quote_string,
      sym_double_quote_string,
      aux_sym_color_code_tailwind_token1,
      aux_sym_color_code_skin_token1,
      aux_sym_color_code_brick_token1,
      aux_sym_color_code_css_token1,
      sym_rbx_asset,
      sym_rbx_content,
  [1481] = 5,
    ACTIONS(5), 1,
      anon_sym_DASH_DASH,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(30), 2,
      sym_macro_call,
      sym_comment,
    ACTIONS(269), 9,
      sym_identifier,
      anon_sym_COLON,
      sym_tag_selector,
      sym_number_ambiguous,
      sym_number_offset,
      sym_div_operator,
      sym_sub_operator,
      sym_color_hex,
      anon_sym_Enum,
    ACTIONS(271), 31,
      sym__block_string_start,
      sym_comma,
      sym_semi_colon,
      sym_equals,
      sym_token,
      sym_static_token,
      sym_static_argument,
      aux_sym_macro_call_token1,
      anon_sym_ATpriority,
      anon_sym_ATname,
      anon_sym_ATderive,
      sym_name_selector,
      aux_sym_state_selector_token1,
      sym_pseudo_selector,
      sym_scope_open,
      sym_scope_close,
      sym_tuple_open,
      sym_number_percent,
      sym_pow_operator,
      sym_floor_div_operator,
      sym_mod_operator,
      sym_mult_operator,
      sym_add_operator,
      sym_single_quote_string,
      sym_double_quote_string,
      aux_sym_color_code_tailwind_token1,
      aux_sym_color_code_skin_token1,
      aux_sym_color_code_brick_token1,
      aux_sym_color_code_css_token1,
      sym_rbx_asset,
      sym_rbx_content,
  [1536] = 5,
    ACTIONS(5), 1,
      anon_sym_DASH_DASH,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(31), 2,
      sym_macro_call,
      sym_comment,
    ACTIONS(273), 9,
      sym_identifier,
      anon_sym_COLON,
      sym_tag_selector,
      sym_number_ambiguous,
      sym_number_offset,
      sym_div_operator,
      sym_sub_operator,
      sym_color_hex,
      anon_sym_Enum,
    ACTIONS(275), 31,
      sym__block_string_start,
      sym_comma,
      sym_semi_colon,
      sym_equals,
      sym_token,
      sym_static_token,
      sym_static_argument,
      aux_sym_macro_call_token1,
      anon_sym_ATpriority,
      anon_sym_ATname,
      anon_sym_ATderive,
      sym_name_selector,
      aux_sym_state_selector_token1,
      sym_pseudo_selector,
      sym_scope_open,
      sym_scope_close,
      sym_tuple_open,
      sym_number_percent,
      sym_pow_operator,
      sym_floor_div_operator,
      sym_mod_operator,
      sym_mult_operator,
      sym_add_operator,
      sym_single_quote_string,
      sym_double_quote_string,
      aux_sym_color_code_tailwind_token1,
      aux_sym_color_code_skin_token1,
      aux_sym_color_code_brick_token1,
      aux_sym_color_code_css_token1,
      sym_rbx_asset,
      sym_rbx_content,
  [1591] = 5,
    ACTIONS(5), 1,
      anon_sym_DASH_DASH,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(32), 2,
      sym_macro_call,
      sym_comment,
    ACTIONS(277), 9,
      sym_identifier,
      anon_sym_COLON,
      sym_tag_selector,
      sym_number_ambiguous,
      sym_number_offset,
      sym_div_operator,
      sym_sub_operator,
      sym_color_hex,
      anon_sym_Enum,
    ACTIONS(279), 31,
      sym__block_string_start,
      sym_comma,
      sym_semi_colon,
      sym_equals,
      sym_token,
      sym_static_token,
      sym_static_argument,
      aux_sym_macro_call_token1,
      anon_sym_ATpriority,
      anon_sym_ATname,
      anon_sym_ATderive,
      sym_name_selector,
      aux_sym_state_selector_token1,
      sym_pseudo_selector,
      sym_scope_open,
      sym_scope_close,
      sym_tuple_open,
      sym_number_percent,
      sym_pow_operator,
      sym_floor_div_operator,
      sym_mod_operator,
      sym_mult_operator,
      sym_add_operator,
      sym_single_quote_string,
      sym_double_quote_string,
      aux_sym_color_code_tailwind_token1,
      aux_sym_color_code_skin_token1,
      aux_sym_color_code_brick_token1,
      aux_sym_color_code_css_token1,
      sym_rbx_asset,
      sym_rbx_content,
  [1646] = 5,
    ACTIONS(5), 1,
      anon_sym_DASH_DASH,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(33), 2,
      sym_macro_call,
      sym_comment,
    ACTIONS(281), 9,
      sym_identifier,
      anon_sym_COLON,
      sym_tag_selector,
      sym_number_ambiguous,
      sym_number_offset,
      sym_div_operator,
      sym_sub_operator,
      sym_color_hex,
      anon_sym_Enum,
    ACTIONS(283), 31,
      sym__block_string_start,
      sym_comma,
      sym_semi_colon,
      sym_equals,
      sym_token,
      sym_static_token,
      sym_static_argument,
      aux_sym_macro_call_token1,
      anon_sym_ATpriority,
      anon_sym_ATname,
      anon_sym_ATderive,
      sym_name_selector,
      aux_sym_state_selector_token1,
      sym_pseudo_selector,
      sym_scope_open,
      sym_scope_close,
      sym_tuple_open,
      sym_number_percent,
      sym_pow_operator,
      sym_floor_div_operator,
      sym_mod_operator,
      sym_mult_operator,
      sym_add_operator,
      sym_single_quote_string,
      sym_double_quote_string,
      aux_sym_color_code_tailwind_token1,
      aux_sym_color_code_skin_token1,
      aux_sym_color_code_brick_token1,
      aux_sym_color_code_css_token1,
      sym_rbx_asset,
      sym_rbx_content,
  [1701] = 5,
    ACTIONS(5), 1,
      anon_sym_DASH_DASH,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(34), 2,
      sym_macro_call,
      sym_comment,
    ACTIONS(285), 9,
      sym_identifier,
      anon_sym_COLON,
      sym_tag_selector,
      sym_number_ambiguous,
      sym_number_offset,
      sym_div_operator,
      sym_sub_operator,
      sym_color_hex,
      anon_sym_Enum,
    ACTIONS(287), 31,
      sym__block_string_start,
      sym_comma,
      sym_semi_colon,
      sym_equals,
      sym_token,
      sym_static_token,
      sym_static_argument,
      aux_sym_macro_call_token1,
      anon_sym_ATpriority,
      anon_sym_ATname,
      anon_sym_ATderive,
      sym_name_selector,
      aux_sym_state_selector_token1,
      sym_pseudo_selector,
      sym_scope_open,
      sym_scope_close,
      sym_tuple_open,
      sym_number_percent,
      sym_pow_operator,
      sym_floor_div_operator,
      sym_mod_operator,
      sym_mult_operator,
      sym_add_operator,
      sym_single_quote_string,
      sym_double_quote_string,
      aux_sym_color_code_tailwind_token1,
      aux_sym_color_code_skin_token1,
      aux_sym_color_code_brick_token1,
      aux_sym_color_code_css_token1,
      sym_rbx_asset,
      sym_rbx_content,
  [1756] = 5,
    ACTIONS(5), 1,
      anon_sym_DASH_DASH,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(35), 2,
      sym_macro_call,
      sym_comment,
    ACTIONS(289), 9,
      sym_identifier,
      anon_sym_COLON,
      sym_tag_selector,
      sym_number_ambiguous,
      sym_number_offset,
      sym_div_operator,
      sym_sub_operator,
      sym_color_hex,
      anon_sym_Enum,
    ACTIONS(291), 31,
      sym__block_string_start,
      sym_comma,
      sym_semi_colon,
      sym_equals,
      sym_token,
      sym_static_token,
      sym_static_argument,
      aux_sym_macro_call_token1,
      anon_sym_ATpriority,
      anon_sym_ATname,
      anon_sym_ATderive,
      sym_name_selector,
      aux_sym_state_selector_token1,
      sym_pseudo_selector,
      sym_scope_open,
      sym_scope_close,
      sym_tuple_open,
      sym_number_percent,
      sym_pow_operator,
      sym_floor_div_operator,
      sym_mod_operator,
      sym_mult_operator,
      sym_add_operator,
      sym_single_quote_string,
      sym_double_quote_string,
      aux_sym_color_code_tailwind_token1,
      aux_sym_color_code_skin_token1,
      aux_sym_color_code_brick_token1,
      aux_sym_color_code_css_token1,
      sym_rbx_asset,
      sym_rbx_content,
  [1811] = 5,
    ACTIONS(5), 1,
      anon_sym_DASH_DASH,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(36), 2,
      sym_macro_call,
      sym_comment,
    ACTIONS(293), 9,
      sym_identifier,
      anon_sym_COLON,
      sym_tag_selector,
      sym_number_ambiguous,
      sym_number_offset,
      sym_div_operator,
      sym_sub_operator,
      sym_color_hex,
      anon_sym_Enum,
    ACTIONS(295), 31,
      sym__block_string_start,
      sym_comma,
      sym_semi_colon,
      sym_equals,
      sym_token,
      sym_static_token,
      sym_static_argument,
      aux_sym_macro_call_token1,
      anon_sym_ATpriority,
      anon_sym_ATname,
      anon_sym_ATderive,
      sym_name_selector,
      aux_sym_state_selector_token1,
      sym_pseudo_selector,
      sym_scope_open,
      sym_scope_close,
      sym_tuple_open,
      sym_number_percent,
      sym_pow_operator,
      sym_floor_div_operator,
      sym_mod_operator,
      sym_mult_operator,
      sym_add_operator,
      sym_single_quote_string,
      sym_double_quote_string,
      aux_sym_color_code_tailwind_token1,
      aux_sym_color_code_skin_token1,
      aux_sym_color_code_brick_token1,
      aux_sym_color_code_css_token1,
      sym_rbx_asset,
      sym_rbx_content,
  [1866] = 5,
    ACTIONS(5), 1,
      anon_sym_DASH_DASH,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(37), 2,
      sym_macro_call,
      sym_comment,
    ACTIONS(297), 9,
      sym_identifier,
      anon_sym_COLON,
      sym_tag_selector,
      sym_number_ambiguous,
      sym_number_offset,
      sym_div_operator,
      sym_sub_operator,
      sym_color_hex,
      anon_sym_Enum,
    ACTIONS(299), 31,
      sym__block_string_start,
      sym_comma,
      sym_semi_colon,
      sym_equals,
      sym_token,
      sym_static_token,
      sym_static_argument,
      aux_sym_macro_call_token1,
      anon_sym_ATpriority,
      anon_sym_ATname,
      anon_sym_ATderive,
      sym_name_selector,
      aux_sym_state_selector_token1,
      sym_pseudo_selector,
      sym_scope_open,
      sym_scope_close,
      sym_tuple_open,
      sym_number_percent,
      sym_pow_operator,
      sym_floor_div_operator,
      sym_mod_operator,
      sym_mult_operator,
      sym_add_operator,
      sym_single_quote_string,
      sym_double_quote_string,
      aux_sym_color_code_tailwind_token1,
      aux_sym_color_code_skin_token1,
      aux_sym_color_code_brick_token1,
      aux_sym_color_code_css_token1,
      sym_rbx_asset,
      sym_rbx_content,
  [1921] = 5,
    ACTIONS(5), 1,
      anon_sym_DASH_DASH,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(38), 2,
      sym_macro_call,
      sym_comment,
    ACTIONS(301), 9,
      sym_identifier,
      anon_sym_COLON,
      sym_tag_selector,
      sym_number_ambiguous,
      sym_number_offset,
      sym_div_operator,
      sym_sub_operator,
      sym_color_hex,
      anon_sym_Enum,
    ACTIONS(303), 31,
      sym__block_string_start,
      sym_comma,
      sym_semi_colon,
      sym_equals,
      sym_token,
      sym_static_token,
      sym_static_argument,
      aux_sym_macro_call_token1,
      anon_sym_ATpriority,
      anon_sym_ATname,
      anon_sym_ATderive,
      sym_name_selector,
      aux_sym_state_selector_token1,
      sym_pseudo_selector,
      sym_scope_open,
      sym_scope_close,
      sym_tuple_open,
      sym_number_percent,
      sym_pow_operator,
      sym_floor_div_operator,
      sym_mod_operator,
      sym_mult_operator,
      sym_add_operator,
      sym_single_quote_string,
      sym_double_quote_string,
      aux_sym_color_code_tailwind_token1,
      aux_sym_color_code_skin_token1,
      aux_sym_color_code_brick_token1,
      aux_sym_color_code_css_token1,
      sym_rbx_asset,
      sym_rbx_content,
  [1976] = 27,
    ACTIONS(5), 1,
      anon_sym_DASH_DASH,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(213), 1,
      sym_identifier,
    ACTIONS(219), 1,
      aux_sym_macro_call_token1,
    ACTIONS(221), 1,
      aux_sym_state_selector_token1,
    ACTIONS(223), 1,
      sym_tuple_open,
    ACTIONS(229), 1,
      sym_number_percent,
    ACTIONS(233), 1,
      aux_sym_color_code_tailwind_token1,
    ACTIONS(235), 1,
      aux_sym_color_code_skin_token1,
    ACTIONS(237), 1,
      aux_sym_color_code_brick_token1,
    ACTIONS(239), 1,
      aux_sym_color_code_css_token1,
    ACTIONS(241), 1,
      sym_color_hex,
    ACTIONS(243), 1,
      anon_sym_Enum,
    ACTIONS(247), 1,
      sym__block_string_start,
    ACTIONS(307), 1,
      sym_tuple_close,
    STATE(79), 1,
      aux_sym_macro_call_repeat1,
    STATE(175), 1,
      sym_datatype,
    STATE(207), 1,
      sym_block_string,
    ACTIONS(227), 2,
      sym_number_ambiguous,
      sym_number_offset,
    ACTIONS(231), 2,
      sym_single_quote_string,
      sym_double_quote_string,
    ACTIONS(245), 2,
      sym_rbx_asset,
      sym_rbx_content,
    ACTIONS(305), 2,
      sym_comma,
      sym_semi_colon,
    STATE(39), 2,
      sym_macro_call,
      sym_comment,
    STATE(197), 2,
      sym_enum_full,
      sym_enum_shorthand,
    ACTIONS(217), 3,
      sym_token,
      sym_static_token,
      sym_static_argument,
    STATE(195), 4,
      sym_color_code_tailwind,
      sym_color_code_skin,
      sym_color_code_brick,
      sym_color_code_css,
    STATE(212), 7,
      sym_reference,
      sym_tuple,
      sym_number,
      sym_operation,
      sym_string,
      sym_color,
      sym_enum,
  [2075] = 5,
    ACTIONS(5), 1,
      anon_sym_DASH_DASH,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(40), 2,
      sym_macro_call,
      sym_comment,
    ACTIONS(309), 9,
      sym_identifier,
      anon_sym_COLON,
      sym_tag_selector,
      sym_number_ambiguous,
      sym_number_offset,
      sym_div_operator,
      sym_sub_operator,
      sym_color_hex,
      anon_sym_Enum,
    ACTIONS(311), 31,
      sym__block_string_start,
      sym_comma,
      sym_semi_colon,
      sym_equals,
      sym_token,
      sym_static_token,
      sym_static_argument,
      aux_sym_macro_call_token1,
      anon_sym_ATpriority,
      anon_sym_ATname,
      anon_sym_ATderive,
      sym_name_selector,
      aux_sym_state_selector_token1,
      sym_pseudo_selector,
      sym_scope_open,
      sym_scope_close,
      sym_tuple_open,
      sym_number_percent,
      sym_pow_operator,
      sym_floor_div_operator,
      sym_mod_operator,
      sym_mult_operator,
      sym_add_operator,
      sym_single_quote_string,
      sym_double_quote_string,
      aux_sym_color_code_tailwind_token1,
      aux_sym_color_code_skin_token1,
      aux_sym_color_code_brick_token1,
      aux_sym_color_code_css_token1,
      sym_rbx_asset,
      sym_rbx_content,
  [2130] = 5,
    ACTIONS(5), 1,
      anon_sym_DASH_DASH,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(41), 2,
      sym_macro_call,
      sym_comment,
    ACTIONS(313), 9,
      sym_identifier,
      anon_sym_COLON,
      sym_tag_selector,
      sym_number_ambiguous,
      sym_number_offset,
      sym_div_operator,
      sym_sub_operator,
      sym_color_hex,
      anon_sym_Enum,
    ACTIONS(315), 31,
      sym__block_string_start,
      sym_comma,
      sym_semi_colon,
      sym_equals,
      sym_token,
      sym_static_token,
      sym_static_argument,
      aux_sym_macro_call_token1,
      anon_sym_ATpriority,
      anon_sym_ATname,
      anon_sym_ATderive,
      sym_name_selector,
      aux_sym_state_selector_token1,
      sym_pseudo_selector,
      sym_scope_open,
      sym_scope_close,
      sym_tuple_open,
      sym_number_percent,
      sym_pow_operator,
      sym_floor_div_operator,
      sym_mod_operator,
      sym_mult_operator,
      sym_add_operator,
      sym_single_quote_string,
      sym_double_quote_string,
      aux_sym_color_code_tailwind_token1,
      aux_sym_color_code_skin_token1,
      aux_sym_color_code_brick_token1,
      aux_sym_color_code_css_token1,
      sym_rbx_asset,
      sym_rbx_content,
  [2185] = 5,
    ACTIONS(5), 1,
      anon_sym_DASH_DASH,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(42), 2,
      sym_macro_call,
      sym_comment,
    ACTIONS(317), 9,
      sym_identifier,
      anon_sym_COLON,
      sym_tag_selector,
      sym_number_ambiguous,
      sym_number_offset,
      sym_div_operator,
      sym_sub_operator,
      sym_color_hex,
      anon_sym_Enum,
    ACTIONS(319), 31,
      sym__block_string_start,
      sym_comma,
      sym_semi_colon,
      sym_equals,
      sym_token,
      sym_static_token,
      sym_static_argument,
      aux_sym_macro_call_token1,
      anon_sym_ATpriority,
      anon_sym_ATname,
      anon_sym_ATderive,
      sym_name_selector,
      aux_sym_state_selector_token1,
      sym_pseudo_selector,
      sym_scope_open,
      sym_scope_close,
      sym_tuple_open,
      sym_number_percent,
      sym_pow_operator,
      sym_floor_div_operator,
      sym_mod_operator,
      sym_mult_operator,
      sym_add_operator,
      sym_single_quote_string,
      sym_double_quote_string,
      aux_sym_color_code_tailwind_token1,
      aux_sym_color_code_skin_token1,
      aux_sym_color_code_brick_token1,
      aux_sym_color_code_css_token1,
      sym_rbx_asset,
      sym_rbx_content,
  [2240] = 5,
    ACTIONS(5), 1,
      anon_sym_DASH_DASH,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(43), 2,
      sym_macro_call,
      sym_comment,
    ACTIONS(321), 9,
      sym_identifier,
      anon_sym_COLON,
      sym_tag_selector,
      sym_number_ambiguous,
      sym_number_offset,
      sym_div_operator,
      sym_sub_operator,
      sym_color_hex,
      anon_sym_Enum,
    ACTIONS(323), 31,
      sym__block_string_start,
      sym_comma,
      sym_semi_colon,
      sym_equals,
      sym_token,
      sym_static_token,
      sym_static_argument,
      aux_sym_macro_call_token1,
      anon_sym_ATpriority,
      anon_sym_ATname,
      anon_sym_ATderive,
      sym_name_selector,
      aux_sym_state_selector_token1,
      sym_pseudo_selector,
      sym_scope_open,
      sym_scope_close,
      sym_tuple_open,
      sym_number_percent,
      sym_pow_operator,
      sym_floor_div_operator,
      sym_mod_operator,
      sym_mult_operator,
      sym_add_operator,
      sym_single_quote_string,
      sym_double_quote_string,
      aux_sym_color_code_tailwind_token1,
      aux_sym_color_code_skin_token1,
      aux_sym_color_code_brick_token1,
      aux_sym_color_code_css_token1,
      sym_rbx_asset,
      sym_rbx_content,
  [2295] = 27,
    ACTIONS(5), 1,
      anon_sym_DASH_DASH,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(213), 1,
      sym_identifier,
    ACTIONS(219), 1,
      aux_sym_macro_call_token1,
    ACTIONS(221), 1,
      aux_sym_state_selector_token1,
    ACTIONS(223), 1,
      sym_tuple_open,
    ACTIONS(229), 1,
      sym_number_percent,
    ACTIONS(233), 1,
      aux_sym_color_code_tailwind_token1,
    ACTIONS(235), 1,
      aux_sym_color_code_skin_token1,
    ACTIONS(237), 1,
      aux_sym_color_code_brick_token1,
    ACTIONS(239), 1,
      aux_sym_color_code_css_token1,
    ACTIONS(241), 1,
      sym_color_hex,
    ACTIONS(243), 1,
      anon_sym_Enum,
    ACTIONS(247), 1,
      sym__block_string_start,
    ACTIONS(327), 1,
      sym_tuple_close,
    STATE(87), 1,
      aux_sym_macro_call_repeat1,
    STATE(165), 1,
      sym_datatype,
    STATE(207), 1,
      sym_block_string,
    ACTIONS(227), 2,
      sym_number_ambiguous,
      sym_number_offset,
    ACTIONS(231), 2,
      sym_single_quote_string,
      sym_double_quote_string,
    ACTIONS(245), 2,
      sym_rbx_asset,
      sym_rbx_content,
    ACTIONS(325), 2,
      sym_comma,
      sym_semi_colon,
    STATE(44), 2,
      sym_macro_call,
      sym_comment,
    STATE(197), 2,
      sym_enum_full,
      sym_enum_shorthand,
    ACTIONS(217), 3,
      sym_token,
      sym_static_token,
      sym_static_argument,
    STATE(195), 4,
      sym_color_code_tailwind,
      sym_color_code_skin,
      sym_color_code_brick,
      sym_color_code_css,
    STATE(212), 7,
      sym_reference,
      sym_tuple,
      sym_number,
      sym_operation,
      sym_string,
      sym_color,
      sym_enum,
  [2394] = 5,
    ACTIONS(5), 1,
      anon_sym_DASH_DASH,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(45), 2,
      sym_macro_call,
      sym_comment,
    ACTIONS(329), 9,
      sym_identifier,
      anon_sym_COLON,
      sym_tag_selector,
      sym_number_ambiguous,
      sym_number_offset,
      sym_div_operator,
      sym_sub_operator,
      sym_color_hex,
      anon_sym_Enum,
    ACTIONS(331), 31,
      sym__block_string_start,
      sym_comma,
      sym_semi_colon,
      sym_equals,
      sym_token,
      sym_static_token,
      sym_static_argument,
      aux_sym_macro_call_token1,
      anon_sym_ATpriority,
      anon_sym_ATname,
      anon_sym_ATderive,
      sym_name_selector,
      aux_sym_state_selector_token1,
      sym_pseudo_selector,
      sym_scope_open,
      sym_scope_close,
      sym_tuple_open,
      sym_number_percent,
      sym_pow_operator,
      sym_floor_div_operator,
      sym_mod_operator,
      sym_mult_operator,
      sym_add_operator,
      sym_single_quote_string,
      sym_double_quote_string,
      aux_sym_color_code_tailwind_token1,
      aux_sym_color_code_skin_token1,
      aux_sym_color_code_brick_token1,
      aux_sym_color_code_css_token1,
      sym_rbx_asset,
      sym_rbx_content,
  [2449] = 5,
    ACTIONS(5), 1,
      anon_sym_DASH_DASH,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(46), 2,
      sym_macro_call,
      sym_comment,
    ACTIONS(333), 9,
      sym_identifier,
      anon_sym_COLON,
      sym_tag_selector,
      sym_number_ambiguous,
      sym_number_offset,
      sym_div_operator,
      sym_sub_operator,
      sym_color_hex,
      anon_sym_Enum,
    ACTIONS(335), 31,
      sym__block_string_start,
      sym_comma,
      sym_semi_colon,
      sym_equals,
      sym_token,
      sym_static_token,
      sym_static_argument,
      aux_sym_macro_call_token1,
      anon_sym_ATpriority,
      anon_sym_ATname,
      anon_sym_ATderive,
      sym_name_selector,
      aux_sym_state_selector_token1,
      sym_pseudo_selector,
      sym_scope_open,
      sym_scope_close,
      sym_tuple_open,
      sym_number_percent,
      sym_pow_operator,
      sym_floor_div_operator,
      sym_mod_operator,
      sym_mult_operator,
      sym_add_operator,
      sym_single_quote_string,
      sym_double_quote_string,
      aux_sym_color_code_tailwind_token1,
      aux_sym_color_code_skin_token1,
      aux_sym_color_code_brick_token1,
      aux_sym_color_code_css_token1,
      sym_rbx_asset,
      sym_rbx_content,
  [2504] = 5,
    ACTIONS(5), 1,
      anon_sym_DASH_DASH,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(47), 2,
      sym_macro_call,
      sym_comment,
    ACTIONS(337), 9,
      sym_identifier,
      anon_sym_COLON,
      sym_tag_selector,
      sym_number_ambiguous,
      sym_number_offset,
      sym_div_operator,
      sym_sub_operator,
      sym_color_hex,
      anon_sym_Enum,
    ACTIONS(339), 31,
      sym__block_string_start,
      sym_comma,
      sym_semi_colon,
      sym_equals,
      sym_token,
      sym_static_token,
      sym_static_argument,
      aux_sym_macro_call_token1,
      anon_sym_ATpriority,
      anon_sym_ATname,
      anon_sym_ATderive,
      sym_name_selector,
      aux_sym_state_selector_token1,
      sym_pseudo_selector,
      sym_scope_open,
      sym_scope_close,
      sym_tuple_open,
      sym_number_percent,
      sym_pow_operator,
      sym_floor_div_operator,
      sym_mod_operator,
      sym_mult_operator,
      sym_add_operator,
      sym_single_quote_string,
      sym_double_quote_string,
      aux_sym_color_code_tailwind_token1,
      aux_sym_color_code_skin_token1,
      aux_sym_color_code_brick_token1,
      aux_sym_color_code_css_token1,
      sym_rbx_asset,
      sym_rbx_content,
  [2559] = 5,
    ACTIONS(5), 1,
      anon_sym_DASH_DASH,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(48), 2,
      sym_macro_call,
      sym_comment,
    ACTIONS(341), 9,
      sym_identifier,
      anon_sym_COLON,
      sym_tag_selector,
      sym_number_ambiguous,
      sym_number_offset,
      sym_div_operator,
      sym_sub_operator,
      sym_color_hex,
      anon_sym_Enum,
    ACTIONS(343), 31,
      sym__block_string_start,
      sym_comma,
      sym_semi_colon,
      sym_equals,
      sym_token,
      sym_static_token,
      sym_static_argument,
      aux_sym_macro_call_token1,
      anon_sym_ATpriority,
      anon_sym_ATname,
      anon_sym_ATderive,
      sym_name_selector,
      aux_sym_state_selector_token1,
      sym_pseudo_selector,
      sym_scope_open,
      sym_scope_close,
      sym_tuple_open,
      sym_number_percent,
      sym_pow_operator,
      sym_floor_div_operator,
      sym_mod_operator,
      sym_mult_operator,
      sym_add_operator,
      sym_single_quote_string,
      sym_double_quote_string,
      aux_sym_color_code_tailwind_token1,
      aux_sym_color_code_skin_token1,
      aux_sym_color_code_brick_token1,
      aux_sym_color_code_css_token1,
      sym_rbx_asset,
      sym_rbx_content,
  [2614] = 5,
    ACTIONS(5), 1,
      anon_sym_DASH_DASH,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(49), 2,
      sym_macro_call,
      sym_comment,
    ACTIONS(345), 9,
      sym_identifier,
      anon_sym_COLON,
      sym_tag_selector,
      sym_number_ambiguous,
      sym_number_offset,
      sym_div_operator,
      sym_sub_operator,
      sym_color_hex,
      anon_sym_Enum,
    ACTIONS(347), 31,
      sym__block_string_start,
      sym_comma,
      sym_semi_colon,
      sym_equals,
      sym_token,
      sym_static_token,
      sym_static_argument,
      aux_sym_macro_call_token1,
      anon_sym_ATpriority,
      anon_sym_ATname,
      anon_sym_ATderive,
      sym_name_selector,
      aux_sym_state_selector_token1,
      sym_pseudo_selector,
      sym_scope_open,
      sym_scope_close,
      sym_tuple_open,
      sym_number_percent,
      sym_pow_operator,
      sym_floor_div_operator,
      sym_mod_operator,
      sym_mult_operator,
      sym_add_operator,
      sym_single_quote_string,
      sym_double_quote_string,
      aux_sym_color_code_tailwind_token1,
      aux_sym_color_code_skin_token1,
      aux_sym_color_code_brick_token1,
      aux_sym_color_code_css_token1,
      sym_rbx_asset,
      sym_rbx_content,
  [2669] = 5,
    ACTIONS(5), 1,
      anon_sym_DASH_DASH,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(50), 2,
      sym_macro_call,
      sym_comment,
    ACTIONS(349), 9,
      sym_identifier,
      anon_sym_COLON,
      sym_tag_selector,
      sym_number_ambiguous,
      sym_number_offset,
      sym_div_operator,
      sym_sub_operator,
      sym_color_hex,
      anon_sym_Enum,
    ACTIONS(351), 31,
      sym__block_string_start,
      sym_comma,
      sym_semi_colon,
      sym_equals,
      sym_token,
      sym_static_token,
      sym_static_argument,
      aux_sym_macro_call_token1,
      anon_sym_ATpriority,
      anon_sym_ATname,
      anon_sym_ATderive,
      sym_name_selector,
      aux_sym_state_selector_token1,
      sym_pseudo_selector,
      sym_scope_open,
      sym_scope_close,
      sym_tuple_open,
      sym_number_percent,
      sym_pow_operator,
      sym_floor_div_operator,
      sym_mod_operator,
      sym_mult_operator,
      sym_add_operator,
      sym_single_quote_string,
      sym_double_quote_string,
      aux_sym_color_code_tailwind_token1,
      aux_sym_color_code_skin_token1,
      aux_sym_color_code_brick_token1,
      aux_sym_color_code_css_token1,
      sym_rbx_asset,
      sym_rbx_content,
  [2724] = 5,
    ACTIONS(5), 1,
      anon_sym_DASH_DASH,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(51), 2,
      sym_macro_call,
      sym_comment,
    ACTIONS(353), 9,
      sym_identifier,
      anon_sym_COLON,
      sym_tag_selector,
      sym_number_ambiguous,
      sym_number_offset,
      sym_div_operator,
      sym_sub_operator,
      sym_color_hex,
      anon_sym_Enum,
    ACTIONS(355), 31,
      sym__block_string_start,
      sym_comma,
      sym_semi_colon,
      sym_equals,
      sym_token,
      sym_static_token,
      sym_static_argument,
      aux_sym_macro_call_token1,
      anon_sym_ATpriority,
      anon_sym_ATname,
      anon_sym_ATderive,
      sym_name_selector,
      aux_sym_state_selector_token1,
      sym_pseudo_selector,
      sym_scope_open,
      sym_scope_close,
      sym_tuple_open,
      sym_number_percent,
      sym_pow_operator,
      sym_floor_div_operator,
      sym_mod_operator,
      sym_mult_operator,
      sym_add_operator,
      sym_single_quote_string,
      sym_double_quote_string,
      aux_sym_color_code_tailwind_token1,
      aux_sym_color_code_skin_token1,
      aux_sym_color_code_brick_token1,
      aux_sym_color_code_css_token1,
      sym_rbx_asset,
      sym_rbx_content,
  [2779] = 5,
    ACTIONS(5), 1,
      anon_sym_DASH_DASH,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(52), 2,
      sym_macro_call,
      sym_comment,
    ACTIONS(357), 9,
      sym_identifier,
      anon_sym_COLON,
      sym_tag_selector,
      sym_number_ambiguous,
      sym_number_offset,
      sym_div_operator,
      sym_sub_operator,
      sym_color_hex,
      anon_sym_Enum,
    ACTIONS(359), 31,
      sym__block_string_start,
      sym_comma,
      sym_semi_colon,
      sym_equals,
      sym_token,
      sym_static_token,
      sym_static_argument,
      aux_sym_macro_call_token1,
      anon_sym_ATpriority,
      anon_sym_ATname,
      anon_sym_ATderive,
      sym_name_selector,
      aux_sym_state_selector_token1,
      sym_pseudo_selector,
      sym_scope_open,
      sym_scope_close,
      sym_tuple_open,
      sym_number_percent,
      sym_pow_operator,
      sym_floor_div_operator,
      sym_mod_operator,
      sym_mult_operator,
      sym_add_operator,
      sym_single_quote_string,
      sym_double_quote_string,
      aux_sym_color_code_tailwind_token1,
      aux_sym_color_code_skin_token1,
      aux_sym_color_code_brick_token1,
      aux_sym_color_code_css_token1,
      sym_rbx_asset,
      sym_rbx_content,
  [2834] = 5,
    ACTIONS(5), 1,
      anon_sym_DASH_DASH,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(53), 2,
      sym_macro_call,
      sym_comment,
    ACTIONS(361), 9,
      sym_identifier,
      anon_sym_COLON,
      sym_tag_selector,
      sym_number_ambiguous,
      sym_number_offset,
      sym_div_operator,
      sym_sub_operator,
      sym_color_hex,
      anon_sym_Enum,
    ACTIONS(363), 31,
      sym__block_string_start,
      sym_comma,
      sym_semi_colon,
      sym_equals,
      sym_token,
      sym_static_token,
      sym_static_argument,
      aux_sym_macro_call_token1,
      anon_sym_ATpriority,
      anon_sym_ATname,
      anon_sym_ATderive,
      sym_name_selector,
      aux_sym_state_selector_token1,
      sym_pseudo_selector,
      sym_scope_open,
      sym_scope_close,
      sym_tuple_open,
      sym_number_percent,
      sym_pow_operator,
      sym_floor_div_operator,
      sym_mod_operator,
      sym_mult_operator,
      sym_add_operator,
      sym_single_quote_string,
      sym_double_quote_string,
      aux_sym_color_code_tailwind_token1,
      aux_sym_color_code_skin_token1,
      aux_sym_color_code_brick_token1,
      aux_sym_color_code_css_token1,
      sym_rbx_asset,
      sym_rbx_content,
  [2889] = 5,
    ACTIONS(5), 1,
      anon_sym_DASH_DASH,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(54), 2,
      sym_macro_call,
      sym_comment,
    ACTIONS(365), 9,
      sym_identifier,
      anon_sym_COLON,
      sym_tag_selector,
      sym_number_ambiguous,
      sym_number_offset,
      sym_div_operator,
      sym_sub_operator,
      sym_color_hex,
      anon_sym_Enum,
    ACTIONS(367), 31,
      sym__block_string_start,
      sym_comma,
      sym_semi_colon,
      sym_equals,
      sym_token,
      sym_static_token,
      sym_static_argument,
      aux_sym_macro_call_token1,
      anon_sym_ATpriority,
      anon_sym_ATname,
      anon_sym_ATderive,
      sym_name_selector,
      aux_sym_state_selector_token1,
      sym_pseudo_selector,
      sym_scope_open,
      sym_scope_close,
      sym_tuple_open,
      sym_number_percent,
      sym_pow_operator,
      sym_floor_div_operator,
      sym_mod_operator,
      sym_mult_operator,
      sym_add_operator,
      sym_single_quote_string,
      sym_double_quote_string,
      aux_sym_color_code_tailwind_token1,
      aux_sym_color_code_skin_token1,
      aux_sym_color_code_brick_token1,
      aux_sym_color_code_css_token1,
      sym_rbx_asset,
      sym_rbx_content,
  [2944] = 5,
    ACTIONS(5), 1,
      anon_sym_DASH_DASH,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(55), 2,
      sym_macro_call,
      sym_comment,
    ACTIONS(369), 9,
      sym_identifier,
      anon_sym_COLON,
      sym_tag_selector,
      sym_number_ambiguous,
      sym_number_offset,
      sym_div_operator,
      sym_sub_operator,
      sym_color_hex,
      anon_sym_Enum,
    ACTIONS(371), 31,
      sym__block_string_start,
      sym_comma,
      sym_semi_colon,
      sym_equals,
      sym_token,
      sym_static_token,
      sym_static_argument,
      aux_sym_macro_call_token1,
      anon_sym_ATpriority,
      anon_sym_ATname,
      anon_sym_ATderive,
      sym_name_selector,
      aux_sym_state_selector_token1,
      sym_pseudo_selector,
      sym_scope_open,
      sym_scope_close,
      sym_tuple_open,
      sym_number_percent,
      sym_pow_operator,
      sym_floor_div_operator,
      sym_mod_operator,
      sym_mult_operator,
      sym_add_operator,
      sym_single_quote_string,
      sym_double_quote_string,
      aux_sym_color_code_tailwind_token1,
      aux_sym_color_code_skin_token1,
      aux_sym_color_code_brick_token1,
      aux_sym_color_code_css_token1,
      sym_rbx_asset,
      sym_rbx_content,
  [2999] = 5,
    ACTIONS(5), 1,
      anon_sym_DASH_DASH,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(56), 2,
      sym_macro_call,
      sym_comment,
    ACTIONS(373), 9,
      sym_identifier,
      anon_sym_COLON,
      sym_tag_selector,
      sym_number_ambiguous,
      sym_number_offset,
      sym_div_operator,
      sym_sub_operator,
      sym_color_hex,
      anon_sym_Enum,
    ACTIONS(375), 31,
      sym__block_string_start,
      sym_comma,
      sym_semi_colon,
      sym_equals,
      sym_token,
      sym_static_token,
      sym_static_argument,
      aux_sym_macro_call_token1,
      anon_sym_ATpriority,
      anon_sym_ATname,
      anon_sym_ATderive,
      sym_name_selector,
      aux_sym_state_selector_token1,
      sym_pseudo_selector,
      sym_scope_open,
      sym_scope_close,
      sym_tuple_open,
      sym_number_percent,
      sym_pow_operator,
      sym_floor_div_operator,
      sym_mod_operator,
      sym_mult_operator,
      sym_add_operator,
      sym_single_quote_string,
      sym_double_quote_string,
      aux_sym_color_code_tailwind_token1,
      aux_sym_color_code_skin_token1,
      aux_sym_color_code_brick_token1,
      aux_sym_color_code_css_token1,
      sym_rbx_asset,
      sym_rbx_content,
  [3054] = 5,
    ACTIONS(5), 1,
      anon_sym_DASH_DASH,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(57), 2,
      sym_macro_call,
      sym_comment,
    ACTIONS(187), 9,
      sym_identifier,
      anon_sym_COLON,
      sym_tag_selector,
      sym_number_ambiguous,
      sym_number_offset,
      sym_div_operator,
      sym_sub_operator,
      sym_color_hex,
      anon_sym_Enum,
    ACTIONS(189), 31,
      sym__block_string_start,
      sym_comma,
      sym_semi_colon,
      sym_equals,
      sym_token,
      sym_static_token,
      sym_static_argument,
      aux_sym_macro_call_token1,
      anon_sym_ATpriority,
      anon_sym_ATname,
      anon_sym_ATderive,
      sym_name_selector,
      aux_sym_state_selector_token1,
      sym_pseudo_selector,
      sym_scope_open,
      sym_scope_close,
      sym_tuple_open,
      sym_number_percent,
      sym_pow_operator,
      sym_floor_div_operator,
      sym_mod_operator,
      sym_mult_operator,
      sym_add_operator,
      sym_single_quote_string,
      sym_double_quote_string,
      aux_sym_color_code_tailwind_token1,
      aux_sym_color_code_skin_token1,
      aux_sym_color_code_brick_token1,
      aux_sym_color_code_css_token1,
      sym_rbx_asset,
      sym_rbx_content,
  [3109] = 5,
    ACTIONS(5), 1,
      anon_sym_DASH_DASH,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(58), 2,
      sym_macro_call,
      sym_comment,
    ACTIONS(377), 9,
      sym_identifier,
      anon_sym_COLON,
      sym_tag_selector,
      sym_number_ambiguous,
      sym_number_offset,
      sym_div_operator,
      sym_sub_operator,
      sym_color_hex,
      anon_sym_Enum,
    ACTIONS(379), 31,
      sym__block_string_start,
      sym_comma,
      sym_semi_colon,
      sym_equals,
      sym_token,
      sym_static_token,
      sym_static_argument,
      aux_sym_macro_call_token1,
      anon_sym_ATpriority,
      anon_sym_ATname,
      anon_sym_ATderive,
      sym_name_selector,
      aux_sym_state_selector_token1,
      sym_pseudo_selector,
      sym_scope_open,
      sym_scope_close,
      sym_tuple_open,
      sym_number_percent,
      sym_pow_operator,
      sym_floor_div_operator,
      sym_mod_operator,
      sym_mult_operator,
      sym_add_operator,
      sym_single_quote_string,
      sym_double_quote_string,
      aux_sym_color_code_tailwind_token1,
      aux_sym_color_code_skin_token1,
      aux_sym_color_code_brick_token1,
      aux_sym_color_code_css_token1,
      sym_rbx_asset,
      sym_rbx_content,
  [3164] = 5,
    ACTIONS(5), 1,
      anon_sym_DASH_DASH,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(59), 2,
      sym_macro_call,
      sym_comment,
    ACTIONS(381), 9,
      sym_identifier,
      anon_sym_COLON,
      sym_tag_selector,
      sym_number_ambiguous,
      sym_number_offset,
      sym_div_operator,
      sym_sub_operator,
      sym_color_hex,
      anon_sym_Enum,
    ACTIONS(383), 31,
      sym__block_string_start,
      sym_comma,
      sym_semi_colon,
      sym_equals,
      sym_token,
      sym_static_token,
      sym_static_argument,
      aux_sym_macro_call_token1,
      anon_sym_ATpriority,
      anon_sym_ATname,
      anon_sym_ATderive,
      sym_name_selector,
      aux_sym_state_selector_token1,
      sym_pseudo_selector,
      sym_scope_open,
      sym_scope_close,
      sym_tuple_open,
      sym_number_percent,
      sym_pow_operator,
      sym_floor_div_operator,
      sym_mod_operator,
      sym_mult_operator,
      sym_add_operator,
      sym_single_quote_string,
      sym_double_quote_string,
      aux_sym_color_code_tailwind_token1,
      aux_sym_color_code_skin_token1,
      aux_sym_color_code_brick_token1,
      aux_sym_color_code_css_token1,
      sym_rbx_asset,
      sym_rbx_content,
  [3219] = 5,
    ACTIONS(5), 1,
      anon_sym_DASH_DASH,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(60), 2,
      sym_macro_call,
      sym_comment,
    ACTIONS(385), 9,
      sym_identifier,
      anon_sym_COLON,
      sym_tag_selector,
      sym_number_ambiguous,
      sym_number_offset,
      sym_div_operator,
      sym_sub_operator,
      sym_color_hex,
      anon_sym_Enum,
    ACTIONS(387), 31,
      sym__block_string_start,
      sym_comma,
      sym_semi_colon,
      sym_equals,
      sym_token,
      sym_static_token,
      sym_static_argument,
      aux_sym_macro_call_token1,
      anon_sym_ATpriority,
      anon_sym_ATname,
      anon_sym_ATderive,
      sym_name_selector,
      aux_sym_state_selector_token1,
      sym_pseudo_selector,
      sym_scope_open,
      sym_scope_close,
      sym_tuple_open,
      sym_number_percent,
      sym_pow_operator,
      sym_floor_div_operator,
      sym_mod_operator,
      sym_mult_operator,
      sym_add_operator,
      sym_single_quote_string,
      sym_double_quote_string,
      aux_sym_color_code_tailwind_token1,
      aux_sym_color_code_skin_token1,
      aux_sym_color_code_brick_token1,
      aux_sym_color_code_css_token1,
      sym_rbx_asset,
      sym_rbx_content,
  [3274] = 5,
    ACTIONS(5), 1,
      anon_sym_DASH_DASH,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(61), 2,
      sym_macro_call,
      sym_comment,
    ACTIONS(389), 9,
      sym_identifier,
      anon_sym_COLON,
      sym_tag_selector,
      sym_number_ambiguous,
      sym_number_offset,
      sym_div_operator,
      sym_sub_operator,
      sym_color_hex,
      anon_sym_Enum,
    ACTIONS(391), 31,
      sym__block_string_start,
      sym_comma,
      sym_semi_colon,
      sym_equals,
      sym_token,
      sym_static_token,
      sym_static_argument,
      aux_sym_macro_call_token1,
      anon_sym_ATpriority,
      anon_sym_ATname,
      anon_sym_ATderive,
      sym_name_selector,
      aux_sym_state_selector_token1,
      sym_pseudo_selector,
      sym_scope_open,
      sym_scope_close,
      sym_tuple_open,
      sym_number_percent,
      sym_pow_operator,
      sym_floor_div_operator,
      sym_mod_operator,
      sym_mult_operator,
      sym_add_operator,
      sym_single_quote_string,
      sym_double_quote_string,
      aux_sym_color_code_tailwind_token1,
      aux_sym_color_code_skin_token1,
      aux_sym_color_code_brick_token1,
      aux_sym_color_code_css_token1,
      sym_rbx_asset,
      sym_rbx_content,
  [3329] = 5,
    ACTIONS(5), 1,
      anon_sym_DASH_DASH,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(62), 2,
      sym_macro_call,
      sym_comment,
    ACTIONS(393), 9,
      sym_identifier,
      anon_sym_COLON,
      sym_tag_selector,
      sym_number_ambiguous,
      sym_number_offset,
      sym_div_operator,
      sym_sub_operator,
      sym_color_hex,
      anon_sym_Enum,
    ACTIONS(395), 31,
      sym__block_string_start,
      sym_comma,
      sym_semi_colon,
      sym_equals,
      sym_token,
      sym_static_token,
      sym_static_argument,
      aux_sym_macro_call_token1,
      anon_sym_ATpriority,
      anon_sym_ATname,
      anon_sym_ATderive,
      sym_name_selector,
      aux_sym_state_selector_token1,
      sym_pseudo_selector,
      sym_scope_open,
      sym_scope_close,
      sym_tuple_open,
      sym_number_percent,
      sym_pow_operator,
      sym_floor_div_operator,
      sym_mod_operator,
      sym_mult_operator,
      sym_add_operator,
      sym_single_quote_string,
      sym_double_quote_string,
      aux_sym_color_code_tailwind_token1,
      aux_sym_color_code_skin_token1,
      aux_sym_color_code_brick_token1,
      aux_sym_color_code_css_token1,
      sym_rbx_asset,
      sym_rbx_content,
  [3384] = 5,
    ACTIONS(5), 1,
      anon_sym_DASH_DASH,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(63), 2,
      sym_macro_call,
      sym_comment,
    ACTIONS(397), 9,
      sym_identifier,
      anon_sym_COLON,
      sym_tag_selector,
      sym_number_ambiguous,
      sym_number_offset,
      sym_div_operator,
      sym_sub_operator,
      sym_color_hex,
      anon_sym_Enum,
    ACTIONS(399), 31,
      sym__block_string_start,
      sym_comma,
      sym_semi_colon,
      sym_equals,
      sym_token,
      sym_static_token,
      sym_static_argument,
      aux_sym_macro_call_token1,
      anon_sym_ATpriority,
      anon_sym_ATname,
      anon_sym_ATderive,
      sym_name_selector,
      aux_sym_state_selector_token1,
      sym_pseudo_selector,
      sym_scope_open,
      sym_scope_close,
      sym_tuple_open,
      sym_number_percent,
      sym_pow_operator,
      sym_floor_div_operator,
      sym_mod_operator,
      sym_mult_operator,
      sym_add_operator,
      sym_single_quote_string,
      sym_double_quote_string,
      aux_sym_color_code_tailwind_token1,
      aux_sym_color_code_skin_token1,
      aux_sym_color_code_brick_token1,
      aux_sym_color_code_css_token1,
      sym_rbx_asset,
      sym_rbx_content,
  [3439] = 5,
    ACTIONS(5), 1,
      anon_sym_DASH_DASH,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(64), 2,
      sym_macro_call,
      sym_comment,
    ACTIONS(401), 9,
      sym_identifier,
      anon_sym_COLON,
      sym_tag_selector,
      sym_number_ambiguous,
      sym_number_offset,
      sym_div_operator,
      sym_sub_operator,
      sym_color_hex,
      anon_sym_Enum,
    ACTIONS(403), 31,
      sym__block_string_start,
      sym_comma,
      sym_semi_colon,
      sym_equals,
      sym_token,
      sym_static_token,
      sym_static_argument,
      aux_sym_macro_call_token1,
      anon_sym_ATpriority,
      anon_sym_ATname,
      anon_sym_ATderive,
      sym_name_selector,
      aux_sym_state_selector_token1,
      sym_pseudo_selector,
      sym_scope_open,
      sym_scope_close,
      sym_tuple_open,
      sym_number_percent,
      sym_pow_operator,
      sym_floor_div_operator,
      sym_mod_operator,
      sym_mult_operator,
      sym_add_operator,
      sym_single_quote_string,
      sym_double_quote_string,
      aux_sym_color_code_tailwind_token1,
      aux_sym_color_code_skin_token1,
      aux_sym_color_code_brick_token1,
      aux_sym_color_code_css_token1,
      sym_rbx_asset,
      sym_rbx_content,
  [3494] = 5,
    ACTIONS(5), 1,
      anon_sym_DASH_DASH,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(65), 2,
      sym_macro_call,
      sym_comment,
    ACTIONS(405), 9,
      sym_identifier,
      anon_sym_COLON,
      sym_tag_selector,
      sym_number_ambiguous,
      sym_number_offset,
      sym_div_operator,
      sym_sub_operator,
      sym_color_hex,
      anon_sym_Enum,
    ACTIONS(407), 31,
      sym__block_string_start,
      sym_comma,
      sym_semi_colon,
      sym_equals,
      sym_token,
      sym_static_token,
      sym_static_argument,
      aux_sym_macro_call_token1,
      anon_sym_ATpriority,
      anon_sym_ATname,
      anon_sym_ATderive,
      sym_name_selector,
      aux_sym_state_selector_token1,
      sym_pseudo_selector,
      sym_scope_open,
      sym_scope_close,
      sym_tuple_open,
      sym_number_percent,
      sym_pow_operator,
      sym_floor_div_operator,
      sym_mod_operator,
      sym_mult_operator,
      sym_add_operator,
      sym_single_quote_string,
      sym_double_quote_string,
      aux_sym_color_code_tailwind_token1,
      aux_sym_color_code_skin_token1,
      aux_sym_color_code_brick_token1,
      aux_sym_color_code_css_token1,
      sym_rbx_asset,
      sym_rbx_content,
  [3549] = 5,
    ACTIONS(5), 1,
      anon_sym_DASH_DASH,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(66), 2,
      sym_macro_call,
      sym_comment,
    ACTIONS(409), 9,
      sym_identifier,
      anon_sym_COLON,
      sym_tag_selector,
      sym_number_ambiguous,
      sym_number_offset,
      sym_div_operator,
      sym_sub_operator,
      sym_color_hex,
      anon_sym_Enum,
    ACTIONS(411), 31,
      sym__block_string_start,
      sym_comma,
      sym_semi_colon,
      sym_equals,
      sym_token,
      sym_static_token,
      sym_static_argument,
      aux_sym_macro_call_token1,
      anon_sym_ATpriority,
      anon_sym_ATname,
      anon_sym_ATderive,
      sym_name_selector,
      aux_sym_state_selector_token1,
      sym_pseudo_selector,
      sym_scope_open,
      sym_scope_close,
      sym_tuple_open,
      sym_number_percent,
      sym_pow_operator,
      sym_floor_div_operator,
      sym_mod_operator,
      sym_mult_operator,
      sym_add_operator,
      sym_single_quote_string,
      sym_double_quote_string,
      aux_sym_color_code_tailwind_token1,
      aux_sym_color_code_skin_token1,
      aux_sym_color_code_brick_token1,
      aux_sym_color_code_css_token1,
      sym_rbx_asset,
      sym_rbx_content,
  [3604] = 5,
    ACTIONS(5), 1,
      anon_sym_DASH_DASH,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(67), 2,
      sym_macro_call,
      sym_comment,
    ACTIONS(413), 9,
      sym_identifier,
      anon_sym_COLON,
      sym_tag_selector,
      sym_number_ambiguous,
      sym_number_offset,
      sym_div_operator,
      sym_sub_operator,
      sym_color_hex,
      anon_sym_Enum,
    ACTIONS(415), 31,
      sym__block_string_start,
      sym_comma,
      sym_semi_colon,
      sym_equals,
      sym_token,
      sym_static_token,
      sym_static_argument,
      aux_sym_macro_call_token1,
      anon_sym_ATpriority,
      anon_sym_ATname,
      anon_sym_ATderive,
      sym_name_selector,
      aux_sym_state_selector_token1,
      sym_pseudo_selector,
      sym_scope_open,
      sym_scope_close,
      sym_tuple_open,
      sym_number_percent,
      sym_pow_operator,
      sym_floor_div_operator,
      sym_mod_operator,
      sym_mult_operator,
      sym_add_operator,
      sym_single_quote_string,
      sym_double_quote_string,
      aux_sym_color_code_tailwind_token1,
      aux_sym_color_code_skin_token1,
      aux_sym_color_code_brick_token1,
      aux_sym_color_code_css_token1,
      sym_rbx_asset,
      sym_rbx_content,
  [3659] = 5,
    ACTIONS(5), 1,
      anon_sym_DASH_DASH,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(68), 2,
      sym_macro_call,
      sym_comment,
    ACTIONS(417), 9,
      sym_identifier,
      anon_sym_COLON,
      sym_tag_selector,
      sym_number_ambiguous,
      sym_number_offset,
      sym_div_operator,
      sym_sub_operator,
      sym_color_hex,
      anon_sym_Enum,
    ACTIONS(419), 31,
      sym__block_string_start,
      sym_comma,
      sym_semi_colon,
      sym_equals,
      sym_token,
      sym_static_token,
      sym_static_argument,
      aux_sym_macro_call_token1,
      anon_sym_ATpriority,
      anon_sym_ATname,
      anon_sym_ATderive,
      sym_name_selector,
      aux_sym_state_selector_token1,
      sym_pseudo_selector,
      sym_scope_open,
      sym_scope_close,
      sym_tuple_open,
      sym_number_percent,
      sym_pow_operator,
      sym_floor_div_operator,
      sym_mod_operator,
      sym_mult_operator,
      sym_add_operator,
      sym_single_quote_string,
      sym_double_quote_string,
      aux_sym_color_code_tailwind_token1,
      aux_sym_color_code_skin_token1,
      aux_sym_color_code_brick_token1,
      aux_sym_color_code_css_token1,
      sym_rbx_asset,
      sym_rbx_content,
  [3714] = 5,
    ACTIONS(5), 1,
      anon_sym_DASH_DASH,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(69), 2,
      sym_macro_call,
      sym_comment,
    ACTIONS(421), 9,
      sym_identifier,
      anon_sym_COLON,
      sym_tag_selector,
      sym_number_ambiguous,
      sym_number_offset,
      sym_div_operator,
      sym_sub_operator,
      sym_color_hex,
      anon_sym_Enum,
    ACTIONS(423), 31,
      sym__block_string_start,
      sym_comma,
      sym_semi_colon,
      sym_equals,
      sym_token,
      sym_static_token,
      sym_static_argument,
      aux_sym_macro_call_token1,
      anon_sym_ATpriority,
      anon_sym_ATname,
      anon_sym_ATderive,
      sym_name_selector,
      aux_sym_state_selector_token1,
      sym_pseudo_selector,
      sym_scope_open,
      sym_scope_close,
      sym_tuple_open,
      sym_number_percent,
      sym_pow_operator,
      sym_floor_div_operator,
      sym_mod_operator,
      sym_mult_operator,
      sym_add_operator,
      sym_single_quote_string,
      sym_double_quote_string,
      aux_sym_color_code_tailwind_token1,
      aux_sym_color_code_skin_token1,
      aux_sym_color_code_brick_token1,
      aux_sym_color_code_css_token1,
      sym_rbx_asset,
      sym_rbx_content,
  [3769] = 5,
    ACTIONS(5), 1,
      anon_sym_DASH_DASH,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(70), 2,
      sym_macro_call,
      sym_comment,
    ACTIONS(425), 9,
      sym_identifier,
      anon_sym_COLON,
      sym_tag_selector,
      sym_number_ambiguous,
      sym_number_offset,
      sym_div_operator,
      sym_sub_operator,
      sym_color_hex,
      anon_sym_Enum,
    ACTIONS(427), 31,
      sym__block_string_start,
      sym_comma,
      sym_semi_colon,
      sym_equals,
      sym_token,
      sym_static_token,
      sym_static_argument,
      aux_sym_macro_call_token1,
      anon_sym_ATpriority,
      anon_sym_ATname,
      anon_sym_ATderive,
      sym_name_selector,
      aux_sym_state_selector_token1,
      sym_pseudo_selector,
      sym_scope_open,
      sym_scope_close,
      sym_tuple_open,
      sym_number_percent,
      sym_pow_operator,
      sym_floor_div_operator,
      sym_mod_operator,
      sym_mult_operator,
      sym_add_operator,
      sym_single_quote_string,
      sym_double_quote_string,
      aux_sym_color_code_tailwind_token1,
      aux_sym_color_code_skin_token1,
      aux_sym_color_code_brick_token1,
      aux_sym_color_code_css_token1,
      sym_rbx_asset,
      sym_rbx_content,
  [3824] = 5,
    ACTIONS(5), 1,
      anon_sym_DASH_DASH,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(71), 2,
      sym_macro_call,
      sym_comment,
    ACTIONS(429), 9,
      sym_identifier,
      anon_sym_COLON,
      sym_tag_selector,
      sym_number_ambiguous,
      sym_number_offset,
      sym_div_operator,
      sym_sub_operator,
      sym_color_hex,
      anon_sym_Enum,
    ACTIONS(431), 31,
      sym__block_string_start,
      sym_comma,
      sym_semi_colon,
      sym_equals,
      sym_token,
      sym_static_token,
      sym_static_argument,
      aux_sym_macro_call_token1,
      anon_sym_ATpriority,
      anon_sym_ATname,
      anon_sym_ATderive,
      sym_name_selector,
      aux_sym_state_selector_token1,
      sym_pseudo_selector,
      sym_scope_open,
      sym_scope_close,
      sym_tuple_open,
      sym_number_percent,
      sym_pow_operator,
      sym_floor_div_operator,
      sym_mod_operator,
      sym_mult_operator,
      sym_add_operator,
      sym_single_quote_string,
      sym_double_quote_string,
      aux_sym_color_code_tailwind_token1,
      aux_sym_color_code_skin_token1,
      aux_sym_color_code_brick_token1,
      aux_sym_color_code_css_token1,
      sym_rbx_asset,
      sym_rbx_content,
  [3879] = 5,
    ACTIONS(5), 1,
      anon_sym_DASH_DASH,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(72), 2,
      sym_macro_call,
      sym_comment,
    ACTIONS(433), 9,
      sym_identifier,
      anon_sym_COLON,
      sym_tag_selector,
      sym_number_ambiguous,
      sym_number_offset,
      sym_div_operator,
      sym_sub_operator,
      sym_color_hex,
      anon_sym_Enum,
    ACTIONS(435), 31,
      sym__block_string_start,
      sym_comma,
      sym_semi_colon,
      sym_equals,
      sym_token,
      sym_static_token,
      sym_static_argument,
      aux_sym_macro_call_token1,
      anon_sym_ATpriority,
      anon_sym_ATname,
      anon_sym_ATderive,
      sym_name_selector,
      aux_sym_state_selector_token1,
      sym_pseudo_selector,
      sym_scope_open,
      sym_scope_close,
      sym_tuple_open,
      sym_number_percent,
      sym_pow_operator,
      sym_floor_div_operator,
      sym_mod_operator,
      sym_mult_operator,
      sym_add_operator,
      sym_single_quote_string,
      sym_double_quote_string,
      aux_sym_color_code_tailwind_token1,
      aux_sym_color_code_skin_token1,
      aux_sym_color_code_brick_token1,
      aux_sym_color_code_css_token1,
      sym_rbx_asset,
      sym_rbx_content,
  [3934] = 27,
    ACTIONS(5), 1,
      anon_sym_DASH_DASH,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(213), 1,
      sym_identifier,
    ACTIONS(219), 1,
      aux_sym_macro_call_token1,
    ACTIONS(221), 1,
      aux_sym_state_selector_token1,
    ACTIONS(223), 1,
      sym_tuple_open,
    ACTIONS(229), 1,
      sym_number_percent,
    ACTIONS(233), 1,
      aux_sym_color_code_tailwind_token1,
    ACTIONS(235), 1,
      aux_sym_color_code_skin_token1,
    ACTIONS(237), 1,
      aux_sym_color_code_brick_token1,
    ACTIONS(239), 1,
      aux_sym_color_code_css_token1,
    ACTIONS(241), 1,
      sym_color_hex,
    ACTIONS(243), 1,
      anon_sym_Enum,
    ACTIONS(247), 1,
      sym__block_string_start,
    ACTIONS(439), 1,
      sym_tuple_close,
    STATE(95), 1,
      aux_sym_macro_call_repeat1,
    STATE(166), 1,
      sym_datatype,
    STATE(207), 1,
      sym_block_string,
    ACTIONS(227), 2,
      sym_number_ambiguous,
      sym_number_offset,
    ACTIONS(231), 2,
      sym_single_quote_string,
      sym_double_quote_string,
    ACTIONS(245), 2,
      sym_rbx_asset,
      sym_rbx_content,
    ACTIONS(437), 2,
      sym_comma,
      sym_semi_colon,
    STATE(73), 2,
      sym_macro_call,
      sym_comment,
    STATE(197), 2,
      sym_enum_full,
      sym_enum_shorthand,
    ACTIONS(217), 3,
      sym_token,
      sym_static_token,
      sym_static_argument,
    STATE(195), 4,
      sym_color_code_tailwind,
      sym_color_code_skin,
      sym_color_code_brick,
      sym_color_code_css,
    STATE(212), 7,
      sym_reference,
      sym_tuple,
      sym_number,
      sym_operation,
      sym_string,
      sym_color,
      sym_enum,
  [4033] = 6,
    ACTIONS(5), 1,
      anon_sym_DASH_DASH,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(441), 1,
      sym_equals,
    STATE(74), 2,
      sym_macro_call,
      sym_comment,
    ACTIONS(199), 9,
      sym_identifier,
      anon_sym_COLON,
      sym_tag_selector,
      sym_number_ambiguous,
      sym_number_offset,
      sym_div_operator,
      sym_sub_operator,
      sym_color_hex,
      anon_sym_Enum,
    ACTIONS(201), 30,
      sym__block_string_start,
      sym_comma,
      sym_semi_colon,
      sym_token,
      sym_static_token,
      sym_static_argument,
      aux_sym_macro_call_token1,
      anon_sym_ATpriority,
      anon_sym_ATname,
      anon_sym_ATderive,
      sym_name_selector,
      aux_sym_state_selector_token1,
      sym_pseudo_selector,
      sym_scope_open,
      sym_scope_close,
      sym_tuple_open,
      sym_number_percent,
      sym_pow_operator,
      sym_floor_div_operator,
      sym_mod_operator,
      sym_mult_operator,
      sym_add_operator,
      sym_single_quote_string,
      sym_double_quote_string,
      aux_sym_color_code_tailwind_token1,
      aux_sym_color_code_skin_token1,
      aux_sym_color_code_brick_token1,
      aux_sym_color_code_css_token1,
      sym_rbx_asset,
      sym_rbx_content,
  [4090] = 27,
    ACTIONS(5), 1,
      anon_sym_DASH_DASH,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(213), 1,
      sym_identifier,
    ACTIONS(219), 1,
      aux_sym_macro_call_token1,
    ACTIONS(221), 1,
      aux_sym_state_selector_token1,
    ACTIONS(223), 1,
      sym_tuple_open,
    ACTIONS(229), 1,
      sym_number_percent,
    ACTIONS(233), 1,
      aux_sym_color_code_tailwind_token1,
    ACTIONS(235), 1,
      aux_sym_color_code_skin_token1,
    ACTIONS(237), 1,
      aux_sym_color_code_brick_token1,
    ACTIONS(239), 1,
      aux_sym_color_code_css_token1,
    ACTIONS(241), 1,
      sym_color_hex,
    ACTIONS(243), 1,
      anon_sym_Enum,
    ACTIONS(247), 1,
      sym__block_string_start,
    ACTIONS(445), 1,
      sym_tuple_close,
    STATE(98), 1,
      aux_sym_macro_call_repeat1,
    STATE(168), 1,
      sym_datatype,
    STATE(207), 1,
      sym_block_string,
    ACTIONS(227), 2,
      sym_number_ambiguous,
      sym_number_offset,
    ACTIONS(231), 2,
      sym_single_quote_string,
      sym_double_quote_string,
    ACTIONS(245), 2,
      sym_rbx_asset,
      sym_rbx_content,
    ACTIONS(443), 2,
      sym_comma,
      sym_semi_colon,
    STATE(75), 2,
      sym_macro_call,
      sym_comment,
    STATE(197), 2,
      sym_enum_full,
      sym_enum_shorthand,
    ACTIONS(217), 3,
      sym_token,
      sym_static_token,
      sym_static_argument,
    STATE(195), 4,
      sym_color_code_tailwind,
      sym_color_code_skin,
      sym_color_code_brick,
      sym_color_code_css,
    STATE(212), 7,
      sym_reference,
      sym_tuple,
      sym_number,
      sym_operation,
      sym_string,
      sym_color,
      sym_enum,
  [4189] = 27,
    ACTIONS(5), 1,
      anon_sym_DASH_DASH,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(213), 1,
      sym_identifier,
    ACTIONS(219), 1,
      aux_sym_macro_call_token1,
    ACTIONS(221), 1,
      aux_sym_state_selector_token1,
    ACTIONS(223), 1,
      sym_tuple_open,
    ACTIONS(229), 1,
      sym_number_percent,
    ACTIONS(233), 1,
      aux_sym_color_code_tailwind_token1,
    ACTIONS(235), 1,
      aux_sym_color_code_skin_token1,
    ACTIONS(237), 1,
      aux_sym_color_code_brick_token1,
    ACTIONS(239), 1,
      aux_sym_color_code_css_token1,
    ACTIONS(241), 1,
      sym_color_hex,
    ACTIONS(243), 1,
      anon_sym_Enum,
    ACTIONS(247), 1,
      sym__block_string_start,
    ACTIONS(449), 1,
      sym_tuple_close,
    STATE(100), 1,
      aux_sym_macro_call_repeat1,
    STATE(169), 1,
      sym_datatype,
    STATE(207), 1,
      sym_block_string,
    ACTIONS(227), 2,
      sym_number_ambiguous,
      sym_number_offset,
    ACTIONS(231), 2,
      sym_single_quote_string,
      sym_double_quote_string,
    ACTIONS(245), 2,
      sym_rbx_asset,
      sym_rbx_content,
    ACTIONS(447), 2,
      sym_comma,
      sym_semi_colon,
    STATE(76), 2,
      sym_macro_call,
      sym_comment,
    STATE(197), 2,
      sym_enum_full,
      sym_enum_shorthand,
    ACTIONS(217), 3,
      sym_token,
      sym_static_token,
      sym_static_argument,
    STATE(195), 4,
      sym_color_code_tailwind,
      sym_color_code_skin,
      sym_color_code_brick,
      sym_color_code_css,
    STATE(212), 7,
      sym_reference,
      sym_tuple,
      sym_number,
      sym_operation,
      sym_string,
      sym_color,
      sym_enum,
  [4288] = 5,
    ACTIONS(5), 1,
      anon_sym_DASH_DASH,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(77), 2,
      sym_macro_call,
      sym_comment,
    ACTIONS(451), 9,
      sym_identifier,
      anon_sym_COLON,
      sym_tag_selector,
      sym_number_ambiguous,
      sym_number_offset,
      sym_div_operator,
      sym_sub_operator,
      sym_color_hex,
      anon_sym_Enum,
    ACTIONS(453), 31,
      sym__block_string_start,
      sym_comma,
      sym_semi_colon,
      sym_equals,
      sym_token,
      sym_static_token,
      sym_static_argument,
      aux_sym_macro_call_token1,
      anon_sym_ATpriority,
      anon_sym_ATname,
      anon_sym_ATderive,
      sym_name_selector,
      aux_sym_state_selector_token1,
      sym_pseudo_selector,
      sym_scope_open,
      sym_scope_close,
      sym_tuple_open,
      sym_number_percent,
      sym_pow_operator,
      sym_floor_div_operator,
      sym_mod_operator,
      sym_mult_operator,
      sym_add_operator,
      sym_single_quote_string,
      sym_double_quote_string,
      aux_sym_color_code_tailwind_token1,
      aux_sym_color_code_skin_token1,
      aux_sym_color_code_brick_token1,
      aux_sym_color_code_css_token1,
      sym_rbx_asset,
      sym_rbx_content,
  [4343] = 25,
    ACTIONS(5), 1,
      anon_sym_DASH_DASH,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(213), 1,
      sym_identifier,
    ACTIONS(219), 1,
      aux_sym_macro_call_token1,
    ACTIONS(221), 1,
      aux_sym_state_selector_token1,
    ACTIONS(223), 1,
      sym_tuple_open,
    ACTIONS(229), 1,
      sym_number_percent,
    ACTIONS(233), 1,
      aux_sym_color_code_tailwind_token1,
    ACTIONS(235), 1,
      aux_sym_color_code_skin_token1,
    ACTIONS(237), 1,
      aux_sym_color_code_brick_token1,
    ACTIONS(239), 1,
      aux_sym_color_code_css_token1,
    ACTIONS(241), 1,
      sym_color_hex,
    ACTIONS(243), 1,
      anon_sym_Enum,
    ACTIONS(247), 1,
      sym__block_string_start,
    ACTIONS(455), 1,
      sym_semi_colon,
    STATE(207), 1,
      sym_block_string,
    STATE(218), 1,
      sym_datatype,
    ACTIONS(227), 2,
      sym_number_ambiguous,
      sym_number_offset,
    ACTIONS(231), 2,
      sym_single_quote_string,
      sym_double_quote_string,
    ACTIONS(245), 2,
      sym_rbx_asset,
      sym_rbx_content,
    STATE(78), 2,
      sym_macro_call,
      sym_comment,
    STATE(197), 2,
      sym_enum_full,
      sym_enum_shorthand,
    ACTIONS(217), 3,
      sym_token,
      sym_static_token,
      sym_static_argument,
    STATE(195), 4,
      sym_color_code_tailwind,
      sym_color_code_skin,
      sym_color_code_brick,
      sym_color_code_css,
    STATE(212), 7,
      sym_reference,
      sym_tuple,
      sym_number,
      sym_operation,
      sym_string,
      sym_color,
      sym_enum,
  [4435] = 25,
    ACTIONS(5), 1,
      anon_sym_DASH_DASH,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(213), 1,
      sym_identifier,
    ACTIONS(219), 1,
      aux_sym_macro_call_token1,
    ACTIONS(221), 1,
      aux_sym_state_selector_token1,
    ACTIONS(223), 1,
      sym_tuple_open,
    ACTIONS(229), 1,
      sym_number_percent,
    ACTIONS(233), 1,
      aux_sym_color_code_tailwind_token1,
    ACTIONS(235), 1,
      aux_sym_color_code_skin_token1,
    ACTIONS(237), 1,
      aux_sym_color_code_brick_token1,
    ACTIONS(239), 1,
      aux_sym_color_code_css_token1,
    ACTIONS(241), 1,
      sym_color_hex,
    ACTIONS(243), 1,
      anon_sym_Enum,
    ACTIONS(247), 1,
      sym__block_string_start,
    STATE(83), 1,
      aux_sym_macro_call_repeat1,
    STATE(176), 1,
      sym_datatype,
    STATE(207), 1,
      sym_block_string,
    ACTIONS(227), 2,
      sym_number_ambiguous,
      sym_number_offset,
    ACTIONS(231), 2,
      sym_single_quote_string,
      sym_double_quote_string,
    ACTIONS(245), 2,
      sym_rbx_asset,
      sym_rbx_content,
    STATE(79), 2,
      sym_macro_call,
      sym_comment,
    STATE(197), 2,
      sym_enum_full,
      sym_enum_shorthand,
    ACTIONS(217), 3,
      sym_token,
      sym_static_token,
      sym_static_argument,
    STATE(195), 4,
      sym_color_code_tailwind,
      sym_color_code_skin,
      sym_color_code_brick,
      sym_color_code_css,
    STATE(212), 7,
      sym_reference,
      sym_tuple,
      sym_number,
      sym_operation,
      sym_string,
      sym_color,
      sym_enum,
  [4527] = 25,
    ACTIONS(5), 1,
      anon_sym_DASH_DASH,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(213), 1,
      sym_identifier,
    ACTIONS(219), 1,
      aux_sym_macro_call_token1,
    ACTIONS(221), 1,
      aux_sym_state_selector_token1,
    ACTIONS(223), 1,
      sym_tuple_open,
    ACTIONS(229), 1,
      sym_number_percent,
    ACTIONS(233), 1,
      aux_sym_color_code_tailwind_token1,
    ACTIONS(235), 1,
      aux_sym_color_code_skin_token1,
    ACTIONS(237), 1,
      aux_sym_color_code_brick_token1,
    ACTIONS(239), 1,
      aux_sym_color_code_css_token1,
    ACTIONS(241), 1,
      sym_color_hex,
    ACTIONS(243), 1,
      anon_sym_Enum,
    ACTIONS(247), 1,
      sym__block_string_start,
    STATE(83), 1,
      aux_sym_macro_call_repeat1,
    STATE(156), 1,
      sym_datatype,
    STATE(207), 1,
      sym_block_string,
    ACTIONS(227), 2,
      sym_number_ambiguous,
      sym_number_offset,
    ACTIONS(231), 2,
      sym_single_quote_string,
      sym_double_quote_string,
    ACTIONS(245), 2,
      sym_rbx_asset,
      sym_rbx_content,
    STATE(80), 2,
      sym_macro_call,
      sym_comment,
    STATE(197), 2,
      sym_enum_full,
      sym_enum_shorthand,
    ACTIONS(217), 3,
      sym_token,
      sym_static_token,
      sym_static_argument,
    STATE(195), 4,
      sym_color_code_tailwind,
      sym_color_code_skin,
      sym_color_code_brick,
      sym_color_code_css,
    STATE(212), 7,
      sym_reference,
      sym_tuple,
      sym_number,
      sym_operation,
      sym_string,
      sym_color,
      sym_enum,
  [4619] = 25,
    ACTIONS(5), 1,
      anon_sym_DASH_DASH,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(213), 1,
      sym_identifier,
    ACTIONS(219), 1,
      aux_sym_macro_call_token1,
    ACTIONS(221), 1,
      aux_sym_state_selector_token1,
    ACTIONS(223), 1,
      sym_tuple_open,
    ACTIONS(229), 1,
      sym_number_percent,
    ACTIONS(233), 1,
      aux_sym_color_code_tailwind_token1,
    ACTIONS(235), 1,
      aux_sym_color_code_skin_token1,
    ACTIONS(237), 1,
      aux_sym_color_code_brick_token1,
    ACTIONS(239), 1,
      aux_sym_color_code_css_token1,
    ACTIONS(241), 1,
      sym_color_hex,
    ACTIONS(243), 1,
      anon_sym_Enum,
    ACTIONS(247), 1,
      sym__block_string_start,
    STATE(90), 1,
      aux_sym_macro_call_repeat1,
    STATE(179), 1,
      sym_datatype,
    STATE(207), 1,
      sym_block_string,
    ACTIONS(227), 2,
      sym_number_ambiguous,
      sym_number_offset,
    ACTIONS(231), 2,
      sym_single_quote_string,
      sym_double_quote_string,
    ACTIONS(245), 2,
      sym_rbx_asset,
      sym_rbx_content,
    STATE(81), 2,
      sym_macro_call,
      sym_comment,
    STATE(197), 2,
      sym_enum_full,
      sym_enum_shorthand,
    ACTIONS(217), 3,
      sym_token,
      sym_static_token,
      sym_static_argument,
    STATE(195), 4,
      sym_color_code_tailwind,
      sym_color_code_skin,
      sym_color_code_brick,
      sym_color_code_css,
    STATE(212), 7,
      sym_reference,
      sym_tuple,
      sym_number,
      sym_operation,
      sym_string,
      sym_color,
      sym_enum,
  [4711] = 25,
    ACTIONS(5), 1,
      anon_sym_DASH_DASH,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(213), 1,
      sym_identifier,
    ACTIONS(219), 1,
      aux_sym_macro_call_token1,
    ACTIONS(221), 1,
      aux_sym_state_selector_token1,
    ACTIONS(223), 1,
      sym_tuple_open,
    ACTIONS(229), 1,
      sym_number_percent,
    ACTIONS(233), 1,
      aux_sym_color_code_tailwind_token1,
    ACTIONS(235), 1,
      aux_sym_color_code_skin_token1,
    ACTIONS(237), 1,
      aux_sym_color_code_brick_token1,
    ACTIONS(239), 1,
      aux_sym_color_code_css_token1,
    ACTIONS(241), 1,
      sym_color_hex,
    ACTIONS(243), 1,
      anon_sym_Enum,
    ACTIONS(247), 1,
      sym__block_string_start,
    STATE(83), 1,
      aux_sym_macro_call_repeat1,
    STATE(179), 1,
      sym_datatype,
    STATE(207), 1,
      sym_block_string,
    ACTIONS(227), 2,
      sym_number_ambiguous,
      sym_number_offset,
    ACTIONS(231), 2,
      sym_single_quote_string,
      sym_double_quote_string,
    ACTIONS(245), 2,
      sym_rbx_asset,
      sym_rbx_content,
    STATE(82), 2,
      sym_macro_call,
      sym_comment,
    STATE(197), 2,
      sym_enum_full,
      sym_enum_shorthand,
    ACTIONS(217), 3,
      sym_token,
      sym_static_token,
      sym_static_argument,
    STATE(195), 4,
      sym_color_code_tailwind,
      sym_color_code_skin,
      sym_color_code_brick,
      sym_color_code_css,
    STATE(212), 7,
      sym_reference,
      sym_tuple,
      sym_number,
      sym_operation,
      sym_string,
      sym_color,
      sym_enum,
  [4803] = 24,
    ACTIONS(5), 1,
      anon_sym_DASH_DASH,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(457), 1,
      sym_identifier,
    ACTIONS(463), 1,
      aux_sym_macro_call_token1,
    ACTIONS(466), 1,
      aux_sym_state_selector_token1,
    ACTIONS(469), 1,
      sym_tuple_open,
    ACTIONS(475), 1,
      sym_number_percent,
    ACTIONS(481), 1,
      aux_sym_color_code_tailwind_token1,
    ACTIONS(484), 1,
      aux_sym_color_code_skin_token1,
    ACTIONS(487), 1,
      aux_sym_color_code_brick_token1,
    ACTIONS(490), 1,
      aux_sym_color_code_css_token1,
    ACTIONS(493), 1,
      sym_color_hex,
    ACTIONS(496), 1,
      anon_sym_Enum,
    ACTIONS(502), 1,
      sym__block_string_start,
    STATE(183), 1,
      sym_datatype,
    STATE(207), 1,
      sym_block_string,
    ACTIONS(472), 2,
      sym_number_ambiguous,
      sym_number_offset,
    ACTIONS(478), 2,
      sym_single_quote_string,
      sym_double_quote_string,
    ACTIONS(499), 2,
      sym_rbx_asset,
      sym_rbx_content,
    STATE(197), 2,
      sym_enum_full,
      sym_enum_shorthand,
    ACTIONS(460), 3,
      sym_token,
      sym_static_token,
      sym_static_argument,
    STATE(83), 3,
      sym_macro_call,
      sym_comment,
      aux_sym_macro_call_repeat1,
    STATE(195), 4,
      sym_color_code_tailwind,
      sym_color_code_skin,
      sym_color_code_brick,
      sym_color_code_css,
    STATE(212), 7,
      sym_reference,
      sym_tuple,
      sym_number,
      sym_operation,
      sym_string,
      sym_color,
      sym_enum,
  [4893] = 25,
    ACTIONS(5), 1,
      anon_sym_DASH_DASH,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(213), 1,
      sym_identifier,
    ACTIONS(219), 1,
      aux_sym_macro_call_token1,
    ACTIONS(221), 1,
      aux_sym_state_selector_token1,
    ACTIONS(223), 1,
      sym_tuple_open,
    ACTIONS(229), 1,
      sym_number_percent,
    ACTIONS(233), 1,
      aux_sym_color_code_tailwind_token1,
    ACTIONS(235), 1,
      aux_sym_color_code_skin_token1,
    ACTIONS(237), 1,
      aux_sym_color_code_brick_token1,
    ACTIONS(239), 1,
      aux_sym_color_code_css_token1,
    ACTIONS(241), 1,
      sym_color_hex,
    ACTIONS(243), 1,
      anon_sym_Enum,
    ACTIONS(247), 1,
      sym__block_string_start,
    ACTIONS(505), 1,
      sym_semi_colon,
    STATE(207), 1,
      sym_block_string,
    STATE(217), 1,
      sym_datatype,
    ACTIONS(227), 2,
      sym_number_ambiguous,
      sym_number_offset,
    ACTIONS(231), 2,
      sym_single_quote_string,
      sym_double_quote_string,
    ACTIONS(245), 2,
      sym_rbx_asset,
      sym_rbx_content,
    STATE(84), 2,
      sym_macro_call,
      sym_comment,
    STATE(197), 2,
      sym_enum_full,
      sym_enum_shorthand,
    ACTIONS(217), 3,
      sym_token,
      sym_static_token,
      sym_static_argument,
    STATE(195), 4,
      sym_color_code_tailwind,
      sym_color_code_skin,
      sym_color_code_brick,
      sym_color_code_css,
    STATE(212), 7,
      sym_reference,
      sym_tuple,
      sym_number,
      sym_operation,
      sym_string,
      sym_color,
      sym_enum,
  [4985] = 25,
    ACTIONS(5), 1,
      anon_sym_DASH_DASH,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(213), 1,
      sym_identifier,
    ACTIONS(219), 1,
      aux_sym_macro_call_token1,
    ACTIONS(221), 1,
      aux_sym_state_selector_token1,
    ACTIONS(223), 1,
      sym_tuple_open,
    ACTIONS(229), 1,
      sym_number_percent,
    ACTIONS(233), 1,
      aux_sym_color_code_tailwind_token1,
    ACTIONS(235), 1,
      aux_sym_color_code_skin_token1,
    ACTIONS(237), 1,
      aux_sym_color_code_brick_token1,
    ACTIONS(239), 1,
      aux_sym_color_code_css_token1,
    ACTIONS(241), 1,
      sym_color_hex,
    ACTIONS(243), 1,
      anon_sym_Enum,
    ACTIONS(247), 1,
      sym__block_string_start,
    ACTIONS(507), 1,
      sym_semi_colon,
    STATE(207), 1,
      sym_block_string,
    STATE(214), 1,
      sym_datatype,
    ACTIONS(227), 2,
      sym_number_ambiguous,
      sym_number_offset,
    ACTIONS(231), 2,
      sym_single_quote_string,
      sym_double_quote_string,
    ACTIONS(245), 2,
      sym_rbx_asset,
      sym_rbx_content,
    STATE(85), 2,
      sym_macro_call,
      sym_comment,
    STATE(197), 2,
      sym_enum_full,
      sym_enum_shorthand,
    ACTIONS(217), 3,
      sym_token,
      sym_static_token,
      sym_static_argument,
    STATE(195), 4,
      sym_color_code_tailwind,
      sym_color_code_skin,
      sym_color_code_brick,
      sym_color_code_css,
    STATE(212), 7,
      sym_reference,
      sym_tuple,
      sym_number,
      sym_operation,
      sym_string,
      sym_color,
      sym_enum,
  [5077] = 25,
    ACTIONS(5), 1,
      anon_sym_DASH_DASH,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(213), 1,
      sym_identifier,
    ACTIONS(219), 1,
      aux_sym_macro_call_token1,
    ACTIONS(221), 1,
      aux_sym_state_selector_token1,
    ACTIONS(223), 1,
      sym_tuple_open,
    ACTIONS(229), 1,
      sym_number_percent,
    ACTIONS(233), 1,
      aux_sym_color_code_tailwind_token1,
    ACTIONS(235), 1,
      aux_sym_color_code_skin_token1,
    ACTIONS(237), 1,
      aux_sym_color_code_brick_token1,
    ACTIONS(239), 1,
      aux_sym_color_code_css_token1,
    ACTIONS(241), 1,
      sym_color_hex,
    ACTIONS(243), 1,
      anon_sym_Enum,
    ACTIONS(247), 1,
      sym__block_string_start,
    STATE(80), 1,
      aux_sym_macro_call_repeat1,
    STATE(161), 1,
      sym_datatype,
    STATE(207), 1,
      sym_block_string,
    ACTIONS(227), 2,
      sym_number_ambiguous,
      sym_number_offset,
    ACTIONS(231), 2,
      sym_single_quote_string,
      sym_double_quote_string,
    ACTIONS(245), 2,
      sym_rbx_asset,
      sym_rbx_content,
    STATE(86), 2,
      sym_macro_call,
      sym_comment,
    STATE(197), 2,
      sym_enum_full,
      sym_enum_shorthand,
    ACTIONS(217), 3,
      sym_token,
      sym_static_token,
      sym_static_argument,
    STATE(195), 4,
      sym_color_code_tailwind,
      sym_color_code_skin,
      sym_color_code_brick,
      sym_color_code_css,
    STATE(212), 7,
      sym_reference,
      sym_tuple,
      sym_number,
      sym_operation,
      sym_string,
      sym_color,
      sym_enum,
  [5169] = 25,
    ACTIONS(5), 1,
      anon_sym_DASH_DASH,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(213), 1,
      sym_identifier,
    ACTIONS(219), 1,
      aux_sym_macro_call_token1,
    ACTIONS(221), 1,
      aux_sym_state_selector_token1,
    ACTIONS(223), 1,
      sym_tuple_open,
    ACTIONS(229), 1,
      sym_number_percent,
    ACTIONS(233), 1,
      aux_sym_color_code_tailwind_token1,
    ACTIONS(235), 1,
      aux_sym_color_code_skin_token1,
    ACTIONS(237), 1,
      aux_sym_color_code_brick_token1,
    ACTIONS(239), 1,
      aux_sym_color_code_css_token1,
    ACTIONS(241), 1,
      sym_color_hex,
    ACTIONS(243), 1,
      anon_sym_Enum,
    ACTIONS(247), 1,
      sym__block_string_start,
    STATE(83), 1,
      aux_sym_macro_call_repeat1,
    STATE(161), 1,
      sym_datatype,
    STATE(207), 1,
      sym_block_string,
    ACTIONS(227), 2,
      sym_number_ambiguous,
      sym_number_offset,
    ACTIONS(231), 2,
      sym_single_quote_string,
      sym_double_quote_string,
    ACTIONS(245), 2,
      sym_rbx_asset,
      sym_rbx_content,
    STATE(87), 2,
      sym_macro_call,
      sym_comment,
    STATE(197), 2,
      sym_enum_full,
      sym_enum_shorthand,
    ACTIONS(217), 3,
      sym_token,
      sym_static_token,
      sym_static_argument,
    STATE(195), 4,
      sym_color_code_tailwind,
      sym_color_code_skin,
      sym_color_code_brick,
      sym_color_code_css,
    STATE(212), 7,
      sym_reference,
      sym_tuple,
      sym_number,
      sym_operation,
      sym_string,
      sym_color,
      sym_enum,
  [5261] = 25,
    ACTIONS(5), 1,
      anon_sym_DASH_DASH,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(213), 1,
      sym_identifier,
    ACTIONS(219), 1,
      aux_sym_macro_call_token1,
    ACTIONS(221), 1,
      aux_sym_state_selector_token1,
    ACTIONS(223), 1,
      sym_tuple_open,
    ACTIONS(229), 1,
      sym_number_percent,
    ACTIONS(233), 1,
      aux_sym_color_code_tailwind_token1,
    ACTIONS(235), 1,
      aux_sym_color_code_skin_token1,
    ACTIONS(237), 1,
      aux_sym_color_code_brick_token1,
    ACTIONS(239), 1,
      aux_sym_color_code_css_token1,
    ACTIONS(241), 1,
      sym_color_hex,
    ACTIONS(243), 1,
      anon_sym_Enum,
    ACTIONS(247), 1,
      sym__block_string_start,
    STATE(93), 1,
      aux_sym_macro_call_repeat1,
    STATE(158), 1,
      sym_datatype,
    STATE(207), 1,
      sym_block_string,
    ACTIONS(227), 2,
      sym_number_ambiguous,
      sym_number_offset,
    ACTIONS(231), 2,
      sym_single_quote_string,
      sym_double_quote_string,
    ACTIONS(245), 2,
      sym_rbx_asset,
      sym_rbx_content,
    STATE(88), 2,
      sym_macro_call,
      sym_comment,
    STATE(197), 2,
      sym_enum_full,
      sym_enum_shorthand,
    ACTIONS(217), 3,
      sym_token,
      sym_static_token,
      sym_static_argument,
    STATE(195), 4,
      sym_color_code_tailwind,
      sym_color_code_skin,
      sym_color_code_brick,
      sym_color_code_css,
    STATE(212), 7,
      sym_reference,
      sym_tuple,
      sym_number,
      sym_operation,
      sym_string,
      sym_color,
      sym_enum,
  [5353] = 25,
    ACTIONS(5), 1,
      anon_sym_DASH_DASH,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(213), 1,
      sym_identifier,
    ACTIONS(219), 1,
      aux_sym_macro_call_token1,
    ACTIONS(221), 1,
      aux_sym_state_selector_token1,
    ACTIONS(223), 1,
      sym_tuple_open,
    ACTIONS(229), 1,
      sym_number_percent,
    ACTIONS(233), 1,
      aux_sym_color_code_tailwind_token1,
    ACTIONS(235), 1,
      aux_sym_color_code_skin_token1,
    ACTIONS(237), 1,
      aux_sym_color_code_brick_token1,
    ACTIONS(239), 1,
      aux_sym_color_code_css_token1,
    ACTIONS(241), 1,
      sym_color_hex,
    ACTIONS(243), 1,
      anon_sym_Enum,
    ACTIONS(247), 1,
      sym__block_string_start,
    STATE(83), 1,
      aux_sym_macro_call_repeat1,
    STATE(177), 1,
      sym_datatype,
    STATE(207), 1,
      sym_block_string,
    ACTIONS(227), 2,
      sym_number_ambiguous,
      sym_number_offset,
    ACTIONS(231), 2,
      sym_single_quote_string,
      sym_double_quote_string,
    ACTIONS(245), 2,
      sym_rbx_asset,
      sym_rbx_content,
    STATE(89), 2,
      sym_macro_call,
      sym_comment,
    STATE(197), 2,
      sym_enum_full,
      sym_enum_shorthand,
    ACTIONS(217), 3,
      sym_token,
      sym_static_token,
      sym_static_argument,
    STATE(195), 4,
      sym_color_code_tailwind,
      sym_color_code_skin,
      sym_color_code_brick,
      sym_color_code_css,
    STATE(212), 7,
      sym_reference,
      sym_tuple,
      sym_number,
      sym_operation,
      sym_string,
      sym_color,
      sym_enum,
  [5445] = 25,
    ACTIONS(5), 1,
      anon_sym_DASH_DASH,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(213), 1,
      sym_identifier,
    ACTIONS(219), 1,
      aux_sym_macro_call_token1,
    ACTIONS(221), 1,
      aux_sym_state_selector_token1,
    ACTIONS(223), 1,
      sym_tuple_open,
    ACTIONS(229), 1,
      sym_number_percent,
    ACTIONS(233), 1,
      aux_sym_color_code_tailwind_token1,
    ACTIONS(235), 1,
      aux_sym_color_code_skin_token1,
    ACTIONS(237), 1,
      aux_sym_color_code_brick_token1,
    ACTIONS(239), 1,
      aux_sym_color_code_css_token1,
    ACTIONS(241), 1,
      sym_color_hex,
    ACTIONS(243), 1,
      anon_sym_Enum,
    ACTIONS(247), 1,
      sym__block_string_start,
    STATE(83), 1,
      aux_sym_macro_call_repeat1,
    STATE(159), 1,
      sym_datatype,
    STATE(207), 1,
      sym_block_string,
    ACTIONS(227), 2,
      sym_number_ambiguous,
      sym_number_offset,
    ACTIONS(231), 2,
      sym_single_quote_string,
      sym_double_quote_string,
    ACTIONS(245), 2,
      sym_rbx_asset,
      sym_rbx_content,
    STATE(90), 2,
      sym_macro_call,
      sym_comment,
    STATE(197), 2,
      sym_enum_full,
      sym_enum_shorthand,
    ACTIONS(217), 3,
      sym_token,
      sym_static_token,
      sym_static_argument,
    STATE(195), 4,
      sym_color_code_tailwind,
      sym_color_code_skin,
      sym_color_code_brick,
      sym_color_code_css,
    STATE(212), 7,
      sym_reference,
      sym_tuple,
      sym_number,
      sym_operation,
      sym_string,
      sym_color,
      sym_enum,
  [5537] = 25,
    ACTIONS(5), 1,
      anon_sym_DASH_DASH,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(213), 1,
      sym_identifier,
    ACTIONS(219), 1,
      aux_sym_macro_call_token1,
    ACTIONS(221), 1,
      aux_sym_state_selector_token1,
    ACTIONS(223), 1,
      sym_tuple_open,
    ACTIONS(229), 1,
      sym_number_percent,
    ACTIONS(233), 1,
      aux_sym_color_code_tailwind_token1,
    ACTIONS(235), 1,
      aux_sym_color_code_skin_token1,
    ACTIONS(237), 1,
      aux_sym_color_code_brick_token1,
    ACTIONS(239), 1,
      aux_sym_color_code_css_token1,
    ACTIONS(241), 1,
      sym_color_hex,
    ACTIONS(243), 1,
      anon_sym_Enum,
    ACTIONS(247), 1,
      sym__block_string_start,
    ACTIONS(509), 1,
      sym_semi_colon,
    STATE(207), 1,
      sym_block_string,
    STATE(215), 1,
      sym_datatype,
    ACTIONS(227), 2,
      sym_number_ambiguous,
      sym_number_offset,
    ACTIONS(231), 2,
      sym_single_quote_string,
      sym_double_quote_string,
    ACTIONS(245), 2,
      sym_rbx_asset,
      sym_rbx_content,
    STATE(91), 2,
      sym_macro_call,
      sym_comment,
    STATE(197), 2,
      sym_enum_full,
      sym_enum_shorthand,
    ACTIONS(217), 3,
      sym_token,
      sym_static_token,
      sym_static_argument,
    STATE(195), 4,
      sym_color_code_tailwind,
      sym_color_code_skin,
      sym_color_code_brick,
      sym_color_code_css,
    STATE(212), 7,
      sym_reference,
      sym_tuple,
      sym_number,
      sym_operation,
      sym_string,
      sym_color,
      sym_enum,
  [5629] = 25,
    ACTIONS(5), 1,
      anon_sym_DASH_DASH,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(213), 1,
      sym_identifier,
    ACTIONS(219), 1,
      aux_sym_macro_call_token1,
    ACTIONS(221), 1,
      aux_sym_state_selector_token1,
    ACTIONS(223), 1,
      sym_tuple_open,
    ACTIONS(229), 1,
      sym_number_percent,
    ACTIONS(233), 1,
      aux_sym_color_code_tailwind_token1,
    ACTIONS(235), 1,
      aux_sym_color_code_skin_token1,
    ACTIONS(237), 1,
      aux_sym_color_code_brick_token1,
    ACTIONS(239), 1,
      aux_sym_color_code_css_token1,
    ACTIONS(241), 1,
      sym_color_hex,
    ACTIONS(243), 1,
      anon_sym_Enum,
    ACTIONS(247), 1,
      sym__block_string_start,
    ACTIONS(511), 1,
      sym_semi_colon,
    STATE(207), 1,
      sym_block_string,
    STATE(213), 1,
      sym_datatype,
    ACTIONS(227), 2,
      sym_number_ambiguous,
      sym_number_offset,
    ACTIONS(231), 2,
      sym_single_quote_string,
      sym_double_quote_string,
    ACTIONS(245), 2,
      sym_rbx_asset,
      sym_rbx_content,
    STATE(92), 2,
      sym_macro_call,
      sym_comment,
    STATE(197), 2,
      sym_enum_full,
      sym_enum_shorthand,
    ACTIONS(217), 3,
      sym_token,
      sym_static_token,
      sym_static_argument,
    STATE(195), 4,
      sym_color_code_tailwind,
      sym_color_code_skin,
      sym_color_code_brick,
      sym_color_code_css,
    STATE(212), 7,
      sym_reference,
      sym_tuple,
      sym_number,
      sym_operation,
      sym_string,
      sym_color,
      sym_enum,
  [5721] = 25,
    ACTIONS(5), 1,
      anon_sym_DASH_DASH,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(213), 1,
      sym_identifier,
    ACTIONS(219), 1,
      aux_sym_macro_call_token1,
    ACTIONS(221), 1,
      aux_sym_state_selector_token1,
    ACTIONS(223), 1,
      sym_tuple_open,
    ACTIONS(229), 1,
      sym_number_percent,
    ACTIONS(233), 1,
      aux_sym_color_code_tailwind_token1,
    ACTIONS(235), 1,
      aux_sym_color_code_skin_token1,
    ACTIONS(237), 1,
      aux_sym_color_code_brick_token1,
    ACTIONS(239), 1,
      aux_sym_color_code_css_token1,
    ACTIONS(241), 1,
      sym_color_hex,
    ACTIONS(243), 1,
      anon_sym_Enum,
    ACTIONS(247), 1,
      sym__block_string_start,
    STATE(83), 1,
      aux_sym_macro_call_repeat1,
    STATE(160), 1,
      sym_datatype,
    STATE(207), 1,
      sym_block_string,
    ACTIONS(227), 2,
      sym_number_ambiguous,
      sym_number_offset,
    ACTIONS(231), 2,
      sym_single_quote_string,
      sym_double_quote_string,
    ACTIONS(245), 2,
      sym_rbx_asset,
      sym_rbx_content,
    STATE(93), 2,
      sym_macro_call,
      sym_comment,
    STATE(197), 2,
      sym_enum_full,
      sym_enum_shorthand,
    ACTIONS(217), 3,
      sym_token,
      sym_static_token,
      sym_static_argument,
    STATE(195), 4,
      sym_color_code_tailwind,
      sym_color_code_skin,
      sym_color_code_brick,
      sym_color_code_css,
    STATE(212), 7,
      sym_reference,
      sym_tuple,
      sym_number,
      sym_operation,
      sym_string,
      sym_color,
      sym_enum,
  [5813] = 25,
    ACTIONS(5), 1,
      anon_sym_DASH_DASH,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(213), 1,
      sym_identifier,
    ACTIONS(219), 1,
      aux_sym_macro_call_token1,
    ACTIONS(221), 1,
      aux_sym_state_selector_token1,
    ACTIONS(223), 1,
      sym_tuple_open,
    ACTIONS(229), 1,
      sym_number_percent,
    ACTIONS(233), 1,
      aux_sym_color_code_tailwind_token1,
    ACTIONS(235), 1,
      aux_sym_color_code_skin_token1,
    ACTIONS(237), 1,
      aux_sym_color_code_brick_token1,
    ACTIONS(239), 1,
      aux_sym_color_code_css_token1,
    ACTIONS(241), 1,
      sym_color_hex,
    ACTIONS(243), 1,
      anon_sym_Enum,
    ACTIONS(247), 1,
      sym__block_string_start,
    STATE(96), 1,
      aux_sym_macro_call_repeat1,
    STATE(167), 1,
      sym_datatype,
    STATE(207), 1,
      sym_block_string,
    ACTIONS(227), 2,
      sym_number_ambiguous,
      sym_number_offset,
    ACTIONS(231), 2,
      sym_single_quote_string,
      sym_double_quote_string,
    ACTIONS(245), 2,
      sym_rbx_asset,
      sym_rbx_content,
    STATE(94), 2,
      sym_macro_call,
      sym_comment,
    STATE(197), 2,
      sym_enum_full,
      sym_enum_shorthand,
    ACTIONS(217), 3,
      sym_token,
      sym_static_token,
      sym_static_argument,
    STATE(195), 4,
      sym_color_code_tailwind,
      sym_color_code_skin,
      sym_color_code_brick,
      sym_color_code_css,
    STATE(212), 7,
      sym_reference,
      sym_tuple,
      sym_number,
      sym_operation,
      sym_string,
      sym_color,
      sym_enum,
  [5905] = 25,
    ACTIONS(5), 1,
      anon_sym_DASH_DASH,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(213), 1,
      sym_identifier,
    ACTIONS(219), 1,
      aux_sym_macro_call_token1,
    ACTIONS(221), 1,
      aux_sym_state_selector_token1,
    ACTIONS(223), 1,
      sym_tuple_open,
    ACTIONS(229), 1,
      sym_number_percent,
    ACTIONS(233), 1,
      aux_sym_color_code_tailwind_token1,
    ACTIONS(235), 1,
      aux_sym_color_code_skin_token1,
    ACTIONS(237), 1,
      aux_sym_color_code_brick_token1,
    ACTIONS(239), 1,
      aux_sym_color_code_css_token1,
    ACTIONS(241), 1,
      sym_color_hex,
    ACTIONS(243), 1,
      anon_sym_Enum,
    ACTIONS(247), 1,
      sym__block_string_start,
    STATE(83), 1,
      aux_sym_macro_call_repeat1,
    STATE(167), 1,
      sym_datatype,
    STATE(207), 1,
      sym_block_string,
    ACTIONS(227), 2,
      sym_number_ambiguous,
      sym_number_offset,
    ACTIONS(231), 2,
      sym_single_quote_string,
      sym_double_quote_string,
    ACTIONS(245), 2,
      sym_rbx_asset,
      sym_rbx_content,
    STATE(95), 2,
      sym_macro_call,
      sym_comment,
    STATE(197), 2,
      sym_enum_full,
      sym_enum_shorthand,
    ACTIONS(217), 3,
      sym_token,
      sym_static_token,
      sym_static_argument,
    STATE(195), 4,
      sym_color_code_tailwind,
      sym_color_code_skin,
      sym_color_code_brick,
      sym_color_code_css,
    STATE(212), 7,
      sym_reference,
      sym_tuple,
      sym_number,
      sym_operation,
      sym_string,
      sym_color,
      sym_enum,
  [5997] = 25,
    ACTIONS(5), 1,
      anon_sym_DASH_DASH,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(213), 1,
      sym_identifier,
    ACTIONS(219), 1,
      aux_sym_macro_call_token1,
    ACTIONS(221), 1,
      aux_sym_state_selector_token1,
    ACTIONS(223), 1,
      sym_tuple_open,
    ACTIONS(229), 1,
      sym_number_percent,
    ACTIONS(233), 1,
      aux_sym_color_code_tailwind_token1,
    ACTIONS(235), 1,
      aux_sym_color_code_skin_token1,
    ACTIONS(237), 1,
      aux_sym_color_code_brick_token1,
    ACTIONS(239), 1,
      aux_sym_color_code_css_token1,
    ACTIONS(241), 1,
      sym_color_hex,
    ACTIONS(243), 1,
      anon_sym_Enum,
    ACTIONS(247), 1,
      sym__block_string_start,
    STATE(83), 1,
      aux_sym_macro_call_repeat1,
    STATE(170), 1,
      sym_datatype,
    STATE(207), 1,
      sym_block_string,
    ACTIONS(227), 2,
      sym_number_ambiguous,
      sym_number_offset,
    ACTIONS(231), 2,
      sym_single_quote_string,
      sym_double_quote_string,
    ACTIONS(245), 2,
      sym_rbx_asset,
      sym_rbx_content,
    STATE(96), 2,
      sym_macro_call,
      sym_comment,
    STATE(197), 2,
      sym_enum_full,
      sym_enum_shorthand,
    ACTIONS(217), 3,
      sym_token,
      sym_static_token,
      sym_static_argument,
    STATE(195), 4,
      sym_color_code_tailwind,
      sym_color_code_skin,
      sym_color_code_brick,
      sym_color_code_css,
    STATE(212), 7,
      sym_reference,
      sym_tuple,
      sym_number,
      sym_operation,
      sym_string,
      sym_color,
      sym_enum,
  [6089] = 25,
    ACTIONS(5), 1,
      anon_sym_DASH_DASH,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(213), 1,
      sym_identifier,
    ACTIONS(219), 1,
      aux_sym_macro_call_token1,
    ACTIONS(221), 1,
      aux_sym_state_selector_token1,
    ACTIONS(223), 1,
      sym_tuple_open,
    ACTIONS(229), 1,
      sym_number_percent,
    ACTIONS(233), 1,
      aux_sym_color_code_tailwind_token1,
    ACTIONS(235), 1,
      aux_sym_color_code_skin_token1,
    ACTIONS(237), 1,
      aux_sym_color_code_brick_token1,
    ACTIONS(239), 1,
      aux_sym_color_code_css_token1,
    ACTIONS(241), 1,
      sym_color_hex,
    ACTIONS(243), 1,
      anon_sym_Enum,
    ACTIONS(247), 1,
      sym__block_string_start,
    STATE(101), 1,
      aux_sym_macro_call_repeat1,
    STATE(171), 1,
      sym_datatype,
    STATE(207), 1,
      sym_block_string,
    ACTIONS(227), 2,
      sym_number_ambiguous,
      sym_number_offset,
    ACTIONS(231), 2,
      sym_single_quote_string,
      sym_double_quote_string,
    ACTIONS(245), 2,
      sym_rbx_asset,
      sym_rbx_content,
    STATE(97), 2,
      sym_macro_call,
      sym_comment,
    STATE(197), 2,
      sym_enum_full,
      sym_enum_shorthand,
    ACTIONS(217), 3,
      sym_token,
      sym_static_token,
      sym_static_argument,
    STATE(195), 4,
      sym_color_code_tailwind,
      sym_color_code_skin,
      sym_color_code_brick,
      sym_color_code_css,
    STATE(212), 7,
      sym_reference,
      sym_tuple,
      sym_number,
      sym_operation,
      sym_string,
      sym_color,
      sym_enum,
  [6181] = 25,
    ACTIONS(5), 1,
      anon_sym_DASH_DASH,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(213), 1,
      sym_identifier,
    ACTIONS(219), 1,
      aux_sym_macro_call_token1,
    ACTIONS(221), 1,
      aux_sym_state_selector_token1,
    ACTIONS(223), 1,
      sym_tuple_open,
    ACTIONS(229), 1,
      sym_number_percent,
    ACTIONS(233), 1,
      aux_sym_color_code_tailwind_token1,
    ACTIONS(235), 1,
      aux_sym_color_code_skin_token1,
    ACTIONS(237), 1,
      aux_sym_color_code_brick_token1,
    ACTIONS(239), 1,
      aux_sym_color_code_css_token1,
    ACTIONS(241), 1,
      sym_color_hex,
    ACTIONS(243), 1,
      anon_sym_Enum,
    ACTIONS(247), 1,
      sym__block_string_start,
    STATE(83), 1,
      aux_sym_macro_call_repeat1,
    STATE(171), 1,
      sym_datatype,
    STATE(207), 1,
      sym_block_string,
    ACTIONS(227), 2,
      sym_number_ambiguous,
      sym_number_offset,
    ACTIONS(231), 2,
      sym_single_quote_string,
      sym_double_quote_string,
    ACTIONS(245), 2,
      sym_rbx_asset,
      sym_rbx_content,
    STATE(98), 2,
      sym_macro_call,
      sym_comment,
    STATE(197), 2,
      sym_enum_full,
      sym_enum_shorthand,
    ACTIONS(217), 3,
      sym_token,
      sym_static_token,
      sym_static_argument,
    STATE(195), 4,
      sym_color_code_tailwind,
      sym_color_code_skin,
      sym_color_code_brick,
      sym_color_code_css,
    STATE(212), 7,
      sym_reference,
      sym_tuple,
      sym_number,
      sym_operation,
      sym_string,
      sym_color,
      sym_enum,
  [6273] = 25,
    ACTIONS(5), 1,
      anon_sym_DASH_DASH,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(213), 1,
      sym_identifier,
    ACTIONS(219), 1,
      aux_sym_macro_call_token1,
    ACTIONS(221), 1,
      aux_sym_state_selector_token1,
    ACTIONS(223), 1,
      sym_tuple_open,
    ACTIONS(229), 1,
      sym_number_percent,
    ACTIONS(233), 1,
      aux_sym_color_code_tailwind_token1,
    ACTIONS(235), 1,
      aux_sym_color_code_skin_token1,
    ACTIONS(237), 1,
      aux_sym_color_code_brick_token1,
    ACTIONS(239), 1,
      aux_sym_color_code_css_token1,
    ACTIONS(241), 1,
      sym_color_hex,
    ACTIONS(243), 1,
      anon_sym_Enum,
    ACTIONS(247), 1,
      sym__block_string_start,
    STATE(102), 1,
      aux_sym_macro_call_repeat1,
    STATE(172), 1,
      sym_datatype,
    STATE(207), 1,
      sym_block_string,
    ACTIONS(227), 2,
      sym_number_ambiguous,
      sym_number_offset,
    ACTIONS(231), 2,
      sym_single_quote_string,
      sym_double_quote_string,
    ACTIONS(245), 2,
      sym_rbx_asset,
      sym_rbx_content,
    STATE(99), 2,
      sym_macro_call,
      sym_comment,
    STATE(197), 2,
      sym_enum_full,
      sym_enum_shorthand,
    ACTIONS(217), 3,
      sym_token,
      sym_static_token,
      sym_static_argument,
    STATE(195), 4,
      sym_color_code_tailwind,
      sym_color_code_skin,
      sym_color_code_brick,
      sym_color_code_css,
    STATE(212), 7,
      sym_reference,
      sym_tuple,
      sym_number,
      sym_operation,
      sym_string,
      sym_color,
      sym_enum,
  [6365] = 25,
    ACTIONS(5), 1,
      anon_sym_DASH_DASH,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(213), 1,
      sym_identifier,
    ACTIONS(219), 1,
      aux_sym_macro_call_token1,
    ACTIONS(221), 1,
      aux_sym_state_selector_token1,
    ACTIONS(223), 1,
      sym_tuple_open,
    ACTIONS(229), 1,
      sym_number_percent,
    ACTIONS(233), 1,
      aux_sym_color_code_tailwind_token1,
    ACTIONS(235), 1,
      aux_sym_color_code_skin_token1,
    ACTIONS(237), 1,
      aux_sym_color_code_brick_token1,
    ACTIONS(239), 1,
      aux_sym_color_code_css_token1,
    ACTIONS(241), 1,
      sym_color_hex,
    ACTIONS(243), 1,
      anon_sym_Enum,
    ACTIONS(247), 1,
      sym__block_string_start,
    STATE(83), 1,
      aux_sym_macro_call_repeat1,
    STATE(172), 1,
      sym_datatype,
    STATE(207), 1,
      sym_block_string,
    ACTIONS(227), 2,
      sym_number_ambiguous,
      sym_number_offset,
    ACTIONS(231), 2,
      sym_single_quote_string,
      sym_double_quote_string,
    ACTIONS(245), 2,
      sym_rbx_asset,
      sym_rbx_content,
    STATE(100), 2,
      sym_macro_call,
      sym_comment,
    STATE(197), 2,
      sym_enum_full,
      sym_enum_shorthand,
    ACTIONS(217), 3,
      sym_token,
      sym_static_token,
      sym_static_argument,
    STATE(195), 4,
      sym_color_code_tailwind,
      sym_color_code_skin,
      sym_color_code_brick,
      sym_color_code_css,
    STATE(212), 7,
      sym_reference,
      sym_tuple,
      sym_number,
      sym_operation,
      sym_string,
      sym_color,
      sym_enum,
  [6457] = 25,
    ACTIONS(5), 1,
      anon_sym_DASH_DASH,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(213), 1,
      sym_identifier,
    ACTIONS(219), 1,
      aux_sym_macro_call_token1,
    ACTIONS(221), 1,
      aux_sym_state_selector_token1,
    ACTIONS(223), 1,
      sym_tuple_open,
    ACTIONS(229), 1,
      sym_number_percent,
    ACTIONS(233), 1,
      aux_sym_color_code_tailwind_token1,
    ACTIONS(235), 1,
      aux_sym_color_code_skin_token1,
    ACTIONS(237), 1,
      aux_sym_color_code_brick_token1,
    ACTIONS(239), 1,
      aux_sym_color_code_css_token1,
    ACTIONS(241), 1,
      sym_color_hex,
    ACTIONS(243), 1,
      anon_sym_Enum,
    ACTIONS(247), 1,
      sym__block_string_start,
    STATE(83), 1,
      aux_sym_macro_call_repeat1,
    STATE(173), 1,
      sym_datatype,
    STATE(207), 1,
      sym_block_string,
    ACTIONS(227), 2,
      sym_number_ambiguous,
      sym_number_offset,
    ACTIONS(231), 2,
      sym_single_quote_string,
      sym_double_quote_string,
    ACTIONS(245), 2,
      sym_rbx_asset,
      sym_rbx_content,
    STATE(101), 2,
      sym_macro_call,
      sym_comment,
    STATE(197), 2,
      sym_enum_full,
      sym_enum_shorthand,
    ACTIONS(217), 3,
      sym_token,
      sym_static_token,
      sym_static_argument,
    STATE(195), 4,
      sym_color_code_tailwind,
      sym_color_code_skin,
      sym_color_code_brick,
      sym_color_code_css,
    STATE(212), 7,
      sym_reference,
      sym_tuple,
      sym_number,
      sym_operation,
      sym_string,
      sym_color,
      sym_enum,
  [6549] = 25,
    ACTIONS(5), 1,
      anon_sym_DASH_DASH,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(213), 1,
      sym_identifier,
    ACTIONS(219), 1,
      aux_sym_macro_call_token1,
    ACTIONS(221), 1,
      aux_sym_state_selector_token1,
    ACTIONS(223), 1,
      sym_tuple_open,
    ACTIONS(229), 1,
      sym_number_percent,
    ACTIONS(233), 1,
      aux_sym_color_code_tailwind_token1,
    ACTIONS(235), 1,
      aux_sym_color_code_skin_token1,
    ACTIONS(237), 1,
      aux_sym_color_code_brick_token1,
    ACTIONS(239), 1,
      aux_sym_color_code_css_token1,
    ACTIONS(241), 1,
      sym_color_hex,
    ACTIONS(243), 1,
      anon_sym_Enum,
    ACTIONS(247), 1,
      sym__block_string_start,
    STATE(83), 1,
      aux_sym_macro_call_repeat1,
    STATE(174), 1,
      sym_datatype,
    STATE(207), 1,
      sym_block_string,
    ACTIONS(227), 2,
      sym_number_ambiguous,
      sym_number_offset,
    ACTIONS(231), 2,
      sym_single_quote_string,
      sym_double_quote_string,
    ACTIONS(245), 2,
      sym_rbx_asset,
      sym_rbx_content,
    STATE(102), 2,
      sym_macro_call,
      sym_comment,
    STATE(197), 2,
      sym_enum_full,
      sym_enum_shorthand,
    ACTIONS(217), 3,
      sym_token,
      sym_static_token,
      sym_static_argument,
    STATE(195), 4,
      sym_color_code_tailwind,
      sym_color_code_skin,
      sym_color_code_brick,
      sym_color_code_css,
    STATE(212), 7,
      sym_reference,
      sym_tuple,
      sym_number,
      sym_operation,
      sym_string,
      sym_color,
      sym_enum,
  [6641] = 25,
    ACTIONS(5), 1,
      anon_sym_DASH_DASH,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(213), 1,
      sym_identifier,
    ACTIONS(219), 1,
      aux_sym_macro_call_token1,
    ACTIONS(221), 1,
      aux_sym_state_selector_token1,
    ACTIONS(223), 1,
      sym_tuple_open,
    ACTIONS(229), 1,
      sym_number_percent,
    ACTIONS(233), 1,
      aux_sym_color_code_tailwind_token1,
    ACTIONS(235), 1,
      aux_sym_color_code_skin_token1,
    ACTIONS(237), 1,
      aux_sym_color_code_brick_token1,
    ACTIONS(239), 1,
      aux_sym_color_code_css_token1,
    ACTIONS(241), 1,
      sym_color_hex,
    ACTIONS(243), 1,
      anon_sym_Enum,
    ACTIONS(247), 1,
      sym__block_string_start,
    STATE(89), 1,
      aux_sym_macro_call_repeat1,
    STATE(176), 1,
      sym_datatype,
    STATE(207), 1,
      sym_block_string,
    ACTIONS(227), 2,
      sym_number_ambiguous,
      sym_number_offset,
    ACTIONS(231), 2,
      sym_single_quote_string,
      sym_double_quote_string,
    ACTIONS(245), 2,
      sym_rbx_asset,
      sym_rbx_content,
    STATE(103), 2,
      sym_macro_call,
      sym_comment,
    STATE(197), 2,
      sym_enum_full,
      sym_enum_shorthand,
    ACTIONS(217), 3,
      sym_token,
      sym_static_token,
      sym_static_argument,
    STATE(195), 4,
      sym_color_code_tailwind,
      sym_color_code_skin,
      sym_color_code_brick,
      sym_color_code_css,
    STATE(212), 7,
      sym_reference,
      sym_tuple,
      sym_number,
      sym_operation,
      sym_string,
      sym_color,
      sym_enum,
  [6733] = 25,
    ACTIONS(5), 1,
      anon_sym_DASH_DASH,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(213), 1,
      sym_identifier,
    ACTIONS(219), 1,
      aux_sym_macro_call_token1,
    ACTIONS(221), 1,
      aux_sym_state_selector_token1,
    ACTIONS(223), 1,
      sym_tuple_open,
    ACTIONS(229), 1,
      sym_number_percent,
    ACTIONS(233), 1,
      aux_sym_color_code_tailwind_token1,
    ACTIONS(235), 1,
      aux_sym_color_code_skin_token1,
    ACTIONS(237), 1,
      aux_sym_color_code_brick_token1,
    ACTIONS(239), 1,
      aux_sym_color_code_css_token1,
    ACTIONS(241), 1,
      sym_color_hex,
    ACTIONS(243), 1,
      anon_sym_Enum,
    ACTIONS(247), 1,
      sym__block_string_start,
    STATE(83), 1,
      aux_sym_macro_call_repeat1,
    STATE(158), 1,
      sym_datatype,
    STATE(207), 1,
      sym_block_string,
    ACTIONS(227), 2,
      sym_number_ambiguous,
      sym_number_offset,
    ACTIONS(231), 2,
      sym_single_quote_string,
      sym_double_quote_string,
    ACTIONS(245), 2,
      sym_rbx_asset,
      sym_rbx_content,
    STATE(104), 2,
      sym_macro_call,
      sym_comment,
    STATE(197), 2,
      sym_enum_full,
      sym_enum_shorthand,
    ACTIONS(217), 3,
      sym_token,
      sym_static_token,
      sym_static_argument,
    STATE(195), 4,
      sym_color_code_tailwind,
      sym_color_code_skin,
      sym_color_code_brick,
      sym_color_code_css,
    STATE(212), 7,
      sym_reference,
      sym_tuple,
      sym_number,
      sym_operation,
      sym_string,
      sym_color,
      sym_enum,
  [6825] = 24,
    ACTIONS(5), 1,
      anon_sym_DASH_DASH,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(47), 1,
      aux_sym_macro_call_token1,
    ACTIONS(65), 1,
      sym_tuple_open,
    ACTIONS(69), 1,
      sym_number_percent,
    ACTIONS(77), 1,
      aux_sym_color_code_tailwind_token1,
    ACTIONS(79), 1,
      aux_sym_color_code_skin_token1,
    ACTIONS(81), 1,
      aux_sym_color_code_brick_token1,
    ACTIONS(83), 1,
      aux_sym_color_code_css_token1,
    ACTIONS(87), 1,
      anon_sym_Enum,
    ACTIONS(91), 1,
      sym__block_string_start,
    ACTIONS(513), 1,
      sym_identifier,
    ACTIONS(517), 1,
      aux_sym_state_selector_token1,
    ACTIONS(519), 1,
      sym_color_hex,
    STATE(7), 1,
      sym_datatype,
    STATE(26), 1,
      sym_block_string,
    ACTIONS(67), 2,
      sym_number_ambiguous,
      sym_number_offset,
    ACTIONS(75), 2,
      sym_single_quote_string,
      sym_double_quote_string,
    ACTIONS(89), 2,
      sym_rbx_asset,
      sym_rbx_content,
    STATE(33), 2,
      sym_enum_full,
      sym_enum_shorthand,
    STATE(105), 2,
      sym_macro_call,
      sym_comment,
    ACTIONS(515), 3,
      sym_token,
      sym_static_token,
      sym_static_argument,
    STATE(31), 4,
      sym_color_code_tailwind,
      sym_color_code_skin,
      sym_color_code_brick,
      sym_color_code_css,
    STATE(32), 7,
      sym_reference,
      sym_tuple,
      sym_number,
      sym_operation,
      sym_string,
      sym_color,
      sym_enum,
  [6914] = 24,
    ACTIONS(5), 1,
      anon_sym_DASH_DASH,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(213), 1,
      sym_identifier,
    ACTIONS(219), 1,
      aux_sym_macro_call_token1,
    ACTIONS(221), 1,
      aux_sym_state_selector_token1,
    ACTIONS(223), 1,
      sym_tuple_open,
    ACTIONS(229), 1,
      sym_number_percent,
    ACTIONS(233), 1,
      aux_sym_color_code_tailwind_token1,
    ACTIONS(235), 1,
      aux_sym_color_code_skin_token1,
    ACTIONS(237), 1,
      aux_sym_color_code_brick_token1,
    ACTIONS(239), 1,
      aux_sym_color_code_css_token1,
    ACTIONS(241), 1,
      sym_color_hex,
    ACTIONS(243), 1,
      anon_sym_Enum,
    ACTIONS(247), 1,
      sym__block_string_start,
    STATE(157), 1,
      sym_datatype,
    STATE(207), 1,
      sym_block_string,
    ACTIONS(227), 2,
      sym_number_ambiguous,
      sym_number_offset,
    ACTIONS(231), 2,
      sym_single_quote_string,
      sym_double_quote_string,
    ACTIONS(245), 2,
      sym_rbx_asset,
      sym_rbx_content,
    STATE(106), 2,
      sym_macro_call,
      sym_comment,
    STATE(197), 2,
      sym_enum_full,
      sym_enum_shorthand,
    ACTIONS(217), 3,
      sym_token,
      sym_static_token,
      sym_static_argument,
    STATE(195), 4,
      sym_color_code_tailwind,
      sym_color_code_skin,
      sym_color_code_brick,
      sym_color_code_css,
    STATE(212), 7,
      sym_reference,
      sym_tuple,
      sym_number,
      sym_operation,
      sym_string,
      sym_color,
      sym_enum,
  [7003] = 20,
    ACTIONS(3), 1,
      aux_sym_macro_call_token1,
    ACTIONS(5), 1,
      anon_sym_DASH_DASH,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(521), 1,
      sym_identifier,
    ACTIONS(524), 1,
      sym_comma,
    ACTIONS(527), 1,
      sym_semi_colon,
    ACTIONS(530), 1,
      sym_token,
    ACTIONS(533), 1,
      sym_static_token,
    ACTIONS(539), 1,
      anon_sym_ATpriority,
    ACTIONS(542), 1,
      anon_sym_ATname,
    ACTIONS(545), 1,
      aux_sym_state_selector_token1,
    ACTIONS(548), 1,
      sym_scope_open,
    ACTIONS(551), 1,
      sym_scope_close,
    STATE(142), 1,
      sym_rule_scope_body,
    STATE(201), 1,
      aux_sym_rule_scope_repeat1,
    STATE(221), 1,
      sym_selector,
    STATE(223), 2,
      sym_class_selector,
      sym_state_selector,
    STATE(107), 3,
      sym_macro_call,
      sym_comment,
      aux_sym_rule_scope_repeat2,
    ACTIONS(536), 4,
      sym_static_argument,
      sym_name_selector,
      sym_tag_selector,
      sym_pseudo_selector,
    STATE(141), 7,
      sym__definition_scope_body,
      sym_property_assignment,
      sym_token_assignment,
      sym_static_token_assignment,
      sym_priority_declaration,
      sym_name_declaration,
      sym_rule_scope,
  [7076] = 21,
    ACTIONS(3), 1,
      aux_sym_macro_call_token1,
    ACTIONS(5), 1,
      anon_sym_DASH_DASH,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(13), 1,
      sym_comma,
    ACTIONS(15), 1,
      sym_semi_colon,
    ACTIONS(17), 1,
      sym_token,
    ACTIONS(19), 1,
      sym_static_token,
    ACTIONS(27), 1,
      aux_sym_state_selector_token1,
    ACTIONS(29), 1,
      sym_scope_open,
    ACTIONS(553), 1,
      sym_identifier,
    ACTIONS(555), 1,
      anon_sym_ATpriority,
    ACTIONS(557), 1,
      anon_sym_ATname,
    ACTIONS(559), 1,
      sym_scope_close,
    STATE(107), 1,
      aux_sym_rule_scope_repeat2,
    STATE(142), 1,
      sym_rule_scope_body,
    STATE(201), 1,
      aux_sym_rule_scope_repeat1,
    STATE(221), 1,
      sym_selector,
    STATE(108), 2,
      sym_macro_call,
      sym_comment,
    STATE(223), 2,
      sym_class_selector,
      sym_state_selector,
    ACTIONS(21), 4,
      sym_static_argument,
      sym_name_selector,
      sym_tag_selector,
      sym_pseudo_selector,
    STATE(141), 7,
      sym__definition_scope_body,
      sym_property_assignment,
      sym_token_assignment,
      sym_static_token_assignment,
      sym_priority_declaration,
      sym_name_declaration,
      sym_rule_scope,
  [7151] = 21,
    ACTIONS(3), 1,
      aux_sym_macro_call_token1,
    ACTIONS(5), 1,
      anon_sym_DASH_DASH,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(13), 1,
      sym_comma,
    ACTIONS(15), 1,
      sym_semi_colon,
    ACTIONS(17), 1,
      sym_token,
    ACTIONS(19), 1,
      sym_static_token,
    ACTIONS(27), 1,
      aux_sym_state_selector_token1,
    ACTIONS(29), 1,
      sym_scope_open,
    ACTIONS(553), 1,
      sym_identifier,
    ACTIONS(555), 1,
      anon_sym_ATpriority,
    ACTIONS(557), 1,
      anon_sym_ATname,
    ACTIONS(561), 1,
      sym_scope_close,
    STATE(116), 1,
      aux_sym_rule_scope_repeat2,
    STATE(142), 1,
      sym_rule_scope_body,
    STATE(201), 1,
      aux_sym_rule_scope_repeat1,
    STATE(221), 1,
      sym_selector,
    STATE(109), 2,
      sym_macro_call,
      sym_comment,
    STATE(223), 2,
      sym_class_selector,
      sym_state_selector,
    ACTIONS(21), 4,
      sym_static_argument,
      sym_name_selector,
      sym_tag_selector,
      sym_pseudo_selector,
    STATE(141), 7,
      sym__definition_scope_body,
      sym_property_assignment,
      sym_token_assignment,
      sym_static_token_assignment,
      sym_priority_declaration,
      sym_name_declaration,
      sym_rule_scope,
  [7226] = 21,
    ACTIONS(3), 1,
      aux_sym_macro_call_token1,
    ACTIONS(5), 1,
      anon_sym_DASH_DASH,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(13), 1,
      sym_comma,
    ACTIONS(15), 1,
      sym_semi_colon,
    ACTIONS(17), 1,
      sym_token,
    ACTIONS(19), 1,
      sym_static_token,
    ACTIONS(27), 1,
      aux_sym_state_selector_token1,
    ACTIONS(29), 1,
      sym_scope_open,
    ACTIONS(553), 1,
      sym_identifier,
    ACTIONS(555), 1,
      anon_sym_ATpriority,
    ACTIONS(557), 1,
      anon_sym_ATname,
    ACTIONS(563), 1,
      sym_scope_close,
    STATE(107), 1,
      aux_sym_rule_scope_repeat2,
    STATE(142), 1,
      sym_rule_scope_body,
    STATE(201), 1,
      aux_sym_rule_scope_repeat1,
    STATE(221), 1,
      sym_selector,
    STATE(110), 2,
      sym_macro_call,
      sym_comment,
    STATE(223), 2,
      sym_class_selector,
      sym_state_selector,
    ACTIONS(21), 4,
      sym_static_argument,
      sym_name_selector,
      sym_tag_selector,
      sym_pseudo_selector,
    STATE(141), 7,
      sym__definition_scope_body,
      sym_property_assignment,
      sym_token_assignment,
      sym_static_token_assignment,
      sym_priority_declaration,
      sym_name_declaration,
      sym_rule_scope,
  [7301] = 21,
    ACTIONS(3), 1,
      aux_sym_macro_call_token1,
    ACTIONS(5), 1,
      anon_sym_DASH_DASH,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(13), 1,
      sym_comma,
    ACTIONS(15), 1,
      sym_semi_colon,
    ACTIONS(17), 1,
      sym_token,
    ACTIONS(19), 1,
      sym_static_token,
    ACTIONS(27), 1,
      aux_sym_state_selector_token1,
    ACTIONS(29), 1,
      sym_scope_open,
    ACTIONS(553), 1,
      sym_identifier,
    ACTIONS(555), 1,
      anon_sym_ATpriority,
    ACTIONS(557), 1,
      anon_sym_ATname,
    ACTIONS(565), 1,
      sym_scope_close,
    STATE(115), 1,
      aux_sym_rule_scope_repeat2,
    STATE(142), 1,
      sym_rule_scope_body,
    STATE(201), 1,
      aux_sym_rule_scope_repeat1,
    STATE(221), 1,
      sym_selector,
    STATE(111), 2,
      sym_macro_call,
      sym_comment,
    STATE(223), 2,
      sym_class_selector,
      sym_state_selector,
    ACTIONS(21), 4,
      sym_static_argument,
      sym_name_selector,
      sym_tag_selector,
      sym_pseudo_selector,
    STATE(141), 7,
      sym__definition_scope_body,
      sym_property_assignment,
      sym_token_assignment,
      sym_static_token_assignment,
      sym_priority_declaration,
      sym_name_declaration,
      sym_rule_scope,
  [7376] = 21,
    ACTIONS(3), 1,
      aux_sym_macro_call_token1,
    ACTIONS(5), 1,
      anon_sym_DASH_DASH,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(13), 1,
      sym_comma,
    ACTIONS(15), 1,
      sym_semi_colon,
    ACTIONS(17), 1,
      sym_token,
    ACTIONS(19), 1,
      sym_static_token,
    ACTIONS(27), 1,
      aux_sym_state_selector_token1,
    ACTIONS(29), 1,
      sym_scope_open,
    ACTIONS(553), 1,
      sym_identifier,
    ACTIONS(555), 1,
      anon_sym_ATpriority,
    ACTIONS(557), 1,
      anon_sym_ATname,
    ACTIONS(567), 1,
      sym_scope_close,
    STATE(114), 1,
      aux_sym_rule_scope_repeat2,
    STATE(142), 1,
      sym_rule_scope_body,
    STATE(201), 1,
      aux_sym_rule_scope_repeat1,
    STATE(221), 1,
      sym_selector,
    STATE(112), 2,
      sym_macro_call,
      sym_comment,
    STATE(223), 2,
      sym_class_selector,
      sym_state_selector,
    ACTIONS(21), 4,
      sym_static_argument,
      sym_name_selector,
      sym_tag_selector,
      sym_pseudo_selector,
    STATE(141), 7,
      sym__definition_scope_body,
      sym_property_assignment,
      sym_token_assignment,
      sym_static_token_assignment,
      sym_priority_declaration,
      sym_name_declaration,
      sym_rule_scope,
  [7451] = 21,
    ACTIONS(3), 1,
      aux_sym_macro_call_token1,
    ACTIONS(5), 1,
      anon_sym_DASH_DASH,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(13), 1,
      sym_comma,
    ACTIONS(15), 1,
      sym_semi_colon,
    ACTIONS(17), 1,
      sym_token,
    ACTIONS(19), 1,
      sym_static_token,
    ACTIONS(27), 1,
      aux_sym_state_selector_token1,
    ACTIONS(29), 1,
      sym_scope_open,
    ACTIONS(553), 1,
      sym_identifier,
    ACTIONS(555), 1,
      anon_sym_ATpriority,
    ACTIONS(557), 1,
      anon_sym_ATname,
    ACTIONS(569), 1,
      sym_scope_close,
    STATE(107), 1,
      aux_sym_rule_scope_repeat2,
    STATE(142), 1,
      sym_rule_scope_body,
    STATE(201), 1,
      aux_sym_rule_scope_repeat1,
    STATE(221), 1,
      sym_selector,
    STATE(113), 2,
      sym_macro_call,
      sym_comment,
    STATE(223), 2,
      sym_class_selector,
      sym_state_selector,
    ACTIONS(21), 4,
      sym_static_argument,
      sym_name_selector,
      sym_tag_selector,
      sym_pseudo_selector,
    STATE(141), 7,
      sym__definition_scope_body,
      sym_property_assignment,
      sym_token_assignment,
      sym_static_token_assignment,
      sym_priority_declaration,
      sym_name_declaration,
      sym_rule_scope,
  [7526] = 21,
    ACTIONS(3), 1,
      aux_sym_macro_call_token1,
    ACTIONS(5), 1,
      anon_sym_DASH_DASH,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(13), 1,
      sym_comma,
    ACTIONS(15), 1,
      sym_semi_colon,
    ACTIONS(17), 1,
      sym_token,
    ACTIONS(19), 1,
      sym_static_token,
    ACTIONS(27), 1,
      aux_sym_state_selector_token1,
    ACTIONS(29), 1,
      sym_scope_open,
    ACTIONS(553), 1,
      sym_identifier,
    ACTIONS(555), 1,
      anon_sym_ATpriority,
    ACTIONS(557), 1,
      anon_sym_ATname,
    ACTIONS(571), 1,
      sym_scope_close,
    STATE(107), 1,
      aux_sym_rule_scope_repeat2,
    STATE(142), 1,
      sym_rule_scope_body,
    STATE(201), 1,
      aux_sym_rule_scope_repeat1,
    STATE(221), 1,
      sym_selector,
    STATE(114), 2,
      sym_macro_call,
      sym_comment,
    STATE(223), 2,
      sym_class_selector,
      sym_state_selector,
    ACTIONS(21), 4,
      sym_static_argument,
      sym_name_selector,
      sym_tag_selector,
      sym_pseudo_selector,
    STATE(141), 7,
      sym__definition_scope_body,
      sym_property_assignment,
      sym_token_assignment,
      sym_static_token_assignment,
      sym_priority_declaration,
      sym_name_declaration,
      sym_rule_scope,
  [7601] = 21,
    ACTIONS(3), 1,
      aux_sym_macro_call_token1,
    ACTIONS(5), 1,
      anon_sym_DASH_DASH,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(13), 1,
      sym_comma,
    ACTIONS(15), 1,
      sym_semi_colon,
    ACTIONS(17), 1,
      sym_token,
    ACTIONS(19), 1,
      sym_static_token,
    ACTIONS(27), 1,
      aux_sym_state_selector_token1,
    ACTIONS(29), 1,
      sym_scope_open,
    ACTIONS(553), 1,
      sym_identifier,
    ACTIONS(555), 1,
      anon_sym_ATpriority,
    ACTIONS(557), 1,
      anon_sym_ATname,
    ACTIONS(573), 1,
      sym_scope_close,
    STATE(107), 1,
      aux_sym_rule_scope_repeat2,
    STATE(142), 1,
      sym_rule_scope_body,
    STATE(201), 1,
      aux_sym_rule_scope_repeat1,
    STATE(221), 1,
      sym_selector,
    STATE(115), 2,
      sym_macro_call,
      sym_comment,
    STATE(223), 2,
      sym_class_selector,
      sym_state_selector,
    ACTIONS(21), 4,
      sym_static_argument,
      sym_name_selector,
      sym_tag_selector,
      sym_pseudo_selector,
    STATE(141), 7,
      sym__definition_scope_body,
      sym_property_assignment,
      sym_token_assignment,
      sym_static_token_assignment,
      sym_priority_declaration,
      sym_name_declaration,
      sym_rule_scope,
  [7676] = 21,
    ACTIONS(3), 1,
      aux_sym_macro_call_token1,
    ACTIONS(5), 1,
      anon_sym_DASH_DASH,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(13), 1,
      sym_comma,
    ACTIONS(15), 1,
      sym_semi_colon,
    ACTIONS(17), 1,
      sym_token,
    ACTIONS(19), 1,
      sym_static_token,
    ACTIONS(27), 1,
      aux_sym_state_selector_token1,
    ACTIONS(29), 1,
      sym_scope_open,
    ACTIONS(553), 1,
      sym_identifier,
    ACTIONS(555), 1,
      anon_sym_ATpriority,
    ACTIONS(557), 1,
      anon_sym_ATname,
    ACTIONS(575), 1,
      sym_scope_close,
    STATE(107), 1,
      aux_sym_rule_scope_repeat2,
    STATE(142), 1,
      sym_rule_scope_body,
    STATE(201), 1,
      aux_sym_rule_scope_repeat1,
    STATE(221), 1,
      sym_selector,
    STATE(116), 2,
      sym_macro_call,
      sym_comment,
    STATE(223), 2,
      sym_class_selector,
      sym_state_selector,
    ACTIONS(21), 4,
      sym_static_argument,
      sym_name_selector,
      sym_tag_selector,
      sym_pseudo_selector,
    STATE(141), 7,
      sym__definition_scope_body,
      sym_property_assignment,
      sym_token_assignment,
      sym_static_token_assignment,
      sym_priority_declaration,
      sym_name_declaration,
      sym_rule_scope,
  [7751] = 21,
    ACTIONS(3), 1,
      aux_sym_macro_call_token1,
    ACTIONS(5), 1,
      anon_sym_DASH_DASH,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(577), 1,
      ts_builtin_sym_end,
    ACTIONS(579), 1,
      sym_identifier,
    ACTIONS(582), 1,
      sym_comma,
    ACTIONS(585), 1,
      sym_semi_colon,
    ACTIONS(588), 1,
      sym_token,
    ACTIONS(591), 1,
      sym_static_token,
    ACTIONS(597), 1,
      anon_sym_ATmacro,
    ACTIONS(600), 1,
      anon_sym_ATderive,
    ACTIONS(603), 1,
      aux_sym_state_selector_token1,
    ACTIONS(606), 1,
      sym_scope_open,
    STATE(151), 1,
      sym__definition,
    STATE(201), 1,
      aux_sym_rule_scope_repeat1,
    STATE(221), 1,
      sym_selector,
    STATE(147), 2,
      sym__definition_no_macro,
      sym_macro_declaration,
    STATE(223), 2,
      sym_class_selector,
      sym_state_selector,
    STATE(117), 3,
      sym_macro_call,
      sym_comment,
      aux_sym_source_file_repeat1,
    ACTIONS(594), 4,
      sym_static_argument,
      sym_name_selector,
      sym_tag_selector,
      sym_pseudo_selector,
    STATE(145), 5,
      sym__definition_scope_body,
      sym_token_assignment,
      sym_static_token_assignment,
      sym_derive_declaration,
      sym_rule_scope,
  [7826] = 21,
    ACTIONS(3), 1,
      aux_sym_macro_call_token1,
    ACTIONS(5), 1,
      anon_sym_DASH_DASH,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(13), 1,
      sym_comma,
    ACTIONS(15), 1,
      sym_semi_colon,
    ACTIONS(17), 1,
      sym_token,
    ACTIONS(19), 1,
      sym_static_token,
    ACTIONS(27), 1,
      aux_sym_state_selector_token1,
    ACTIONS(29), 1,
      sym_scope_open,
    ACTIONS(553), 1,
      sym_identifier,
    ACTIONS(555), 1,
      anon_sym_ATpriority,
    ACTIONS(557), 1,
      anon_sym_ATname,
    ACTIONS(609), 1,
      sym_scope_close,
    STATE(119), 1,
      aux_sym_rule_scope_repeat2,
    STATE(142), 1,
      sym_rule_scope_body,
    STATE(201), 1,
      aux_sym_rule_scope_repeat1,
    STATE(221), 1,
      sym_selector,
    STATE(118), 2,
      sym_macro_call,
      sym_comment,
    STATE(223), 2,
      sym_class_selector,
      sym_state_selector,
    ACTIONS(21), 4,
      sym_static_argument,
      sym_name_selector,
      sym_tag_selector,
      sym_pseudo_selector,
    STATE(141), 7,
      sym__definition_scope_body,
      sym_property_assignment,
      sym_token_assignment,
      sym_static_token_assignment,
      sym_priority_declaration,
      sym_name_declaration,
      sym_rule_scope,
  [7901] = 21,
    ACTIONS(3), 1,
      aux_sym_macro_call_token1,
    ACTIONS(5), 1,
      anon_sym_DASH_DASH,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(13), 1,
      sym_comma,
    ACTIONS(15), 1,
      sym_semi_colon,
    ACTIONS(17), 1,
      sym_token,
    ACTIONS(19), 1,
      sym_static_token,
    ACTIONS(27), 1,
      aux_sym_state_selector_token1,
    ACTIONS(29), 1,
      sym_scope_open,
    ACTIONS(553), 1,
      sym_identifier,
    ACTIONS(555), 1,
      anon_sym_ATpriority,
    ACTIONS(557), 1,
      anon_sym_ATname,
    ACTIONS(611), 1,
      sym_scope_close,
    STATE(107), 1,
      aux_sym_rule_scope_repeat2,
    STATE(142), 1,
      sym_rule_scope_body,
    STATE(201), 1,
      aux_sym_rule_scope_repeat1,
    STATE(221), 1,
      sym_selector,
    STATE(119), 2,
      sym_macro_call,
      sym_comment,
    STATE(223), 2,
      sym_class_selector,
      sym_state_selector,
    ACTIONS(21), 4,
      sym_static_argument,
      sym_name_selector,
      sym_tag_selector,
      sym_pseudo_selector,
    STATE(141), 7,
      sym__definition_scope_body,
      sym_property_assignment,
      sym_token_assignment,
      sym_static_token_assignment,
      sym_priority_declaration,
      sym_name_declaration,
      sym_rule_scope,
  [7976] = 21,
    ACTIONS(3), 1,
      aux_sym_macro_call_token1,
    ACTIONS(5), 1,
      anon_sym_DASH_DASH,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(13), 1,
      sym_comma,
    ACTIONS(15), 1,
      sym_semi_colon,
    ACTIONS(17), 1,
      sym_token,
    ACTIONS(19), 1,
      sym_static_token,
    ACTIONS(27), 1,
      aux_sym_state_selector_token1,
    ACTIONS(29), 1,
      sym_scope_open,
    ACTIONS(553), 1,
      sym_identifier,
    ACTIONS(555), 1,
      anon_sym_ATpriority,
    ACTIONS(557), 1,
      anon_sym_ATname,
    ACTIONS(613), 1,
      sym_scope_close,
    STATE(121), 1,
      aux_sym_rule_scope_repeat2,
    STATE(142), 1,
      sym_rule_scope_body,
    STATE(201), 1,
      aux_sym_rule_scope_repeat1,
    STATE(221), 1,
      sym_selector,
    STATE(120), 2,
      sym_macro_call,
      sym_comment,
    STATE(223), 2,
      sym_class_selector,
      sym_state_selector,
    ACTIONS(21), 4,
      sym_static_argument,
      sym_name_selector,
      sym_tag_selector,
      sym_pseudo_selector,
    STATE(141), 7,
      sym__definition_scope_body,
      sym_property_assignment,
      sym_token_assignment,
      sym_static_token_assignment,
      sym_priority_declaration,
      sym_name_declaration,
      sym_rule_scope,
  [8051] = 21,
    ACTIONS(3), 1,
      aux_sym_macro_call_token1,
    ACTIONS(5), 1,
      anon_sym_DASH_DASH,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(13), 1,
      sym_comma,
    ACTIONS(15), 1,
      sym_semi_colon,
    ACTIONS(17), 1,
      sym_token,
    ACTIONS(19), 1,
      sym_static_token,
    ACTIONS(27), 1,
      aux_sym_state_selector_token1,
    ACTIONS(29), 1,
      sym_scope_open,
    ACTIONS(553), 1,
      sym_identifier,
    ACTIONS(555), 1,
      anon_sym_ATpriority,
    ACTIONS(557), 1,
      anon_sym_ATname,
    ACTIONS(615), 1,
      sym_scope_close,
    STATE(107), 1,
      aux_sym_rule_scope_repeat2,
    STATE(142), 1,
      sym_rule_scope_body,
    STATE(201), 1,
      aux_sym_rule_scope_repeat1,
    STATE(221), 1,
      sym_selector,
    STATE(121), 2,
      sym_macro_call,
      sym_comment,
    STATE(223), 2,
      sym_class_selector,
      sym_state_selector,
    ACTIONS(21), 4,
      sym_static_argument,
      sym_name_selector,
      sym_tag_selector,
      sym_pseudo_selector,
    STATE(141), 7,
      sym__definition_scope_body,
      sym_property_assignment,
      sym_token_assignment,
      sym_static_token_assignment,
      sym_priority_declaration,
      sym_name_declaration,
      sym_rule_scope,
  [8126] = 21,
    ACTIONS(3), 1,
      aux_sym_macro_call_token1,
    ACTIONS(5), 1,
      anon_sym_DASH_DASH,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(13), 1,
      sym_comma,
    ACTIONS(15), 1,
      sym_semi_colon,
    ACTIONS(17), 1,
      sym_token,
    ACTIONS(19), 1,
      sym_static_token,
    ACTIONS(27), 1,
      aux_sym_state_selector_token1,
    ACTIONS(29), 1,
      sym_scope_open,
    ACTIONS(553), 1,
      sym_identifier,
    ACTIONS(555), 1,
      anon_sym_ATpriority,
    ACTIONS(557), 1,
      anon_sym_ATname,
    ACTIONS(617), 1,
      sym_scope_close,
    STATE(108), 1,
      aux_sym_rule_scope_repeat2,
    STATE(142), 1,
      sym_rule_scope_body,
    STATE(201), 1,
      aux_sym_rule_scope_repeat1,
    STATE(221), 1,
      sym_selector,
    STATE(122), 2,
      sym_macro_call,
      sym_comment,
    STATE(223), 2,
      sym_class_selector,
      sym_state_selector,
    ACTIONS(21), 4,
      sym_static_argument,
      sym_name_selector,
      sym_tag_selector,
      sym_pseudo_selector,
    STATE(141), 7,
      sym__definition_scope_body,
      sym_property_assignment,
      sym_token_assignment,
      sym_static_token_assignment,
      sym_priority_declaration,
      sym_name_declaration,
      sym_rule_scope,
  [8201] = 22,
    ACTIONS(3), 1,
      aux_sym_macro_call_token1,
    ACTIONS(5), 1,
      anon_sym_DASH_DASH,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11), 1,
      sym_identifier,
    ACTIONS(13), 1,
      sym_comma,
    ACTIONS(15), 1,
      sym_semi_colon,
    ACTIONS(17), 1,
      sym_token,
    ACTIONS(19), 1,
      sym_static_token,
    ACTIONS(23), 1,
      anon_sym_ATmacro,
    ACTIONS(25), 1,
      anon_sym_ATderive,
    ACTIONS(27), 1,
      aux_sym_state_selector_token1,
    ACTIONS(29), 1,
      sym_scope_open,
    ACTIONS(619), 1,
      ts_builtin_sym_end,
    STATE(117), 1,
      aux_sym_source_file_repeat1,
    STATE(151), 1,
      sym__definition,
    STATE(201), 1,
      aux_sym_rule_scope_repeat1,
    STATE(221), 1,
      sym_selector,
    STATE(123), 2,
      sym_macro_call,
      sym_comment,
    STATE(147), 2,
      sym__definition_no_macro,
      sym_macro_declaration,
    STATE(223), 2,
      sym_class_selector,
      sym_state_selector,
    ACTIONS(21), 4,
      sym_static_argument,
      sym_name_selector,
      sym_tag_selector,
      sym_pseudo_selector,
    STATE(145), 5,
      sym__definition_scope_body,
      sym_token_assignment,
      sym_static_token_assignment,
      sym_derive_declaration,
      sym_rule_scope,
  [8278] = 21,
    ACTIONS(3), 1,
      aux_sym_macro_call_token1,
    ACTIONS(5), 1,
      anon_sym_DASH_DASH,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(13), 1,
      sym_comma,
    ACTIONS(15), 1,
      sym_semi_colon,
    ACTIONS(17), 1,
      sym_token,
    ACTIONS(19), 1,
      sym_static_token,
    ACTIONS(27), 1,
      aux_sym_state_selector_token1,
    ACTIONS(29), 1,
      sym_scope_open,
    ACTIONS(553), 1,
      sym_identifier,
    ACTIONS(555), 1,
      anon_sym_ATpriority,
    ACTIONS(557), 1,
      anon_sym_ATname,
    ACTIONS(621), 1,
      sym_scope_close,
    STATE(110), 1,
      aux_sym_rule_scope_repeat2,
    STATE(142), 1,
      sym_rule_scope_body,
    STATE(201), 1,
      aux_sym_rule_scope_repeat1,
    STATE(221), 1,
      sym_selector,
    STATE(124), 2,
      sym_macro_call,
      sym_comment,
    STATE(223), 2,
      sym_class_selector,
      sym_state_selector,
    ACTIONS(21), 4,
      sym_static_argument,
      sym_name_selector,
      sym_tag_selector,
      sym_pseudo_selector,
    STATE(141), 7,
      sym__definition_scope_body,
      sym_property_assignment,
      sym_token_assignment,
      sym_static_token_assignment,
      sym_priority_declaration,
      sym_name_declaration,
      sym_rule_scope,
  [8353] = 5,
    ACTIONS(5), 1,
      anon_sym_DASH_DASH,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(125), 2,
      sym_macro_call,
      sym_comment,
    ACTIONS(623), 4,
      sym_identifier,
      sym_number_ambiguous,
      sym_number_offset,
      anon_sym_Enum,
    ACTIONS(625), 17,
      sym__block_string_start,
      sym_token,
      sym_static_token,
      sym_static_argument,
      aux_sym_macro_call_token1,
      aux_sym_state_selector_token1,
      sym_tuple_open,
      sym_number_percent,
      sym_single_quote_string,
      sym_double_quote_string,
      aux_sym_color_code_tailwind_token1,
      aux_sym_color_code_skin_token1,
      aux_sym_color_code_brick_token1,
      aux_sym_color_code_css_token1,
      sym_color_hex,
      sym_rbx_asset,
      sym_rbx_content,
  [8389] = 5,
    ACTIONS(5), 1,
      anon_sym_DASH_DASH,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(126), 2,
      sym_macro_call,
      sym_comment,
    ACTIONS(337), 4,
      sym_identifier,
      sym_number_ambiguous,
      sym_number_offset,
      anon_sym_Enum,
    ACTIONS(339), 17,
      sym__block_string_start,
      sym_token,
      sym_static_token,
      sym_static_argument,
      aux_sym_macro_call_token1,
      aux_sym_state_selector_token1,
      sym_tuple_open,
      sym_number_percent,
      sym_single_quote_string,
      sym_double_quote_string,
      aux_sym_color_code_tailwind_token1,
      aux_sym_color_code_skin_token1,
      aux_sym_color_code_brick_token1,
      aux_sym_color_code_css_token1,
      sym_color_hex,
      sym_rbx_asset,
      sym_rbx_content,
  [8425] = 6,
    ACTIONS(3), 1,
      aux_sym_macro_call_token1,
    ACTIONS(5), 1,
      anon_sym_DASH_DASH,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(195), 1,
      sym_identifier,
    STATE(127), 2,
      sym_macro_call,
      sym_comment,
    ACTIONS(197), 16,
      ts_builtin_sym_end,
      sym_comma,
      sym_semi_colon,
      sym_token,
      sym_static_token,
      sym_static_argument,
      anon_sym_ATmacro,
      anon_sym_ATpriority,
      anon_sym_ATname,
      anon_sym_ATderive,
      sym_name_selector,
      sym_tag_selector,
      aux_sym_state_selector_token1,
      sym_pseudo_selector,
      sym_scope_open,
      sym_scope_close,
  [8460] = 6,
    ACTIONS(3), 1,
      aux_sym_macro_call_token1,
    ACTIONS(5), 1,
      anon_sym_DASH_DASH,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(401), 1,
      sym_identifier,
    STATE(128), 2,
      sym_macro_call,
      sym_comment,
    ACTIONS(403), 16,
      ts_builtin_sym_end,
      sym_comma,
      sym_semi_colon,
      sym_token,
      sym_static_token,
      sym_static_argument,
      anon_sym_ATmacro,
      anon_sym_ATpriority,
      anon_sym_ATname,
      anon_sym_ATderive,
      sym_name_selector,
      sym_tag_selector,
      aux_sym_state_selector_token1,
      sym_pseudo_selector,
      sym_scope_open,
      sym_scope_close,
  [8495] = 6,
    ACTIONS(3), 1,
      aux_sym_macro_call_token1,
    ACTIONS(5), 1,
      anon_sym_DASH_DASH,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(377), 1,
      sym_identifier,
    STATE(129), 2,
      sym_macro_call,
      sym_comment,
    ACTIONS(379), 16,
      ts_builtin_sym_end,
      sym_comma,
      sym_semi_colon,
      sym_token,
      sym_static_token,
      sym_static_argument,
      anon_sym_ATmacro,
      anon_sym_ATpriority,
      anon_sym_ATname,
      anon_sym_ATderive,
      sym_name_selector,
      sym_tag_selector,
      aux_sym_state_selector_token1,
      sym_pseudo_selector,
      sym_scope_open,
      sym_scope_close,
  [8530] = 6,
    ACTIONS(3), 1,
      aux_sym_macro_call_token1,
    ACTIONS(5), 1,
      anon_sym_DASH_DASH,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(381), 1,
      sym_identifier,
    STATE(130), 2,
      sym_macro_call,
      sym_comment,
    ACTIONS(383), 16,
      ts_builtin_sym_end,
      sym_comma,
      sym_semi_colon,
      sym_token,
      sym_static_token,
      sym_static_argument,
      anon_sym_ATmacro,
      anon_sym_ATpriority,
      anon_sym_ATname,
      anon_sym_ATderive,
      sym_name_selector,
      sym_tag_selector,
      aux_sym_state_selector_token1,
      sym_pseudo_selector,
      sym_scope_open,
      sym_scope_close,
  [8565] = 6,
    ACTIONS(3), 1,
      aux_sym_macro_call_token1,
    ACTIONS(5), 1,
      anon_sym_DASH_DASH,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(345), 1,
      sym_identifier,
    STATE(131), 2,
      sym_macro_call,
      sym_comment,
    ACTIONS(347), 16,
      ts_builtin_sym_end,
      sym_comma,
      sym_semi_colon,
      sym_token,
      sym_static_token,
      sym_static_argument,
      anon_sym_ATmacro,
      anon_sym_ATpriority,
      anon_sym_ATname,
      anon_sym_ATderive,
      sym_name_selector,
      sym_tag_selector,
      aux_sym_state_selector_token1,
      sym_pseudo_selector,
      sym_scope_open,
      sym_scope_close,
  [8600] = 6,
    ACTIONS(3), 1,
      aux_sym_macro_call_token1,
    ACTIONS(5), 1,
      anon_sym_DASH_DASH,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(349), 1,
      sym_identifier,
    STATE(132), 2,
      sym_macro_call,
      sym_comment,
    ACTIONS(351), 16,
      ts_builtin_sym_end,
      sym_comma,
      sym_semi_colon,
      sym_token,
      sym_static_token,
      sym_static_argument,
      anon_sym_ATmacro,
      anon_sym_ATpriority,
      anon_sym_ATname,
      anon_sym_ATderive,
      sym_name_selector,
      sym_tag_selector,
      aux_sym_state_selector_token1,
      sym_pseudo_selector,
      sym_scope_open,
      sym_scope_close,
  [8635] = 6,
    ACTIONS(3), 1,
      aux_sym_macro_call_token1,
    ACTIONS(5), 1,
      anon_sym_DASH_DASH,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(629), 1,
      sym_identifier,
    STATE(133), 2,
      sym_macro_call,
      sym_comment,
    ACTIONS(627), 16,
      ts_builtin_sym_end,
      sym_comma,
      sym_semi_colon,
      sym_token,
      sym_static_token,
      sym_static_argument,
      anon_sym_ATmacro,
      anon_sym_ATpriority,
      anon_sym_ATname,
      anon_sym_ATderive,
      sym_name_selector,
      sym_tag_selector,
      aux_sym_state_selector_token1,
      sym_pseudo_selector,
      sym_scope_open,
      sym_scope_close,
  [8670] = 6,
    ACTIONS(3), 1,
      aux_sym_macro_call_token1,
    ACTIONS(5), 1,
      anon_sym_DASH_DASH,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(265), 1,
      sym_identifier,
    STATE(134), 2,
      sym_macro_call,
      sym_comment,
    ACTIONS(267), 16,
      ts_builtin_sym_end,
      sym_comma,
      sym_semi_colon,
      sym_token,
      sym_static_token,
      sym_static_argument,
      anon_sym_ATmacro,
      anon_sym_ATpriority,
      anon_sym_ATname,
      anon_sym_ATderive,
      sym_name_selector,
      sym_tag_selector,
      aux_sym_state_selector_token1,
      sym_pseudo_selector,
      sym_scope_open,
      sym_scope_close,
  [8705] = 6,
    ACTIONS(3), 1,
      aux_sym_macro_call_token1,
    ACTIONS(5), 1,
      anon_sym_DASH_DASH,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(285), 1,
      sym_identifier,
    STATE(135), 2,
      sym_macro_call,
      sym_comment,
    ACTIONS(287), 16,
      ts_builtin_sym_end,
      sym_comma,
      sym_semi_colon,
      sym_token,
      sym_static_token,
      sym_static_argument,
      anon_sym_ATmacro,
      anon_sym_ATpriority,
      anon_sym_ATname,
      anon_sym_ATderive,
      sym_name_selector,
      sym_tag_selector,
      aux_sym_state_selector_token1,
      sym_pseudo_selector,
      sym_scope_open,
      sym_scope_close,
  [8740] = 6,
    ACTIONS(3), 1,
      aux_sym_macro_call_token1,
    ACTIONS(5), 1,
      anon_sym_DASH_DASH,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(289), 1,
      sym_identifier,
    STATE(136), 2,
      sym_macro_call,
      sym_comment,
    ACTIONS(291), 16,
      ts_builtin_sym_end,
      sym_comma,
      sym_semi_colon,
      sym_token,
      sym_static_token,
      sym_static_argument,
      anon_sym_ATmacro,
      anon_sym_ATpriority,
      anon_sym_ATname,
      anon_sym_ATderive,
      sym_name_selector,
      sym_tag_selector,
      aux_sym_state_selector_token1,
      sym_pseudo_selector,
      sym_scope_open,
      sym_scope_close,
  [8775] = 6,
    ACTIONS(3), 1,
      aux_sym_macro_call_token1,
    ACTIONS(5), 1,
      anon_sym_DASH_DASH,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(341), 1,
      sym_identifier,
    STATE(137), 2,
      sym_macro_call,
      sym_comment,
    ACTIONS(343), 16,
      ts_builtin_sym_end,
      sym_comma,
      sym_semi_colon,
      sym_token,
      sym_static_token,
      sym_static_argument,
      anon_sym_ATmacro,
      anon_sym_ATpriority,
      anon_sym_ATname,
      anon_sym_ATderive,
      sym_name_selector,
      sym_tag_selector,
      aux_sym_state_selector_token1,
      sym_pseudo_selector,
      sym_scope_open,
      sym_scope_close,
  [8810] = 6,
    ACTIONS(3), 1,
      aux_sym_macro_call_token1,
    ACTIONS(5), 1,
      anon_sym_DASH_DASH,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(365), 1,
      sym_identifier,
    STATE(138), 2,
      sym_macro_call,
      sym_comment,
    ACTIONS(367), 16,
      ts_builtin_sym_end,
      sym_comma,
      sym_semi_colon,
      sym_token,
      sym_static_token,
      sym_static_argument,
      anon_sym_ATmacro,
      anon_sym_ATpriority,
      anon_sym_ATname,
      anon_sym_ATderive,
      sym_name_selector,
      sym_tag_selector,
      aux_sym_state_selector_token1,
      sym_pseudo_selector,
      sym_scope_open,
      sym_scope_close,
  [8845] = 6,
    ACTIONS(3), 1,
      aux_sym_macro_call_token1,
    ACTIONS(5), 1,
      anon_sym_DASH_DASH,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(309), 1,
      sym_identifier,
    STATE(139), 2,
      sym_macro_call,
      sym_comment,
    ACTIONS(311), 16,
      ts_builtin_sym_end,
      sym_comma,
      sym_semi_colon,
      sym_token,
      sym_static_token,
      sym_static_argument,
      anon_sym_ATmacro,
      anon_sym_ATpriority,
      anon_sym_ATname,
      anon_sym_ATderive,
      sym_name_selector,
      sym_tag_selector,
      aux_sym_state_selector_token1,
      sym_pseudo_selector,
      sym_scope_open,
      sym_scope_close,
  [8880] = 6,
    ACTIONS(3), 1,
      aux_sym_macro_call_token1,
    ACTIONS(5), 1,
      anon_sym_DASH_DASH,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(633), 1,
      sym_identifier,
    STATE(140), 2,
      sym_macro_call,
      sym_comment,
    ACTIONS(631), 13,
      ts_builtin_sym_end,
      sym_comma,
      sym_semi_colon,
      sym_token,
      sym_static_token,
      sym_static_argument,
      anon_sym_ATmacro,
      anon_sym_ATderive,
      sym_name_selector,
      sym_tag_selector,
      aux_sym_state_selector_token1,
      sym_pseudo_selector,
      sym_scope_open,
  [8912] = 6,
    ACTIONS(3), 1,
      aux_sym_macro_call_token1,
    ACTIONS(5), 1,
      anon_sym_DASH_DASH,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(635), 1,
      sym_identifier,
    STATE(141), 2,
      sym_macro_call,
      sym_comment,
    ACTIONS(637), 13,
      sym_comma,
      sym_semi_colon,
      sym_token,
      sym_static_token,
      sym_static_argument,
      anon_sym_ATpriority,
      anon_sym_ATname,
      sym_name_selector,
      sym_tag_selector,
      aux_sym_state_selector_token1,
      sym_pseudo_selector,
      sym_scope_open,
      sym_scope_close,
  [8944] = 6,
    ACTIONS(3), 1,
      aux_sym_macro_call_token1,
    ACTIONS(5), 1,
      anon_sym_DASH_DASH,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(639), 1,
      sym_identifier,
    STATE(142), 2,
      sym_macro_call,
      sym_comment,
    ACTIONS(641), 13,
      sym_comma,
      sym_semi_colon,
      sym_token,
      sym_static_token,
      sym_static_argument,
      anon_sym_ATpriority,
      anon_sym_ATname,
      sym_name_selector,
      sym_tag_selector,
      aux_sym_state_selector_token1,
      sym_pseudo_selector,
      sym_scope_open,
      sym_scope_close,
  [8976] = 6,
    ACTIONS(3), 1,
      aux_sym_macro_call_token1,
    ACTIONS(5), 1,
      anon_sym_DASH_DASH,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(301), 1,
      sym_identifier,
    STATE(143), 2,
      sym_macro_call,
      sym_comment,
    ACTIONS(303), 13,
      sym_comma,
      sym_semi_colon,
      sym_token,
      sym_static_token,
      sym_static_argument,
      anon_sym_ATpriority,
      anon_sym_ATname,
      sym_name_selector,
      sym_tag_selector,
      aux_sym_state_selector_token1,
      sym_pseudo_selector,
      sym_scope_open,
      sym_scope_close,
  [9008] = 6,
    ACTIONS(3), 1,
      aux_sym_macro_call_token1,
    ACTIONS(5), 1,
      anon_sym_DASH_DASH,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(645), 1,
      sym_identifier,
    STATE(144), 2,
      sym_macro_call,
      sym_comment,
    ACTIONS(643), 13,
      ts_builtin_sym_end,
      sym_comma,
      sym_semi_colon,
      sym_token,
      sym_static_token,
      sym_static_argument,
      anon_sym_ATmacro,
      anon_sym_ATderive,
      sym_name_selector,
      sym_tag_selector,
      aux_sym_state_selector_token1,
      sym_pseudo_selector,
      sym_scope_open,
  [9040] = 6,
    ACTIONS(3), 1,
      aux_sym_macro_call_token1,
    ACTIONS(5), 1,
      anon_sym_DASH_DASH,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(257), 1,
      sym_identifier,
    STATE(145), 2,
      sym_macro_call,
      sym_comment,
    ACTIONS(259), 13,
      ts_builtin_sym_end,
      sym_comma,
      sym_semi_colon,
      sym_token,
      sym_static_token,
      sym_static_argument,
      anon_sym_ATmacro,
      anon_sym_ATderive,
      sym_name_selector,
      sym_tag_selector,
      aux_sym_state_selector_token1,
      sym_pseudo_selector,
      sym_scope_open,
  [9072] = 6,
    ACTIONS(3), 1,
      aux_sym_macro_call_token1,
    ACTIONS(5), 1,
      anon_sym_DASH_DASH,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(647), 1,
      sym_identifier,
    STATE(146), 2,
      sym_macro_call,
      sym_comment,
    ACTIONS(649), 13,
      sym_comma,
      sym_semi_colon,
      sym_token,
      sym_static_token,
      sym_static_argument,
      anon_sym_ATpriority,
      anon_sym_ATname,
      sym_name_selector,
      sym_tag_selector,
      aux_sym_state_selector_token1,
      sym_pseudo_selector,
      sym_scope_open,
      sym_scope_close,
  [9104] = 6,
    ACTIONS(3), 1,
      aux_sym_macro_call_token1,
    ACTIONS(5), 1,
      anon_sym_DASH_DASH,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(653), 1,
      sym_identifier,
    STATE(147), 2,
      sym_macro_call,
      sym_comment,
    ACTIONS(651), 13,
      ts_builtin_sym_end,
      sym_comma,
      sym_semi_colon,
      sym_token,
      sym_static_token,
      sym_static_argument,
      anon_sym_ATmacro,
      anon_sym_ATderive,
      sym_name_selector,
      sym_tag_selector,
      aux_sym_state_selector_token1,
      sym_pseudo_selector,
      sym_scope_open,
  [9136] = 6,
    ACTIONS(3), 1,
      aux_sym_macro_call_token1,
    ACTIONS(5), 1,
      anon_sym_DASH_DASH,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(297), 1,
      sym_identifier,
    STATE(148), 2,
      sym_macro_call,
      sym_comment,
    ACTIONS(299), 13,
      sym_comma,
      sym_semi_colon,
      sym_token,
      sym_static_token,
      sym_static_argument,
      anon_sym_ATpriority,
      anon_sym_ATname,
      sym_name_selector,
      sym_tag_selector,
      aux_sym_state_selector_token1,
      sym_pseudo_selector,
      sym_scope_open,
      sym_scope_close,
  [9168] = 6,
    ACTIONS(3), 1,
      aux_sym_macro_call_token1,
    ACTIONS(5), 1,
      anon_sym_DASH_DASH,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(655), 1,
      sym_identifier,
    STATE(149), 2,
      sym_macro_call,
      sym_comment,
    ACTIONS(657), 13,
      sym_comma,
      sym_semi_colon,
      sym_token,
      sym_static_token,
      sym_static_argument,
      anon_sym_ATpriority,
      anon_sym_ATname,
      sym_name_selector,
      sym_tag_selector,
      aux_sym_state_selector_token1,
      sym_pseudo_selector,
      sym_scope_open,
      sym_scope_close,
  [9200] = 6,
    ACTIONS(3), 1,
      aux_sym_macro_call_token1,
    ACTIONS(5), 1,
      anon_sym_DASH_DASH,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(661), 1,
      sym_identifier,
    STATE(150), 2,
      sym_macro_call,
      sym_comment,
    ACTIONS(659), 13,
      ts_builtin_sym_end,
      sym_comma,
      sym_semi_colon,
      sym_token,
      sym_static_token,
      sym_static_argument,
      anon_sym_ATmacro,
      anon_sym_ATderive,
      sym_name_selector,
      sym_tag_selector,
      aux_sym_state_selector_token1,
      sym_pseudo_selector,
      sym_scope_open,
  [9232] = 6,
    ACTIONS(3), 1,
      aux_sym_macro_call_token1,
    ACTIONS(5), 1,
      anon_sym_DASH_DASH,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(665), 1,
      sym_identifier,
    STATE(151), 2,
      sym_macro_call,
      sym_comment,
    ACTIONS(663), 13,
      ts_builtin_sym_end,
      sym_comma,
      sym_semi_colon,
      sym_token,
      sym_static_token,
      sym_static_argument,
      anon_sym_ATmacro,
      anon_sym_ATderive,
      sym_name_selector,
      sym_tag_selector,
      aux_sym_state_selector_token1,
      sym_pseudo_selector,
      sym_scope_open,
  [9264] = 6,
    ACTIONS(3), 1,
      aux_sym_macro_call_token1,
    ACTIONS(5), 1,
      anon_sym_DASH_DASH,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(357), 1,
      sym_identifier,
    STATE(152), 2,
      sym_macro_call,
      sym_comment,
    ACTIONS(359), 13,
      sym_comma,
      sym_semi_colon,
      sym_token,
      sym_static_token,
      sym_static_argument,
      anon_sym_ATpriority,
      anon_sym_ATname,
      sym_name_selector,
      sym_tag_selector,
      aux_sym_state_selector_token1,
      sym_pseudo_selector,
      sym_scope_open,
      sym_scope_close,
  [9296] = 6,
    ACTIONS(3), 1,
      aux_sym_macro_call_token1,
    ACTIONS(5), 1,
      anon_sym_DASH_DASH,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(361), 1,
      sym_identifier,
    STATE(153), 2,
      sym_macro_call,
      sym_comment,
    ACTIONS(363), 13,
      sym_comma,
      sym_semi_colon,
      sym_token,
      sym_static_token,
      sym_static_argument,
      anon_sym_ATpriority,
      anon_sym_ATname,
      sym_name_selector,
      sym_tag_selector,
      aux_sym_state_selector_token1,
      sym_pseudo_selector,
      sym_scope_open,
      sym_scope_close,
  [9328] = 6,
    ACTIONS(3), 1,
      aux_sym_macro_call_token1,
    ACTIONS(5), 1,
      anon_sym_DASH_DASH,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(669), 1,
      sym_identifier,
    STATE(154), 2,
      sym_macro_call,
      sym_comment,
    ACTIONS(667), 13,
      ts_builtin_sym_end,
      sym_comma,
      sym_semi_colon,
      sym_token,
      sym_static_token,
      sym_static_argument,
      anon_sym_ATmacro,
      anon_sym_ATderive,
      sym_name_selector,
      sym_tag_selector,
      aux_sym_state_selector_token1,
      sym_pseudo_selector,
      sym_scope_open,
  [9360] = 6,
    ACTIONS(3), 1,
      aux_sym_macro_call_token1,
    ACTIONS(5), 1,
      anon_sym_DASH_DASH,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(293), 1,
      sym_identifier,
    STATE(155), 2,
      sym_macro_call,
      sym_comment,
    ACTIONS(295), 13,
      ts_builtin_sym_end,
      sym_comma,
      sym_semi_colon,
      sym_token,
      sym_static_token,
      sym_static_argument,
      anon_sym_ATmacro,
      anon_sym_ATderive,
      sym_name_selector,
      sym_tag_selector,
      aux_sym_state_selector_token1,
      sym_pseudo_selector,
      sym_scope_open,
  [9392] = 9,
    ACTIONS(3), 1,
      aux_sym_macro_call_token1,
    ACTIONS(5), 1,
      anon_sym_DASH_DASH,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(673), 1,
      sym_tuple_close,
    STATE(106), 1,
      sym_operator,
    ACTIONS(671), 2,
      sym_comma,
      sym_semi_colon,
    ACTIONS(677), 2,
      sym_div_operator,
      sym_sub_operator,
    STATE(156), 2,
      sym_macro_call,
      sym_comment,
    ACTIONS(675), 5,
      sym_pow_operator,
      sym_floor_div_operator,
      sym_mod_operator,
      sym_mult_operator,
      sym_add_operator,
  [9427] = 7,
    ACTIONS(3), 1,
      aux_sym_macro_call_token1,
    ACTIONS(5), 1,
      anon_sym_DASH_DASH,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(106), 1,
      sym_operator,
    ACTIONS(191), 2,
      sym_div_operator,
      sym_sub_operator,
    STATE(157), 2,
      sym_macro_call,
      sym_comment,
    ACTIONS(193), 8,
      sym_comma,
      sym_semi_colon,
      sym_tuple_close,
      sym_pow_operator,
      sym_floor_div_operator,
      sym_mod_operator,
      sym_mult_operator,
      sym_add_operator,
  [9458] = 9,
    ACTIONS(3), 1,
      aux_sym_macro_call_token1,
    ACTIONS(5), 1,
      anon_sym_DASH_DASH,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(679), 1,
      sym_tuple_close,
    STATE(106), 1,
      sym_operator,
    ACTIONS(671), 2,
      sym_comma,
      sym_semi_colon,
    ACTIONS(677), 2,
      sym_div_operator,
      sym_sub_operator,
    STATE(158), 2,
      sym_macro_call,
      sym_comment,
    ACTIONS(675), 5,
      sym_pow_operator,
      sym_floor_div_operator,
      sym_mod_operator,
      sym_mult_operator,
      sym_add_operator,
  [9493] = 9,
    ACTIONS(3), 1,
      aux_sym_macro_call_token1,
    ACTIONS(5), 1,
      anon_sym_DASH_DASH,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(681), 1,
      sym_tuple_close,
    STATE(106), 1,
      sym_operator,
    ACTIONS(671), 2,
      sym_comma,
      sym_semi_colon,
    ACTIONS(677), 2,
      sym_div_operator,
      sym_sub_operator,
    STATE(159), 2,
      sym_macro_call,
      sym_comment,
    ACTIONS(675), 5,
      sym_pow_operator,
      sym_floor_div_operator,
      sym_mod_operator,
      sym_mult_operator,
      sym_add_operator,
  [9528] = 9,
    ACTIONS(3), 1,
      aux_sym_macro_call_token1,
    ACTIONS(5), 1,
      anon_sym_DASH_DASH,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(683), 1,
      sym_tuple_close,
    STATE(106), 1,
      sym_operator,
    ACTIONS(671), 2,
      sym_comma,
      sym_semi_colon,
    ACTIONS(677), 2,
      sym_div_operator,
      sym_sub_operator,
    STATE(160), 2,
      sym_macro_call,
      sym_comment,
    ACTIONS(675), 5,
      sym_pow_operator,
      sym_floor_div_operator,
      sym_mod_operator,
      sym_mult_operator,
      sym_add_operator,
  [9563] = 9,
    ACTIONS(3), 1,
      aux_sym_macro_call_token1,
    ACTIONS(5), 1,
      anon_sym_DASH_DASH,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(685), 1,
      sym_tuple_close,
    STATE(106), 1,
      sym_operator,
    ACTIONS(671), 2,
      sym_comma,
      sym_semi_colon,
    ACTIONS(677), 2,
      sym_div_operator,
      sym_sub_operator,
    STATE(161), 2,
      sym_macro_call,
      sym_comment,
    ACTIONS(675), 5,
      sym_pow_operator,
      sym_floor_div_operator,
      sym_mod_operator,
      sym_mult_operator,
      sym_add_operator,
  [9598] = 9,
    ACTIONS(3), 1,
      aux_sym_macro_call_token1,
    ACTIONS(5), 1,
      anon_sym_DASH_DASH,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(687), 1,
      sym_tuple_close,
    STATE(106), 1,
      sym_operator,
    ACTIONS(671), 2,
      sym_comma,
      sym_semi_colon,
    ACTIONS(677), 2,
      sym_div_operator,
      sym_sub_operator,
    STATE(162), 2,
      sym_macro_call,
      sym_comment,
    ACTIONS(675), 5,
      sym_pow_operator,
      sym_floor_div_operator,
      sym_mod_operator,
      sym_mult_operator,
      sym_add_operator,
  [9633] = 7,
    ACTIONS(3), 1,
      aux_sym_macro_call_token1,
    ACTIONS(5), 1,
      anon_sym_DASH_DASH,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(689), 1,
      anon_sym_COLON,
    ACTIONS(317), 2,
      sym_div_operator,
      sym_sub_operator,
    STATE(163), 2,
      sym_macro_call,
      sym_comment,
    ACTIONS(319), 8,
      sym_comma,
      sym_semi_colon,
      sym_tuple_close,
      sym_pow_operator,
      sym_floor_div_operator,
      sym_mod_operator,
      sym_mult_operator,
      sym_add_operator,
  [9664] = 7,
    ACTIONS(3), 1,
      aux_sym_macro_call_token1,
    ACTIONS(5), 1,
      anon_sym_DASH_DASH,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(691), 1,
      anon_sym_COLON,
    ACTIONS(321), 2,
      sym_div_operator,
      sym_sub_operator,
    STATE(164), 2,
      sym_macro_call,
      sym_comment,
    ACTIONS(323), 8,
      sym_comma,
      sym_semi_colon,
      sym_tuple_close,
      sym_pow_operator,
      sym_floor_div_operator,
      sym_mod_operator,
      sym_mult_operator,
      sym_add_operator,
  [9695] = 9,
    ACTIONS(3), 1,
      aux_sym_macro_call_token1,
    ACTIONS(5), 1,
      anon_sym_DASH_DASH,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(693), 1,
      sym_tuple_close,
    STATE(106), 1,
      sym_operator,
    ACTIONS(671), 2,
      sym_comma,
      sym_semi_colon,
    ACTIONS(677), 2,
      sym_div_operator,
      sym_sub_operator,
    STATE(165), 2,
      sym_macro_call,
      sym_comment,
    ACTIONS(675), 5,
      sym_pow_operator,
      sym_floor_div_operator,
      sym_mod_operator,
      sym_mult_operator,
      sym_add_operator,
  [9730] = 9,
    ACTIONS(3), 1,
      aux_sym_macro_call_token1,
    ACTIONS(5), 1,
      anon_sym_DASH_DASH,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(695), 1,
      sym_tuple_close,
    STATE(106), 1,
      sym_operator,
    ACTIONS(671), 2,
      sym_comma,
      sym_semi_colon,
    ACTIONS(677), 2,
      sym_div_operator,
      sym_sub_operator,
    STATE(166), 2,
      sym_macro_call,
      sym_comment,
    ACTIONS(675), 5,
      sym_pow_operator,
      sym_floor_div_operator,
      sym_mod_operator,
      sym_mult_operator,
      sym_add_operator,
  [9765] = 9,
    ACTIONS(3), 1,
      aux_sym_macro_call_token1,
    ACTIONS(5), 1,
      anon_sym_DASH_DASH,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(697), 1,
      sym_tuple_close,
    STATE(106), 1,
      sym_operator,
    ACTIONS(671), 2,
      sym_comma,
      sym_semi_colon,
    ACTIONS(677), 2,
      sym_div_operator,
      sym_sub_operator,
    STATE(167), 2,
      sym_macro_call,
      sym_comment,
    ACTIONS(675), 5,
      sym_pow_operator,
      sym_floor_div_operator,
      sym_mod_operator,
      sym_mult_operator,
      sym_add_operator,
  [9800] = 9,
    ACTIONS(3), 1,
      aux_sym_macro_call_token1,
    ACTIONS(5), 1,
      anon_sym_DASH_DASH,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(699), 1,
      sym_tuple_close,
    STATE(106), 1,
      sym_operator,
    ACTIONS(671), 2,
      sym_comma,
      sym_semi_colon,
    ACTIONS(677), 2,
      sym_div_operator,
      sym_sub_operator,
    STATE(168), 2,
      sym_macro_call,
      sym_comment,
    ACTIONS(675), 5,
      sym_pow_operator,
      sym_floor_div_operator,
      sym_mod_operator,
      sym_mult_operator,
      sym_add_operator,
  [9835] = 9,
    ACTIONS(3), 1,
      aux_sym_macro_call_token1,
    ACTIONS(5), 1,
      anon_sym_DASH_DASH,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(701), 1,
      sym_tuple_close,
    STATE(106), 1,
      sym_operator,
    ACTIONS(671), 2,
      sym_comma,
      sym_semi_colon,
    ACTIONS(677), 2,
      sym_div_operator,
      sym_sub_operator,
    STATE(169), 2,
      sym_macro_call,
      sym_comment,
    ACTIONS(675), 5,
      sym_pow_operator,
      sym_floor_div_operator,
      sym_mod_operator,
      sym_mult_operator,
      sym_add_operator,
  [9870] = 9,
    ACTIONS(3), 1,
      aux_sym_macro_call_token1,
    ACTIONS(5), 1,
      anon_sym_DASH_DASH,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(703), 1,
      sym_tuple_close,
    STATE(106), 1,
      sym_operator,
    ACTIONS(671), 2,
      sym_comma,
      sym_semi_colon,
    ACTIONS(677), 2,
      sym_div_operator,
      sym_sub_operator,
    STATE(170), 2,
      sym_macro_call,
      sym_comment,
    ACTIONS(675), 5,
      sym_pow_operator,
      sym_floor_div_operator,
      sym_mod_operator,
      sym_mult_operator,
      sym_add_operator,
  [9905] = 9,
    ACTIONS(3), 1,
      aux_sym_macro_call_token1,
    ACTIONS(5), 1,
      anon_sym_DASH_DASH,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(705), 1,
      sym_tuple_close,
    STATE(106), 1,
      sym_operator,
    ACTIONS(671), 2,
      sym_comma,
      sym_semi_colon,
    ACTIONS(677), 2,
      sym_div_operator,
      sym_sub_operator,
    STATE(171), 2,
      sym_macro_call,
      sym_comment,
    ACTIONS(675), 5,
      sym_pow_operator,
      sym_floor_div_operator,
      sym_mod_operator,
      sym_mult_operator,
      sym_add_operator,
  [9940] = 9,
    ACTIONS(3), 1,
      aux_sym_macro_call_token1,
    ACTIONS(5), 1,
      anon_sym_DASH_DASH,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(707), 1,
      sym_tuple_close,
    STATE(106), 1,
      sym_operator,
    ACTIONS(671), 2,
      sym_comma,
      sym_semi_colon,
    ACTIONS(677), 2,
      sym_div_operator,
      sym_sub_operator,
    STATE(172), 2,
      sym_macro_call,
      sym_comment,
    ACTIONS(675), 5,
      sym_pow_operator,
      sym_floor_div_operator,
      sym_mod_operator,
      sym_mult_operator,
      sym_add_operator,
  [9975] = 9,
    ACTIONS(3), 1,
      aux_sym_macro_call_token1,
    ACTIONS(5), 1,
      anon_sym_DASH_DASH,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(709), 1,
      sym_tuple_close,
    STATE(106), 1,
      sym_operator,
    ACTIONS(671), 2,
      sym_comma,
      sym_semi_colon,
    ACTIONS(677), 2,
      sym_div_operator,
      sym_sub_operator,
    STATE(173), 2,
      sym_macro_call,
      sym_comment,
    ACTIONS(675), 5,
      sym_pow_operator,
      sym_floor_div_operator,
      sym_mod_operator,
      sym_mult_operator,
      sym_add_operator,
  [10010] = 9,
    ACTIONS(3), 1,
      aux_sym_macro_call_token1,
    ACTIONS(5), 1,
      anon_sym_DASH_DASH,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(711), 1,
      sym_tuple_close,
    STATE(106), 1,
      sym_operator,
    ACTIONS(671), 2,
      sym_comma,
      sym_semi_colon,
    ACTIONS(677), 2,
      sym_div_operator,
      sym_sub_operator,
    STATE(174), 2,
      sym_macro_call,
      sym_comment,
    ACTIONS(675), 5,
      sym_pow_operator,
      sym_floor_div_operator,
      sym_mod_operator,
      sym_mult_operator,
      sym_add_operator,
  [10045] = 9,
    ACTIONS(3), 1,
      aux_sym_macro_call_token1,
    ACTIONS(5), 1,
      anon_sym_DASH_DASH,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(713), 1,
      sym_tuple_close,
    STATE(106), 1,
      sym_operator,
    ACTIONS(671), 2,
      sym_comma,
      sym_semi_colon,
    ACTIONS(677), 2,
      sym_div_operator,
      sym_sub_operator,
    STATE(175), 2,
      sym_macro_call,
      sym_comment,
    ACTIONS(675), 5,
      sym_pow_operator,
      sym_floor_div_operator,
      sym_mod_operator,
      sym_mult_operator,
      sym_add_operator,
  [10080] = 9,
    ACTIONS(3), 1,
      aux_sym_macro_call_token1,
    ACTIONS(5), 1,
      anon_sym_DASH_DASH,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(715), 1,
      sym_tuple_close,
    STATE(106), 1,
      sym_operator,
    ACTIONS(671), 2,
      sym_comma,
      sym_semi_colon,
    ACTIONS(677), 2,
      sym_div_operator,
      sym_sub_operator,
    STATE(176), 2,
      sym_macro_call,
      sym_comment,
    ACTIONS(675), 5,
      sym_pow_operator,
      sym_floor_div_operator,
      sym_mod_operator,
      sym_mult_operator,
      sym_add_operator,
  [10115] = 9,
    ACTIONS(3), 1,
      aux_sym_macro_call_token1,
    ACTIONS(5), 1,
      anon_sym_DASH_DASH,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(717), 1,
      sym_tuple_close,
    STATE(106), 1,
      sym_operator,
    ACTIONS(671), 2,
      sym_comma,
      sym_semi_colon,
    ACTIONS(677), 2,
      sym_div_operator,
      sym_sub_operator,
    STATE(177), 2,
      sym_macro_call,
      sym_comment,
    ACTIONS(675), 5,
      sym_pow_operator,
      sym_floor_div_operator,
      sym_mod_operator,
      sym_mult_operator,
      sym_add_operator,
  [10150] = 9,
    ACTIONS(3), 1,
      aux_sym_macro_call_token1,
    ACTIONS(5), 1,
      anon_sym_DASH_DASH,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(719), 1,
      sym_tuple_close,
    STATE(106), 1,
      sym_operator,
    ACTIONS(671), 2,
      sym_comma,
      sym_semi_colon,
    ACTIONS(677), 2,
      sym_div_operator,
      sym_sub_operator,
    STATE(178), 2,
      sym_macro_call,
      sym_comment,
    ACTIONS(675), 5,
      sym_pow_operator,
      sym_floor_div_operator,
      sym_mod_operator,
      sym_mult_operator,
      sym_add_operator,
  [10185] = 9,
    ACTIONS(3), 1,
      aux_sym_macro_call_token1,
    ACTIONS(5), 1,
      anon_sym_DASH_DASH,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(721), 1,
      sym_tuple_close,
    STATE(106), 1,
      sym_operator,
    ACTIONS(671), 2,
      sym_comma,
      sym_semi_colon,
    ACTIONS(677), 2,
      sym_div_operator,
      sym_sub_operator,
    STATE(179), 2,
      sym_macro_call,
      sym_comment,
    ACTIONS(675), 5,
      sym_pow_operator,
      sym_floor_div_operator,
      sym_mod_operator,
      sym_mult_operator,
      sym_add_operator,
  [10220] = 6,
    ACTIONS(3), 1,
      aux_sym_macro_call_token1,
    ACTIONS(5), 1,
      anon_sym_DASH_DASH,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(269), 2,
      sym_div_operator,
      sym_sub_operator,
    STATE(180), 2,
      sym_macro_call,
      sym_comment,
    ACTIONS(271), 8,
      sym_comma,
      sym_semi_colon,
      sym_tuple_close,
      sym_pow_operator,
      sym_floor_div_operator,
      sym_mod_operator,
      sym_mult_operator,
      sym_add_operator,
  [10248] = 6,
    ACTIONS(3), 1,
      aux_sym_macro_call_token1,
    ACTIONS(5), 1,
      anon_sym_DASH_DASH,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(313), 2,
      sym_div_operator,
      sym_sub_operator,
    STATE(181), 2,
      sym_macro_call,
      sym_comment,
    ACTIONS(315), 8,
      sym_comma,
      sym_semi_colon,
      sym_tuple_close,
      sym_pow_operator,
      sym_floor_div_operator,
      sym_mod_operator,
      sym_mult_operator,
      sym_add_operator,
  [10276] = 10,
    ACTIONS(3), 1,
      aux_sym_macro_call_token1,
    ACTIONS(5), 1,
      anon_sym_DASH_DASH,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11), 1,
      sym_identifier,
    ACTIONS(27), 1,
      aux_sym_state_selector_token1,
    STATE(188), 1,
      aux_sym_rule_scope_repeat1,
    STATE(220), 1,
      sym_selector,
    STATE(182), 2,
      sym_macro_call,
      sym_comment,
    STATE(223), 2,
      sym_class_selector,
      sym_state_selector,
    ACTIONS(21), 4,
      sym_static_argument,
      sym_name_selector,
      sym_tag_selector,
      sym_pseudo_selector,
  [10312] = 8,
    ACTIONS(3), 1,
      aux_sym_macro_call_token1,
    ACTIONS(5), 1,
      anon_sym_DASH_DASH,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(106), 1,
      sym_operator,
    ACTIONS(671), 2,
      sym_comma,
      sym_semi_colon,
    ACTIONS(677), 2,
      sym_div_operator,
      sym_sub_operator,
    STATE(183), 2,
      sym_macro_call,
      sym_comment,
    ACTIONS(675), 5,
      sym_pow_operator,
      sym_floor_div_operator,
      sym_mod_operator,
      sym_mult_operator,
      sym_add_operator,
  [10344] = 6,
    ACTIONS(3), 1,
      aux_sym_macro_call_token1,
    ACTIONS(5), 1,
      anon_sym_DASH_DASH,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(385), 2,
      sym_div_operator,
      sym_sub_operator,
    STATE(184), 2,
      sym_macro_call,
      sym_comment,
    ACTIONS(387), 8,
      sym_comma,
      sym_semi_colon,
      sym_tuple_close,
      sym_pow_operator,
      sym_floor_div_operator,
      sym_mod_operator,
      sym_mult_operator,
      sym_add_operator,
  [10372] = 6,
    ACTIONS(3), 1,
      aux_sym_macro_call_token1,
    ACTIONS(5), 1,
      anon_sym_DASH_DASH,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(451), 2,
      sym_div_operator,
      sym_sub_operator,
    STATE(185), 2,
      sym_macro_call,
      sym_comment,
    ACTIONS(453), 8,
      sym_comma,
      sym_semi_colon,
      sym_tuple_close,
      sym_pow_operator,
      sym_floor_div_operator,
      sym_mod_operator,
      sym_mult_operator,
      sym_add_operator,
  [10400] = 6,
    ACTIONS(3), 1,
      aux_sym_macro_call_token1,
    ACTIONS(5), 1,
      anon_sym_DASH_DASH,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(329), 2,
      sym_div_operator,
      sym_sub_operator,
    STATE(186), 2,
      sym_macro_call,
      sym_comment,
    ACTIONS(331), 8,
      sym_comma,
      sym_semi_colon,
      sym_tuple_close,
      sym_pow_operator,
      sym_floor_div_operator,
      sym_mod_operator,
      sym_mult_operator,
      sym_add_operator,
  [10428] = 6,
    ACTIONS(3), 1,
      aux_sym_macro_call_token1,
    ACTIONS(5), 1,
      anon_sym_DASH_DASH,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(425), 2,
      sym_div_operator,
      sym_sub_operator,
    STATE(187), 2,
      sym_macro_call,
      sym_comment,
    ACTIONS(427), 8,
      sym_comma,
      sym_semi_colon,
      sym_tuple_close,
      sym_pow_operator,
      sym_floor_div_operator,
      sym_mod_operator,
      sym_mult_operator,
      sym_add_operator,
  [10456] = 9,
    ACTIONS(3), 1,
      aux_sym_macro_call_token1,
    ACTIONS(5), 1,
      anon_sym_DASH_DASH,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(723), 1,
      sym_identifier,
    ACTIONS(729), 1,
      aux_sym_state_selector_token1,
    STATE(231), 1,
      sym_selector,
    STATE(223), 2,
      sym_class_selector,
      sym_state_selector,
    STATE(188), 3,
      sym_macro_call,
      sym_comment,
      aux_sym_rule_scope_repeat1,
    ACTIONS(726), 4,
      sym_static_argument,
      sym_name_selector,
      sym_tag_selector,
      sym_pseudo_selector,
  [10490] = 6,
    ACTIONS(3), 1,
      aux_sym_macro_call_token1,
    ACTIONS(5), 1,
      anon_sym_DASH_DASH,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(429), 2,
      sym_div_operator,
      sym_sub_operator,
    STATE(189), 2,
      sym_macro_call,
      sym_comment,
    ACTIONS(431), 8,
      sym_comma,
      sym_semi_colon,
      sym_tuple_close,
      sym_pow_operator,
      sym_floor_div_operator,
      sym_mod_operator,
      sym_mult_operator,
      sym_add_operator,
  [10518] = 6,
    ACTIONS(3), 1,
      aux_sym_macro_call_token1,
    ACTIONS(5), 1,
      anon_sym_DASH_DASH,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(199), 2,
      sym_div_operator,
      sym_sub_operator,
    STATE(190), 2,
      sym_macro_call,
      sym_comment,
    ACTIONS(201), 8,
      sym_comma,
      sym_semi_colon,
      sym_tuple_close,
      sym_pow_operator,
      sym_floor_div_operator,
      sym_mod_operator,
      sym_mult_operator,
      sym_add_operator,
  [10546] = 6,
    ACTIONS(3), 1,
      aux_sym_macro_call_token1,
    ACTIONS(5), 1,
      anon_sym_DASH_DASH,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(205), 2,
      sym_div_operator,
      sym_sub_operator,
    STATE(191), 2,
      sym_macro_call,
      sym_comment,
    ACTIONS(207), 8,
      sym_comma,
      sym_semi_colon,
      sym_tuple_close,
      sym_pow_operator,
      sym_floor_div_operator,
      sym_mod_operator,
      sym_mult_operator,
      sym_add_operator,
  [10574] = 10,
    ACTIONS(3), 1,
      aux_sym_macro_call_token1,
    ACTIONS(5), 1,
      anon_sym_DASH_DASH,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11), 1,
      sym_identifier,
    ACTIONS(27), 1,
      aux_sym_state_selector_token1,
    STATE(182), 1,
      aux_sym_rule_scope_repeat1,
    STATE(219), 1,
      sym_selector,
    STATE(192), 2,
      sym_macro_call,
      sym_comment,
    STATE(223), 2,
      sym_class_selector,
      sym_state_selector,
    ACTIONS(21), 4,
      sym_static_argument,
      sym_name_selector,
      sym_tag_selector,
      sym_pseudo_selector,
  [10610] = 10,
    ACTIONS(3), 1,
      aux_sym_macro_call_token1,
    ACTIONS(5), 1,
      anon_sym_DASH_DASH,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11), 1,
      sym_identifier,
    ACTIONS(27), 1,
      aux_sym_state_selector_token1,
    STATE(188), 1,
      aux_sym_rule_scope_repeat1,
    STATE(227), 1,
      sym_selector,
    STATE(193), 2,
      sym_macro_call,
      sym_comment,
    STATE(223), 2,
      sym_class_selector,
      sym_state_selector,
    ACTIONS(21), 4,
      sym_static_argument,
      sym_name_selector,
      sym_tag_selector,
      sym_pseudo_selector,
  [10646] = 6,
    ACTIONS(3), 1,
      aux_sym_macro_call_token1,
    ACTIONS(5), 1,
      anon_sym_DASH_DASH,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(433), 2,
      sym_div_operator,
      sym_sub_operator,
    STATE(194), 2,
      sym_macro_call,
      sym_comment,
    ACTIONS(435), 8,
      sym_comma,
      sym_semi_colon,
      sym_tuple_close,
      sym_pow_operator,
      sym_floor_div_operator,
      sym_mod_operator,
      sym_mult_operator,
      sym_add_operator,
  [10674] = 6,
    ACTIONS(3), 1,
      aux_sym_macro_call_token1,
    ACTIONS(5), 1,
      anon_sym_DASH_DASH,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(273), 2,
      sym_div_operator,
      sym_sub_operator,
    STATE(195), 2,
      sym_macro_call,
      sym_comment,
    ACTIONS(275), 8,
      sym_comma,
      sym_semi_colon,
      sym_tuple_close,
      sym_pow_operator,
      sym_floor_div_operator,
      sym_mod_operator,
      sym_mult_operator,
      sym_add_operator,
  [10702] = 6,
    ACTIONS(3), 1,
      aux_sym_macro_call_token1,
    ACTIONS(5), 1,
      anon_sym_DASH_DASH,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(389), 2,
      sym_div_operator,
      sym_sub_operator,
    STATE(196), 2,
      sym_macro_call,
      sym_comment,
    ACTIONS(391), 8,
      sym_comma,
      sym_semi_colon,
      sym_tuple_close,
      sym_pow_operator,
      sym_floor_div_operator,
      sym_mod_operator,
      sym_mult_operator,
      sym_add_operator,
  [10730] = 6,
    ACTIONS(3), 1,
      aux_sym_macro_call_token1,
    ACTIONS(5), 1,
      anon_sym_DASH_DASH,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(281), 2,
      sym_div_operator,
      sym_sub_operator,
    STATE(197), 2,
      sym_macro_call,
      sym_comment,
    ACTIONS(283), 8,
      sym_comma,
      sym_semi_colon,
      sym_tuple_close,
      sym_pow_operator,
      sym_floor_div_operator,
      sym_mod_operator,
      sym_mult_operator,
      sym_add_operator,
  [10758] = 6,
    ACTIONS(3), 1,
      aux_sym_macro_call_token1,
    ACTIONS(5), 1,
      anon_sym_DASH_DASH,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(353), 2,
      sym_div_operator,
      sym_sub_operator,
    STATE(198), 2,
      sym_macro_call,
      sym_comment,
    ACTIONS(355), 8,
      sym_comma,
      sym_semi_colon,
      sym_tuple_close,
      sym_pow_operator,
      sym_floor_div_operator,
      sym_mod_operator,
      sym_mult_operator,
      sym_add_operator,
  [10786] = 6,
    ACTIONS(3), 1,
      aux_sym_macro_call_token1,
    ACTIONS(5), 1,
      anon_sym_DASH_DASH,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(393), 2,
      sym_div_operator,
      sym_sub_operator,
    STATE(199), 2,
      sym_macro_call,
      sym_comment,
    ACTIONS(395), 8,
      sym_comma,
      sym_semi_colon,
      sym_tuple_close,
      sym_pow_operator,
      sym_floor_div_operator,
      sym_mod_operator,
      sym_mult_operator,
      sym_add_operator,
  [10814] = 6,
    ACTIONS(3), 1,
      aux_sym_macro_call_token1,
    ACTIONS(5), 1,
      anon_sym_DASH_DASH,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(397), 2,
      sym_div_operator,
      sym_sub_operator,
    STATE(200), 2,
      sym_macro_call,
      sym_comment,
    ACTIONS(399), 8,
      sym_comma,
      sym_semi_colon,
      sym_tuple_close,
      sym_pow_operator,
      sym_floor_div_operator,
      sym_mod_operator,
      sym_mult_operator,
      sym_add_operator,
  [10842] = 10,
    ACTIONS(3), 1,
      aux_sym_macro_call_token1,
    ACTIONS(5), 1,
      anon_sym_DASH_DASH,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11), 1,
      sym_identifier,
    ACTIONS(27), 1,
      aux_sym_state_selector_token1,
    STATE(188), 1,
      aux_sym_rule_scope_repeat1,
    STATE(219), 1,
      sym_selector,
    STATE(201), 2,
      sym_macro_call,
      sym_comment,
    STATE(223), 2,
      sym_class_selector,
      sym_state_selector,
    ACTIONS(21), 4,
      sym_static_argument,
      sym_name_selector,
      sym_tag_selector,
      sym_pseudo_selector,
  [10878] = 6,
    ACTIONS(3), 1,
      aux_sym_macro_call_token1,
    ACTIONS(5), 1,
      anon_sym_DASH_DASH,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(405), 2,
      sym_div_operator,
      sym_sub_operator,
    STATE(202), 2,
      sym_macro_call,
      sym_comment,
    ACTIONS(407), 8,
      sym_comma,
      sym_semi_colon,
      sym_tuple_close,
      sym_pow_operator,
      sym_floor_div_operator,
      sym_mod_operator,
      sym_mult_operator,
      sym_add_operator,
  [10906] = 6,
    ACTIONS(3), 1,
      aux_sym_macro_call_token1,
    ACTIONS(5), 1,
      anon_sym_DASH_DASH,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(409), 2,
      sym_div_operator,
      sym_sub_operator,
    STATE(203), 2,
      sym_macro_call,
      sym_comment,
    ACTIONS(411), 8,
      sym_comma,
      sym_semi_colon,
      sym_tuple_close,
      sym_pow_operator,
      sym_floor_div_operator,
      sym_mod_operator,
      sym_mult_operator,
      sym_add_operator,
  [10934] = 6,
    ACTIONS(3), 1,
      aux_sym_macro_call_token1,
    ACTIONS(5), 1,
      anon_sym_DASH_DASH,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(413), 2,
      sym_div_operator,
      sym_sub_operator,
    STATE(204), 2,
      sym_macro_call,
      sym_comment,
    ACTIONS(415), 8,
      sym_comma,
      sym_semi_colon,
      sym_tuple_close,
      sym_pow_operator,
      sym_floor_div_operator,
      sym_mod_operator,
      sym_mult_operator,
      sym_add_operator,
  [10962] = 6,
    ACTIONS(3), 1,
      aux_sym_macro_call_token1,
    ACTIONS(5), 1,
      anon_sym_DASH_DASH,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(417), 2,
      sym_div_operator,
      sym_sub_operator,
    STATE(205), 2,
      sym_macro_call,
      sym_comment,
    ACTIONS(419), 8,
      sym_comma,
      sym_semi_colon,
      sym_tuple_close,
      sym_pow_operator,
      sym_floor_div_operator,
      sym_mod_operator,
      sym_mult_operator,
      sym_add_operator,
  [10990] = 12,
    ACTIONS(3), 1,
      aux_sym_macro_call_token1,
    ACTIONS(5), 1,
      anon_sym_DASH_DASH,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(247), 1,
      sym__block_string_start,
    ACTIONS(734), 1,
      sym_tuple_open,
    STATE(207), 1,
      sym_block_string,
    STATE(222), 1,
      aux_sym_derive_value_repeat1,
    STATE(250), 1,
      sym_derive_value,
    STATE(255), 1,
      sym_string,
    ACTIONS(231), 2,
      sym_single_quote_string,
      sym_double_quote_string,
    ACTIONS(732), 2,
      sym_comma,
      sym_semi_colon,
    STATE(206), 2,
      sym_macro_call,
      sym_comment,
  [11030] = 6,
    ACTIONS(3), 1,
      aux_sym_macro_call_token1,
    ACTIONS(5), 1,
      anon_sym_DASH_DASH,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(261), 2,
      sym_div_operator,
      sym_sub_operator,
    STATE(207), 2,
      sym_macro_call,
      sym_comment,
    ACTIONS(263), 8,
      sym_comma,
      sym_semi_colon,
      sym_tuple_close,
      sym_pow_operator,
      sym_floor_div_operator,
      sym_mod_operator,
      sym_mult_operator,
      sym_add_operator,
  [11058] = 6,
    ACTIONS(3), 1,
      aux_sym_macro_call_token1,
    ACTIONS(5), 1,
      anon_sym_DASH_DASH,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(369), 2,
      sym_div_operator,
      sym_sub_operator,
    STATE(208), 2,
      sym_macro_call,
      sym_comment,
    ACTIONS(371), 8,
      sym_comma,
      sym_semi_colon,
      sym_tuple_close,
      sym_pow_operator,
      sym_floor_div_operator,
      sym_mod_operator,
      sym_mult_operator,
      sym_add_operator,
  [11086] = 6,
    ACTIONS(3), 1,
      aux_sym_macro_call_token1,
    ACTIONS(5), 1,
      anon_sym_DASH_DASH,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(421), 2,
      sym_div_operator,
      sym_sub_operator,
    STATE(209), 2,
      sym_macro_call,
      sym_comment,
    ACTIONS(423), 8,
      sym_comma,
      sym_semi_colon,
      sym_tuple_close,
      sym_pow_operator,
      sym_floor_div_operator,
      sym_mod_operator,
      sym_mult_operator,
      sym_add_operator,
  [11114] = 6,
    ACTIONS(3), 1,
      aux_sym_macro_call_token1,
    ACTIONS(5), 1,
      anon_sym_DASH_DASH,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(373), 2,
      sym_div_operator,
      sym_sub_operator,
    STATE(210), 2,
      sym_macro_call,
      sym_comment,
    ACTIONS(375), 8,
      sym_comma,
      sym_semi_colon,
      sym_tuple_close,
      sym_pow_operator,
      sym_floor_div_operator,
      sym_mod_operator,
      sym_mult_operator,
      sym_add_operator,
  [11142] = 10,
    ACTIONS(3), 1,
      aux_sym_macro_call_token1,
    ACTIONS(5), 1,
      anon_sym_DASH_DASH,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11), 1,
      sym_identifier,
    ACTIONS(27), 1,
      aux_sym_state_selector_token1,
    STATE(188), 1,
      aux_sym_rule_scope_repeat1,
    STATE(228), 1,
      sym_selector,
    STATE(211), 2,
      sym_macro_call,
      sym_comment,
    STATE(223), 2,
      sym_class_selector,
      sym_state_selector,
    ACTIONS(21), 4,
      sym_static_argument,
      sym_name_selector,
      sym_tag_selector,
      sym_pseudo_selector,
  [11178] = 6,
    ACTIONS(3), 1,
      aux_sym_macro_call_token1,
    ACTIONS(5), 1,
      anon_sym_DASH_DASH,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(277), 2,
      sym_div_operator,
      sym_sub_operator,
    STATE(212), 2,
      sym_macro_call,
      sym_comment,
    ACTIONS(279), 8,
      sym_comma,
      sym_semi_colon,
      sym_tuple_close,
      sym_pow_operator,
      sym_floor_div_operator,
      sym_mod_operator,
      sym_mult_operator,
      sym_add_operator,
  [11206] = 8,
    ACTIONS(3), 1,
      aux_sym_macro_call_token1,
    ACTIONS(5), 1,
      anon_sym_DASH_DASH,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(736), 1,
      sym_semi_colon,
    STATE(106), 1,
      sym_operator,
    ACTIONS(677), 2,
      sym_div_operator,
      sym_sub_operator,
    STATE(213), 2,
      sym_macro_call,
      sym_comment,
    ACTIONS(675), 5,
      sym_pow_operator,
      sym_floor_div_operator,
      sym_mod_operator,
      sym_mult_operator,
      sym_add_operator,
  [11237] = 8,
    ACTIONS(3), 1,
      aux_sym_macro_call_token1,
    ACTIONS(5), 1,
      anon_sym_DASH_DASH,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(738), 1,
      sym_semi_colon,
    STATE(106), 1,
      sym_operator,
    ACTIONS(677), 2,
      sym_div_operator,
      sym_sub_operator,
    STATE(214), 2,
      sym_macro_call,
      sym_comment,
    ACTIONS(675), 5,
      sym_pow_operator,
      sym_floor_div_operator,
      sym_mod_operator,
      sym_mult_operator,
      sym_add_operator,
  [11268] = 8,
    ACTIONS(3), 1,
      aux_sym_macro_call_token1,
    ACTIONS(5), 1,
      anon_sym_DASH_DASH,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(740), 1,
      sym_semi_colon,
    STATE(106), 1,
      sym_operator,
    ACTIONS(677), 2,
      sym_div_operator,
      sym_sub_operator,
    STATE(215), 2,
      sym_macro_call,
      sym_comment,
    ACTIONS(675), 5,
      sym_pow_operator,
      sym_floor_div_operator,
      sym_mod_operator,
      sym_mult_operator,
      sym_add_operator,
  [11299] = 7,
    ACTIONS(3), 1,
      aux_sym_macro_call_token1,
    ACTIONS(5), 1,
      anon_sym_DASH_DASH,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(742), 1,
      sym_identifier,
    ACTIONS(746), 1,
      sym_equals,
    STATE(216), 2,
      sym_macro_call,
      sym_comment,
    ACTIONS(744), 7,
      sym_comma,
      sym_static_argument,
      sym_name_selector,
      sym_tag_selector,
      aux_sym_state_selector_token1,
      sym_pseudo_selector,
      sym_scope_open,
  [11328] = 8,
    ACTIONS(3), 1,
      aux_sym_macro_call_token1,
    ACTIONS(5), 1,
      anon_sym_DASH_DASH,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(748), 1,
      sym_semi_colon,
    STATE(106), 1,
      sym_operator,
    ACTIONS(677), 2,
      sym_div_operator,
      sym_sub_operator,
    STATE(217), 2,
      sym_macro_call,
      sym_comment,
    ACTIONS(675), 5,
      sym_pow_operator,
      sym_floor_div_operator,
      sym_mod_operator,
      sym_mult_operator,
      sym_add_operator,
  [11359] = 8,
    ACTIONS(3), 1,
      aux_sym_macro_call_token1,
    ACTIONS(5), 1,
      anon_sym_DASH_DASH,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(750), 1,
      sym_semi_colon,
    STATE(106), 1,
      sym_operator,
    ACTIONS(677), 2,
      sym_div_operator,
      sym_sub_operator,
    STATE(218), 2,
      sym_macro_call,
      sym_comment,
    ACTIONS(675), 5,
      sym_pow_operator,
      sym_floor_div_operator,
      sym_mod_operator,
      sym_mult_operator,
      sym_add_operator,
  [11390] = 8,
    ACTIONS(3), 1,
      aux_sym_macro_call_token1,
    ACTIONS(5), 1,
      anon_sym_DASH_DASH,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(752), 1,
      sym_identifier,
    ACTIONS(754), 1,
      sym_comma,
    ACTIONS(758), 1,
      sym_scope_open,
    STATE(219), 2,
      sym_macro_call,
      sym_comment,
    ACTIONS(756), 5,
      sym_static_argument,
      sym_name_selector,
      sym_tag_selector,
      aux_sym_state_selector_token1,
      sym_pseudo_selector,
  [11420] = 8,
    ACTIONS(3), 1,
      aux_sym_macro_call_token1,
    ACTIONS(5), 1,
      anon_sym_DASH_DASH,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(752), 1,
      sym_identifier,
    ACTIONS(754), 1,
      sym_comma,
    ACTIONS(760), 1,
      sym_scope_open,
    STATE(220), 2,
      sym_macro_call,
      sym_comment,
    ACTIONS(756), 5,
      sym_static_argument,
      sym_name_selector,
      sym_tag_selector,
      aux_sym_state_selector_token1,
      sym_pseudo_selector,
  [11450] = 8,
    ACTIONS(3), 1,
      aux_sym_macro_call_token1,
    ACTIONS(5), 1,
      anon_sym_DASH_DASH,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(752), 1,
      sym_identifier,
    ACTIONS(754), 1,
      sym_comma,
    ACTIONS(762), 1,
      sym_scope_open,
    STATE(221), 2,
      sym_macro_call,
      sym_comment,
    ACTIONS(756), 5,
      sym_static_argument,
      sym_name_selector,
      sym_tag_selector,
      aux_sym_state_selector_token1,
      sym_pseudo_selector,
  [11480] = 11,
    ACTIONS(3), 1,
      aux_sym_macro_call_token1,
    ACTIONS(5), 1,
      anon_sym_DASH_DASH,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(247), 1,
      sym__block_string_start,
    ACTIONS(734), 1,
      sym_tuple_open,
    STATE(207), 1,
      sym_block_string,
    STATE(230), 1,
      aux_sym_derive_value_repeat1,
    STATE(247), 1,
      sym_derive_value,
    STATE(255), 1,
      sym_string,
    ACTIONS(231), 2,
      sym_single_quote_string,
      sym_double_quote_string,
    STATE(222), 2,
      sym_macro_call,
      sym_comment,
  [11516] = 6,
    ACTIONS(3), 1,
      aux_sym_macro_call_token1,
    ACTIONS(5), 1,
      anon_sym_DASH_DASH,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(253), 1,
      sym_identifier,
    STATE(223), 2,
      sym_macro_call,
      sym_comment,
    ACTIONS(255), 7,
      sym_comma,
      sym_static_argument,
      sym_name_selector,
      sym_tag_selector,
      aux_sym_state_selector_token1,
      sym_pseudo_selector,
      sym_scope_open,
  [11542] = 11,
    ACTIONS(3), 1,
      aux_sym_macro_call_token1,
    ACTIONS(5), 1,
      anon_sym_DASH_DASH,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(247), 1,
      sym__block_string_start,
    ACTIONS(734), 1,
      sym_tuple_open,
    STATE(207), 1,
      sym_block_string,
    STATE(225), 1,
      aux_sym_derive_value_repeat1,
    STATE(247), 1,
      sym_derive_value,
    STATE(255), 1,
      sym_string,
    ACTIONS(231), 2,
      sym_single_quote_string,
      sym_double_quote_string,
    STATE(224), 2,
      sym_macro_call,
      sym_comment,
  [11578] = 11,
    ACTIONS(3), 1,
      aux_sym_macro_call_token1,
    ACTIONS(5), 1,
      anon_sym_DASH_DASH,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(247), 1,
      sym__block_string_start,
    ACTIONS(734), 1,
      sym_tuple_open,
    STATE(207), 1,
      sym_block_string,
    STATE(230), 1,
      aux_sym_derive_value_repeat1,
    STATE(248), 1,
      sym_derive_value,
    STATE(255), 1,
      sym_string,
    ACTIONS(231), 2,
      sym_single_quote_string,
      sym_double_quote_string,
    STATE(225), 2,
      sym_macro_call,
      sym_comment,
  [11614] = 6,
    ACTIONS(3), 1,
      aux_sym_macro_call_token1,
    ACTIONS(5), 1,
      anon_sym_DASH_DASH,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(742), 1,
      sym_identifier,
    STATE(226), 2,
      sym_macro_call,
      sym_comment,
    ACTIONS(744), 7,
      sym_comma,
      sym_static_argument,
      sym_name_selector,
      sym_tag_selector,
      aux_sym_state_selector_token1,
      sym_pseudo_selector,
      sym_scope_open,
  [11640] = 8,
    ACTIONS(3), 1,
      aux_sym_macro_call_token1,
    ACTIONS(5), 1,
      anon_sym_DASH_DASH,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(752), 1,
      sym_identifier,
    ACTIONS(754), 1,
      sym_comma,
    ACTIONS(764), 1,
      sym_scope_open,
    STATE(227), 2,
      sym_macro_call,
      sym_comment,
    ACTIONS(756), 5,
      sym_static_argument,
      sym_name_selector,
      sym_tag_selector,
      aux_sym_state_selector_token1,
      sym_pseudo_selector,
  [11670] = 8,
    ACTIONS(3), 1,
      aux_sym_macro_call_token1,
    ACTIONS(5), 1,
      anon_sym_DASH_DASH,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(752), 1,
      sym_identifier,
    ACTIONS(754), 1,
      sym_comma,
    ACTIONS(766), 1,
      sym_scope_open,
    STATE(228), 2,
      sym_macro_call,
      sym_comment,
    ACTIONS(756), 5,
      sym_static_argument,
      sym_name_selector,
      sym_tag_selector,
      aux_sym_state_selector_token1,
      sym_pseudo_selector,
  [11700] = 6,
    ACTIONS(3), 1,
      aux_sym_macro_call_token1,
    ACTIONS(5), 1,
      anon_sym_DASH_DASH,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(768), 1,
      sym_identifier,
    STATE(229), 2,
      sym_macro_call,
      sym_comment,
    ACTIONS(770), 7,
      sym_comma,
      sym_static_argument,
      sym_name_selector,
      sym_tag_selector,
      aux_sym_state_selector_token1,
      sym_pseudo_selector,
      sym_scope_open,
  [11726] = 10,
    ACTIONS(3), 1,
      aux_sym_macro_call_token1,
    ACTIONS(5), 1,
      anon_sym_DASH_DASH,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(772), 1,
      sym_tuple_open,
    ACTIONS(778), 1,
      sym__block_string_start,
    STATE(207), 1,
      sym_block_string,
    STATE(255), 1,
      sym_string,
    STATE(264), 1,
      sym_derive_value,
    ACTIONS(775), 2,
      sym_single_quote_string,
      sym_double_quote_string,
    STATE(230), 3,
      sym_macro_call,
      sym_comment,
      aux_sym_derive_value_repeat1,
  [11760] = 7,
    ACTIONS(3), 1,
      aux_sym_macro_call_token1,
    ACTIONS(5), 1,
      anon_sym_DASH_DASH,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(752), 1,
      sym_identifier,
    ACTIONS(754), 1,
      sym_comma,
    STATE(231), 2,
      sym_macro_call,
      sym_comment,
    ACTIONS(756), 5,
      sym_static_argument,
      sym_name_selector,
      sym_tag_selector,
      aux_sym_state_selector_token1,
      sym_pseudo_selector,
  [11787] = 10,
    ACTIONS(3), 1,
      aux_sym_macro_call_token1,
    ACTIONS(5), 1,
      anon_sym_DASH_DASH,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(247), 1,
      sym__block_string_start,
    ACTIONS(734), 1,
      sym_tuple_open,
    STATE(207), 1,
      sym_block_string,
    STATE(255), 1,
      sym_string,
    STATE(293), 1,
      sym_derive_value,
    ACTIONS(231), 2,
      sym_single_quote_string,
      sym_double_quote_string,
    STATE(232), 2,
      sym_macro_call,
      sym_comment,
  [11820] = 10,
    ACTIONS(3), 1,
      aux_sym_macro_call_token1,
    ACTIONS(5), 1,
      anon_sym_DASH_DASH,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(247), 1,
      sym__block_string_start,
    ACTIONS(734), 1,
      sym_tuple_open,
    STATE(207), 1,
      sym_block_string,
    STATE(255), 1,
      sym_string,
    STATE(290), 1,
      sym_derive_value,
    ACTIONS(231), 2,
      sym_single_quote_string,
      sym_double_quote_string,
    STATE(233), 2,
      sym_macro_call,
      sym_comment,
  [11853] = 6,
    ACTIONS(3), 1,
      aux_sym_macro_call_token1,
    ACTIONS(5), 1,
      anon_sym_DASH_DASH,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(781), 1,
      sym_identifier,
    STATE(234), 2,
      sym_macro_call,
      sym_comment,
    ACTIONS(783), 5,
      sym_static_argument,
      sym_name_selector,
      sym_tag_selector,
      aux_sym_state_selector_token1,
      sym_pseudo_selector,
  [11877] = 9,
    ACTIONS(3), 1,
      aux_sym_macro_call_token1,
    ACTIONS(5), 1,
      anon_sym_DASH_DASH,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(247), 1,
      sym__block_string_start,
    ACTIONS(785), 1,
      sym_semi_colon,
    STATE(207), 1,
      sym_block_string,
    STATE(273), 1,
      sym_string,
    ACTIONS(231), 2,
      sym_single_quote_string,
      sym_double_quote_string,
    STATE(235), 2,
      sym_macro_call,
      sym_comment,
  [11907] = 9,
    ACTIONS(3), 1,
      aux_sym_macro_call_token1,
    ACTIONS(5), 1,
      anon_sym_DASH_DASH,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(247), 1,
      sym__block_string_start,
    ACTIONS(787), 1,
      sym_semi_colon,
    STATE(207), 1,
      sym_block_string,
    STATE(296), 1,
      sym_string,
    ACTIONS(231), 2,
      sym_single_quote_string,
      sym_double_quote_string,
    STATE(236), 2,
      sym_macro_call,
      sym_comment,
  [11937] = 8,
    ACTIONS(3), 1,
      aux_sym_macro_call_token1,
    ACTIONS(5), 1,
      anon_sym_DASH_DASH,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(229), 1,
      sym_number_percent,
    ACTIONS(789), 1,
      sym_semi_colon,
    STATE(295), 1,
      sym_number,
    ACTIONS(227), 2,
      sym_number_ambiguous,
      sym_number_offset,
    STATE(237), 2,
      sym_macro_call,
      sym_comment,
  [11964] = 8,
    ACTIONS(3), 1,
      aux_sym_macro_call_token1,
    ACTIONS(5), 1,
      anon_sym_DASH_DASH,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(229), 1,
      sym_number_percent,
    ACTIONS(791), 1,
      sym_semi_colon,
    STATE(281), 1,
      sym_number,
    ACTIONS(227), 2,
      sym_number_ambiguous,
      sym_number_offset,
    STATE(238), 2,
      sym_macro_call,
      sym_comment,
  [11991] = 8,
    ACTIONS(3), 1,
      aux_sym_macro_call_token1,
    ACTIONS(5), 1,
      anon_sym_DASH_DASH,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(795), 1,
      sym_static_argument,
    ACTIONS(797), 1,
      sym_tuple_close,
    STATE(256), 1,
      aux_sym_macro_args_repeat1,
    ACTIONS(793), 2,
      sym_comma,
      sym_semi_colon,
    STATE(239), 2,
      sym_macro_call,
      sym_comment,
  [12018] = 8,
    ACTIONS(3), 1,
      aux_sym_macro_call_token1,
    ACTIONS(5), 1,
      anon_sym_DASH_DASH,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(799), 1,
      sym_scope_open,
    ACTIONS(801), 1,
      sym_tuple_open,
    STATE(150), 1,
      sym_macro_scope,
    STATE(268), 1,
      sym_macro_args,
    STATE(240), 2,
      sym_macro_call,
      sym_comment,
  [12044] = 5,
    ACTIONS(3), 1,
      aux_sym_macro_call_token1,
    ACTIONS(5), 1,
      anon_sym_DASH_DASH,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(241), 2,
      sym_macro_call,
      sym_comment,
    ACTIONS(803), 4,
      sym__block_string_start,
      sym_tuple_open,
      sym_single_quote_string,
      sym_double_quote_string,
  [12064] = 5,
    ACTIONS(3), 1,
      aux_sym_macro_call_token1,
    ACTIONS(5), 1,
      anon_sym_DASH_DASH,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(242), 2,
      sym_macro_call,
      sym_comment,
    ACTIONS(805), 3,
      sym_comma,
      sym_semi_colon,
      sym_tuple_close,
  [12083] = 7,
    ACTIONS(3), 1,
      aux_sym_macro_call_token1,
    ACTIONS(5), 1,
      anon_sym_DASH_DASH,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(807), 1,
      anon_sym_COLON,
    ACTIONS(809), 1,
      anon_sym_DOT,
    STATE(302), 1,
      sym_colon,
    STATE(243), 2,
      sym_macro_call,
      sym_comment,
  [12106] = 5,
    ACTIONS(3), 1,
      aux_sym_macro_call_token1,
    ACTIONS(5), 1,
      anon_sym_DASH_DASH,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(244), 2,
      sym_macro_call,
      sym_comment,
    ACTIONS(811), 3,
      sym_comma,
      sym_semi_colon,
      sym_tuple_close,
  [12125] = 6,
    ACTIONS(3), 1,
      aux_sym_macro_call_token1,
    ACTIONS(5), 1,
      anon_sym_DASH_DASH,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(815), 1,
      sym_tuple_close,
    ACTIONS(813), 2,
      sym_comma,
      sym_semi_colon,
    STATE(245), 2,
      sym_macro_call,
      sym_comment,
  [12146] = 6,
    ACTIONS(3), 1,
      aux_sym_macro_call_token1,
    ACTIONS(5), 1,
      anon_sym_DASH_DASH,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(817), 1,
      sym_tuple_close,
    ACTIONS(813), 2,
      sym_comma,
      sym_semi_colon,
    STATE(246), 2,
      sym_macro_call,
      sym_comment,
  [12167] = 6,
    ACTIONS(3), 1,
      aux_sym_macro_call_token1,
    ACTIONS(5), 1,
      anon_sym_DASH_DASH,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(821), 1,
      sym_tuple_close,
    ACTIONS(819), 2,
      sym_comma,
      sym_semi_colon,
    STATE(247), 2,
      sym_macro_call,
      sym_comment,
  [12188] = 6,
    ACTIONS(3), 1,
      aux_sym_macro_call_token1,
    ACTIONS(5), 1,
      anon_sym_DASH_DASH,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(823), 1,
      sym_tuple_close,
    ACTIONS(819), 2,
      sym_comma,
      sym_semi_colon,
    STATE(248), 2,
      sym_macro_call,
      sym_comment,
  [12209] = 5,
    ACTIONS(3), 1,
      aux_sym_macro_call_token1,
    ACTIONS(5), 1,
      anon_sym_DASH_DASH,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(249), 2,
      sym_macro_call,
      sym_comment,
    ACTIONS(825), 3,
      sym_comma,
      sym_semi_colon,
      sym_tuple_close,
  [12228] = 6,
    ACTIONS(3), 1,
      aux_sym_macro_call_token1,
    ACTIONS(5), 1,
      anon_sym_DASH_DASH,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(827), 1,
      sym_tuple_close,
    ACTIONS(819), 2,
      sym_comma,
      sym_semi_colon,
    STATE(250), 2,
      sym_macro_call,
      sym_comment,
  [12249] = 7,
    ACTIONS(3), 1,
      aux_sym_macro_call_token1,
    ACTIONS(5), 1,
      anon_sym_DASH_DASH,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(807), 1,
      anon_sym_COLON,
    ACTIONS(829), 1,
      anon_sym_DOT,
    STATE(304), 1,
      sym_colon,
    STATE(251), 2,
      sym_macro_call,
      sym_comment,
  [12272] = 6,
    ACTIONS(3), 1,
      aux_sym_macro_call_token1,
    ACTIONS(5), 1,
      anon_sym_DASH_DASH,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(831), 1,
      sym_tuple_close,
    ACTIONS(813), 2,
      sym_comma,
      sym_semi_colon,
    STATE(252), 2,
      sym_macro_call,
      sym_comment,
  [12293] = 7,
    ACTIONS(3), 1,
      aux_sym_macro_call_token1,
    ACTIONS(5), 1,
      anon_sym_DASH_DASH,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(807), 1,
      anon_sym_COLON,
    ACTIONS(833), 1,
      anon_sym_DOT,
    STATE(279), 1,
      sym_colon,
    STATE(253), 2,
      sym_macro_call,
      sym_comment,
  [12316] = 7,
    ACTIONS(3), 1,
      aux_sym_macro_call_token1,
    ACTIONS(5), 1,
      anon_sym_DASH_DASH,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(807), 1,
      anon_sym_COLON,
    ACTIONS(835), 1,
      anon_sym_DOT,
    STATE(292), 1,
      sym_colon,
    STATE(254), 2,
      sym_macro_call,
      sym_comment,
  [12339] = 5,
    ACTIONS(3), 1,
      aux_sym_macro_call_token1,
    ACTIONS(5), 1,
      anon_sym_DASH_DASH,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(255), 2,
      sym_macro_call,
      sym_comment,
    ACTIONS(837), 3,
      sym_comma,
      sym_semi_colon,
      sym_tuple_close,
  [12358] = 6,
    ACTIONS(3), 1,
      aux_sym_macro_call_token1,
    ACTIONS(5), 1,
      anon_sym_DASH_DASH,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(839), 1,
      sym_static_argument,
    STATE(271), 1,
      aux_sym_macro_args_repeat1,
    STATE(256), 2,
      sym_macro_call,
      sym_comment,
  [12378] = 6,
    ACTIONS(3), 1,
      aux_sym_macro_call_token1,
    ACTIONS(5), 1,
      anon_sym_DASH_DASH,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(839), 1,
      sym_static_argument,
    STATE(267), 1,
      aux_sym_macro_args_repeat1,
    STATE(257), 2,
      sym_macro_call,
      sym_comment,
  [12398] = 5,
    ACTIONS(3), 1,
      aux_sym_macro_call_token1,
    ACTIONS(5), 1,
      anon_sym_DASH_DASH,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(813), 2,
      sym_comma,
      sym_semi_colon,
    STATE(258), 2,
      sym_macro_call,
      sym_comment,
  [12416] = 5,
    ACTIONS(3), 1,
      aux_sym_macro_call_token1,
    ACTIONS(5), 1,
      anon_sym_DASH_DASH,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(841), 2,
      sym_static_argument,
      sym__color_code_shade,
    STATE(259), 2,
      sym_macro_call,
      sym_comment,
  [12434] = 6,
    ACTIONS(3), 1,
      aux_sym_macro_call_token1,
    ACTIONS(5), 1,
      anon_sym_DASH_DASH,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(843), 1,
      sym_static_argument,
    ACTIONS(845), 1,
      aux_sym_color_code_tailwind_token2,
    STATE(260), 2,
      sym_macro_call,
      sym_comment,
  [12454] = 6,
    ACTIONS(3), 1,
      aux_sym_macro_call_token1,
    ACTIONS(5), 1,
      anon_sym_DASH_DASH,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(847), 1,
      sym_static_argument,
    ACTIONS(849), 1,
      aux_sym_color_code_skin_token2,
    STATE(261), 2,
      sym_macro_call,
      sym_comment,
  [12474] = 6,
    ACTIONS(3), 1,
      aux_sym_macro_call_token1,
    ACTIONS(5), 1,
      anon_sym_DASH_DASH,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(851), 1,
      sym_static_argument,
    ACTIONS(853), 1,
      aux_sym_color_code_brick_token2,
    STATE(262), 2,
      sym_macro_call,
      sym_comment,
  [12494] = 6,
    ACTIONS(3), 1,
      aux_sym_macro_call_token1,
    ACTIONS(5), 1,
      anon_sym_DASH_DASH,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(855), 1,
      sym_static_argument,
    ACTIONS(857), 1,
      aux_sym_color_code_css_token2,
    STATE(263), 2,
      sym_macro_call,
      sym_comment,
  [12514] = 5,
    ACTIONS(3), 1,
      aux_sym_macro_call_token1,
    ACTIONS(5), 1,
      anon_sym_DASH_DASH,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(819), 2,
      sym_comma,
      sym_semi_colon,
    STATE(264), 2,
      sym_macro_call,
      sym_comment,
  [12532] = 6,
    ACTIONS(3), 1,
      aux_sym_macro_call_token1,
    ACTIONS(5), 1,
      anon_sym_DASH_DASH,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(859), 1,
      sym_static_argument,
    ACTIONS(861), 1,
      aux_sym_color_code_tailwind_token2,
    STATE(265), 2,
      sym_macro_call,
      sym_comment,
  [12552] = 6,
    ACTIONS(3), 1,
      aux_sym_macro_call_token1,
    ACTIONS(5), 1,
      anon_sym_DASH_DASH,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(863), 1,
      sym_static_argument,
    ACTIONS(865), 1,
      aux_sym_color_code_skin_token2,
    STATE(266), 2,
      sym_macro_call,
      sym_comment,
  [12572] = 6,
    ACTIONS(3), 1,
      aux_sym_macro_call_token1,
    ACTIONS(5), 1,
      anon_sym_DASH_DASH,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(867), 1,
      sym_static_argument,
    STATE(271), 1,
      aux_sym_macro_args_repeat1,
    STATE(267), 2,
      sym_macro_call,
      sym_comment,
  [12592] = 6,
    ACTIONS(3), 1,
      aux_sym_macro_call_token1,
    ACTIONS(5), 1,
      anon_sym_DASH_DASH,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(799), 1,
      sym_scope_open,
    STATE(154), 1,
      sym_macro_scope,
    STATE(268), 2,
      sym_macro_call,
      sym_comment,
  [12612] = 6,
    ACTIONS(3), 1,
      aux_sym_macro_call_token1,
    ACTIONS(5), 1,
      anon_sym_DASH_DASH,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(869), 1,
      sym_static_argument,
    ACTIONS(871), 1,
      aux_sym_color_code_brick_token2,
    STATE(269), 2,
      sym_macro_call,
      sym_comment,
  [12632] = 6,
    ACTIONS(3), 1,
      aux_sym_macro_call_token1,
    ACTIONS(5), 1,
      anon_sym_DASH_DASH,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(873), 1,
      sym_static_argument,
    ACTIONS(875), 1,
      aux_sym_color_code_css_token2,
    STATE(270), 2,
      sym_macro_call,
      sym_comment,
  [12652] = 5,
    ACTIONS(3), 1,
      aux_sym_macro_call_token1,
    ACTIONS(5), 1,
      anon_sym_DASH_DASH,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(877), 1,
      sym_static_argument,
    STATE(271), 3,
      sym_macro_call,
      sym_comment,
      aux_sym_macro_args_repeat1,
  [12670] = 5,
    ACTIONS(3), 1,
      aux_sym_macro_call_token1,
    ACTIONS(5), 1,
      anon_sym_DASH_DASH,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(880), 2,
      sym_static_argument,
      sym__color_code_shade,
    STATE(272), 2,
      sym_macro_call,
      sym_comment,
  [12688] = 5,
    ACTIONS(3), 1,
      aux_sym_macro_call_token1,
    ACTIONS(5), 1,
      anon_sym_DASH_DASH,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(882), 1,
      sym_semi_colon,
    STATE(273), 2,
      sym_macro_call,
      sym_comment,
  [12705] = 5,
    ACTIONS(3), 1,
      aux_sym_macro_call_token1,
    ACTIONS(5), 1,
      anon_sym_DASH_DASH,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(884), 1,
      sym_tuple_open,
    STATE(274), 2,
      sym_macro_call,
      sym_comment,
  [12722] = 5,
    ACTIONS(3), 1,
      aux_sym_macro_call_token1,
    ACTIONS(5), 1,
      anon_sym_DASH_DASH,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(886), 1,
      sym_equals,
    STATE(275), 2,
      sym_macro_call,
      sym_comment,
  [12739] = 5,
    ACTIONS(3), 1,
      aux_sym_macro_call_token1,
    ACTIONS(5), 1,
      anon_sym_DASH_DASH,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(888), 1,
      ts_builtin_sym_end,
    STATE(276), 2,
      sym_macro_call,
      sym_comment,
  [12756] = 5,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(890), 1,
      aux_sym_macro_call_token1,
    ACTIONS(892), 1,
      anon_sym_DASH_DASH,
    ACTIONS(894), 1,
      aux_sym_comment_token1,
    STATE(277), 2,
      sym_macro_call,
      sym_comment,
  [12773] = 5,
    ACTIONS(3), 1,
      aux_sym_macro_call_token1,
    ACTIONS(5), 1,
      anon_sym_DASH_DASH,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(896), 1,
      sym__block_comment_content,
    STATE(278), 2,
      sym_macro_call,
      sym_comment,
  [12790] = 5,
    ACTIONS(3), 1,
      aux_sym_macro_call_token1,
    ACTIONS(5), 1,
      anon_sym_DASH_DASH,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(898), 1,
      sym_identifier,
    STATE(279), 2,
      sym_macro_call,
      sym_comment,
  [12807] = 5,
    ACTIONS(3), 1,
      aux_sym_macro_call_token1,
    ACTIONS(5), 1,
      anon_sym_DASH_DASH,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(900), 1,
      sym_tuple_open,
    STATE(280), 2,
      sym_macro_call,
      sym_comment,
  [12824] = 5,
    ACTIONS(3), 1,
      aux_sym_macro_call_token1,
    ACTIONS(5), 1,
      anon_sym_DASH_DASH,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(902), 1,
      sym_semi_colon,
    STATE(281), 2,
      sym_macro_call,
      sym_comment,
  [12841] = 5,
    ACTIONS(3), 1,
      aux_sym_macro_call_token1,
    ACTIONS(5), 1,
      anon_sym_DASH_DASH,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(904), 1,
      sym__block_string_content,
    STATE(282), 2,
      sym_macro_call,
      sym_comment,
  [12858] = 5,
    ACTIONS(3), 1,
      aux_sym_macro_call_token1,
    ACTIONS(5), 1,
      anon_sym_DASH_DASH,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(333), 1,
      sym_identifier,
    STATE(283), 2,
      sym_macro_call,
      sym_comment,
  [12875] = 5,
    ACTIONS(3), 1,
      aux_sym_macro_call_token1,
    ACTIONS(5), 1,
      anon_sym_DASH_DASH,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(906), 1,
      sym_tuple_open,
    STATE(284), 2,
      sym_macro_call,
      sym_comment,
  [12892] = 5,
    ACTIONS(3), 1,
      aux_sym_macro_call_token1,
    ACTIONS(5), 1,
      anon_sym_DASH_DASH,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(908), 1,
      sym_static_argument,
    STATE(285), 2,
      sym_macro_call,
      sym_comment,
  [12909] = 5,
    ACTIONS(3), 1,
      aux_sym_macro_call_token1,
    ACTIONS(5), 1,
      anon_sym_DASH_DASH,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(910), 1,
      sym__block_string_content,
    STATE(286), 2,
      sym_macro_call,
      sym_comment,
  [12926] = 5,
    ACTIONS(3), 1,
      aux_sym_macro_call_token1,
    ACTIONS(5), 1,
      anon_sym_DASH_DASH,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(912), 1,
      sym_tuple_open,
    STATE(287), 2,
      sym_macro_call,
      sym_comment,
  [12943] = 5,
    ACTIONS(3), 1,
      aux_sym_macro_call_token1,
    ACTIONS(5), 1,
      anon_sym_DASH_DASH,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(914), 1,
      sym__block_comment_end,
    STATE(288), 2,
      sym_macro_call,
      sym_comment,
  [12960] = 5,
    ACTIONS(3), 1,
      aux_sym_macro_call_token1,
    ACTIONS(5), 1,
      anon_sym_DASH_DASH,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(916), 1,
      sym_scope_open,
    STATE(289), 2,
      sym_macro_call,
      sym_comment,
  [12977] = 5,
    ACTIONS(3), 1,
      aux_sym_macro_call_token1,
    ACTIONS(5), 1,
      anon_sym_DASH_DASH,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(918), 1,
      sym_semi_colon,
    STATE(290), 2,
      sym_macro_call,
      sym_comment,
  [12994] = 5,
    ACTIONS(3), 1,
      aux_sym_macro_call_token1,
    ACTIONS(5), 1,
      anon_sym_DASH_DASH,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(920), 1,
      sym_equals,
    STATE(291), 2,
      sym_macro_call,
      sym_comment,
  [13011] = 5,
    ACTIONS(3), 1,
      aux_sym_macro_call_token1,
    ACTIONS(5), 1,
      anon_sym_DASH_DASH,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(922), 1,
      sym_identifier,
    STATE(292), 2,
      sym_macro_call,
      sym_comment,
  [13028] = 5,
    ACTIONS(3), 1,
      aux_sym_macro_call_token1,
    ACTIONS(5), 1,
      anon_sym_DASH_DASH,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(924), 1,
      sym_semi_colon,
    STATE(293), 2,
      sym_macro_call,
      sym_comment,
  [13045] = 5,
    ACTIONS(3), 1,
      aux_sym_macro_call_token1,
    ACTIONS(5), 1,
      anon_sym_DASH_DASH,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(926), 1,
      sym__block_string_end,
    STATE(294), 2,
      sym_macro_call,
      sym_comment,
  [13062] = 5,
    ACTIONS(3), 1,
      aux_sym_macro_call_token1,
    ACTIONS(5), 1,
      anon_sym_DASH_DASH,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(928), 1,
      sym_semi_colon,
    STATE(295), 2,
      sym_macro_call,
      sym_comment,
  [13079] = 5,
    ACTIONS(3), 1,
      aux_sym_macro_call_token1,
    ACTIONS(5), 1,
      anon_sym_DASH_DASH,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(930), 1,
      sym_semi_colon,
    STATE(296), 2,
      sym_macro_call,
      sym_comment,
  [13096] = 5,
    ACTIONS(3), 1,
      aux_sym_macro_call_token1,
    ACTIONS(5), 1,
      anon_sym_DASH_DASH,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(932), 1,
      sym_tuple_open,
    STATE(297), 2,
      sym_macro_call,
      sym_comment,
  [13113] = 5,
    ACTIONS(3), 1,
      aux_sym_macro_call_token1,
    ACTIONS(5), 1,
      anon_sym_DASH_DASH,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(934), 1,
      sym_scope_open,
    STATE(298), 2,
      sym_macro_call,
      sym_comment,
  [13130] = 5,
    ACTIONS(3), 1,
      aux_sym_macro_call_token1,
    ACTIONS(5), 1,
      anon_sym_DASH_DASH,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(936), 1,
      sym_identifier,
    STATE(299), 2,
      sym_macro_call,
      sym_comment,
  [13147] = 5,
    ACTIONS(3), 1,
      aux_sym_macro_call_token1,
    ACTIONS(5), 1,
      anon_sym_DASH_DASH,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(938), 1,
      sym__block_string_end,
    STATE(300), 2,
      sym_macro_call,
      sym_comment,
  [13164] = 5,
    ACTIONS(3), 1,
      aux_sym_macro_call_token1,
    ACTIONS(5), 1,
      anon_sym_DASH_DASH,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(940), 1,
      sym_scope_open,
    STATE(301), 2,
      sym_macro_call,
      sym_comment,
  [13181] = 5,
    ACTIONS(3), 1,
      aux_sym_macro_call_token1,
    ACTIONS(5), 1,
      anon_sym_DASH_DASH,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(942), 1,
      sym_identifier,
    STATE(302), 2,
      sym_macro_call,
      sym_comment,
  [13198] = 5,
    ACTIONS(3), 1,
      aux_sym_macro_call_token1,
    ACTIONS(5), 1,
      anon_sym_DASH_DASH,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(944), 1,
      sym_scope_open,
    STATE(303), 2,
      sym_macro_call,
      sym_comment,
  [13215] = 5,
    ACTIONS(3), 1,
      aux_sym_macro_call_token1,
    ACTIONS(5), 1,
      anon_sym_DASH_DASH,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(946), 1,
      sym_identifier,
    STATE(304), 2,
      sym_macro_call,
      sym_comment,
  [13232] = 1,
    ACTIONS(431), 1,
      ts_builtin_sym_end,
  [13236] = 1,
    ACTIONS(435), 1,
      ts_builtin_sym_end,
  [13240] = 1,
    ACTIONS(948), 1,
      ts_builtin_sym_end,
  [13244] = 1,
    ACTIONS(427), 1,
      ts_builtin_sym_end,
  [13248] = 1,
    ACTIONS(423), 1,
      ts_builtin_sym_end,
  [13252] = 1,
    ACTIONS(950), 1,
      ts_builtin_sym_end,
};

static const uint32_t ts_small_parse_table_map[] = {
  [SMALL_STATE(5)] = 0,
  [SMALL_STATE(6)] = 65,
  [SMALL_STATE(7)] = 123,
  [SMALL_STATE(8)] = 181,
  [SMALL_STATE(9)] = 236,
  [SMALL_STATE(10)] = 293,
  [SMALL_STATE(11)] = 348,
  [SMALL_STATE(12)] = 403,
  [SMALL_STATE(13)] = 458,
  [SMALL_STATE(14)] = 513,
  [SMALL_STATE(15)] = 568,
  [SMALL_STATE(16)] = 623,
  [SMALL_STATE(17)] = 678,
  [SMALL_STATE(18)] = 733,
  [SMALL_STATE(19)] = 788,
  [SMALL_STATE(20)] = 887,
  [SMALL_STATE(21)] = 942,
  [SMALL_STATE(22)] = 997,
  [SMALL_STATE(23)] = 1096,
  [SMALL_STATE(24)] = 1151,
  [SMALL_STATE(25)] = 1206,
  [SMALL_STATE(26)] = 1261,
  [SMALL_STATE(27)] = 1316,
  [SMALL_STATE(28)] = 1371,
  [SMALL_STATE(29)] = 1426,
  [SMALL_STATE(30)] = 1481,
  [SMALL_STATE(31)] = 1536,
  [SMALL_STATE(32)] = 1591,
  [SMALL_STATE(33)] = 1646,
  [SMALL_STATE(34)] = 1701,
  [SMALL_STATE(35)] = 1756,
  [SMALL_STATE(36)] = 1811,
  [SMALL_STATE(37)] = 1866,
  [SMALL_STATE(38)] = 1921,
  [SMALL_STATE(39)] = 1976,
  [SMALL_STATE(40)] = 2075,
  [SMALL_STATE(41)] = 2130,
  [SMALL_STATE(42)] = 2185,
  [SMALL_STATE(43)] = 2240,
  [SMALL_STATE(44)] = 2295,
  [SMALL_STATE(45)] = 2394,
  [SMALL_STATE(46)] = 2449,
  [SMALL_STATE(47)] = 2504,
  [SMALL_STATE(48)] = 2559,
  [SMALL_STATE(49)] = 2614,
  [SMALL_STATE(50)] = 2669,
  [SMALL_STATE(51)] = 2724,
  [SMALL_STATE(52)] = 2779,
  [SMALL_STATE(53)] = 2834,
  [SMALL_STATE(54)] = 2889,
  [SMALL_STATE(55)] = 2944,
  [SMALL_STATE(56)] = 2999,
  [SMALL_STATE(57)] = 3054,
  [SMALL_STATE(58)] = 3109,
  [SMALL_STATE(59)] = 3164,
  [SMALL_STATE(60)] = 3219,
  [SMALL_STATE(61)] = 3274,
  [SMALL_STATE(62)] = 3329,
  [SMALL_STATE(63)] = 3384,
  [SMALL_STATE(64)] = 3439,
  [SMALL_STATE(65)] = 3494,
  [SMALL_STATE(66)] = 3549,
  [SMALL_STATE(67)] = 3604,
  [SMALL_STATE(68)] = 3659,
  [SMALL_STATE(69)] = 3714,
  [SMALL_STATE(70)] = 3769,
  [SMALL_STATE(71)] = 3824,
  [SMALL_STATE(72)] = 3879,
  [SMALL_STATE(73)] = 3934,
  [SMALL_STATE(74)] = 4033,
  [SMALL_STATE(75)] = 4090,
  [SMALL_STATE(76)] = 4189,
  [SMALL_STATE(77)] = 4288,
  [SMALL_STATE(78)] = 4343,
  [SMALL_STATE(79)] = 4435,
  [SMALL_STATE(80)] = 4527,
  [SMALL_STATE(81)] = 4619,
  [SMALL_STATE(82)] = 4711,
  [SMALL_STATE(83)] = 4803,
  [SMALL_STATE(84)] = 4893,
  [SMALL_STATE(85)] = 4985,
  [SMALL_STATE(86)] = 5077,
  [SMALL_STATE(87)] = 5169,
  [SMALL_STATE(88)] = 5261,
  [SMALL_STATE(89)] = 5353,
  [SMALL_STATE(90)] = 5445,
  [SMALL_STATE(91)] = 5537,
  [SMALL_STATE(92)] = 5629,
  [SMALL_STATE(93)] = 5721,
  [SMALL_STATE(94)] = 5813,
  [SMALL_STATE(95)] = 5905,
  [SMALL_STATE(96)] = 5997,
  [SMALL_STATE(97)] = 6089,
  [SMALL_STATE(98)] = 6181,
  [SMALL_STATE(99)] = 6273,
  [SMALL_STATE(100)] = 6365,
  [SMALL_STATE(101)] = 6457,
  [SMALL_STATE(102)] = 6549,
  [SMALL_STATE(103)] = 6641,
  [SMALL_STATE(104)] = 6733,
  [SMALL_STATE(105)] = 6825,
  [SMALL_STATE(106)] = 6914,
  [SMALL_STATE(107)] = 7003,
  [SMALL_STATE(108)] = 7076,
  [SMALL_STATE(109)] = 7151,
  [SMALL_STATE(110)] = 7226,
  [SMALL_STATE(111)] = 7301,
  [SMALL_STATE(112)] = 7376,
  [SMALL_STATE(113)] = 7451,
  [SMALL_STATE(114)] = 7526,
  [SMALL_STATE(115)] = 7601,
  [SMALL_STATE(116)] = 7676,
  [SMALL_STATE(117)] = 7751,
  [SMALL_STATE(118)] = 7826,
  [SMALL_STATE(119)] = 7901,
  [SMALL_STATE(120)] = 7976,
  [SMALL_STATE(121)] = 8051,
  [SMALL_STATE(122)] = 8126,
  [SMALL_STATE(123)] = 8201,
  [SMALL_STATE(124)] = 8278,
  [SMALL_STATE(125)] = 8353,
  [SMALL_STATE(126)] = 8389,
  [SMALL_STATE(127)] = 8425,
  [SMALL_STATE(128)] = 8460,
  [SMALL_STATE(129)] = 8495,
  [SMALL_STATE(130)] = 8530,
  [SMALL_STATE(131)] = 8565,
  [SMALL_STATE(132)] = 8600,
  [SMALL_STATE(133)] = 8635,
  [SMALL_STATE(134)] = 8670,
  [SMALL_STATE(135)] = 8705,
  [SMALL_STATE(136)] = 8740,
  [SMALL_STATE(137)] = 8775,
  [SMALL_STATE(138)] = 8810,
  [SMALL_STATE(139)] = 8845,
  [SMALL_STATE(140)] = 8880,
  [SMALL_STATE(141)] = 8912,
  [SMALL_STATE(142)] = 8944,
  [SMALL_STATE(143)] = 8976,
  [SMALL_STATE(144)] = 9008,
  [SMALL_STATE(145)] = 9040,
  [SMALL_STATE(146)] = 9072,
  [SMALL_STATE(147)] = 9104,
  [SMALL_STATE(148)] = 9136,
  [SMALL_STATE(149)] = 9168,
  [SMALL_STATE(150)] = 9200,
  [SMALL_STATE(151)] = 9232,
  [SMALL_STATE(152)] = 9264,
  [SMALL_STATE(153)] = 9296,
  [SMALL_STATE(154)] = 9328,
  [SMALL_STATE(155)] = 9360,
  [SMALL_STATE(156)] = 9392,
  [SMALL_STATE(157)] = 9427,
  [SMALL_STATE(158)] = 9458,
  [SMALL_STATE(159)] = 9493,
  [SMALL_STATE(160)] = 9528,
  [SMALL_STATE(161)] = 9563,
  [SMALL_STATE(162)] = 9598,
  [SMALL_STATE(163)] = 9633,
  [SMALL_STATE(164)] = 9664,
  [SMALL_STATE(165)] = 9695,
  [SMALL_STATE(166)] = 9730,
  [SMALL_STATE(167)] = 9765,
  [SMALL_STATE(168)] = 9800,
  [SMALL_STATE(169)] = 9835,
  [SMALL_STATE(170)] = 9870,
  [SMALL_STATE(171)] = 9905,
  [SMALL_STATE(172)] = 9940,
  [SMALL_STATE(173)] = 9975,
  [SMALL_STATE(174)] = 10010,
  [SMALL_STATE(175)] = 10045,
  [SMALL_STATE(176)] = 10080,
  [SMALL_STATE(177)] = 10115,
  [SMALL_STATE(178)] = 10150,
  [SMALL_STATE(179)] = 10185,
  [SMALL_STATE(180)] = 10220,
  [SMALL_STATE(181)] = 10248,
  [SMALL_STATE(182)] = 10276,
  [SMALL_STATE(183)] = 10312,
  [SMALL_STATE(184)] = 10344,
  [SMALL_STATE(185)] = 10372,
  [SMALL_STATE(186)] = 10400,
  [SMALL_STATE(187)] = 10428,
  [SMALL_STATE(188)] = 10456,
  [SMALL_STATE(189)] = 10490,
  [SMALL_STATE(190)] = 10518,
  [SMALL_STATE(191)] = 10546,
  [SMALL_STATE(192)] = 10574,
  [SMALL_STATE(193)] = 10610,
  [SMALL_STATE(194)] = 10646,
  [SMALL_STATE(195)] = 10674,
  [SMALL_STATE(196)] = 10702,
  [SMALL_STATE(197)] = 10730,
  [SMALL_STATE(198)] = 10758,
  [SMALL_STATE(199)] = 10786,
  [SMALL_STATE(200)] = 10814,
  [SMALL_STATE(201)] = 10842,
  [SMALL_STATE(202)] = 10878,
  [SMALL_STATE(203)] = 10906,
  [SMALL_STATE(204)] = 10934,
  [SMALL_STATE(205)] = 10962,
  [SMALL_STATE(206)] = 10990,
  [SMALL_STATE(207)] = 11030,
  [SMALL_STATE(208)] = 11058,
  [SMALL_STATE(209)] = 11086,
  [SMALL_STATE(210)] = 11114,
  [SMALL_STATE(211)] = 11142,
  [SMALL_STATE(212)] = 11178,
  [SMALL_STATE(213)] = 11206,
  [SMALL_STATE(214)] = 11237,
  [SMALL_STATE(215)] = 11268,
  [SMALL_STATE(216)] = 11299,
  [SMALL_STATE(217)] = 11328,
  [SMALL_STATE(218)] = 11359,
  [SMALL_STATE(219)] = 11390,
  [SMALL_STATE(220)] = 11420,
  [SMALL_STATE(221)] = 11450,
  [SMALL_STATE(222)] = 11480,
  [SMALL_STATE(223)] = 11516,
  [SMALL_STATE(224)] = 11542,
  [SMALL_STATE(225)] = 11578,
  [SMALL_STATE(226)] = 11614,
  [SMALL_STATE(227)] = 11640,
  [SMALL_STATE(228)] = 11670,
  [SMALL_STATE(229)] = 11700,
  [SMALL_STATE(230)] = 11726,
  [SMALL_STATE(231)] = 11760,
  [SMALL_STATE(232)] = 11787,
  [SMALL_STATE(233)] = 11820,
  [SMALL_STATE(234)] = 11853,
  [SMALL_STATE(235)] = 11877,
  [SMALL_STATE(236)] = 11907,
  [SMALL_STATE(237)] = 11937,
  [SMALL_STATE(238)] = 11964,
  [SMALL_STATE(239)] = 11991,
  [SMALL_STATE(240)] = 12018,
  [SMALL_STATE(241)] = 12044,
  [SMALL_STATE(242)] = 12064,
  [SMALL_STATE(243)] = 12083,
  [SMALL_STATE(244)] = 12106,
  [SMALL_STATE(245)] = 12125,
  [SMALL_STATE(246)] = 12146,
  [SMALL_STATE(247)] = 12167,
  [SMALL_STATE(248)] = 12188,
  [SMALL_STATE(249)] = 12209,
  [SMALL_STATE(250)] = 12228,
  [SMALL_STATE(251)] = 12249,
  [SMALL_STATE(252)] = 12272,
  [SMALL_STATE(253)] = 12293,
  [SMALL_STATE(254)] = 12316,
  [SMALL_STATE(255)] = 12339,
  [SMALL_STATE(256)] = 12358,
  [SMALL_STATE(257)] = 12378,
  [SMALL_STATE(258)] = 12398,
  [SMALL_STATE(259)] = 12416,
  [SMALL_STATE(260)] = 12434,
  [SMALL_STATE(261)] = 12454,
  [SMALL_STATE(262)] = 12474,
  [SMALL_STATE(263)] = 12494,
  [SMALL_STATE(264)] = 12514,
  [SMALL_STATE(265)] = 12532,
  [SMALL_STATE(266)] = 12552,
  [SMALL_STATE(267)] = 12572,
  [SMALL_STATE(268)] = 12592,
  [SMALL_STATE(269)] = 12612,
  [SMALL_STATE(270)] = 12632,
  [SMALL_STATE(271)] = 12652,
  [SMALL_STATE(272)] = 12670,
  [SMALL_STATE(273)] = 12688,
  [SMALL_STATE(274)] = 12705,
  [SMALL_STATE(275)] = 12722,
  [SMALL_STATE(276)] = 12739,
  [SMALL_STATE(277)] = 12756,
  [SMALL_STATE(278)] = 12773,
  [SMALL_STATE(279)] = 12790,
  [SMALL_STATE(280)] = 12807,
  [SMALL_STATE(281)] = 12824,
  [SMALL_STATE(282)] = 12841,
  [SMALL_STATE(283)] = 12858,
  [SMALL_STATE(284)] = 12875,
  [SMALL_STATE(285)] = 12892,
  [SMALL_STATE(286)] = 12909,
  [SMALL_STATE(287)] = 12926,
  [SMALL_STATE(288)] = 12943,
  [SMALL_STATE(289)] = 12960,
  [SMALL_STATE(290)] = 12977,
  [SMALL_STATE(291)] = 12994,
  [SMALL_STATE(292)] = 13011,
  [SMALL_STATE(293)] = 13028,
  [SMALL_STATE(294)] = 13045,
  [SMALL_STATE(295)] = 13062,
  [SMALL_STATE(296)] = 13079,
  [SMALL_STATE(297)] = 13096,
  [SMALL_STATE(298)] = 13113,
  [SMALL_STATE(299)] = 13130,
  [SMALL_STATE(300)] = 13147,
  [SMALL_STATE(301)] = 13164,
  [SMALL_STATE(302)] = 13181,
  [SMALL_STATE(303)] = 13198,
  [SMALL_STATE(304)] = 13215,
  [SMALL_STATE(305)] = 13232,
  [SMALL_STATE(306)] = 13236,
  [SMALL_STATE(307)] = 13240,
  [SMALL_STATE(308)] = 13244,
  [SMALL_STATE(309)] = 13248,
  [SMALL_STATE(310)] = 13252,
};

static const TSParseActionEntry ts_parse_actions[] = {
  [0] = {.entry = {.count = 0, .reusable = false}},
  [1] = {.entry = {.count = 1, .reusable = false}}, RECOVER(),
  [3] = {.entry = {.count = 1, .reusable = true}}, SHIFT(274),
  [5] = {.entry = {.count = 1, .reusable = true}}, SHIFT(277),
  [7] = {.entry = {.count = 1, .reusable = true}}, SHIFT(278),
  [9] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_source_file, 0, 0, 0),
  [11] = {.entry = {.count = 1, .reusable = false}}, SHIFT(226),
  [13] = {.entry = {.count = 1, .reusable = true}}, SHIFT(192),
  [15] = {.entry = {.count = 1, .reusable = true}}, SHIFT(133),
  [17] = {.entry = {.count = 1, .reusable = true}}, SHIFT(291),
  [19] = {.entry = {.count = 1, .reusable = true}}, SHIFT(275),
  [21] = {.entry = {.count = 1, .reusable = true}}, SHIFT(223),
  [23] = {.entry = {.count = 1, .reusable = true}}, SHIFT(299),
  [25] = {.entry = {.count = 1, .reusable = true}}, SHIFT(233),
  [27] = {.entry = {.count = 1, .reusable = true}}, SHIFT(229),
  [29] = {.entry = {.count = 1, .reusable = true}}, SHIFT(109),
  [31] = {.entry = {.count = 1, .reusable = false}}, SHIFT(21),
  [33] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5),
  [35] = {.entry = {.count = 1, .reusable = true}}, SHIFT(23),
  [37] = {.entry = {.count = 1, .reusable = false}}, SHIFT(46),
  [39] = {.entry = {.count = 1, .reusable = true}}, SHIFT(57),
  [41] = {.entry = {.count = 1, .reusable = true}}, SHIFT(74),
  [43] = {.entry = {.count = 1, .reusable = true}}, SHIFT(9),
  [45] = {.entry = {.count = 1, .reusable = true}}, SHIFT(10),
  [47] = {.entry = {.count = 1, .reusable = true}}, SHIFT(297),
  [49] = {.entry = {.count = 1, .reusable = true}}, SHIFT(237),
  [51] = {.entry = {.count = 1, .reusable = true}}, SHIFT(236),
  [53] = {.entry = {.count = 1, .reusable = true}}, SHIFT(232),
  [55] = {.entry = {.count = 1, .reusable = true}}, SHIFT(24),
  [57] = {.entry = {.count = 1, .reusable = false}}, SHIFT(24),
  [59] = {.entry = {.count = 1, .reusable = true}}, SHIFT(11),
  [61] = {.entry = {.count = 1, .reusable = true}}, SHIFT(124),
  [63] = {.entry = {.count = 1, .reusable = true}}, SHIFT(144),
  [65] = {.entry = {.count = 1, .reusable = true}}, SHIFT(75),
  [67] = {.entry = {.count = 1, .reusable = false}}, SHIFT(30),
  [69] = {.entry = {.count = 1, .reusable = true}}, SHIFT(30),
  [71] = {.entry = {.count = 1, .reusable = true}}, SHIFT(47),
  [73] = {.entry = {.count = 1, .reusable = false}}, SHIFT(47),
  [75] = {.entry = {.count = 1, .reusable = true}}, SHIFT(26),
  [77] = {.entry = {.count = 1, .reusable = true}}, SHIFT(260),
  [79] = {.entry = {.count = 1, .reusable = true}}, SHIFT(261),
  [81] = {.entry = {.count = 1, .reusable = true}}, SHIFT(262),
  [83] = {.entry = {.count = 1, .reusable = true}}, SHIFT(263),
  [85] = {.entry = {.count = 1, .reusable = false}}, SHIFT(31),
  [87] = {.entry = {.count = 1, .reusable = false}}, SHIFT(243),
  [89] = {.entry = {.count = 1, .reusable = true}}, SHIFT(32),
  [91] = {.entry = {.count = 1, .reusable = true}}, SHIFT(286),
  [93] = {.entry = {.count = 1, .reusable = true}}, SHIFT(140),
  [95] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_macro_scope_repeat1, 2, 0, 0), SHIFT_REPEAT(21),
  [98] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_macro_scope_repeat1, 2, 0, 0), SHIFT_REPEAT(5),
  [101] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_macro_scope_repeat1, 2, 0, 0), SHIFT_REPEAT(23),
  [104] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_macro_scope_repeat1, 2, 0, 0), SHIFT_REPEAT(46),
  [107] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_macro_scope_repeat1, 2, 0, 0), SHIFT_REPEAT(57),
  [110] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_macro_scope_repeat1, 2, 0, 0), SHIFT_REPEAT(74),
  [113] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_macro_scope_repeat1, 2, 0, 0), SHIFT_REPEAT(9),
  [116] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_macro_scope_repeat1, 2, 0, 0), SHIFT_REPEAT(10),
  [119] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_macro_scope_repeat1, 2, 0, 0), SHIFT_REPEAT(297),
  [122] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_macro_scope_repeat1, 2, 0, 0), SHIFT_REPEAT(237),
  [125] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_macro_scope_repeat1, 2, 0, 0), SHIFT_REPEAT(236),
  [128] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_macro_scope_repeat1, 2, 0, 0), SHIFT_REPEAT(232),
  [131] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_macro_scope_repeat1, 2, 0, 0), SHIFT_REPEAT(24),
  [134] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_macro_scope_repeat1, 2, 0, 0), SHIFT_REPEAT(24),
  [137] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_macro_scope_repeat1, 2, 0, 0), SHIFT_REPEAT(11),
  [140] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_macro_scope_repeat1, 2, 0, 0), SHIFT_REPEAT(124),
  [143] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_macro_scope_repeat1, 2, 0, 0),
  [145] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_macro_scope_repeat1, 2, 0, 0), SHIFT_REPEAT(75),
  [148] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_macro_scope_repeat1, 2, 0, 0), SHIFT_REPEAT(30),
  [151] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_macro_scope_repeat1, 2, 0, 0), SHIFT_REPEAT(30),
  [154] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_macro_scope_repeat1, 2, 0, 0), SHIFT_REPEAT(47),
  [157] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_macro_scope_repeat1, 2, 0, 0), SHIFT_REPEAT(47),
  [160] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_macro_scope_repeat1, 2, 0, 0), SHIFT_REPEAT(26),
  [163] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_macro_scope_repeat1, 2, 0, 0), SHIFT_REPEAT(260),
  [166] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_macro_scope_repeat1, 2, 0, 0), SHIFT_REPEAT(261),
  [169] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_macro_scope_repeat1, 2, 0, 0), SHIFT_REPEAT(262),
  [172] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_macro_scope_repeat1, 2, 0, 0), SHIFT_REPEAT(263),
  [175] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_macro_scope_repeat1, 2, 0, 0), SHIFT_REPEAT(31),
  [178] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_macro_scope_repeat1, 2, 0, 0), SHIFT_REPEAT(243),
  [181] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_macro_scope_repeat1, 2, 0, 0), SHIFT_REPEAT(32),
  [184] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_macro_scope_repeat1, 2, 0, 0), SHIFT_REPEAT(286),
  [187] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_macro_body, 1, 0, 0),
  [189] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_macro_body, 1, 0, 0),
  [191] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_operation, 3, 0, 0),
  [193] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_operation, 3, 0, 0),
  [195] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_rule_scope, 3, 0, 6),
  [197] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_rule_scope, 3, 0, 6),
  [199] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_reference, 1, 0, 0),
  [201] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_reference, 1, 0, 0),
  [203] = {.entry = {.count = 1, .reusable = true}}, SHIFT(91),
  [205] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_enum_shorthand, 1, 0, 0),
  [207] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_enum_shorthand, 1, 0, 0),
  [209] = {.entry = {.count = 1, .reusable = false}}, REDUCE(aux_sym_macro_scope_repeat1, 1, 0, 0),
  [211] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_macro_scope_repeat1, 1, 0, 0),
  [213] = {.entry = {.count = 1, .reusable = false}}, SHIFT(280),
  [215] = {.entry = {.count = 1, .reusable = true}}, SHIFT(88),
  [217] = {.entry = {.count = 1, .reusable = true}}, SHIFT(190),
  [219] = {.entry = {.count = 1, .reusable = true}}, SHIFT(284),
  [221] = {.entry = {.count = 1, .reusable = true}}, SHIFT(191),
  [223] = {.entry = {.count = 1, .reusable = true}}, SHIFT(22),
  [225] = {.entry = {.count = 1, .reusable = true}}, SHIFT(208),
  [227] = {.entry = {.count = 1, .reusable = false}}, SHIFT(180),
  [229] = {.entry = {.count = 1, .reusable = true}}, SHIFT(180),
  [231] = {.entry = {.count = 1, .reusable = true}}, SHIFT(207),
  [233] = {.entry = {.count = 1, .reusable = true}}, SHIFT(265),
  [235] = {.entry = {.count = 1, .reusable = true}}, SHIFT(266),
  [237] = {.entry = {.count = 1, .reusable = true}}, SHIFT(269),
  [239] = {.entry = {.count = 1, .reusable = true}}, SHIFT(270),
  [241] = {.entry = {.count = 1, .reusable = true}}, SHIFT(195),
  [243] = {.entry = {.count = 1, .reusable = false}}, SHIFT(254),
  [245] = {.entry = {.count = 1, .reusable = true}}, SHIFT(212),
  [247] = {.entry = {.count = 1, .reusable = true}}, SHIFT(282),
  [249] = {.entry = {.count = 1, .reusable = true}}, SHIFT(81),
  [251] = {.entry = {.count = 1, .reusable = true}}, SHIFT(181),
  [253] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_selector, 1, 0, 0),
  [255] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_selector, 1, 0, 0),
  [257] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__definition_no_macro, 1, 0, 0),
  [259] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__definition_no_macro, 1, 0, 0),
  [261] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_string, 1, 0, 0),
  [263] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_string, 1, 0, 0),
  [265] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_rule_scope, 2, 0, 0),
  [267] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_rule_scope, 2, 0, 0),
  [269] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_number, 1, 0, 0),
  [271] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_number, 1, 0, 0),
  [273] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_color, 1, 0, 0),
  [275] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_color, 1, 0, 0),
  [277] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_datatype, 1, 0, 0),
  [279] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_datatype, 1, 0, 0),
  [281] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_enum, 1, 0, 0),
  [283] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_enum, 1, 0, 0),
  [285] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_token_assignment, 3, 0, 4),
  [287] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_token_assignment, 3, 0, 4),
  [289] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_static_token_assignment, 3, 0, 5),
  [291] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_static_token_assignment, 3, 0, 5),
  [293] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_derive_declaration, 3, 0, 0),
  [295] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_derive_declaration, 3, 0, 0),
  [297] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_priority_declaration, 2, 0, 0),
  [299] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_priority_declaration, 2, 0, 0),
  [301] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_name_declaration, 2, 0, 0),
  [303] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_name_declaration, 2, 0, 0),
  [305] = {.entry = {.count = 1, .reusable = true}}, SHIFT(103),
  [307] = {.entry = {.count = 1, .reusable = true}}, SHIFT(69),
  [309] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_rule_scope, 3, 0, 7),
  [311] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_rule_scope, 3, 0, 7),
  [313] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_tuple, 2, 0, 0),
  [315] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_tuple, 2, 0, 0),
  [317] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_color_code_tailwind, 2, 0, 8),
  [319] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_color_code_tailwind, 2, 0, 8),
  [321] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_color_code_skin, 2, 0, 8),
  [323] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_color_code_skin, 2, 0, 8),
  [325] = {.entry = {.count = 1, .reusable = true}}, SHIFT(86),
  [327] = {.entry = {.count = 1, .reusable = true}}, SHIFT(309),
  [329] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_color_code_css, 2, 0, 8),
  [331] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_color_code_css, 2, 0, 8),
  [333] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_colon, 1, 0, 0),
  [335] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_colon, 1, 0, 0),
  [337] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_operator, 1, 0, 0),
  [339] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_operator, 1, 0, 0),
  [341] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_rule_scope, 4, 0, 9),
  [343] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_rule_scope, 4, 0, 9),
  [345] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_token_assignment, 4, 0, 10),
  [347] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_token_assignment, 4, 0, 10),
  [349] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_static_token_assignment, 4, 0, 11),
  [351] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_static_token_assignment, 4, 0, 11),
  [353] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_block_string, 3, 0, 2),
  [355] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_block_string, 3, 0, 2),
  [357] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_priority_declaration, 3, 0, 0),
  [359] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_priority_declaration, 3, 0, 0),
  [361] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_name_declaration, 3, 0, 0),
  [363] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_name_declaration, 3, 0, 0),
  [365] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_rule_scope, 4, 0, 13),
  [367] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_rule_scope, 4, 0, 13),
  [369] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_tuple, 3, 0, 3),
  [371] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_tuple, 3, 0, 3),
  [373] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_tuple, 3, 0, 0),
  [375] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_tuple, 3, 0, 0),
  [377] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_rule_scope, 5, 0, 14),
  [379] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_rule_scope, 5, 0, 14),
  [381] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_rule_scope, 5, 0, 15),
  [383] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_rule_scope, 5, 0, 15),
  [385] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_tuple, 4, 0, 3),
  [387] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_tuple, 4, 0, 3),
  [389] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_tuple, 4, 0, 0),
  [391] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_tuple, 4, 0, 0),
  [393] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_color_code_tailwind, 4, 0, 17),
  [395] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_color_code_tailwind, 4, 0, 17),
  [397] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_color_code_skin, 4, 0, 17),
  [399] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_color_code_skin, 4, 0, 17),
  [401] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_rule_scope, 6, 0, 18),
  [403] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_rule_scope, 6, 0, 18),
  [405] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_tuple, 5, 0, 3),
  [407] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_tuple, 5, 0, 3),
  [409] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_tuple, 5, 0, 0),
  [411] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_tuple, 5, 0, 0),
  [413] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_enum_full, 5, 0, 0),
  [415] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_enum_full, 5, 0, 0),
  [417] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_tuple, 6, 0, 3),
  [419] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_tuple, 6, 0, 3),
  [421] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_macro_call, 3, 0, 3),
  [423] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_macro_call, 3, 0, 3),
  [425] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_macro_call, 4, 0, 3),
  [427] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_macro_call, 4, 0, 3),
  [429] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_macro_call, 5, 0, 3),
  [431] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_macro_call, 5, 0, 3),
  [433] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_macro_call, 6, 0, 3),
  [435] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_macro_call, 6, 0, 3),
  [437] = {.entry = {.count = 1, .reusable = true}}, SHIFT(94),
  [439] = {.entry = {.count = 1, .reusable = true}}, SHIFT(209),
  [441] = {.entry = {.count = 1, .reusable = true}}, SHIFT(78),
  [443] = {.entry = {.count = 1, .reusable = true}}, SHIFT(97),
  [445] = {.entry = {.count = 1, .reusable = true}}, SHIFT(41),
  [447] = {.entry = {.count = 1, .reusable = true}}, SHIFT(99),
  [449] = {.entry = {.count = 1, .reusable = true}}, SHIFT(55),
  [451] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_color_code_brick, 2, 0, 8),
  [453] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_color_code_brick, 2, 0, 8),
  [455] = {.entry = {.count = 1, .reusable = true}}, SHIFT(34),
  [457] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_macro_call_repeat1, 2, 0, 0), SHIFT_REPEAT(280),
  [460] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_macro_call_repeat1, 2, 0, 0), SHIFT_REPEAT(190),
  [463] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_macro_call_repeat1, 2, 0, 0), SHIFT_REPEAT(284),
  [466] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_macro_call_repeat1, 2, 0, 0), SHIFT_REPEAT(191),
  [469] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_macro_call_repeat1, 2, 0, 0), SHIFT_REPEAT(22),
  [472] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_macro_call_repeat1, 2, 0, 0), SHIFT_REPEAT(180),
  [475] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_macro_call_repeat1, 2, 0, 0), SHIFT_REPEAT(180),
  [478] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_macro_call_repeat1, 2, 0, 0), SHIFT_REPEAT(207),
  [481] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_macro_call_repeat1, 2, 0, 0), SHIFT_REPEAT(265),
  [484] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_macro_call_repeat1, 2, 0, 0), SHIFT_REPEAT(266),
  [487] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_macro_call_repeat1, 2, 0, 0), SHIFT_REPEAT(269),
  [490] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_macro_call_repeat1, 2, 0, 0), SHIFT_REPEAT(270),
  [493] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_macro_call_repeat1, 2, 0, 0), SHIFT_REPEAT(195),
  [496] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_macro_call_repeat1, 2, 0, 0), SHIFT_REPEAT(254),
  [499] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_macro_call_repeat1, 2, 0, 0), SHIFT_REPEAT(212),
  [502] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_macro_call_repeat1, 2, 0, 0), SHIFT_REPEAT(282),
  [505] = {.entry = {.count = 1, .reusable = true}}, SHIFT(135),
  [507] = {.entry = {.count = 1, .reusable = true}}, SHIFT(136),
  [509] = {.entry = {.count = 1, .reusable = true}}, SHIFT(35),
  [511] = {.entry = {.count = 1, .reusable = true}}, SHIFT(149),
  [513] = {.entry = {.count = 1, .reusable = false}}, SHIFT(287),
  [515] = {.entry = {.count = 1, .reusable = true}}, SHIFT(28),
  [517] = {.entry = {.count = 1, .reusable = true}}, SHIFT(29),
  [519] = {.entry = {.count = 1, .reusable = true}}, SHIFT(31),
  [521] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_rule_scope_repeat2, 2, 0, 0), SHIFT_REPEAT(216),
  [524] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_rule_scope_repeat2, 2, 0, 0), SHIFT_REPEAT(192),
  [527] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_rule_scope_repeat2, 2, 0, 0), SHIFT_REPEAT(133),
  [530] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_rule_scope_repeat2, 2, 0, 0), SHIFT_REPEAT(291),
  [533] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_rule_scope_repeat2, 2, 0, 0), SHIFT_REPEAT(275),
  [536] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_rule_scope_repeat2, 2, 0, 0), SHIFT_REPEAT(223),
  [539] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_rule_scope_repeat2, 2, 0, 0), SHIFT_REPEAT(238),
  [542] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_rule_scope_repeat2, 2, 0, 0), SHIFT_REPEAT(235),
  [545] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_rule_scope_repeat2, 2, 0, 0), SHIFT_REPEAT(229),
  [548] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_rule_scope_repeat2, 2, 0, 0), SHIFT_REPEAT(109),
  [551] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_rule_scope_repeat2, 2, 0, 0),
  [553] = {.entry = {.count = 1, .reusable = false}}, SHIFT(216),
  [555] = {.entry = {.count = 1, .reusable = true}}, SHIFT(238),
  [557] = {.entry = {.count = 1, .reusable = true}}, SHIFT(235),
  [559] = {.entry = {.count = 1, .reusable = true}}, SHIFT(128),
  [561] = {.entry = {.count = 1, .reusable = true}}, SHIFT(134),
  [563] = {.entry = {.count = 1, .reusable = true}}, SHIFT(8),
  [565] = {.entry = {.count = 1, .reusable = true}}, SHIFT(59),
  [567] = {.entry = {.count = 1, .reusable = true}}, SHIFT(48),
  [569] = {.entry = {.count = 1, .reusable = true}}, SHIFT(54),
  [571] = {.entry = {.count = 1, .reusable = true}}, SHIFT(58),
  [573] = {.entry = {.count = 1, .reusable = true}}, SHIFT(64),
  [575] = {.entry = {.count = 1, .reusable = true}}, SHIFT(127),
  [577] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_source_file_repeat1, 2, 0, 0),
  [579] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_source_file_repeat1, 2, 0, 0), SHIFT_REPEAT(226),
  [582] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_source_file_repeat1, 2, 0, 0), SHIFT_REPEAT(192),
  [585] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_source_file_repeat1, 2, 0, 0), SHIFT_REPEAT(133),
  [588] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_source_file_repeat1, 2, 0, 0), SHIFT_REPEAT(291),
  [591] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_source_file_repeat1, 2, 0, 0), SHIFT_REPEAT(275),
  [594] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_source_file_repeat1, 2, 0, 0), SHIFT_REPEAT(223),
  [597] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_source_file_repeat1, 2, 0, 0), SHIFT_REPEAT(299),
  [600] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_source_file_repeat1, 2, 0, 0), SHIFT_REPEAT(233),
  [603] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_source_file_repeat1, 2, 0, 0), SHIFT_REPEAT(229),
  [606] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_source_file_repeat1, 2, 0, 0), SHIFT_REPEAT(109),
  [609] = {.entry = {.count = 1, .reusable = true}}, SHIFT(139),
  [611] = {.entry = {.count = 1, .reusable = true}}, SHIFT(138),
  [613] = {.entry = {.count = 1, .reusable = true}}, SHIFT(137),
  [615] = {.entry = {.count = 1, .reusable = true}}, SHIFT(129),
  [617] = {.entry = {.count = 1, .reusable = true}}, SHIFT(130),
  [619] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_source_file, 1, 0, 0),
  [621] = {.entry = {.count = 1, .reusable = true}}, SHIFT(27),
  [623] = {.entry = {.count = 1, .reusable = false}}, REDUCE(aux_sym_macro_call_repeat1, 2, 0, 0),
  [625] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_macro_call_repeat1, 2, 0, 0),
  [627] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__definition_scope_body, 1, 0, 0),
  [629] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__definition_scope_body, 1, 0, 0),
  [631] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_macro_scope, 3, 0, 6),
  [633] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_macro_scope, 3, 0, 6),
  [635] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_rule_scope_body, 1, 0, 0),
  [637] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_rule_scope_body, 1, 0, 0),
  [639] = {.entry = {.count = 1, .reusable = false}}, REDUCE(aux_sym_rule_scope_repeat2, 1, 0, 0),
  [641] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_rule_scope_repeat2, 1, 0, 0),
  [643] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_macro_scope, 2, 0, 0),
  [645] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_macro_scope, 2, 0, 0),
  [647] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_property_assignment, 4, 0, 16),
  [649] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_property_assignment, 4, 0, 16),
  [651] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__definition, 1, 0, 0),
  [653] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__definition, 1, 0, 0),
  [655] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_property_assignment, 3, 0, 12),
  [657] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_property_assignment, 3, 0, 12),
  [659] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_macro_declaration, 3, 0, 0),
  [661] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_macro_declaration, 3, 0, 0),
  [663] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_source_file_repeat1, 1, 0, 0),
  [665] = {.entry = {.count = 1, .reusable = false}}, REDUCE(aux_sym_source_file_repeat1, 1, 0, 0),
  [667] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_macro_declaration, 4, 0, 0),
  [669] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_macro_declaration, 4, 0, 0),
  [671] = {.entry = {.count = 1, .reusable = true}}, SHIFT(125),
  [673] = {.entry = {.count = 1, .reusable = true}}, SHIFT(306),
  [675] = {.entry = {.count = 1, .reusable = true}}, SHIFT(126),
  [677] = {.entry = {.count = 1, .reusable = false}}, SHIFT(126),
  [679] = {.entry = {.count = 1, .reusable = true}}, SHIFT(202),
  [681] = {.entry = {.count = 1, .reusable = true}}, SHIFT(203),
  [683] = {.entry = {.count = 1, .reusable = true}}, SHIFT(205),
  [685] = {.entry = {.count = 1, .reusable = true}}, SHIFT(305),
  [687] = {.entry = {.count = 1, .reusable = true}}, SHIFT(210),
  [689] = {.entry = {.count = 1, .reusable = true}}, SHIFT(272),
  [691] = {.entry = {.count = 1, .reusable = true}}, SHIFT(259),
  [693] = {.entry = {.count = 1, .reusable = true}}, SHIFT(308),
  [695] = {.entry = {.count = 1, .reusable = true}}, SHIFT(187),
  [697] = {.entry = {.count = 1, .reusable = true}}, SHIFT(189),
  [699] = {.entry = {.count = 1, .reusable = true}}, SHIFT(56),
  [701] = {.entry = {.count = 1, .reusable = true}}, SHIFT(60),
  [703] = {.entry = {.count = 1, .reusable = true}}, SHIFT(194),
  [705] = {.entry = {.count = 1, .reusable = true}}, SHIFT(61),
  [707] = {.entry = {.count = 1, .reusable = true}}, SHIFT(65),
  [709] = {.entry = {.count = 1, .reusable = true}}, SHIFT(66),
  [711] = {.entry = {.count = 1, .reusable = true}}, SHIFT(68),
  [713] = {.entry = {.count = 1, .reusable = true}}, SHIFT(70),
  [715] = {.entry = {.count = 1, .reusable = true}}, SHIFT(71),
  [717] = {.entry = {.count = 1, .reusable = true}}, SHIFT(72),
  [719] = {.entry = {.count = 1, .reusable = true}}, SHIFT(184),
  [721] = {.entry = {.count = 1, .reusable = true}}, SHIFT(196),
  [723] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_rule_scope_repeat1, 2, 0, 0), SHIFT_REPEAT(226),
  [726] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_rule_scope_repeat1, 2, 0, 0), SHIFT_REPEAT(223),
  [729] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_rule_scope_repeat1, 2, 0, 0), SHIFT_REPEAT(229),
  [732] = {.entry = {.count = 1, .reusable = true}}, SHIFT(224),
  [734] = {.entry = {.count = 1, .reusable = true}}, SHIFT(206),
  [736] = {.entry = {.count = 1, .reusable = true}}, SHIFT(146),
  [738] = {.entry = {.count = 1, .reusable = true}}, SHIFT(132),
  [740] = {.entry = {.count = 1, .reusable = true}}, SHIFT(50),
  [742] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_class_selector, 1, 0, 0),
  [744] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_class_selector, 1, 0, 0),
  [746] = {.entry = {.count = 1, .reusable = true}}, SHIFT(92),
  [748] = {.entry = {.count = 1, .reusable = true}}, SHIFT(131),
  [750] = {.entry = {.count = 1, .reusable = true}}, SHIFT(49),
  [752] = {.entry = {.count = 1, .reusable = false}}, REDUCE(aux_sym_rule_scope_repeat1, 1, 0, 0),
  [754] = {.entry = {.count = 1, .reusable = true}}, SHIFT(234),
  [756] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_rule_scope_repeat1, 1, 0, 0),
  [758] = {.entry = {.count = 1, .reusable = true}}, SHIFT(120),
  [760] = {.entry = {.count = 1, .reusable = true}}, SHIFT(122),
  [762] = {.entry = {.count = 1, .reusable = true}}, SHIFT(118),
  [764] = {.entry = {.count = 1, .reusable = true}}, SHIFT(111),
  [766] = {.entry = {.count = 1, .reusable = true}}, SHIFT(112),
  [768] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_state_selector, 1, 0, 0),
  [770] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_state_selector, 1, 0, 0),
  [772] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_derive_value_repeat1, 2, 0, 0), SHIFT_REPEAT(206),
  [775] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_derive_value_repeat1, 2, 0, 0), SHIFT_REPEAT(207),
  [778] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_derive_value_repeat1, 2, 0, 0), SHIFT_REPEAT(282),
  [781] = {.entry = {.count = 1, .reusable = false}}, REDUCE(aux_sym_rule_scope_repeat1, 2, 0, 0),
  [783] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_rule_scope_repeat1, 2, 0, 0),
  [785] = {.entry = {.count = 1, .reusable = true}}, SHIFT(143),
  [787] = {.entry = {.count = 1, .reusable = true}}, SHIFT(38),
  [789] = {.entry = {.count = 1, .reusable = true}}, SHIFT(37),
  [791] = {.entry = {.count = 1, .reusable = true}}, SHIFT(148),
  [793] = {.entry = {.count = 1, .reusable = true}}, SHIFT(257),
  [795] = {.entry = {.count = 1, .reusable = true}}, SHIFT(245),
  [797] = {.entry = {.count = 1, .reusable = true}}, SHIFT(303),
  [799] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2),
  [801] = {.entry = {.count = 1, .reusable = true}}, SHIFT(239),
  [803] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_derive_value_repeat1, 2, 0, 0),
  [805] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_derive_value, 4, 0, 0),
  [807] = {.entry = {.count = 1, .reusable = true}}, SHIFT(283),
  [809] = {.entry = {.count = 1, .reusable = true}}, SHIFT(302),
  [811] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_derive_value, 5, 0, 0),
  [813] = {.entry = {.count = 1, .reusable = true}}, SHIFT(285),
  [815] = {.entry = {.count = 1, .reusable = true}}, SHIFT(289),
  [817] = {.entry = {.count = 1, .reusable = true}}, SHIFT(301),
  [819] = {.entry = {.count = 1, .reusable = true}}, SHIFT(241),
  [821] = {.entry = {.count = 1, .reusable = true}}, SHIFT(242),
  [823] = {.entry = {.count = 1, .reusable = true}}, SHIFT(244),
  [825] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_derive_value, 3, 0, 0),
  [827] = {.entry = {.count = 1, .reusable = true}}, SHIFT(249),
  [829] = {.entry = {.count = 1, .reusable = true}}, SHIFT(304),
  [831] = {.entry = {.count = 1, .reusable = true}}, SHIFT(298),
  [833] = {.entry = {.count = 1, .reusable = true}}, SHIFT(279),
  [835] = {.entry = {.count = 1, .reusable = true}}, SHIFT(292),
  [837] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_derive_value, 1, 0, 0),
  [839] = {.entry = {.count = 1, .reusable = true}}, SHIFT(252),
  [841] = {.entry = {.count = 1, .reusable = true}}, SHIFT(200),
  [843] = {.entry = {.count = 1, .reusable = true}}, SHIFT(42),
  [845] = {.entry = {.count = 1, .reusable = false}}, SHIFT(42),
  [847] = {.entry = {.count = 1, .reusable = true}}, SHIFT(43),
  [849] = {.entry = {.count = 1, .reusable = false}}, SHIFT(43),
  [851] = {.entry = {.count = 1, .reusable = true}}, SHIFT(77),
  [853] = {.entry = {.count = 1, .reusable = false}}, SHIFT(77),
  [855] = {.entry = {.count = 1, .reusable = true}}, SHIFT(45),
  [857] = {.entry = {.count = 1, .reusable = false}}, SHIFT(45),
  [859] = {.entry = {.count = 1, .reusable = true}}, SHIFT(163),
  [861] = {.entry = {.count = 1, .reusable = false}}, SHIFT(163),
  [863] = {.entry = {.count = 1, .reusable = true}}, SHIFT(164),
  [865] = {.entry = {.count = 1, .reusable = false}}, SHIFT(164),
  [867] = {.entry = {.count = 1, .reusable = true}}, SHIFT(246),
  [869] = {.entry = {.count = 1, .reusable = true}}, SHIFT(185),
  [871] = {.entry = {.count = 1, .reusable = false}}, SHIFT(185),
  [873] = {.entry = {.count = 1, .reusable = true}}, SHIFT(186),
  [875] = {.entry = {.count = 1, .reusable = false}}, SHIFT(186),
  [877] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_macro_args_repeat1, 2, 0, 0), SHIFT_REPEAT(258),
  [880] = {.entry = {.count = 1, .reusable = true}}, SHIFT(199),
  [882] = {.entry = {.count = 1, .reusable = true}}, SHIFT(153),
  [884] = {.entry = {.count = 1, .reusable = true}}, SHIFT(44),
  [886] = {.entry = {.count = 1, .reusable = true}}, SHIFT(85),
  [888] = {.entry = {.count = 1, .reusable = true}},  ACCEPT_INPUT(),
  [890] = {.entry = {.count = 1, .reusable = false}}, SHIFT(274),
  [892] = {.entry = {.count = 1, .reusable = false}}, SHIFT(277),
  [894] = {.entry = {.count = 1, .reusable = false}}, SHIFT(310),
  [896] = {.entry = {.count = 1, .reusable = true}}, SHIFT(288),
  [898] = {.entry = {.count = 1, .reusable = false}}, SHIFT(67),
  [900] = {.entry = {.count = 1, .reusable = true}}, SHIFT(19),
  [902] = {.entry = {.count = 1, .reusable = true}}, SHIFT(152),
  [904] = {.entry = {.count = 1, .reusable = true}}, SHIFT(300),
  [906] = {.entry = {.count = 1, .reusable = true}}, SHIFT(73),
  [908] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_macro_args_repeat1, 2, 0, 0),
  [910] = {.entry = {.count = 1, .reusable = true}}, SHIFT(294),
  [912] = {.entry = {.count = 1, .reusable = true}}, SHIFT(76),
  [914] = {.entry = {.count = 1, .reusable = true}}, SHIFT(307),
  [916] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_macro_args, 3, 0, 0),
  [918] = {.entry = {.count = 1, .reusable = true}}, SHIFT(155),
  [920] = {.entry = {.count = 1, .reusable = true}}, SHIFT(84),
  [922] = {.entry = {.count = 1, .reusable = false}}, SHIFT(251),
  [924] = {.entry = {.count = 1, .reusable = true}}, SHIFT(36),
  [926] = {.entry = {.count = 1, .reusable = true}}, SHIFT(51),
  [928] = {.entry = {.count = 1, .reusable = true}}, SHIFT(52),
  [930] = {.entry = {.count = 1, .reusable = true}}, SHIFT(53),
  [932] = {.entry = {.count = 1, .reusable = true}}, SHIFT(39),
  [934] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_macro_args, 4, 0, 0),
  [936] = {.entry = {.count = 1, .reusable = false}}, SHIFT(240),
  [938] = {.entry = {.count = 1, .reusable = true}}, SHIFT(198),
  [940] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_macro_args, 5, 0, 0),
  [942] = {.entry = {.count = 1, .reusable = false}}, SHIFT(253),
  [944] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_macro_args, 2, 0, 0),
  [946] = {.entry = {.count = 1, .reusable = false}}, SHIFT(204),
  [948] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_comment, 3, 0, 2),
  [950] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_comment, 2, 0, 1),
};

enum ts_external_scanner_symbol_identifiers {
  ts_external_token__block_comment_start = 0,
  ts_external_token__block_comment_content = 1,
  ts_external_token__block_comment_end = 2,
  ts_external_token__block_string_start = 3,
  ts_external_token__block_string_content = 4,
  ts_external_token__block_string_end = 5,
};

static const TSSymbol ts_external_scanner_symbol_map[EXTERNAL_TOKEN_COUNT] = {
  [ts_external_token__block_comment_start] = sym__block_comment_start,
  [ts_external_token__block_comment_content] = sym__block_comment_content,
  [ts_external_token__block_comment_end] = sym__block_comment_end,
  [ts_external_token__block_string_start] = sym__block_string_start,
  [ts_external_token__block_string_content] = sym__block_string_content,
  [ts_external_token__block_string_end] = sym__block_string_end,
};

static const bool ts_external_scanner_states[8][EXTERNAL_TOKEN_COUNT] = {
  [1] = {
    [ts_external_token__block_comment_start] = true,
    [ts_external_token__block_comment_content] = true,
    [ts_external_token__block_comment_end] = true,
    [ts_external_token__block_string_start] = true,
    [ts_external_token__block_string_content] = true,
    [ts_external_token__block_string_end] = true,
  },
  [2] = {
    [ts_external_token__block_comment_start] = true,
  },
  [3] = {
    [ts_external_token__block_comment_start] = true,
    [ts_external_token__block_string_start] = true,
  },
  [4] = {
    [ts_external_token__block_comment_start] = true,
    [ts_external_token__block_comment_content] = true,
  },
  [5] = {
    [ts_external_token__block_comment_start] = true,
    [ts_external_token__block_string_content] = true,
  },
  [6] = {
    [ts_external_token__block_comment_start] = true,
    [ts_external_token__block_comment_end] = true,
  },
  [7] = {
    [ts_external_token__block_comment_start] = true,
    [ts_external_token__block_string_end] = true,
  },
};

#ifdef __cplusplus
extern "C" {
#endif
void *tree_sitter_rsml_external_scanner_create(void);
void tree_sitter_rsml_external_scanner_destroy(void *);
bool tree_sitter_rsml_external_scanner_scan(void *, TSLexer *, const bool *);
unsigned tree_sitter_rsml_external_scanner_serialize(void *, char *);
void tree_sitter_rsml_external_scanner_deserialize(void *, const char *, unsigned);

#ifdef TREE_SITTER_HIDE_SYMBOLS
#define TS_PUBLIC
#elif defined(_WIN32)
#define TS_PUBLIC __declspec(dllexport)
#else
#define TS_PUBLIC __attribute__((visibility("default")))
#endif

TS_PUBLIC const TSLanguage *tree_sitter_rsml(void) {
  static const TSLanguage language = {
    .abi_version = LANGUAGE_VERSION,
    .symbol_count = SYMBOL_COUNT,
    .alias_count = ALIAS_COUNT,
    .token_count = TOKEN_COUNT,
    .external_token_count = EXTERNAL_TOKEN_COUNT,
    .state_count = STATE_COUNT,
    .large_state_count = LARGE_STATE_COUNT,
    .production_id_count = PRODUCTION_ID_COUNT,
    .supertype_count = SUPERTYPE_COUNT,
    .field_count = FIELD_COUNT,
    .max_alias_sequence_length = MAX_ALIAS_SEQUENCE_LENGTH,
    .parse_table = &ts_parse_table[0][0],
    .small_parse_table = ts_small_parse_table,
    .small_parse_table_map = ts_small_parse_table_map,
    .parse_actions = ts_parse_actions,
    .symbol_names = ts_symbol_names,
    .field_names = ts_field_names,
    .field_map_slices = ts_field_map_slices,
    .field_map_entries = ts_field_map_entries,
    .symbol_metadata = ts_symbol_metadata,
    .public_symbol_map = ts_symbol_map,
    .alias_map = ts_non_terminal_alias_map,
    .alias_sequences = &ts_alias_sequences[0][0],
    .lex_modes = (const void*)ts_lex_modes,
    .lex_fn = ts_lex,
    .external_scanner = {
      &ts_external_scanner_states[0][0],
      ts_external_scanner_symbol_map,
      tree_sitter_rsml_external_scanner_create,
      tree_sitter_rsml_external_scanner_destroy,
      tree_sitter_rsml_external_scanner_scan,
      tree_sitter_rsml_external_scanner_serialize,
      tree_sitter_rsml_external_scanner_deserialize,
    },
    .primary_state_ids = ts_primary_state_ids,
    .name = "rsml",
    .max_reserved_word_set_size = 0,
    .metadata = {
      .major_version = 0,
      .minor_version = 1,
      .patch_version = 6,
    },
  };
  return &language;
}
#ifdef __cplusplus
}
#endif
